{-# OPTIONS_GHC -w #-}
{-# OPTIONS -fglasgow-exts -cpp #-}
{-# OPTIONS_HADDOCK hide #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Language.Haskell.Exts.Annotated.Parser
-- Copyright   :  (c) Niklas Broberg 2004-2009,
--                Original (c) Simon Marlow, Sven Panne 1997-2000
-- License     :  BSD-style (see the file LICENSE.txt)
--
-- Maintainer  :  Niklas Broberg, d00nibro@chalmers.se
-- Stability   :  stable
-- Portability :  portable
--
--
-----------------------------------------------------------------------------
module Language.Haskell.Exts.InternalParser (
              -- * General parsing
              ParseMode(..), defaultParseMode, ParseResult(..), fromParseResult,
              -- * Parsing of specific AST elements
              -- ** Modules
              parseModule, parseModuleWithMode, parseModuleWithComments,
              -- ** Expressions
              parseExp, parseExpWithMode, parseExpWithComments,
              -- ** Statements
              parseStmt, parseStmtWithMode, parseStmtWithComments,
              -- ** Patterns
              parsePat, parsePatWithMode, parsePatWithComments,
              -- ** Declarations
              parseDecl, parseDeclWithMode, parseDeclWithComments,
              -- ** Types
              parseType, parseTypeWithMode, parseTypeWithComments,
              -- ** Multiple modules in one file
              parseModules, parseModulesWithMode, parseModulesWithComments,
              -- ** Option pragmas
              getTopPragmas
              ) where
import Language.Haskell.Exts.Annotated.Syntax hiding ( Type(..), Exp(..), Asst(..), XAttr(..), FieldUpdate(..) )
import Language.Haskell.Exts.Annotated.Syntax ( Type, Exp, Asst )
import Language.Haskell.Exts.ParseMonad
import Language.Haskell.Exts.InternalLexer
import Language.Haskell.Exts.ParseUtils
import Language.Haskell.Exts.Annotated.Fixity
import Language.Haskell.Exts.SrcLoc
import Language.Haskell.Exts.Comments ( Comment )
import Language.Haskell.Exts.Extension

import Control.Monad ( liftM, (<=<) )
import qualified Data.Array as Happy_Data_Array
import qualified GHC.Exts as Happy_GHC_Exts

-- parser produced by Happy Version 1.19.3

newtype HappyAbsSyn  = HappyAbsSyn HappyAny
#if __GLASGOW_HASKELL__ >= 607
type HappyAny = Happy_GHC_Exts.Any
#else
type HappyAny = forall a . a
#endif
happyIn11 :: ([Module L]) -> (HappyAbsSyn )
happyIn11 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn11 #-}
happyOut11 :: (HappyAbsSyn ) -> ([Module L])
happyOut11 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut11 #-}
happyIn12 :: ([[ModulePragma L] -> [S] -> L -> Module L]) -> (HappyAbsSyn )
happyIn12 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn12 #-}
happyOut12 :: (HappyAbsSyn ) -> ([[ModulePragma L] -> [S] -> L -> Module L])
happyOut12 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut12 #-}
happyIn13 :: (Module L) -> (HappyAbsSyn )
happyIn13 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn13 #-}
happyOut13 :: (HappyAbsSyn ) -> (Module L)
happyOut13 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut13 #-}
happyIn14 :: (PExp L) -> (HappyAbsSyn )
happyIn14 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn14 #-}
happyOut14 :: (HappyAbsSyn ) -> (PExp L)
happyOut14 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut14 #-}
happyIn15 :: (([ModulePragma L],[S],L)) -> (HappyAbsSyn )
happyIn15 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn15 #-}
happyOut15 :: (HappyAbsSyn ) -> (([ModulePragma L],[S],L))
happyOut15 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut15 #-}
happyIn16 :: (([ModulePragma L],[S],Maybe L)) -> (HappyAbsSyn )
happyIn16 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn16 #-}
happyOut16 :: (HappyAbsSyn ) -> (([ModulePragma L],[S],Maybe L))
happyOut16 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut16 #-}
happyIn17 :: (ModulePragma L) -> (HappyAbsSyn )
happyIn17 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn17 #-}
happyOut17 :: (HappyAbsSyn ) -> (ModulePragma L)
happyOut17 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut17 #-}
happyIn18 :: (([Name L],[S])) -> (HappyAbsSyn )
happyIn18 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn18 #-}
happyOut18 :: (HappyAbsSyn ) -> (([Name L],[S]))
happyOut18 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut18 #-}
happyIn19 :: ([ModulePragma L] -> [S] -> L -> Module L) -> (HappyAbsSyn )
happyIn19 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn19 #-}
happyOut19 :: (HappyAbsSyn ) -> ([ModulePragma L] -> [S] -> L -> Module L)
happyOut19 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut19 #-}
happyIn20 :: (Maybe (ModuleHead L)) -> (HappyAbsSyn )
happyIn20 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn20 #-}
happyOut20 :: (HappyAbsSyn ) -> (Maybe (ModuleHead L))
happyOut20 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut20 #-}
happyIn21 :: (Maybe (WarningText L)) -> (HappyAbsSyn )
happyIn21 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn21 #-}
happyOut21 :: (HappyAbsSyn ) -> (Maybe (WarningText L))
happyOut21 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut21 #-}
happyIn22 :: (([ImportDecl L],[Decl L],[S],L)) -> (HappyAbsSyn )
happyIn22 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn22 #-}
happyOut22 :: (HappyAbsSyn ) -> (([ImportDecl L],[Decl L],[S],L))
happyOut22 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut22 #-}
happyIn23 :: (([ImportDecl L],[Decl L],[S])) -> (HappyAbsSyn )
happyIn23 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn23 #-}
happyOut23 :: (HappyAbsSyn ) -> (([ImportDecl L],[Decl L],[S]))
happyOut23 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut23 #-}
happyIn24 :: ([S]) -> (HappyAbsSyn )
happyIn24 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn24 #-}
happyOut24 :: (HappyAbsSyn ) -> ([S])
happyOut24 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut24 #-}
happyIn25 :: ([S]) -> (HappyAbsSyn )
happyIn25 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn25 #-}
happyOut25 :: (HappyAbsSyn ) -> ([S])
happyOut25 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut25 #-}
happyIn26 :: (Maybe (ExportSpecList L)) -> (HappyAbsSyn )
happyIn26 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn26 #-}
happyOut26 :: (HappyAbsSyn ) -> (Maybe (ExportSpecList L))
happyOut26 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut26 #-}
happyIn27 :: (ExportSpecList L) -> (HappyAbsSyn )
happyIn27 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn27 #-}
happyOut27 :: (HappyAbsSyn ) -> (ExportSpecList L)
happyOut27 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut27 #-}
happyIn28 :: ([S]) -> (HappyAbsSyn )
happyIn28 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn28 #-}
happyOut28 :: (HappyAbsSyn ) -> ([S])
happyOut28 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut28 #-}
happyIn29 :: (([ExportSpec L],[S])) -> (HappyAbsSyn )
happyIn29 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn29 #-}
happyOut29 :: (HappyAbsSyn ) -> (([ExportSpec L],[S]))
happyOut29 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut29 #-}
happyIn30 :: (ExportSpec L) -> (HappyAbsSyn )
happyIn30 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn30 #-}
happyOut30 :: (HappyAbsSyn ) -> (ExportSpec L)
happyOut30 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut30 #-}
happyIn31 :: (([ImportDecl L],[S])) -> (HappyAbsSyn )
happyIn31 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn31 #-}
happyOut31 :: (HappyAbsSyn ) -> (([ImportDecl L],[S]))
happyOut31 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut31 #-}
happyIn32 :: (ImportDecl L) -> (HappyAbsSyn )
happyIn32 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn32 #-}
happyOut32 :: (HappyAbsSyn ) -> (ImportDecl L)
happyOut32 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut32 #-}
happyIn33 :: ((Bool,[S])) -> (HappyAbsSyn )
happyIn33 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn33 #-}
happyOut33 :: (HappyAbsSyn ) -> ((Bool,[S]))
happyOut33 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut33 #-}
happyIn34 :: ((Bool,[S])) -> (HappyAbsSyn )
happyIn34 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn34 #-}
happyOut34 :: (HappyAbsSyn ) -> ((Bool,[S]))
happyOut34 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut34 #-}
happyIn35 :: ((Maybe String,[S])) -> (HappyAbsSyn )
happyIn35 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn35 #-}
happyOut35 :: (HappyAbsSyn ) -> ((Maybe String,[S]))
happyOut35 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut35 #-}
happyIn36 :: ((Maybe (ModuleName L),[S],Maybe L)) -> (HappyAbsSyn )
happyIn36 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn36 #-}
happyOut36 :: (HappyAbsSyn ) -> ((Maybe (ModuleName L),[S],Maybe L))
happyOut36 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut36 #-}
happyIn37 :: (Maybe (ImportSpecList L)) -> (HappyAbsSyn )
happyIn37 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn37 #-}
happyOut37 :: (HappyAbsSyn ) -> (Maybe (ImportSpecList L))
happyOut37 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut37 #-}
happyIn38 :: (ImportSpecList L) -> (HappyAbsSyn )
happyIn38 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn38 #-}
happyOut38 :: (HappyAbsSyn ) -> (ImportSpecList L)
happyOut38 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut38 #-}
happyIn39 :: ((Bool, Maybe L,[S])) -> (HappyAbsSyn )
happyIn39 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn39 #-}
happyOut39 :: (HappyAbsSyn ) -> ((Bool, Maybe L,[S]))
happyOut39 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut39 #-}
happyIn40 :: (([ImportSpec L],[S])) -> (HappyAbsSyn )
happyIn40 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn40 #-}
happyOut40 :: (HappyAbsSyn ) -> (([ImportSpec L],[S]))
happyOut40 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut40 #-}
happyIn41 :: (ImportSpec L) -> (HappyAbsSyn )
happyIn41 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn41 #-}
happyOut41 :: (HappyAbsSyn ) -> (ImportSpec L)
happyOut41 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut41 #-}
happyIn42 :: (([CName L],[S])) -> (HappyAbsSyn )
happyIn42 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn42 #-}
happyOut42 :: (HappyAbsSyn ) -> (([CName L],[S]))
happyOut42 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut42 #-}
happyIn43 :: (CName L) -> (HappyAbsSyn )
happyIn43 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn43 #-}
happyOut43 :: (HappyAbsSyn ) -> (CName L)
happyOut43 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut43 #-}
happyIn44 :: (Decl L) -> (HappyAbsSyn )
happyIn44 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn44 #-}
happyOut44 :: (HappyAbsSyn ) -> (Decl L)
happyOut44 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut44 #-}
happyIn45 :: ((Maybe Int, [S])) -> (HappyAbsSyn )
happyIn45 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn45 #-}
happyOut45 :: (HappyAbsSyn ) -> ((Maybe Int, [S]))
happyOut45 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut45 #-}
happyIn46 :: (Assoc L) -> (HappyAbsSyn )
happyIn46 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn46 #-}
happyOut46 :: (HappyAbsSyn ) -> (Assoc L)
happyOut46 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut46 #-}
happyIn47 :: (([Op L],[S],L)) -> (HappyAbsSyn )
happyIn47 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn47 #-}
happyOut47 :: (HappyAbsSyn ) -> (([Op L],[S],L))
happyOut47 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut47 #-}
happyIn48 :: (([Decl L],[S])) -> (HappyAbsSyn )
happyIn48 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn48 #-}
happyOut48 :: (HappyAbsSyn ) -> (([Decl L],[S]))
happyOut48 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut48 #-}
happyIn49 :: (([Decl L],[S])) -> (HappyAbsSyn )
happyIn49 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn49 #-}
happyOut49 :: (HappyAbsSyn ) -> (([Decl L],[S]))
happyOut49 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut49 #-}
happyIn50 :: (Decl L) -> (HappyAbsSyn )
happyIn50 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn50 #-}
happyOut50 :: (HappyAbsSyn ) -> (Decl L)
happyOut50 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut50 #-}
happyIn51 :: (DataOrNew L) -> (HappyAbsSyn )
happyIn51 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn51 #-}
happyOut51 :: (HappyAbsSyn ) -> (DataOrNew L)
happyOut51 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut51 #-}
happyIn52 :: (([Type L],[S])) -> (HappyAbsSyn )
happyIn52 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn52 #-}
happyOut52 :: (HappyAbsSyn ) -> (([Type L],[S]))
happyOut52 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut52 #-}
happyIn53 :: (([Decl L],[S])) -> (HappyAbsSyn )
happyIn53 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn53 #-}
happyOut53 :: (HappyAbsSyn ) -> (([Decl L],[S]))
happyOut53 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut53 #-}
happyIn54 :: (([Decl L],[S])) -> (HappyAbsSyn )
happyIn54 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn54 #-}
happyOut54 :: (HappyAbsSyn ) -> (([Decl L],[S]))
happyOut54 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut54 #-}
happyIn55 :: (Decl L) -> (HappyAbsSyn )
happyIn55 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn55 #-}
happyOut55 :: (HappyAbsSyn ) -> (Decl L)
happyOut55 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut55 #-}
happyIn56 :: (Binds L) -> (HappyAbsSyn )
happyIn56 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn56 #-}
happyOut56 :: (HappyAbsSyn ) -> (Binds L)
happyOut56 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut56 #-}
happyIn57 :: (Decl L) -> (HappyAbsSyn )
happyIn57 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn57 #-}
happyOut57 :: (HappyAbsSyn ) -> (Decl L)
happyOut57 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut57 #-}
happyIn58 :: (Decl L) -> (HappyAbsSyn )
happyIn58 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn58 #-}
happyOut58 :: (HappyAbsSyn ) -> (Decl L)
happyOut58 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut58 #-}
happyIn59 :: (([Type L],[S])) -> (HappyAbsSyn )
happyIn59 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn59 #-}
happyOut59 :: (HappyAbsSyn ) -> (([Type L],[S]))
happyOut59 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut59 #-}
happyIn60 :: (Type L) -> (HappyAbsSyn )
happyIn60 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn60 #-}
happyOut60 :: (HappyAbsSyn ) -> (Type L)
happyOut60 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut60 #-}
happyIn61 :: (Binds L) -> (HappyAbsSyn )
happyIn61 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn61 #-}
happyOut61 :: (HappyAbsSyn ) -> (Binds L)
happyOut61 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut61 #-}
happyIn62 :: (([Name L],[S],L)) -> (HappyAbsSyn )
happyIn62 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn62 #-}
happyOut62 :: (HappyAbsSyn ) -> (([Name L],[S],L))
happyOut62 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut62 #-}
happyIn63 :: (CallConv L) -> (HappyAbsSyn )
happyIn63 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn63 #-}
happyOut63 :: (HappyAbsSyn ) -> (CallConv L)
happyOut63 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut63 #-}
happyIn64 :: (Maybe (Safety L)) -> (HappyAbsSyn )
happyIn64 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn64 #-}
happyOut64 :: (HappyAbsSyn ) -> (Maybe (Safety L))
happyOut64 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut64 #-}
happyIn65 :: ((Maybe String, Name L, Type L, [S])) -> (HappyAbsSyn )
happyIn65 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn65 #-}
happyOut65 :: (HappyAbsSyn ) -> ((Maybe String, Name L, Type L, [S]))
happyOut65 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut65 #-}
happyIn66 :: ([Rule L]) -> (HappyAbsSyn )
happyIn66 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn66 #-}
happyOut66 :: (HappyAbsSyn ) -> ([Rule L])
happyOut66 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut66 #-}
happyIn67 :: (Rule L) -> (HappyAbsSyn )
happyIn67 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn67 #-}
happyOut67 :: (HappyAbsSyn ) -> (Rule L)
happyOut67 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut67 #-}
happyIn68 :: (Maybe (Activation L)) -> (HappyAbsSyn )
happyIn68 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn68 #-}
happyOut68 :: (HappyAbsSyn ) -> (Maybe (Activation L))
happyOut68 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut68 #-}
happyIn69 :: ((Maybe [RuleVar L],[S])) -> (HappyAbsSyn )
happyIn69 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn69 #-}
happyOut69 :: (HappyAbsSyn ) -> ((Maybe [RuleVar L],[S]))
happyOut69 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut69 #-}
happyIn70 :: ([RuleVar L]) -> (HappyAbsSyn )
happyIn70 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn70 #-}
happyOut70 :: (HappyAbsSyn ) -> ([RuleVar L])
happyOut70 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut70 #-}
happyIn71 :: (RuleVar L) -> (HappyAbsSyn )
happyIn71 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn71 #-}
happyOut71 :: (HappyAbsSyn ) -> (RuleVar L)
happyOut71 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut71 #-}
happyIn72 :: (([([Name L],String)],[S])) -> (HappyAbsSyn )
happyIn72 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn72 #-}
happyOut72 :: (HappyAbsSyn ) -> (([([Name L],String)],[S]))
happyOut72 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut72 #-}
happyIn73 :: ((([Name L], String),[S])) -> (HappyAbsSyn )
happyIn73 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn73 #-}
happyOut73 :: (HappyAbsSyn ) -> ((([Name L], String),[S]))
happyOut73 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut73 #-}
happyIn74 :: (([Name L],[S])) -> (HappyAbsSyn )
happyIn74 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn74 #-}
happyOut74 :: (HappyAbsSyn ) -> (([Name L],[S]))
happyOut74 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut74 #-}
happyIn75 :: (Name L) -> (HappyAbsSyn )
happyIn75 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn75 #-}
happyOut75 :: (HappyAbsSyn ) -> (Name L)
happyOut75 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut75 #-}
happyIn76 :: (Annotation L) -> (HappyAbsSyn )
happyIn76 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn76 #-}
happyOut76 :: (HappyAbsSyn ) -> (Annotation L)
happyOut76 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut76 #-}
happyIn77 :: (Type L) -> (HappyAbsSyn )
happyIn77 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn77 #-}
happyOut77 :: (HappyAbsSyn ) -> (Type L)
happyOut77 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut77 #-}
happyIn78 :: (PType L) -> (HappyAbsSyn )
happyIn78 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn78 #-}
happyOut78 :: (HappyAbsSyn ) -> (PType L)
happyOut78 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut78 #-}
happyIn79 :: (Type L) -> (HappyAbsSyn )
happyIn79 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn79 #-}
happyOut79 :: (HappyAbsSyn ) -> (Type L)
happyOut79 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut79 #-}
happyIn80 :: (PType L) -> (HappyAbsSyn )
happyIn80 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn80 #-}
happyOut80 :: (HappyAbsSyn ) -> (PType L)
happyOut80 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut80 #-}
happyIn81 :: (Type L) -> (HappyAbsSyn )
happyIn81 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn81 #-}
happyOut81 :: (HappyAbsSyn ) -> (Type L)
happyOut81 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut81 #-}
happyIn82 :: (PType L) -> (HappyAbsSyn )
happyIn82 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn82 #-}
happyOut82 :: (HappyAbsSyn ) -> (PType L)
happyOut82 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut82 #-}
happyIn83 :: (Type L) -> (HappyAbsSyn )
happyIn83 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn83 #-}
happyOut83 :: (HappyAbsSyn ) -> (Type L)
happyOut83 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut83 #-}
happyIn84 :: (PType L) -> (HappyAbsSyn )
happyIn84 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn84 #-}
happyOut84 :: (HappyAbsSyn ) -> (PType L)
happyOut84 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut84 #-}
happyIn85 :: (Promoted L) -> (HappyAbsSyn )
happyIn85 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn85 #-}
happyOut85 :: (HappyAbsSyn ) -> (Promoted L)
happyOut85 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut85 #-}
happyIn86 :: (Promoted L) -> (HappyAbsSyn )
happyIn86 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn86 #-}
happyOut86 :: (HappyAbsSyn ) -> (Promoted L)
happyOut86 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut86 #-}
happyIn87 :: (([Promoted L],[S])) -> (HappyAbsSyn )
happyIn87 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn87 #-}
happyOut87 :: (HappyAbsSyn ) -> (([Promoted L],[S]))
happyOut87 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut87 #-}
happyIn88 :: (([Promoted L],[S])) -> (HappyAbsSyn )
happyIn88 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn88 #-}
happyOut88 :: (HappyAbsSyn ) -> (([Promoted L],[S]))
happyOut88 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut88 #-}
happyIn89 :: (QName L) -> (HappyAbsSyn )
happyIn89 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn89 #-}
happyOut89 :: (HappyAbsSyn ) -> (QName L)
happyOut89 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut89 #-}
happyIn90 :: (QName L) -> (HappyAbsSyn )
happyIn90 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn90 #-}
happyOut90 :: (HappyAbsSyn ) -> (QName L)
happyOut90 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut90 #-}
happyIn91 :: (QName L) -> (HappyAbsSyn )
happyIn91 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn91 #-}
happyOut91 :: (HappyAbsSyn ) -> (QName L)
happyOut91 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut91 #-}
happyIn92 :: (Type L) -> (HappyAbsSyn )
happyIn92 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn92 #-}
happyOut92 :: (HappyAbsSyn ) -> (Type L)
happyOut92 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut92 #-}
happyIn93 :: (PType L) -> (HappyAbsSyn )
happyIn93 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn93 #-}
happyOut93 :: (HappyAbsSyn ) -> (PType L)
happyOut93 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut93 #-}
happyIn94 :: (PContext L) -> (HappyAbsSyn )
happyIn94 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn94 #-}
happyOut94 :: (HappyAbsSyn ) -> (PContext L)
happyOut94 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut94 #-}
happyIn95 :: (([PType L],[S])) -> (HappyAbsSyn )
happyIn95 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn95 #-}
happyOut95 :: (HappyAbsSyn ) -> (([PType L],[S]))
happyOut95 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut95 #-}
happyIn96 :: (([PType L],[S])) -> (HappyAbsSyn )
happyIn96 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn96 #-}
happyOut96 :: (HappyAbsSyn ) -> (([PType L],[S]))
happyOut96 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut96 #-}
happyIn97 :: (([TyVarBind L],Maybe L)) -> (HappyAbsSyn )
happyIn97 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn97 #-}
happyOut97 :: (HappyAbsSyn ) -> (([TyVarBind L],Maybe L))
happyOut97 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut97 #-}
happyIn98 :: (TyVarBind L) -> (HappyAbsSyn )
happyIn98 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn98 #-}
happyOut98 :: (HappyAbsSyn ) -> (TyVarBind L)
happyOut98 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut98 #-}
happyIn99 :: (([Name L],Maybe L)) -> (HappyAbsSyn )
happyIn99 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn99 #-}
happyOut99 :: (HappyAbsSyn ) -> (([Name L],Maybe L))
happyOut99 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut99 #-}
happyIn100 :: (([Name L],L)) -> (HappyAbsSyn )
happyIn100 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn100 #-}
happyOut100 :: (HappyAbsSyn ) -> (([Name L],L))
happyOut100 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut100 #-}
happyIn101 :: (([FunDep L],[S],Maybe L)) -> (HappyAbsSyn )
happyIn101 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn101 #-}
happyOut101 :: (HappyAbsSyn ) -> (([FunDep L],[S],Maybe L))
happyOut101 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut101 #-}
happyIn102 :: (([FunDep L],[S],L)) -> (HappyAbsSyn )
happyIn102 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn102 #-}
happyOut102 :: (HappyAbsSyn ) -> (([FunDep L],[S],L))
happyOut102 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut102 #-}
happyIn103 :: (FunDep L) -> (HappyAbsSyn )
happyIn103 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn103 #-}
happyOut103 :: (HappyAbsSyn ) -> (FunDep L)
happyOut103 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut103 #-}
happyIn104 :: (([GadtDecl L],[S],Maybe L)) -> (HappyAbsSyn )
happyIn104 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn104 #-}
happyOut104 :: (HappyAbsSyn ) -> (([GadtDecl L],[S],Maybe L))
happyOut104 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut104 #-}
happyIn105 :: (([GadtDecl L],[S])) -> (HappyAbsSyn )
happyIn105 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn105 #-}
happyOut105 :: (HappyAbsSyn ) -> (([GadtDecl L],[S]))
happyOut105 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut105 #-}
happyIn106 :: (([GadtDecl L],[S])) -> (HappyAbsSyn )
happyIn106 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn106 #-}
happyOut106 :: (HappyAbsSyn ) -> (([GadtDecl L],[S]))
happyOut106 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut106 #-}
happyIn107 :: (GadtDecl L) -> (HappyAbsSyn )
happyIn107 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn107 #-}
happyOut107 :: (HappyAbsSyn ) -> (GadtDecl L)
happyOut107 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut107 #-}
happyIn108 :: (([QualConDecl L],[S],Maybe L)) -> (HappyAbsSyn )
happyIn108 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn108 #-}
happyOut108 :: (HappyAbsSyn ) -> (([QualConDecl L],[S],Maybe L))
happyOut108 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut108 #-}
happyIn109 :: (([QualConDecl L],[S],L)) -> (HappyAbsSyn )
happyIn109 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn109 #-}
happyOut109 :: (HappyAbsSyn ) -> (([QualConDecl L],[S],L))
happyOut109 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut109 #-}
happyIn110 :: (QualConDecl L) -> (HappyAbsSyn )
happyIn110 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn110 #-}
happyOut110 :: (HappyAbsSyn ) -> (QualConDecl L)
happyOut110 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut110 #-}
happyIn111 :: ((Maybe [TyVarBind L], [S], Maybe L)) -> (HappyAbsSyn )
happyIn111 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn111 #-}
happyOut111 :: (HappyAbsSyn ) -> ((Maybe [TyVarBind L], [S], Maybe L))
happyOut111 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut111 #-}
happyIn112 :: (ConDecl L) -> (HappyAbsSyn )
happyIn112 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn112 #-}
happyOut112 :: (HappyAbsSyn ) -> (ConDecl L)
happyOut112 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut112 #-}
happyIn113 :: ((Name L, [BangType L], L)) -> (HappyAbsSyn )
happyIn113 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn113 #-}
happyOut113 :: (HappyAbsSyn ) -> ((Name L, [BangType L], L))
happyOut113 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut113 #-}
happyIn114 :: ((Name L, [BangType L],L)) -> (HappyAbsSyn )
happyIn114 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn114 #-}
happyOut114 :: (HappyAbsSyn ) -> ((Name L, [BangType L],L))
happyOut114 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut114 #-}
happyIn115 :: (BangType L) -> (HappyAbsSyn )
happyIn115 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn115 #-}
happyOut115 :: (HappyAbsSyn ) -> (BangType L)
happyOut115 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut115 #-}
happyIn116 :: (BangType L) -> (HappyAbsSyn )
happyIn116 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn116 #-}
happyOut116 :: (HappyAbsSyn ) -> (BangType L)
happyOut116 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut116 #-}
happyIn117 :: (([FieldDecl L],[S])) -> (HappyAbsSyn )
happyIn117 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn117 #-}
happyOut117 :: (HappyAbsSyn ) -> (([FieldDecl L],[S]))
happyOut117 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut117 #-}
happyIn118 :: (FieldDecl L) -> (HappyAbsSyn )
happyIn118 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn118 #-}
happyOut118 :: (HappyAbsSyn ) -> (FieldDecl L)
happyOut118 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut118 #-}
happyIn119 :: (BangType L) -> (HappyAbsSyn )
happyIn119 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn119 #-}
happyOut119 :: (HappyAbsSyn ) -> (BangType L)
happyOut119 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut119 #-}
happyIn120 :: (Maybe (Deriving L)) -> (HappyAbsSyn )
happyIn120 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn120 #-}
happyOut120 :: (HappyAbsSyn ) -> (Maybe (Deriving L))
happyOut120 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut120 #-}
happyIn121 :: (([InstHead L],[S])) -> (HappyAbsSyn )
happyIn121 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn121 #-}
happyOut121 :: (HappyAbsSyn ) -> (([InstHead L],[S]))
happyOut121 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut121 #-}
happyIn122 :: (QName L) -> (HappyAbsSyn )
happyIn122 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn122 #-}
happyOut122 :: (HappyAbsSyn ) -> (QName L)
happyOut122 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut122 #-}
happyIn123 :: (Kind L) -> (HappyAbsSyn )
happyIn123 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn123 #-}
happyOut123 :: (HappyAbsSyn ) -> (Kind L)
happyOut123 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut123 #-}
happyIn124 :: (Kind L) -> (HappyAbsSyn )
happyIn124 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn124 #-}
happyOut124 :: (HappyAbsSyn ) -> (Kind L)
happyOut124 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut124 #-}
happyIn125 :: (Kind L) -> (HappyAbsSyn )
happyIn125 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn125 #-}
happyOut125 :: (HappyAbsSyn ) -> (Kind L)
happyOut125 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut125 #-}
happyIn126 :: (Kind L) -> (HappyAbsSyn )
happyIn126 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn126 #-}
happyOut126 :: (HappyAbsSyn ) -> (Kind L)
happyOut126 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut126 #-}
happyIn127 :: (([Kind L],[S])) -> (HappyAbsSyn )
happyIn127 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn127 #-}
happyOut127 :: (HappyAbsSyn ) -> (([Kind L],[S]))
happyOut127 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut127 #-}
happyIn128 :: (([Kind L],[S])) -> (HappyAbsSyn )
happyIn128 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn128 #-}
happyOut128 :: (HappyAbsSyn ) -> (([Kind L],[S]))
happyOut128 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut128 #-}
happyIn129 :: ((Maybe (Kind L), [S])) -> (HappyAbsSyn )
happyIn129 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn129 #-}
happyOut129 :: (HappyAbsSyn ) -> ((Maybe (Kind L), [S]))
happyOut129 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut129 #-}
happyIn130 :: ((Maybe [ClassDecl L],[S],Maybe L)) -> (HappyAbsSyn )
happyIn130 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn130 #-}
happyOut130 :: (HappyAbsSyn ) -> ((Maybe [ClassDecl L],[S],Maybe L))
happyOut130 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut130 #-}
happyIn131 :: (([ClassDecl L],[S])) -> (HappyAbsSyn )
happyIn131 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn131 #-}
happyOut131 :: (HappyAbsSyn ) -> (([ClassDecl L],[S]))
happyOut131 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut131 #-}
happyIn132 :: (([ClassDecl L],[S])) -> (HappyAbsSyn )
happyIn132 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn132 #-}
happyOut132 :: (HappyAbsSyn ) -> (([ClassDecl L],[S]))
happyOut132 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut132 #-}
happyIn133 :: (ClassDecl L) -> (HappyAbsSyn )
happyIn133 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn133 #-}
happyOut133 :: (HappyAbsSyn ) -> (ClassDecl L)
happyOut133 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut133 #-}
happyIn134 :: (ClassDecl L) -> (HappyAbsSyn )
happyIn134 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn134 #-}
happyOut134 :: (HappyAbsSyn ) -> (ClassDecl L)
happyOut134 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut134 #-}
happyIn135 :: ((Maybe [InstDecl L],[S],Maybe L)) -> (HappyAbsSyn )
happyIn135 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn135 #-}
happyOut135 :: (HappyAbsSyn ) -> ((Maybe [InstDecl L],[S],Maybe L))
happyOut135 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut135 #-}
happyIn136 :: (([InstDecl L],[S])) -> (HappyAbsSyn )
happyIn136 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn136 #-}
happyOut136 :: (HappyAbsSyn ) -> (([InstDecl L],[S]))
happyOut136 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut136 #-}
happyIn137 :: (([InstDecl L],[S])) -> (HappyAbsSyn )
happyIn137 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn137 #-}
happyOut137 :: (HappyAbsSyn ) -> (([InstDecl L],[S]))
happyOut137 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut137 #-}
happyIn138 :: (InstDecl L) -> (HappyAbsSyn )
happyIn138 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn138 #-}
happyOut138 :: (HappyAbsSyn ) -> (InstDecl L)
happyOut138 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut138 #-}
happyIn139 :: (InstDecl L) -> (HappyAbsSyn )
happyIn139 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn139 #-}
happyOut139 :: (HappyAbsSyn ) -> (InstDecl L)
happyOut139 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut139 #-}
happyIn140 :: (Decl L) -> (HappyAbsSyn )
happyIn140 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn140 #-}
happyOut140 :: (HappyAbsSyn ) -> (Decl L)
happyOut140 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut140 #-}
happyIn141 :: ((Maybe (Binds L),[S])) -> (HappyAbsSyn )
happyIn141 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn141 #-}
happyOut141 :: (HappyAbsSyn ) -> ((Maybe (Binds L),[S]))
happyOut141 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut141 #-}
happyIn142 :: ((Maybe (Type L),[S])) -> (HappyAbsSyn )
happyIn142 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn142 #-}
happyOut142 :: (HappyAbsSyn ) -> ((Maybe (Type L),[S]))
happyOut142 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut142 #-}
happyIn143 :: (Rhs L) -> (HappyAbsSyn )
happyIn143 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn143 #-}
happyOut143 :: (HappyAbsSyn ) -> (Rhs L)
happyOut143 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut143 #-}
happyIn144 :: (([GuardedRhs L],L)) -> (HappyAbsSyn )
happyIn144 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn144 #-}
happyOut144 :: (HappyAbsSyn ) -> (([GuardedRhs L],L))
happyOut144 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut144 #-}
happyIn145 :: (GuardedRhs L) -> (HappyAbsSyn )
happyIn145 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn145 #-}
happyOut145 :: (HappyAbsSyn ) -> (GuardedRhs L)
happyOut145 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut145 #-}
happyIn146 :: (Exp L) -> (HappyAbsSyn )
happyIn146 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn146 #-}
happyOut146 :: (HappyAbsSyn ) -> (Exp L)
happyOut146 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut146 #-}
happyIn147 :: (PExp L) -> (HappyAbsSyn )
happyIn147 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn147 #-}
happyOut147 :: (HappyAbsSyn ) -> (PExp L)
happyOut147 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut147 #-}
happyIn148 :: (PExp L) -> (HappyAbsSyn )
happyIn148 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn148 #-}
happyOut148 :: (HappyAbsSyn ) -> (PExp L)
happyOut148 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut148 #-}
happyIn149 :: (PExp L) -> (HappyAbsSyn )
happyIn149 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn149 #-}
happyOut149 :: (HappyAbsSyn ) -> (PExp L)
happyOut149 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut149 #-}
happyIn150 :: (PExp L) -> (HappyAbsSyn )
happyIn150 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn150 #-}
happyOut150 :: (HappyAbsSyn ) -> (PExp L)
happyOut150 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut150 #-}
happyIn151 :: (PExp L) -> (HappyAbsSyn )
happyIn151 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn151 #-}
happyOut151 :: (HappyAbsSyn ) -> (PExp L)
happyOut151 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut151 #-}
happyIn152 :: ([S]) -> (HappyAbsSyn )
happyIn152 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn152 #-}
happyOut152 :: (HappyAbsSyn ) -> ([S])
happyOut152 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut152 #-}
happyIn153 :: ([S]) -> (HappyAbsSyn )
happyIn153 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn153 #-}
happyOut153 :: (HappyAbsSyn ) -> ([S])
happyOut153 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut153 #-}
happyIn154 :: (PExp L) -> (HappyAbsSyn )
happyIn154 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn154 #-}
happyOut154 :: (HappyAbsSyn ) -> (PExp L)
happyOut154 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut154 #-}
happyIn155 :: (PExp L) -> (HappyAbsSyn )
happyIn155 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn155 #-}
happyOut155 :: (HappyAbsSyn ) -> (PExp L)
happyOut155 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut155 #-}
happyIn156 :: (PExp L) -> (HappyAbsSyn )
happyIn156 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn156 #-}
happyOut156 :: (HappyAbsSyn ) -> (PExp L)
happyOut156 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut156 #-}
happyIn157 :: ([Pat L]) -> (HappyAbsSyn )
happyIn157 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn157 #-}
happyOut157 :: (HappyAbsSyn ) -> ([Pat L])
happyOut157 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut157 #-}
happyIn158 :: (Pat L) -> (HappyAbsSyn )
happyIn158 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn158 #-}
happyOut158 :: (HappyAbsSyn ) -> (Pat L)
happyOut158 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut158 #-}
happyIn159 :: (PExp L) -> (HappyAbsSyn )
happyIn159 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn159 #-}
happyOut159 :: (HappyAbsSyn ) -> (PExp L)
happyOut159 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut159 #-}
happyIn160 :: (PExp L) -> (HappyAbsSyn )
happyIn160 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn160 #-}
happyOut160 :: (HappyAbsSyn ) -> (PExp L)
happyOut160 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut160 #-}
happyIn161 :: (PExp L) -> (HappyAbsSyn )
happyIn161 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn161 #-}
happyOut161 :: (HappyAbsSyn ) -> (PExp L)
happyOut161 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut161 #-}
happyIn162 :: ([S]) -> (HappyAbsSyn )
happyIn162 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn162 #-}
happyOut162 :: (HappyAbsSyn ) -> ([S])
happyOut162 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut162 #-}
happyIn163 :: (PExp L) -> (HappyAbsSyn )
happyIn163 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn163 #-}
happyOut163 :: (HappyAbsSyn ) -> (PExp L)
happyOut163 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut163 #-}
happyIn164 :: (([Maybe (PExp L)],[S])) -> (HappyAbsSyn )
happyIn164 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn164 #-}
happyOut164 :: (HappyAbsSyn ) -> (([Maybe (PExp L)],[S]))
happyOut164 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut164 #-}
happyIn165 :: (([Maybe (PExp L)],[S])) -> (HappyAbsSyn )
happyIn165 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn165 #-}
happyOut165 :: (HappyAbsSyn ) -> (([Maybe (PExp L)],[S]))
happyOut165 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut165 #-}
happyIn166 :: (([PExp L],[S])) -> (HappyAbsSyn )
happyIn166 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn166 #-}
happyOut166 :: (HappyAbsSyn ) -> (([PExp L],[S]))
happyOut166 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut166 #-}
happyIn167 :: (PExp L) -> (HappyAbsSyn )
happyIn167 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn167 #-}
happyOut167 :: (HappyAbsSyn ) -> (PExp L)
happyOut167 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut167 #-}
happyIn168 :: (PExp L) -> (HappyAbsSyn )
happyIn168 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn168 #-}
happyOut168 :: (HappyAbsSyn ) -> (PExp L)
happyOut168 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut168 #-}
happyIn169 :: ([PExp L]) -> (HappyAbsSyn )
happyIn169 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn169 #-}
happyOut169 :: (HappyAbsSyn ) -> ([PExp L])
happyOut169 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut169 #-}
happyIn170 :: (PExp L) -> (HappyAbsSyn )
happyIn170 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn170 #-}
happyOut170 :: (HappyAbsSyn ) -> (PExp L)
happyOut170 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut170 #-}
happyIn171 :: (XName L) -> (HappyAbsSyn )
happyIn171 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn171 #-}
happyOut171 :: (HappyAbsSyn ) -> (XName L)
happyOut171 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut171 #-}
happyIn172 :: (Loc String) -> (HappyAbsSyn )
happyIn172 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn172 #-}
happyOut172 :: (HappyAbsSyn ) -> (Loc String)
happyOut172 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut172 #-}
happyIn173 :: (Loc String) -> (HappyAbsSyn )
happyIn173 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn173 #-}
happyOut173 :: (HappyAbsSyn ) -> (Loc String)
happyOut173 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut173 #-}
happyIn174 :: ([ParseXAttr L]) -> (HappyAbsSyn )
happyIn174 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn174 #-}
happyOut174 :: (HappyAbsSyn ) -> ([ParseXAttr L])
happyOut174 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut174 #-}
happyIn175 :: (ParseXAttr L) -> (HappyAbsSyn )
happyIn175 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn175 #-}
happyOut175 :: (HappyAbsSyn ) -> (ParseXAttr L)
happyOut175 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut175 #-}
happyIn176 :: (Maybe (PExp L)) -> (HappyAbsSyn )
happyIn176 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn176 #-}
happyOut176 :: (HappyAbsSyn ) -> (Maybe (PExp L))
happyOut176 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut176 #-}
happyIn177 :: (L -> PExp L) -> (HappyAbsSyn )
happyIn177 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn177 #-}
happyOut177 :: (HappyAbsSyn ) -> (L -> PExp L)
happyOut177 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut177 #-}
happyIn178 :: (([PExp L],[S])) -> (HappyAbsSyn )
happyIn178 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn178 #-}
happyOut178 :: (HappyAbsSyn ) -> (([PExp L],[S]))
happyOut178 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut178 #-}
happyIn179 :: (([[QualStmt L]],[S])) -> (HappyAbsSyn )
happyIn179 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn179 #-}
happyOut179 :: (HappyAbsSyn ) -> (([[QualStmt L]],[S]))
happyOut179 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut179 #-}
happyIn180 :: (([QualStmt L],[S])) -> (HappyAbsSyn )
happyIn180 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn180 #-}
happyOut180 :: (HappyAbsSyn ) -> (([QualStmt L],[S]))
happyOut180 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut180 #-}
happyIn181 :: (QualStmt L) -> (HappyAbsSyn )
happyIn181 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn181 #-}
happyOut181 :: (HappyAbsSyn ) -> (QualStmt L)
happyOut181 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut181 #-}
happyIn182 :: (QualStmt L) -> (HappyAbsSyn )
happyIn182 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn182 #-}
happyOut182 :: (HappyAbsSyn ) -> (QualStmt L)
happyOut182 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut182 #-}
happyIn183 :: (([Stmt L],[S])) -> (HappyAbsSyn )
happyIn183 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn183 #-}
happyOut183 :: (HappyAbsSyn ) -> (([Stmt L],[S]))
happyOut183 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut183 #-}
happyIn184 :: (Stmt L) -> (HappyAbsSyn )
happyIn184 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn184 #-}
happyOut184 :: (HappyAbsSyn ) -> (Stmt L)
happyOut184 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut184 #-}
happyIn185 :: (([Alt L],L,[S])) -> (HappyAbsSyn )
happyIn185 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn185 #-}
happyOut185 :: (HappyAbsSyn ) -> (([Alt L],L,[S]))
happyOut185 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut185 #-}
happyIn186 :: (([Alt L],[S])) -> (HappyAbsSyn )
happyIn186 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn186 #-}
happyOut186 :: (HappyAbsSyn ) -> (([Alt L],[S]))
happyOut186 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut186 #-}
happyIn187 :: (([Alt L],[S])) -> (HappyAbsSyn )
happyIn187 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn187 #-}
happyOut187 :: (HappyAbsSyn ) -> (([Alt L],[S]))
happyOut187 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut187 #-}
happyIn188 :: (Alt L) -> (HappyAbsSyn )
happyIn188 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn188 #-}
happyOut188 :: (HappyAbsSyn ) -> (Alt L)
happyOut188 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut188 #-}
happyIn189 :: (GuardedAlts L) -> (HappyAbsSyn )
happyIn189 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn189 #-}
happyOut189 :: (HappyAbsSyn ) -> (GuardedAlts L)
happyOut189 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut189 #-}
happyIn190 :: (([GuardedAlt L],L)) -> (HappyAbsSyn )
happyIn190 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn190 #-}
happyOut190 :: (HappyAbsSyn ) -> (([GuardedAlt L],L))
happyOut190 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut190 #-}
happyIn191 :: (GuardedAlt L) -> (HappyAbsSyn )
happyIn191 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn191 #-}
happyOut191 :: (HappyAbsSyn ) -> (GuardedAlt L)
happyOut191 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut191 #-}
happyIn192 :: (Pat L) -> (HappyAbsSyn )
happyIn192 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn192 #-}
happyOut192 :: (HappyAbsSyn ) -> (Pat L)
happyOut192 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut192 #-}
happyIn193 :: (([IfAlt L], L, [S])) -> (HappyAbsSyn )
happyIn193 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn193 #-}
happyOut193 :: (HappyAbsSyn ) -> (([IfAlt L], L, [S]))
happyOut193 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut193 #-}
happyIn194 :: (([IfAlt L], [S])) -> (HappyAbsSyn )
happyIn194 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn194 #-}
happyOut194 :: (HappyAbsSyn ) -> (([IfAlt L], [S]))
happyOut194 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut194 #-}
happyIn195 :: (([IfAlt L], [S])) -> (HappyAbsSyn )
happyIn195 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn195 #-}
happyOut195 :: (HappyAbsSyn ) -> (([IfAlt L], [S]))
happyOut195 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut195 #-}
happyIn196 :: (IfAlt L) -> (HappyAbsSyn )
happyIn196 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn196 #-}
happyOut196 :: (HappyAbsSyn ) -> (IfAlt L)
happyOut196 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut196 #-}
happyIn197 :: (([Stmt L],L,[S])) -> (HappyAbsSyn )
happyIn197 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn197 #-}
happyOut197 :: (HappyAbsSyn ) -> (([Stmt L],L,[S]))
happyOut197 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut197 #-}
happyIn198 :: (([Stmt L],[S])) -> (HappyAbsSyn )
happyIn198 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn198 #-}
happyOut198 :: (HappyAbsSyn ) -> (([Stmt L],[S]))
happyOut198 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut198 #-}
happyIn199 :: (([Stmt L],[S])) -> (HappyAbsSyn )
happyIn199 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn199 #-}
happyOut199 :: (HappyAbsSyn ) -> (([Stmt L],[S]))
happyOut199 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut199 #-}
happyIn200 :: (Stmt L) -> (HappyAbsSyn )
happyIn200 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn200 #-}
happyOut200 :: (HappyAbsSyn ) -> (Stmt L)
happyOut200 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut200 #-}
happyIn201 :: (([PFieldUpdate L],[S])) -> (HappyAbsSyn )
happyIn201 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn201 #-}
happyOut201 :: (HappyAbsSyn ) -> (([PFieldUpdate L],[S]))
happyOut201 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut201 #-}
happyIn202 :: (PFieldUpdate L) -> (HappyAbsSyn )
happyIn202 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn202 #-}
happyOut202 :: (HappyAbsSyn ) -> (PFieldUpdate L)
happyOut202 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut202 #-}
happyIn203 :: (([IPBind L],[S])) -> (HappyAbsSyn )
happyIn203 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn203 #-}
happyOut203 :: (HappyAbsSyn ) -> (([IPBind L],[S]))
happyOut203 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut203 #-}
happyIn204 :: (([IPBind L],[S])) -> (HappyAbsSyn )
happyIn204 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn204 #-}
happyOut204 :: (HappyAbsSyn ) -> (([IPBind L],[S]))
happyOut204 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut204 #-}
happyIn205 :: (IPBind L) -> (HappyAbsSyn )
happyIn205 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn205 #-}
happyOut205 :: (HappyAbsSyn ) -> (IPBind L)
happyOut205 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut205 #-}
happyIn206 :: (PExp L) -> (HappyAbsSyn )
happyIn206 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn206 #-}
happyOut206 :: (HappyAbsSyn ) -> (PExp L)
happyOut206 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut206 #-}
happyIn207 :: (Name L) -> (HappyAbsSyn )
happyIn207 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn207 #-}
happyOut207 :: (HappyAbsSyn ) -> (Name L)
happyOut207 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut207 #-}
happyIn208 :: (Name L) -> (HappyAbsSyn )
happyIn208 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn208 #-}
happyOut208 :: (HappyAbsSyn ) -> (Name L)
happyOut208 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut208 #-}
happyIn209 :: (QName L) -> (HappyAbsSyn )
happyIn209 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn209 #-}
happyOut209 :: (HappyAbsSyn ) -> (QName L)
happyOut209 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut209 #-}
happyIn210 :: (IPName L) -> (HappyAbsSyn )
happyIn210 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn210 #-}
happyOut210 :: (HappyAbsSyn ) -> (IPName L)
happyOut210 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut210 #-}
happyIn211 :: (Name L) -> (HappyAbsSyn )
happyIn211 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn211 #-}
happyOut211 :: (HappyAbsSyn ) -> (Name L)
happyOut211 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut211 #-}
happyIn212 :: (QName L) -> (HappyAbsSyn )
happyIn212 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn212 #-}
happyOut212 :: (HappyAbsSyn ) -> (QName L)
happyOut212 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut212 #-}
happyIn213 :: (Name L) -> (HappyAbsSyn )
happyIn213 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn213 #-}
happyOut213 :: (HappyAbsSyn ) -> (Name L)
happyOut213 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut213 #-}
happyIn214 :: (QName L) -> (HappyAbsSyn )
happyIn214 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn214 #-}
happyOut214 :: (HappyAbsSyn ) -> (QName L)
happyOut214 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut214 #-}
happyIn215 :: (QName L) -> (HappyAbsSyn )
happyIn215 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn215 #-}
happyOut215 :: (HappyAbsSyn ) -> (QName L)
happyOut215 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut215 #-}
happyIn216 :: (Name L) -> (HappyAbsSyn )
happyIn216 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn216 #-}
happyOut216 :: (HappyAbsSyn ) -> (Name L)
happyOut216 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut216 #-}
happyIn217 :: (QName L) -> (HappyAbsSyn )
happyIn217 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn217 #-}
happyOut217 :: (HappyAbsSyn ) -> (QName L)
happyOut217 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut217 #-}
happyIn218 :: (Op L) -> (HappyAbsSyn )
happyIn218 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn218 #-}
happyOut218 :: (HappyAbsSyn ) -> (Op L)
happyOut218 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut218 #-}
happyIn219 :: (QOp L) -> (HappyAbsSyn )
happyIn219 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn219 #-}
happyOut219 :: (HappyAbsSyn ) -> (QOp L)
happyOut219 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut219 #-}
happyIn220 :: (QOp L) -> (HappyAbsSyn )
happyIn220 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn220 #-}
happyOut220 :: (HappyAbsSyn ) -> (QOp L)
happyOut220 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut220 #-}
happyIn221 :: (QName L) -> (HappyAbsSyn )
happyIn221 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn221 #-}
happyOut221 :: (HappyAbsSyn ) -> (QName L)
happyOut221 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut221 #-}
happyIn222 :: (QName L) -> (HappyAbsSyn )
happyIn222 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn222 #-}
happyOut222 :: (HappyAbsSyn ) -> (QName L)
happyOut222 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut222 #-}
happyIn223 :: (Name L) -> (HappyAbsSyn )
happyIn223 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn223 #-}
happyOut223 :: (HappyAbsSyn ) -> (Name L)
happyOut223 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut223 #-}
happyIn224 :: (Name L) -> (HappyAbsSyn )
happyIn224 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn224 #-}
happyOut224 :: (HappyAbsSyn ) -> (Name L)
happyOut224 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut224 #-}
happyIn225 :: (IPName L) -> (HappyAbsSyn )
happyIn225 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn225 #-}
happyOut225 :: (HappyAbsSyn ) -> (IPName L)
happyOut225 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut225 #-}
happyIn226 :: (QName L) -> (HappyAbsSyn )
happyIn226 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn226 #-}
happyOut226 :: (HappyAbsSyn ) -> (QName L)
happyOut226 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut226 #-}
happyIn227 :: (Name L) -> (HappyAbsSyn )
happyIn227 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn227 #-}
happyOut227 :: (HappyAbsSyn ) -> (Name L)
happyOut227 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut227 #-}
happyIn228 :: (QName L) -> (HappyAbsSyn )
happyIn228 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn228 #-}
happyOut228 :: (HappyAbsSyn ) -> (QName L)
happyOut228 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut228 #-}
happyIn229 :: (Name L) -> (HappyAbsSyn )
happyIn229 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn229 #-}
happyOut229 :: (HappyAbsSyn ) -> (Name L)
happyOut229 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut229 #-}
happyIn230 :: (QName L) -> (HappyAbsSyn )
happyIn230 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn230 #-}
happyOut230 :: (HappyAbsSyn ) -> (QName L)
happyOut230 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut230 #-}
happyIn231 :: (QName L) -> (HappyAbsSyn )
happyIn231 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn231 #-}
happyOut231 :: (HappyAbsSyn ) -> (QName L)
happyOut231 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut231 #-}
happyIn232 :: (Name L) -> (HappyAbsSyn )
happyIn232 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn232 #-}
happyOut232 :: (HappyAbsSyn ) -> (Name L)
happyOut232 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut232 #-}
happyIn233 :: (Name L) -> (HappyAbsSyn )
happyIn233 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn233 #-}
happyOut233 :: (HappyAbsSyn ) -> (Name L)
happyOut233 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut233 #-}
happyIn234 :: (QName L) -> (HappyAbsSyn )
happyIn234 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn234 #-}
happyOut234 :: (HappyAbsSyn ) -> (QName L)
happyOut234 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut234 #-}
happyIn235 :: (Literal L) -> (HappyAbsSyn )
happyIn235 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn235 #-}
happyOut235 :: (HappyAbsSyn ) -> (Literal L)
happyOut235 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut235 #-}
happyIn236 :: (S) -> (HappyAbsSyn )
happyIn236 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn236 #-}
happyOut236 :: (HappyAbsSyn ) -> (S)
happyOut236 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut236 #-}
happyIn237 :: (S) -> (HappyAbsSyn )
happyIn237 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn237 #-}
happyOut237 :: (HappyAbsSyn ) -> (S)
happyOut237 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut237 #-}
happyIn238 :: (ModuleName L) -> (HappyAbsSyn )
happyIn238 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn238 #-}
happyOut238 :: (HappyAbsSyn ) -> (ModuleName L)
happyOut238 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut238 #-}
happyIn239 :: (Name L) -> (HappyAbsSyn )
happyIn239 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn239 #-}
happyOut239 :: (HappyAbsSyn ) -> (Name L)
happyOut239 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut239 #-}
happyIn240 :: (QName L) -> (HappyAbsSyn )
happyIn240 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn240 #-}
happyOut240 :: (HappyAbsSyn ) -> (QName L)
happyOut240 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut240 #-}
happyIn241 :: (Name L) -> (HappyAbsSyn )
happyIn241 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn241 #-}
happyOut241 :: (HappyAbsSyn ) -> (Name L)
happyOut241 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut241 #-}
happyIn242 :: (Name L) -> (HappyAbsSyn )
happyIn242 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn242 #-}
happyOut242 :: (HappyAbsSyn ) -> (Name L)
happyOut242 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut242 #-}
happyIn243 :: (QName L) -> (HappyAbsSyn )
happyIn243 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn243 #-}
happyOut243 :: (HappyAbsSyn ) -> (QName L)
happyOut243 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut243 #-}
happyIn244 :: (Name L) -> (HappyAbsSyn )
happyIn244 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyIn244 #-}
happyOut244 :: (HappyAbsSyn ) -> (Name L)
happyOut244 x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOut244 #-}
happyInTok :: (Loc Token) -> (HappyAbsSyn )
happyInTok x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyInTok #-}
happyOutTok :: (HappyAbsSyn ) -> (Loc Token)
happyOutTok x = Happy_GHC_Exts.unsafeCoerce# x
{-# INLINE happyOutTok #-}


happyActOffsets :: HappyAddr
happyActOffsets = HappyA# "\x00\x00\xa1\x16\x28\x16\x67\x0b\x59\x1f\x52\x13\x00\x00\x00\x00\x00\x00\x82\x07\xa4\x04\xf0\x07\x5f\x07\x00\x00\xbe\x07\x00\x00\x00\x00\x21\x22\x00\x00\x00\x00\x00\x00\xf9\x1b\x00\x00\xcc\x07\x00\x00\x00\x00\xb8\x07\x56\x07\x00\x00\x4b\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xf1\x0f\x6f\x0f\x73\x10\x00\x00\x41\x1b\xf9\x1b\xf9\x1b\xf9\x1b\xa1\x16\x00\x00\xa1\x16\xa1\x16\xa1\x16\x59\x1f\x00\x00\x1b\x22\xec\x20\x00\x00\x20\x1c\xa1\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa1\x16\xc9\x07\x00\x00\x00\x00\x00\x00\xaf\x15\xc0\x07\xb9\x07\x9d\x1b\xb9\x07\x00\x00\xc8\x07\xc7\x07\xc6\x07\x00\x00\x00\x00\xfb\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x48\x07\x00\x00\x59\x1f\xad\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x99\x1c\xdc\x1d\x4c\x1f\x1a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xc5\x07\x42\x07\x56\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x1c\xf9\x1b\x1e\x00\xf0\x1e\x6d\x07\xb5\x07\x53\x07\x00\x00\x00\x00\x00\x00\xf0\x1e\xac\x07\x00\x00\x00\x20\xa5\x07\xa5\x07\x29\x00\xa5\x07\xab\x07\x4d\x23\x4d\x23\xe8\x21\x00\x00\x32\x07\x32\x07\x00\x00\x32\x07\x04\x00\x00\x00\x00\x00\x93\x07\x20\x1c\x47\x07\x67\x06\xf9\x1b\x2f\x07\x00\x00\x00\x00\x00\x00\x00\x00\x91\x02\xf9\x1b\xa2\x07\x36\x00\x00\x00\x97\x07\x84\x07\x33\x00\x2c\x00\x00\x00\x81\x07\x34\x23\x4f\x00\x34\x23\xf0\x1e\x34\x23\x34\x23\x76\x07\x53\x1d\xf3\x1f\x55\x21\x00\x00\x2d\x07\x00\x00\x00\x00\x1d\x07\xf0\x1e\xf0\x1e\xf0\x1e\x6b\x07\x23\x05\x23\x05\x8d\x05\x8d\x05\x00\x00\x00\x00\xa1\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x34\x23\xec\x0c\x00\x00\x00\x00\xf0\x1e\x00\x00\x00\x00\x00\x00\x42\x06\xf0\x1e\x53\x03\x00\x00\x0f\x24\x00\x00\x00\x00\xd9\x0c\x63\x02\x70\x07\x68\x07\x00\x00\x6e\x07\x64\x07\xbb\x00\x00\x00\x99\x1c\x00\x00\x00\x00\xe0\x02\x89\x02\x00\x00\x00\x00\x99\x01\x6c\x07\x62\x07\x50\x02\x6a\x07\x69\x07\x00\x00\x67\x07\x55\x21\x00\x00\x00\x00\x55\x21\x00\x00\x55\x21\x00\x00\x00\x00\xbc\x1e\xf0\x1e\x55\x21\x00\x00\x6f\x07\xf6\x06\xf4\x06\x00\x00\x6e\x11\x6e\x11\x51\x07\x00\x00\xf9\x1b\x00\x00\x10\x07\x5e\x07\x00\x00\x00\x00\x00\x00\x00\x00\x02\x07\xe6\x03\x25\x07\x00\x00\x4b\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5c\x17\x5f\x02\x4f\x07\x00\x00\x00\x00\x62\x11\x67\x0b\x2e\x07\x2a\x07\x70\x0c\x29\x07\x50\x07\x43\x07\x3f\x00\x00\x00\xf9\x1b\x00\x00\xe5\x1a\x00\x00\x4a\x07\x3f\x07\x91\x01\x41\x07\x3e\x07\x00\x00\x00\x00\xa1\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xec\x0c\x00\x00\x00\x00\x00\x00\xed\x0e\x2e\x02\x00\x00\x73\x05\x6b\x0e\x4b\x02\x40\x07\x3c\x07\x3a\x07\x37\x07\x35\x07\x00\x00\x33\x07\xf9\x1b\x31\x07\x30\x07\xf9\x1b\xf9\x1b\xa1\x16\xc1\x21\x00\x00\xa1\x16\xf0\x1e\xa1\x16\xa1\x16\xa1\x16\xa1\x16\x62\x00\x2c\x07\x00\x00\x3d\x07\x7c\x21\x00\x00\x17\x00\x00\x00\xb6\x06\x00\x00\x1e\x07\x00\x00\x09\x00\xa4\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xee\x02\x00\x00\x1c\x07\x54\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xe9\x0d\x00\x00\x00\x00\x0b\x02\x00\x00\x00\x00\xa1\x16\xa1\x16\x67\x0d\x00\x00\x00\x00\xe7\x01\x00\x00\x1b\x07\x18\x07\x00\x00\xf5\x10\x00\x00\xf5\x10\xa1\x16\xd9\x12\x00\x00\x00\x00\x00\x00\x00\x00\xa1\x16\xa1\x16\x00\x00\x36\x15\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x00\x00\x00\x00\x21\x07\x20\x1c\xfd\x19\x00\x00\xf1\x06\x00\x00\x00\x00\x00\x00\x00\x00\xa1\x16\x19\x07\x52\x01\x17\x07\x62\x00\xbd\x06\x00\x00\xa1\x16\x00\x00\xa1\x16\x12\x07\x11\x07\x6e\x11\x62\x00\xa1\x16\xa1\x16\x0a\x07\xd2\x20\x00\x00\x08\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xf0\x1e\x00\x00\x00\x00\x76\x20\x00\x00\x00\x00\xf0\x1e\x00\x07\x63\x02\x00\x00\x00\x00\x00\x00\x47\x06\x6a\x01\x63\x02\xd1\x01\x00\x00\x00\x00\x00\x00\x69\x20\xf0\x1e\x09\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xd9\x23\x39\x06\xc4\x06\xae\x06\x76\x20\xbb\x06\x61\x05\xf3\x06\x64\x02\x00\x00\xaa\x06\xf5\x06\x00\x00\xa1\x16\x36\x15\xa8\x06\x36\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x55\x07\x9e\x06\x00\x00\xec\x06\xf7\x06\x00\x00\xf0\x06\x00\x00\x00\x00\x00\x00\xee\x06\x91\x18\xeb\x06\xe7\x06\x62\x00\x62\x00\xd7\x06\x00\x00\xda\x06\x55\x21\xf0\x1e\x76\x06\x70\x06\x6e\x06\xcf\x06\xd5\x06\xe3\x06\xc5\x06\x8f\x06\xd8\x06\x00\x00\x18\x23\x00\x00\x18\x23\x00\x00\x00\x00\xf9\x1b\x00\x00\xd3\x06\xce\x06\x00\x00\x00\x00\xda\x05\x00\x00\x00\x00\xa2\x06\x00\x00\x00\x00\x00\x00\x00\x00\xc7\x06\x57\x0a\x62\x00\xfd\x19\x9c\x06\xcb\x06\xd0\x06\xcc\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa1\x16\xab\x23\xf0\x1e\xb7\x06\x00\x00\xf0\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x55\x21\x00\x00\xf0\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xe5\x1d\x00\x00\x00\x00\xac\x06\x00\x00\x75\x06\x00\x00\x00\x00\x00\x00\x00\x00\x4e\x22\xad\x06\x00\x00\x00\x00\xb1\x06\x00\x00\xa9\x06\x00\x00\x39\x05\xdc\x01\x55\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xaf\x06\x00\x00\xc3\x01\x00\x00\xa1\x06\xa7\x06\x00\x00\x00\x00\x00\x00\x92\x23\xf0\x1e\x00\x00\x9d\x06\x00\x00\x08\x0d\x00\x00\x00\x00\x00\x00\xd3\x1f\x00\x00\x00\x00\x00\x00\x80\x1d\x76\x20\x00\x00\x00\x00\x61\x06\xa5\x06\x00\x00\x30\x02\x5f\x06\x46\x06\x9b\x06\x9a\x06\x53\x03\x00\x00\x95\x06\x00\x00\x63\x02\x00\x00\x99\x06\x88\x00\x00\x00\xa0\x06\x96\x06\x00\x00\x00\x00\xa6\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6e\x11\x00\x00\x00\x00\x00\x00\xa1\x16\x00\x00\x00\x00\x00\x00\x00\x00\xa1\x16\x00\x00\x00\x00\x0d\x00\x64\x06\x00\x00\x84\x06\x00\x00\x01\x06\x3a\x05\x36\x05\x35\x05\x2f\x05\x1f\x05\x10\x05\x09\x05\xf7\x04\xdd\x04\xcf\x04\xc1\x04\xbc\x04\xa0\x04\x97\x04\x7f\x04\x34\x04\x2d\x04\x00\x00\x67\x0b\x8c\x06\x66\x06\x1a\x00\x00\x00\x00\x00\x60\x12\x87\x06\x62\x00\x7e\x06\x7a\x06\x00\x00\x00\x00\x00\x00\xbd\x14\x00\x00\xa9\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xec\x00\x00\x00\x00\x00\x79\x06\x00\x00\x00\x00\x00\x00\xbe\x01\x00\x00\xa1\x16\x00\x00\x97\x22\x00\x00\x00\x00\x94\x06\xf9\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa1\x16\x3d\x06\xf8\xff\xe7\x11\xe7\x11\x00\x00\x00\x00\x00\x00\x00\x00\xea\x03\x44\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xf9\x1b\x00\x00\x00\x00\x6b\x06\x52\x01\x7d\x06\x00\x00\x65\x06\x00\x00\x00\x00\x76\x20\x00\x00\x00\x00\x00\x00\x35\x06\x00\x00\x00\x00\x00\x00\xfa\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x1d\x06\x7c\x06\xd0\x1e\x7b\x06\x71\x06\x00\x00\x00\x00\x76\x20\xf2\x23\x00\x00\xeb\x0b\x08\x0d\x00\x00\x00\x00\xac\x0c\x57\x00\x73\x06\x55\x21\x06\x06\x26\x06\x00\x00\x00\x00\xdc\x01\x1f\x06\xa1\x16\xa1\x16\x00\x00\x00\x00\x6d\x06\x59\x06\x55\x21\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x14\x18\x5e\x06\x62\x00\xa1\x16\x2e\x06\x5c\x06\x76\x1a\x80\x19\x5c\x06\x00\x00\xec\x05\x54\x06\x00\x00\x00\x00\xea\x05\x4c\x06\x79\x23\x00\x00\xff\x05\x1a\x06\xdb\x05\xc8\x05\xd4\x05\x00\x00\xf9\x20\x00\x00\xe0\x05\x00\x00\x00\x00\x00\x00\x00\x00\xb9\x05\x00\x00\xc5\x05\xb0\x05\xdf\x0a\x1b\x06\x00\x00\x00\x00\x1c\x06\x0d\x06\x00\x00\x00\x00\x14\x06\x00\x00\x18\x06\x00\x00\x00\x00\x00\x00\xa1\x16\xfd\x05\x00\x00\x00\x00\x00\x00\x76\x1e\x00\x00\x00\x00\x00\x00\xfa\x05\x00\x00\x00\x00\x00\x00\x76\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x16\x1e\x00\x00\xdf\x20\x69\x20\x00\x00\x00\x00\x1a\x17\x03\x06\x62\x00\x00\x00\xdf\x20\x00\x00\x00\x00\x00\x00\x00\x00\xf5\x05\x00\x00\x00\x00\x7b\x22\x80\x0c\x0f\x06\x00\x00\x00\x00\xdf\x20\x88\x05\x54\x0c\x00\x00\xdf\x20\xdf\x20\x70\x05\x00\x00\x00\x00\x76\x20\x00\x00\x00\x00\x00\x00\x76\x20\x35\x01\xf3\x05\x62\x00\xed\x05\xe6\x05\x00\x00\x00\x00\xd9\x05\xe2\x05\x99\x05\x00\x00\xa1\x16\x00\x00\xe6\x03\x00\x00\x6f\x05\xc1\x05\x00\x00\x44\x14\xa1\x16\xcb\x13\xc4\x05\xc0\x05\x00\x00\xa1\x16\xa1\x16\xa1\x16\x00\x00\x00\x00\x00\x00\x58\x05\x00\x00\x00\x00\x44\x01\x00\x00\x00\x00\x97\x05\x00\x00\xa1\x16\xaf\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa6\x05\x00\x00\xaa\x05\x98\x05\x00\x00\x5c\x20\x8f\x05\x00\x00\xdf\x20\x00\x00\x04\x01\xba\x02\x00\x00\x00\x00\xdf\x20\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x1e\x79\x1f\x6c\x1d\x8e\x05\x76\x1e\x03\x19\x8a\x05\xd0\x05\x00\x00\x76\x1e\x00\x00\x00\x00\x62\x21\x89\x05\x2f\x20\x00\x00\x44\x04\x00\x00\x00\x00\x00\x00\x8c\x05\x00\x00\x05\x06\x00\x00\x20\x1c\x00\x00\x00\x00\x7e\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7b\x05\x6c\x05\x2b\x05\x05\x05\x00\x00\x00\x00\x76\x1e\x4b\x05\x44\x05\x40\x05\x40\x05\x97\x17\x49\x05\x00\x00\x00\x00\xf1\x22\xf8\x0b\xdf\x20\xdf\x20\x76\x1e\xed\x01\x49\x05\x47\x05\x00\x00\x20\x1c\xa1\x16\xa1\x16\x00\x00\x00\x00\xff\x04\xc7\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xdf\x20\xc3\x04\x00\x00\x00\x00\x00\x00\x00\x00\x76\x1e\x00\x00\xde\x04\x00\x00\x00\x00\x00\x00\x00\x00\xbe\x22\xee\x04\xd8\x04\x12\x00\x05\x06\x00\x00\x00\x00\x00\x00\x00\x00\xf6\x04\x00\x00\xa1\x16\x00\x00\x00\x00\xdf\x20\x00\x00\x00\x00\x00\x00\x02\x05\x00\x00\x62\x22\x00\x00\xfb\x04\xef\x04\x00\x00\x00\x00\x00\x00\xf2\x04\xde\x1c\xc6\x04\xbe\x22\x00\x00\x00\x00\x00\x00\x77\x01\x00\x00\xbe\x04\x00\x00\x00\x00\x00\x00"#

happyGotoOffsets :: HappyAddr
happyGotoOffsets = HappyA# "\x02\x00\x86\x2a\x37\x2e\xc8\x01\xb3\x06\xb5\x26\x01\x00\xff\xff\xfe\xff\x51\x03\xef\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x52\x19\x00\x00\x00\x00\x00\x00\x9a\x35\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xd8\x2a\x4f\x2b\xf6\x2a\x00\x00\x01\x33\x84\x35\xeb\x32\x48\x35\xe5\x2d\x00\x00\x6e\x2a\x1c\x2a\xaf\x32\x9f\x06\xed\x03\xd7\x02\xf5\x03\x00\x00\xfc\x03\xf5\x31\x22\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xdf\x31\x9a\xff\x00\x00\x00\x00\x00\x00\xcb\x26\x0c\x00\x8f\xff\x53\x33\x62\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xc5\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1b\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x13\x04\xae\x04\xd1\x03\x8c\xff\x00\x00\x00\x00\x00\x00\x65\x04\x00\x00\x8f\x04\x00\x00\x07\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4f\x0c\x32\x35\x00\x00\xf3\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xdf\x07\x0b\x00\x00\x00\x75\x08\x75\x04\x58\x04\x36\x04\x32\x04\xae\x05\x41\x05\x42\x04\xb8\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xf5\x02\x00\x00\x00\x00\xfc\xff\xa0\x03\xd2\x05\x91\x03\xf6\x34\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x56\x02\xe0\x34\x7d\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x14\x04\x7a\x02\x00\x00\x52\x02\xc4\x07\x0f\x02\xa0\x01\x00\x00\xc5\x08\x5a\x08\xa6\x03\x00\x00\x00\x00\x37\x00\x2e\x00\xd5\x03\xb0\x07\xe2\xff\x9c\x07\xd9\x03\x07\x04\x04\x04\x00\x04\xf3\x03\x00\x00\x00\x00\x4e\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb8\x00\x9c\x04\x00\x00\x00\x00\x8b\x06\x00\x00\x00\x00\x00\x00\xf0\x00\x34\x06\x8e\x00\x00\x00\x5d\x00\x00\x00\x00\x00\x23\x01\x8c\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xbc\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x46\x08\x00\x00\x00\x00\x32\x08\x00\x00\x27\x04\x00\x00\x00\x00\xde\x01\x49\x07\x18\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x48\x26\xf6\x25\x00\x00\x00\x00\xa4\x34\x00\x00\x00\x00\xa2\x03\x00\x00\x21\x00\x07\x00\x00\x00\x00\x00\xb0\x01\x00\x00\x88\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xf1\x01\x7b\x03\x00\x00\x00\x00\x00\x00\x05\x04\x3b\x01\x00\x00\x00\x00\x17\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x8e\x34\x00\x00\x3d\x33\x00\x00\x7c\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5d\x32\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x04\x00\x00\x00\x00\x00\x00\xb4\x2c\x24\x02\x00\x00\x00\x00\x5b\x2c\x6b\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xeb\x32\x00\x00\x00\x00\x52\x34\x3c\x34\x04\x2a\x28\x03\x00\x00\x4e\x09\xf1\x05\x8d\x31\x77\x31\x25\x31\x0f\x31\x40\x03\x00\x00\xcb\x05\x0f\x00\x41\x01\x00\x00\x29\x03\x00\x00\x00\x00\xb8\x05\x00\x00\x00\x00\x00\x00\xbe\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x46\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x2c\x00\x00\x00\x00\xe1\x02\x00\x00\x00\x00\xcf\x2d\xbd\x30\xe4\x2b\x00\x00\x00\x00\xfc\x01\x00\x00\x00\x00\x00\x00\x00\x00\xc6\x2b\x00\x00\x6d\x2b\xa7\x30\xe7\x23\x00\x00\x25\x00\x13\x00\x00\x00\x55\x30\x3f\x30\x00\x00\xcf\x25\x00\x00\x00\x00\x00\x00\x00\x00\x37\x03\xac\x05\x00\x00\x00\x00\x0e\x05\xae\x33\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x2d\x71\xff\x20\x04\x00\x00\x25\x03\x00\x00\x00\x00\xed\x2f\x00\x00\xd7\x2f\x00\x00\x42\x03\x7d\x25\x03\x03\x85\x2f\x6f\x2f\x00\x00\x70\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x06\x07\x00\x00\x00\x00\xea\x06\x00\x00\x00\x00\xf2\x06\x00\x00\xcf\x00\x00\x00\x00\x00\x00\x00\x00\x00\xd4\x01\xe9\x00\x00\x00\x00\x00\x00\x00\x00\x00\xc8\x00\xde\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xbc\x02\xef\x00\x70\x03\x7c\x03\xd5\x05\xe3\x03\x00\x00\x00\x00\x00\x00\x00\x00\x52\x03\x27\x03\x00\x00\xb2\x29\x56\x25\x3e\x03\x85\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xe7\xff\x39\x03\xf2\x00\x30\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb7\x02\x16\x02\x00\x00\x00\x00\xb0\x02\xad\x02\x00\x00\x00\x00\x08\x03\x01\x09\xdd\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x03\x18\x03\x00\x00\xe3\x04\x00\x00\x6b\x01\x00\x00\x00\x00\x00\x34\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x56\x03\x00\x00\x00\x00\x00\x00\x95\x02\x00\x00\xe5\x03\xde\x03\x00\x00\x69\x00\x5d\x02\x98\x33\x2b\x03\xe8\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x47\x32\x2f\x00\x78\x03\x00\x00\x00\x00\x50\x03\x00\x00\x00\x00\x00\x00\x00\x00\x70\x09\x00\x00\xc9\x05\x00\x00\x00\x00\x00\x00\x00\x00\x92\x04\x00\x00\x54\x0b\xff\x03\x00\x00\x00\x00\x00\x00\x00\x00\x34\x00\x32\x00\x00\x00\x00\x00\xc6\x00\x00\x00\x00\x00\x00\x00\x47\x02\x00\x00\x00\x00\x00\x00\x5a\x00\x3f\x02\xd6\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xec\xff\x00\x00\x00\x00\x00\x00\x00\x00\xe5\xff\x00\x00\x00\x00\x00\x00\xc9\xff\xb5\x05\x00\x00\x00\x00\x00\x00\xbe\x08\xc1\x02\x00\x00\x00\x00\xab\x0b\x00\x00\x00\x00\x00\x00\x35\x1a\x0b\x1a\x00\x00\x00\x00\xa3\x02\x28\x02\x00\x00\xa9\xff\x83\x02\x96\x02\x00\x00\x00\x00\x25\x02\x00\x00\x00\x00\x00\x00\xb1\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x04\x25\x00\x00\x00\x00\x00\x00\x1d\x2f\x00\x00\x00\x00\xbf\x03\x00\x00\x9a\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x51\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0d\x2d\x00\x00\xef\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x48\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x87\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xbf\x02\x00\x00\x07\x2f\x00\x00\xfd\x02\x00\x00\x00\x00\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb5\x2e\x00\x00\x00\x00\x8b\x24\x39\x24\x00\x00\x00\x00\xbb\x03\x00\x00\x2e\x03\x63\x26\x00\x00\x00\x00\x00\x00\x12\x02\x00\x00\xea\x33\x00\x00\x00\x00\x00\x00\xdf\x01\x17\x02\x00\x00\x00\x00\x00\x00\x00\x00\xd8\x03\x00\x00\x00\x00\x00\x00\x01\x02\x00\x00\x00\x00\x00\x00\x81\x04\x41\x00\x3e\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x5e\x09\x5d\x00\x00\x00\x70\x09\xed\x08\x00\x00\x00\x00\x21\x09\x4b\x00\x00\x00\x69\x09\x00\x00\x61\x03\x00\x00\x00\x00\x83\x01\x00\x00\x30\x29\xde\x28\x00\x00\x00\x00\x00\x00\x00\x00\x95\x03\x5b\x00\x58\x00\xae\x01\x00\x00\x43\x03\xb6\x02\x00\x00\x98\x01\xc6\x28\xaf\x00\x00\x00\x99\x32\x8e\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1b\x00\x00\x00\x93\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xf7\xff\x00\x00\x00\x00\x00\x00\x3b\x03\x00\x00\x00\x00\x40\x02\x00\x00\x05\x02\x00\x00\x7f\x00\x00\x00\xb2\x01\x00\x00\x00\x00\x29\x02\x00\x00\x00\x00\x00\x00\x00\x00\x0a\x01\x00\x00\x00\x00\x00\x00\x74\x28\x00\x00\x00\x00\x00\x00\x00\x00\x56\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9a\x05\x00\x00\x13\x03\x00\x00\x00\x00\x00\x00\x54\x0b\x00\x00\x8c\x03\x73\x00\x00\x00\x00\x00\x3e\x02\x00\x00\xfd\x00\x00\x00\x34\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xce\x00\x9a\x08\xe0\x00\x00\x00\x00\x00\x58\x09\x00\x00\x70\x09\x00\x00\x86\x08\x42\x09\x00\x00\x00\x00\x00\x00\x64\x09\x00\x00\x00\x00\x00\x00\xce\x04\xb1\xff\x00\x00\xf1\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5c\x28\x00\x00\x95\x01\x00\x00\x31\x01\xdc\x00\x00\x00\xdd\x24\x0a\x28\x5f\x2d\x00\x00\x00\x00\x00\x00\xf2\x27\xa0\x27\x88\x27\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9f\x2e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x78\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x70\x09\x00\x00\x00\x00\x70\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x35\x09\x00\x00\x00\x00\x00\x00\x00\x00\xbb\x01\x00\x00\x00\x00\xca\x06\x74\x01\x00\x00\x00\x00\x84\x05\xde\x02\x00\x00\xd6\x00\x00\x00\x1b\x05\x00\x00\x00\x00\x53\x00\x00\x00\x11\x00\x00\x00\x49\x02\x00\x00\x00\x00\x00\x00\x72\x01\x00\x00\x99\x00\x00\x00\x17\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb3\x00\xaa\x00\x00\x00\x00\x00\xea\x04\x00\x00\x00\x00\x8c\x00\x63\x00\x66\x02\x00\x00\x00\x00\x00\x00\xbe\x00\xeb\x03\x2c\x09\xf5\x08\xd6\x04\xc2\xff\x00\x00\x00\x00\x00\x00\xa2\x02\x36\x27\x1e\x27\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xc2\x04\x00\x00\x43\x00\x00\x00\x00\x00\x00\x00\x00\x00\xef\xff\x00\x00\x7a\x00\xac\x02\x77\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x2e\x00\x00\x00\x00\x22\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xfd\xff\x00\x00\x00\x00\x64\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4e\x00\x00\x00\x49\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"#

happyDefActions :: HappyAddr
happyDefActions = HappyA# "\x7a\xfd\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\xfd\x7a\xfd\x00\x00\xe5\xff\xed\xff\x00\x00\x00\x00\xdb\xfd\xa9\xfe\xa7\xfe\xa1\xfe\xa0\xfe\x9e\xfe\x9c\xfe\x96\xfe\x8c\xfe\x87\xfe\x7f\xfe\x7c\xfe\x6a\xfe\x00\x00\x00\x00\x79\xfe\x7a\xfe\x7b\xfe\xcb\xfd\xc6\xfd\xa0\xfd\xad\xfd\xc4\xfd\xc1\xfd\x98\xfd\x78\xfe\xab\xfd\xac\xfd\x9a\xfd\x99\xfd\x96\xfd\x97\xfd\x84\xfd\x82\xfd\x83\xfd\x81\xfd\x80\xfd\x7f\xfd\x7e\xfd\x7d\xfd\x7c\xfd\x7b\xfd\x00\x00\x00\x00\x00\x00\x6e\xfe\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x69\xfe\x00\x00\x00\x00\x00\x00\x00\x00\x7a\xfd\x00\x00\x00\x00\x5f\xfe\x00\x00\x00\x00\x4a\xfe\xa7\xfd\x9f\xfd\x9e\xfd\x9d\xfd\xa6\xfd\xa5\xfd\xa4\xfd\xa3\xfd\xa2\xfd\xa1\xfd\xaa\xfd\x00\x00\x7a\xfd\x9b\xfd\x9c\xfd\xa8\xfd\x7a\xfd\x7a\xfd\x7a\xfd\x00\x00\x7a\xfd\xa9\xfd\x00\x00\x00\x00\x00\x00\x41\xff\x1a\xff\x48\xff\x3e\xff\x35\xff\x3c\xff\x28\xff\x00\x00\x1d\xff\x00\x00\x00\x00\x72\xfd\x21\xff\x3b\xff\x73\xfd\x2f\xff\x2e\xff\x00\x00\x00\x00\x00\x00\x00\x00\x71\xfd\x70\xfd\x6f\xfd\x13\xff\x85\xff\xad\xff\x00\x00\x00\x00\x90\xff\x86\xff\x7f\xff\x84\xff\x97\xff\xa0\xfe\x00\x00\x00\x00\x00\x00\x8f\xff\x00\x00\x00\x00\xab\xff\xaa\xff\xa9\xff\x00\x00\x7a\xfd\x8e\xff\x00\x00\x5e\xff\x5e\xff\x5e\xff\x5e\xff\x60\xff\x52\xff\x52\xff\x00\x00\xec\xfd\x00\x00\x00\x00\xa9\xfe\x00\x00\xe5\xff\xf4\xff\xf2\xff\x7a\xfd\x00\x00\xe5\xff\x00\x00\x00\x00\x00\x00\x4d\xff\x4e\xff\xca\xfd\xc3\xfd\x00\x00\x00\x00\x00\x00\x00\x00\x53\xff\x00\x00\x50\xff\x00\x00\x00\x00\x61\xff\x5e\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x48\xff\x00\x00\x00\x00\x76\xff\x00\x00\xd9\xff\xd9\xff\xbf\xfe\x00\x00\x8b\xff\x00\x00\x0d\xff\x00\x00\x00\x00\x00\x00\x00\x00\xb3\xfd\xb2\xfd\x00\x00\xb7\xfd\xae\xfd\x95\xfd\xbd\xfd\x92\xfd\x91\xfd\x8e\xfd\x93\xfd\x85\xfd\x94\xfd\x00\x00\x00\x00\x8b\xfd\xaf\xfd\x00\x00\x8d\xfd\x8c\xfd\x8a\xfd\xcf\xfe\x00\x00\x00\x00\xac\xff\x00\x00\x74\xfd\x30\xff\x00\x00\x00\x00\x00\x00\x35\xff\x2a\xff\x00\x00\x00\x00\x21\xff\x2c\xff\x00\x00\x25\xff\x16\xff\x00\x00\x00\x00\x23\xff\x5d\xfe\x16\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x27\xff\x00\x00\x00\x00\x1b\xff\x3f\xff\x00\x00\x1e\xff\x00\x00\x6d\xfd\x6c\xfd\x00\x00\x00\x00\x00\x00\x19\xff\x00\x00\x00\x00\x00\x00\xda\xfd\xe0\xfd\xe0\xfd\x00\x00\x84\xfe\x00\x00\x8d\xfe\xdd\xfd\x94\xfe\x98\xfe\xd9\xff\xd9\xff\x8e\xfe\x00\x00\x92\xfe\x00\x00\x14\xfe\x45\xfe\x41\xfe\x44\xfe\x43\xfe\x42\xfe\x3d\xfe\x3c\xfe\x3b\xfe\x3a\xfe\x39\xfe\x38\xfe\x37\xfe\x36\xfe\x35\xfe\x34\xfe\x33\xfe\x32\xfe\x31\xfe\x30\xfe\x3f\xfe\x3e\xfe\x2f\xfe\x2e\xfe\x2d\xfe\x2c\xfe\x2b\xfe\x2a\xfe\x29\xfe\x28\xfe\x27\xfe\x26\xfe\x25\xfe\x24\xfe\x23\xfe\x22\xfe\x21\xfe\x20\xfe\x1f\xfe\x1e\xfe\x1d\xfe\x1c\xfe\x1b\xfe\x1a\xfe\x19\xfe\x40\xfe\x18\xfe\x17\xfe\x16\xfe\x60\xfe\x61\xfe\x00\x00\x00\x00\x00\x00\x63\xfe\x62\xfe\x00\x00\x00\x00\x00\x00\x00\x00\xa0\xfe\x00\x00\x00\x00\x52\xfe\x00\x00\xeb\xfd\x8f\xfe\x80\xfe\x00\x00\x85\xfe\x7a\xfd\x5c\xfe\x10\xfe\x00\x00\x0f\xfe\xb1\xfd\xb0\xfd\x00\x00\xbb\xfd\x90\xfd\x8f\xfd\x89\xfd\xcf\xfd\x00\x00\x87\xfd\x88\xfd\x86\xfd\x00\x00\x00\x00\xcd\xfd\x5c\xfe\x00\x00\x00\x00\x00\x00\xb7\xfd\x00\x00\x8f\xfd\x89\xfd\xd0\xfd\x87\xfd\x8d\xfd\x88\xfd\x86\xfd\x00\x00\x00\x00\x00\x00\x00\x00\x88\xfe\xa6\xfe\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xd9\xff\x00\x00\x00\x00\xf7\xff\xe5\xff\xf6\xff\x00\x00\xd9\xff\xe8\xff\xda\xff\x00\x00\xed\xff\xef\xff\x78\xfd\x79\xfd\xa2\xfe\xa3\xfe\xa4\xfe\xa5\xfe\xa8\xfe\x9f\xfe\x9d\xfe\x00\x00\xd8\xfd\xd6\xfd\x00\x00\x7e\xfe\xd5\xfd\xdc\xfd\x81\xfe\x82\xfe\xc5\xfd\xc0\xfd\x6d\xfe\x00\x00\x76\xfe\x77\xfe\x00\x00\xce\xfd\x5e\xfe\x00\x00\x00\x00\x00\x00\x73\xfe\x72\xfe\x00\x00\xcc\xfd\x00\x00\x00\x00\x5b\xfe\x00\x00\x6f\xfe\x00\x00\x0e\xfe\x00\x00\x90\xfe\xd9\xff\xd9\xff\x86\xfe\x00\x00\x00\x00\x6c\xfe\x00\x00\x68\xfe\x67\xfe\x66\xfe\x65\xfe\x00\x00\xd9\xff\xa4\xff\x00\x00\x00\x00\x11\xfe\x4d\xfe\x00\x00\x47\xfe\x4b\xfe\x93\xfe\x49\xfe\x00\x00\x7a\xfd\x00\x00\x00\x00\x00\x00\x00\x00\x95\xfe\x00\x00\x83\xfe\x00\x00\x00\x00\xde\xfd\xe0\xfd\x00\x00\x00\x00\x00\x00\x00\x00\x44\xff\x45\xff\x00\x00\x46\xff\x47\xff\x42\xff\x26\xff\x1f\xff\x20\xff\x24\xff\x00\x00\x3a\xff\x37\xff\x00\x00\x22\xff\x39\xff\x00\x00\x00\x00\x00\x00\x33\xff\x38\xff\x2d\xff\x00\x00\x00\x00\x00\x00\x00\x00\x31\xff\x14\xff\x12\xff\x00\x00\x00\x00\xae\xff\xb5\xfd\xb4\xfd\xa7\xff\xb9\xfd\xbf\xfd\x00\x00\xcf\xfe\xe5\xfe\x06\xff\x00\x00\xfb\xfe\x81\xff\x00\x00\x00\x00\x72\xff\xb1\xfe\xad\xfe\xab\xfe\x00\x00\x00\x00\xb1\xfe\x66\xff\x71\xff\x70\xff\x6f\xff\x6e\xff\x6d\xff\x6c\xff\x6b\xff\x00\x00\xcb\xfe\x0f\xff\xcf\xfe\x00\x00\x8c\xff\x43\xff\x8d\xff\x99\xff\x9a\xff\x7a\xfd\x89\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x49\xff\xcf\xfe\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5b\xff\x62\xff\x94\xff\x54\xff\x93\xff\x00\x00\x51\xff\x92\xff\x00\x00\x4b\xff\x00\x00\x00\x00\x91\xff\x4a\xff\xe2\xff\x77\xfd\x76\xfd\x00\x00\x14\xfe\xe7\xff\xd9\xff\xd9\xff\x00\x00\xdc\xff\x00\x00\x11\xfe\x00\x00\xd7\xff\x00\x00\x00\x00\xc9\xfd\xc2\xfd\x4c\xff\x4f\xff\x55\xff\x63\xff\x00\x00\x00\x00\x00\x00\x00\x00\x5d\xff\x00\x00\x7a\xff\x7d\xff\x7e\xff\xa3\xff\x44\xff\xa2\xff\x00\x00\x74\xff\x82\xff\x75\xff\x83\xff\xd9\xff\x87\xff\xaf\xfe\xd9\xff\xd2\xfd\x7b\xfe\xdb\xff\x00\x00\xd9\xff\xd9\xff\x98\xff\x9e\xff\x00\x00\x0c\xff\x0a\xff\x9b\xff\x7a\xfd\x95\xff\x00\x00\xc8\xfd\x00\x00\x00\x00\x00\x00\x6a\xff\x69\xff\x68\xff\x67\xff\xb3\xfe\x7a\xfd\xf9\xfd\x00\x00\xfb\xfd\x00\x00\x7a\xfd\xae\xfe\xac\xfe\xb4\xfe\x00\x00\x00\x00\xbc\xfd\x01\xff\xff\xfe\x00\x00\x13\xff\xce\xfe\xdf\xfe\xde\xfe\xd8\xfe\xd7\xfe\xd6\xfe\x00\x00\x00\x00\xda\xfe\xdb\xfe\xe5\xfe\x7a\xfd\xa0\xff\x00\x00\xe5\xfe\x06\xff\x00\x00\x00\x00\x00\x00\x1c\xff\x00\x00\x32\xff\x00\x00\x34\xff\x2b\xff\x20\xff\x15\xff\x00\x00\x17\xff\x6e\xfd\x18\xff\x00\x00\x8a\xfe\x8b\xfe\xe3\xfd\xe1\xfd\xe2\xfd\xe0\xfd\xe4\xfd\x97\xfe\x9a\xfe\x00\x00\xe9\xfd\xea\xfd\xd9\xff\xe6\xfd\x00\x00\x91\xfe\x52\xfe\x00\x00\x00\x00\x12\xfe\x00\x00\x15\xfe\x00\x00\xab\xfd\x96\xfd\xa7\xfd\x9f\xfd\x9e\xfd\x9d\xfd\xa6\xfd\xa5\xfd\xa4\xfd\xa3\xfd\xa2\xfd\xa1\xfd\xaa\xfd\x9b\xfd\x9c\xfd\xa8\xfd\xa9\xfd\x46\xfe\xda\xff\xa6\xff\x00\x00\x00\x00\x53\xfe\x9b\xfe\x00\x00\x00\x00\x00\x00\x0a\xfe\x06\xfe\x04\xfe\x03\xfe\x02\xfe\x00\x00\x0c\xfe\x5c\xfe\x08\xfe\x09\xfe\xb6\xfd\xba\xfd\x70\xfe\x71\xfe\x00\x00\x54\xfe\x5a\xfe\x50\xfe\x51\xfe\x74\xfe\x75\xfe\x00\x00\x57\xfe\x00\x00\x7d\xfe\x00\x00\xee\xff\xeb\xff\x00\x00\x00\x00\xea\xff\xec\xff\xe9\xff\xd9\xfd\xd7\xfd\x59\xfe\x58\xfe\x56\xfe\x55\xfe\x0d\xfe\x01\xfe\x00\x00\x00\x00\x00\x00\xf6\xfd\xf7\xfd\xd9\xff\xf3\xfd\x00\x00\x00\x00\x6b\xfe\x64\xfe\xa5\xff\x4a\xfe\x4e\xfe\x00\x00\x4c\xfe\x48\xfe\x00\x00\xe8\xfd\x94\xfe\xdf\xfd\x00\x00\x36\xff\x29\xff\x00\x00\xa8\xff\xb8\xfd\xbe\xfd\xe5\xfe\x9d\xff\xe4\xfe\xe0\xfe\x00\x00\xd9\xff\xd9\xff\x9f\xff\xd1\xfe\x00\x00\x00\x00\xde\xfe\x00\x00\x00\x00\xd5\xfe\xdd\xfe\x00\x00\x00\x00\xee\xfe\xf6\xfe\x00\x00\xfd\xfe\xfa\xfe\xf5\xfe\x00\x00\x00\x00\x00\x00\x00\x00\xfb\xfe\x80\xff\x73\xff\x00\x00\xf8\xfd\x00\x00\x00\x00\xb2\xfe\x96\xff\x00\x00\x00\x00\x00\x00\xd9\xff\xd9\xff\x0f\xff\x10\xff\x0f\xff\xbd\xfe\x00\x00\x00\x00\x00\x00\xda\xff\xd4\xfd\x00\x00\xda\xff\x8a\xff\xa1\xff\x00\x00\x79\xff\x77\xff\x5c\xff\x00\x00\x00\x00\x59\xff\x57\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xd8\xff\xd3\xff\xf3\xff\x00\x00\xe0\xff\xd9\xff\xc9\xff\xde\xff\xc6\xff\xe1\xff\xc4\xff\x00\x00\xda\xff\xdd\xff\x4a\xfe\xf0\xff\x00\x00\xd3\xff\xd1\xff\xd0\xff\xcf\xff\xd4\xff\x00\x00\xe6\xff\xe4\xff\xe3\xff\x00\x00\x00\x00\x58\xff\x5a\xff\x7b\xff\x00\x00\x7c\xff\x88\xff\xd3\xfd\x00\x00\xd1\xfd\xc0\xfe\xc1\xfe\x00\x00\xb8\xfe\xd9\xff\xbb\xfe\xb9\xfe\xba\xfe\xaf\xfe\x8f\xff\x00\x00\x00\x00\x09\xff\x0b\xff\xc9\xfe\x00\x00\x00\x00\x64\xff\x00\x00\xc7\xfd\xaa\xfe\xfa\xfd\x00\xff\x00\x00\xed\xfe\x3d\xff\x00\x00\x00\x00\x00\x00\xf1\xfe\xf2\xfe\x00\x00\x00\x00\xf6\xfe\xfe\xfe\x00\x00\x00\x00\x00\x00\xfc\xfe\xdc\xfe\x00\x00\xd4\xfe\xd9\xfe\xd3\xfe\x00\x00\x00\x00\x00\x00\x00\x00\xe1\xfe\x00\x00\xe3\xfe\x9c\xff\x00\x00\x00\x00\x00\x00\xe7\xfd\x00\x00\x13\xfe\x92\xfe\xfc\xfd\xb1\xfe\xf0\xfd\xee\xfd\x00\x00\x00\x00\xda\xff\xf5\xfd\x07\xfe\x05\xfe\x00\x00\x00\x00\x00\x00\x0b\xfe\x00\xfe\xfe\xfd\xff\xfd\xf4\xfd\xf1\xfd\x00\x00\xef\xfd\xf2\xfd\x00\x00\xe5\xfd\x00\x00\x00\x00\x11\xff\xe2\xfe\x07\xff\x08\xff\xd9\xff\x03\xff\x00\x00\xd0\xfe\xd2\xfe\x00\x00\xf4\xfe\x00\x00\x00\x00\xf0\xfe\x40\xff\xf9\xfe\x00\x00\x00\x00\xea\xfe\xf8\xfe\x00\x00\x65\xff\xcc\xfe\xcd\xfe\xc6\xfe\xd9\xff\xc7\xfe\xc5\xfe\x00\x00\x00\x00\x0e\xff\x00\x00\x00\x00\xda\xff\xbe\xfe\xcf\xfe\x78\xff\x00\x00\x5f\xff\xcb\xff\x00\x00\x00\x00\xd4\xff\xd5\xff\x00\x00\xca\xff\xdf\xff\xc7\xff\xc2\xff\xc5\xff\x00\x00\xc3\xff\x00\x00\xd2\xff\xd6\xff\x00\x00\xb1\xff\xb0\xff\xaf\xff\xcd\xff\x00\x00\x00\x00\xe5\xfe\x06\xff\xbc\xfe\xb0\xfe\x00\x00\x00\x00\x41\xff\xcf\xfe\xcf\xfe\xda\xff\xca\xfe\xec\xfe\xf7\xfe\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xda\xff\x05\xff\x00\x00\x99\xfe\x00\x00\x00\x00\x00\x00\xfd\xfd\xed\xfd\x00\x00\x00\x00\x04\xff\x02\xff\xf3\xfe\xef\xfe\xe8\xfe\xe9\xfe\x00\x00\x00\x00\xeb\xfe\xc8\xfe\xc2\xfe\xc4\xfe\x00\x00\xb7\xfe\xe5\xfe\xb6\xfe\x56\xff\xce\xff\xcc\xff\x00\x00\x00\x00\xc0\xff\xbe\xff\x00\x00\xf1\xff\xb2\xff\xb5\xfe\xc3\xfe\x00\x00\xe7\xfe\x00\x00\x4f\xfe\x89\xfe\x00\x00\xc1\xff\xc8\xff\xbf\xff\x00\x00\xbb\xff\xd3\xff\xe6\xfe\x00\x00\xd3\xff\xb8\xff\xb7\xff\x75\xfd\xb6\xff\x00\x00\x00\x00\xd4\xff\xbc\xff\xb9\xff\xbd\xff\x00\x00\xb4\xff\x00\x00\xb5\xff\xb3\xff"#

happyCheck :: HappyAddr
happyCheck = HappyA# "\xff\xff\x00\x00\x04\x00\x04\x00\x02\x00\x04\x00\x04\x00\x0b\x00\x11\x00\x12\x00\x13\x00\x29\x00\x36\x00\x07\x00\x11\x00\x20\x00\x5f\x00\x60\x00\x2d\x00\x1a\x00\x0d\x00\x0e\x00\x07\x00\x32\x00\x6f\x00\x2d\x00\x1d\x00\x1e\x00\xba\x00\x36\x00\x32\x00\xae\x00\x0d\x00\x0e\x00\x60\x00\x1a\x00\x13\x00\x43\x00\x44\x00\x45\x00\x16\x00\x47\x00\xae\x00\x49\x00\x4a\x00\x20\x00\x0d\x00\x0e\x00\x4e\x00\x4f\x00\x0d\x00\x0e\x00\x52\x00\x53\x00\x54\x00\x55\x00\x2d\x00\x2d\x00\x20\x00\x0d\x00\x0e\x00\x32\x00\x32\x00\x0d\x00\x0e\x00\x0d\x00\x0e\x00\xe1\x00\x0d\x00\x0e\x00\x1a\x00\x1e\x00\x44\x00\xba\x00\x46\x00\x0d\x00\x0e\x00\x1a\x00\x0d\x00\x0e\x00\x1a\x00\x37\x00\xe1\x00\x5b\x00\xba\x00\xc9\x00\x3b\x00\x3c\x00\x2a\x00\x4c\x00\xe3\x00\x0c\x00\x08\x00\xe1\x00\x0a\x00\x20\x00\x09\x00\x2a\x00\x00\x00\xd7\x00\xd8\x00\x0d\x00\x0e\x00\x1e\x00\x0d\x00\x0e\x00\x3b\x00\x3c\x00\x4e\x00\x1f\x00\x20\x00\x77\x00\xe1\x00\xe5\x00\x1f\x00\x20\x00\x70\x00\x11\x00\x37\x00\x23\x00\x66\x00\x22\x00\xc9\x00\xe1\x00\x2d\x00\x14\x00\x15\x00\x1d\x00\xd7\x00\xd8\x00\x89\x00\x2f\x00\x30\x00\x31\x00\x68\x00\x70\x00\xd7\x00\xd8\x00\x21\x00\xc9\x00\x23\x00\xc4\x00\x25\x00\x26\x00\x27\x00\x28\x00\x89\x00\x19\x00\x15\x00\x2c\x00\x6d\x00\x2e\x00\x2f\x00\xd7\x00\xd8\x00\xc5\x00\x5e\x00\xd4\x00\xd5\x00\x5e\x00\x21\x00\x8a\x00\x23\x00\x1b\x00\x25\x00\x26\x00\x27\x00\x28\x00\x20\x00\xc7\x00\xd4\x00\x2c\x00\xc5\x00\x2e\x00\x2f\x00\x7d\x00\x6d\x00\x7d\x00\x24\x00\xc4\x00\x57\x00\x89\x00\xd4\x00\xc8\x00\xd6\x00\xd7\x00\xd8\x00\xd4\x00\x89\x00\xc6\x00\xb7\x00\x89\x00\xc9\x00\xc4\x00\xaf\x00\xd4\x00\xd5\x00\xc8\x00\xe1\x00\xd8\x00\xe6\x00\xe7\x00\xd3\x00\xd4\x00\xd5\x00\xe1\x00\xd7\x00\xd8\x00\x78\x00\xd4\x00\xd5\x00\x78\x00\xaf\x00\xd8\x00\x1b\x00\xc6\x00\xb7\x00\x76\x00\xc9\x00\x20\x00\xe5\x00\xe1\x00\xd8\x00\xe1\x00\xe1\x00\xe4\x00\xe1\x00\xe1\x00\xd3\x00\xd4\x00\xd5\x00\xd8\x00\xd7\x00\xd8\x00\x81\x00\x33\x00\xe1\x00\xe1\x00\xc0\x00\xd4\x00\xd5\x00\x33\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xe5\x00\xc0\x00\x8f\x00\x90\x00\x91\x00\x4a\x00\x4b\x00\x94\x00\x95\x00\x96\x00\x81\x00\x33\x00\x76\x00\xd4\x00\xd5\x00\x19\x00\x9d\x00\x5d\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\x20\x00\xc4\x00\x8f\x00\x90\x00\x91\x00\xc8\x00\xc4\x00\x94\x00\x95\x00\x96\x00\xc8\x00\xc4\x00\xcd\x00\x4a\x00\x4b\x00\xc8\x00\x9d\x00\xd4\x00\xd5\x00\xc5\x00\x6d\x00\xd8\x00\xd4\x00\xd5\x00\x20\x00\xda\x00\xd8\x00\xd4\x00\xd5\x00\x6b\x00\x26\x00\xd8\x00\xc3\x00\xe4\x00\xd4\x00\xc6\x00\xc7\x00\xd4\x00\xc9\x00\x4a\x00\x4b\x00\x4c\x00\x4d\x00\x61\x00\x6a\x00\x6b\x00\x05\x00\x06\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xe6\x00\xe7\x00\xc6\x00\xc7\x00\xd4\x00\xc9\x00\xe0\x00\x58\x00\x16\x00\x76\x00\x5b\x00\x5c\x00\x1a\x00\x61\x00\x61\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xca\x00\xe6\x00\xe7\x00\xcd\x00\x21\x00\xcf\x00\x23\x00\xe0\x00\x25\x00\x26\x00\x27\x00\x28\x00\x20\x00\x76\x00\x76\x00\x2c\x00\xda\x00\x2e\x00\x2f\x00\xdd\x00\x1a\x00\x4a\x00\x4b\x00\x2b\x00\x4d\x00\xc2\x00\x21\x00\x09\x00\x23\x00\x0b\x00\xc7\x00\x2c\x00\x27\x00\x28\x00\xc9\x00\x29\x00\xe3\x00\x2c\x00\xc6\x00\x2e\x00\x2f\x00\x40\x00\x41\x00\x38\x00\xc6\x00\xd6\x00\x30\x00\x31\x00\xd7\x00\xd8\x00\x18\x00\xd3\x00\xd4\x00\xd5\x00\x17\x00\x25\x00\xb4\x00\xd3\x00\xd4\x00\xd5\x00\xc6\x00\x17\x00\xe5\x00\x20\x00\xc9\x00\x43\x00\xd4\x00\x45\x00\xd4\x00\x47\x00\x20\x00\x49\x00\x4a\x00\xd3\x00\xd4\x00\xd5\x00\x4e\x00\x4f\x00\xd7\x00\xd8\x00\x52\x00\x53\x00\x3f\x00\x40\x00\xe6\x00\xe7\x00\xe6\x00\xe7\x00\x17\x00\x20\x00\xc9\x00\x82\x00\xe5\x00\x24\x00\x42\x00\x43\x00\xd8\x00\x45\x00\x29\x00\x47\x00\x81\x00\x49\x00\x4a\x00\x26\x00\xd7\x00\xd8\x00\x4e\x00\x4f\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\x0d\x00\x0e\x00\x8f\x00\x90\x00\x91\x00\x24\x00\xe5\x00\x94\x00\x95\x00\x96\x00\x81\x00\xe2\x00\x2b\x00\x17\x00\x4a\x00\x4b\x00\x9d\x00\x4d\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\x20\x00\xe2\x00\x8f\x00\x90\x00\x91\x00\x20\x00\x08\x00\x94\x00\x95\x00\x96\x00\x17\x00\x21\x00\x27\x00\x23\x00\xc9\x00\xe3\x00\x9d\x00\x27\x00\x28\x00\x20\x00\x05\x00\x06\x00\x2c\x00\xd2\x00\x2e\x00\x2f\x00\x40\x00\x41\x00\xd7\x00\xd8\x00\xd9\x00\xda\x00\xc3\x00\x23\x00\x19\x00\xc6\x00\xc7\x00\x16\x00\xc9\x00\xc4\x00\x58\x00\x20\x00\xe5\x00\xc8\x00\x5c\x00\x2f\x00\x30\x00\x31\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xd4\x00\xd5\x00\xc6\x00\xc7\x00\xd8\x00\xc9\x00\xe0\x00\x17\x00\xc7\x00\x97\x00\x05\x00\x06\x00\x9a\x00\x17\x00\x8e\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xd4\x00\x20\x00\xd6\x00\xd7\x00\xd8\x00\xc4\x00\x16\x00\xe0\x00\x9d\x00\xc8\x00\x9f\x00\x05\x00\x06\x00\x21\x00\x1e\x00\x23\x00\x11\x00\xc7\x00\xe6\x00\xe7\x00\x8e\x00\xd4\x00\xd5\x00\x2b\x00\x2c\x00\xd8\x00\x2e\x00\x2f\x00\x16\x00\x19\x00\xd4\x00\x81\x00\xd6\x00\xd7\x00\xd8\x00\x9d\x00\x20\x00\x9f\x00\x9e\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xc9\x00\x16\x00\x8f\x00\x90\x00\x91\x00\xe6\x00\xe7\x00\x94\x00\x95\x00\x96\x00\x21\x00\xdd\x00\x23\x00\x17\x00\xd7\x00\xd8\x00\x9d\x00\xc6\x00\x17\x00\x05\x00\x06\x00\x2c\x00\x20\x00\x2e\x00\x2f\x00\x6d\x00\x0c\x00\x20\x00\xe5\x00\x0f\x00\xd3\x00\xd4\x00\xd5\x00\x42\x00\x43\x00\x19\x00\x16\x00\xe2\x00\x47\x00\xc4\x00\x49\x00\x4a\x00\x20\x00\xc8\x00\x1e\x00\x4e\x00\x4f\x00\x20\x00\x0d\x00\x0e\x00\x21\x00\x97\x00\x23\x00\x26\x00\xc3\x00\xd4\x00\xd5\x00\xc6\x00\xc7\x00\xd8\x00\xc9\x00\x2c\x00\x97\x00\x2e\x00\x2f\x00\x9a\x00\x81\x00\xb9\x00\x08\x00\x09\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x8b\x00\x19\x00\x40\x00\x8d\x00\x8f\x00\xd2\x00\x91\x00\xe0\x00\x20\x00\x94\x00\x95\x00\x96\x00\xd9\x00\xda\x00\x21\x00\x9e\x00\x23\x00\xd4\x00\x9d\x00\x23\x00\xd7\x00\xd8\x00\x79\x00\x7a\x00\x7b\x00\x2c\x00\x97\x00\x2e\x00\x2f\x00\x9a\x00\x81\x00\x2f\x00\x30\x00\x31\x00\xd2\x00\xe6\x00\xe7\x00\x1a\x00\x1b\x00\x1c\x00\x8b\x00\xd9\x00\xda\x00\xdb\x00\x8f\x00\xdd\x00\x91\x00\xdf\x00\xe2\x00\x94\x00\x95\x00\x96\x00\xc6\x00\x1c\x00\xc1\x00\xc2\x00\xc3\x00\x20\x00\x9d\x00\xc6\x00\xc7\x00\x28\x00\xc9\x00\x7a\x00\x7b\x00\xd3\x00\xd4\x00\xd5\x00\x2f\x00\x9d\x00\x81\x00\x9f\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xca\x00\x6d\x00\x8b\x00\xcd\x00\x5d\x00\xcf\x00\x8f\x00\xe0\x00\x91\x00\x03\x00\x19\x00\x94\x00\x95\x00\x96\x00\x08\x00\x09\x00\xda\x00\x20\x00\xc3\x00\xdd\x00\x9d\x00\xc6\x00\xc7\x00\x28\x00\xc9\x00\xd4\x00\xe1\x00\x1c\x00\xd7\x00\xd8\x00\x2f\x00\x20\x00\x81\x00\x6d\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x56\x00\xc6\x00\x8b\x00\xe6\x00\xe7\x00\xdd\x00\x8f\x00\xe0\x00\x91\x00\x0d\x00\x0e\x00\x94\x00\x95\x00\x96\x00\xd3\x00\xd4\x00\xd5\x00\xe1\x00\xc3\x00\x01\x00\x9d\x00\xc6\x00\xc7\x00\x03\x00\xc9\x00\xda\x00\x08\x00\x09\x00\xdd\x00\x7e\x00\x7f\x00\x80\x00\x81\x00\xa3\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xe2\x00\xc6\x00\x8b\x00\xa0\x00\xa1\x00\xa2\x00\x8f\x00\xe0\x00\x91\x00\x0d\x00\x0e\x00\x94\x00\x95\x00\x96\x00\xd3\x00\xd4\x00\xd5\x00\x38\x00\xc3\x00\x01\x00\x9d\x00\xc6\x00\xc7\x00\x97\x00\xc9\x00\x99\x00\x08\x00\x09\x00\x08\x00\x09\x00\x7f\x00\x80\x00\x81\x00\x0a\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xd4\x00\xd5\x00\x8b\x00\x48\x00\x49\x00\x4a\x00\x8f\x00\xe0\x00\x91\x00\x4e\x00\x4f\x00\x94\x00\x95\x00\x96\x00\x22\x00\x23\x00\x3a\x00\x97\x00\xc3\x00\x99\x00\x9d\x00\xc6\x00\xc7\x00\x76\x00\xc9\x00\x30\x00\x31\x00\x2f\x00\x30\x00\x31\x00\x4f\x00\x50\x00\x51\x00\x52\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xe2\x00\xd4\x00\xd5\x00\xe2\x00\x43\x00\xd8\x00\x45\x00\xe0\x00\x47\x00\xe1\x00\x49\x00\x4a\x00\x58\x00\x59\x00\xc6\x00\x4e\x00\x4f\x00\xc9\x00\xc3\x00\x52\x00\x53\x00\xc6\x00\xc7\x00\x76\x00\xc9\x00\x30\x00\x31\x00\xd3\x00\xd4\x00\xd5\x00\x86\x00\xd7\x00\xd8\x00\x77\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xa0\x00\xa1\x00\xa2\x00\x35\x00\x43\x00\xbf\x00\x45\x00\xe0\x00\x47\x00\x82\x00\x49\x00\x4a\x00\xc6\x00\x63\x00\x64\x00\x4e\x00\x4f\x00\x0d\x00\x0e\x00\x52\x00\x53\x00\x0d\x00\x0e\x00\x42\x00\x43\x00\xd3\x00\xd4\x00\xd5\x00\x47\x00\x82\x00\x49\x00\x4a\x00\x42\x00\x43\x00\x5d\x00\x4e\x00\x4f\x00\x47\x00\x6d\x00\x49\x00\x4a\x00\xb2\x00\xb3\x00\xb4\x00\x4e\x00\x4f\x00\xe2\x00\xbe\x00\xbf\x00\x42\x00\x43\x00\x0c\x00\x0d\x00\x0e\x00\x47\x00\xc6\x00\x49\x00\x4a\x00\x0c\x00\x0d\x00\x0e\x00\x4e\x00\x4f\x00\xd4\x00\x0f\x00\x10\x00\xd7\x00\xd8\x00\xd3\x00\xd4\x00\xd5\x00\xbc\x00\x43\x00\x1a\x00\x45\x00\x97\x00\x47\x00\x99\x00\x49\x00\x4a\x00\xe2\x00\xe6\x00\xe7\x00\x4e\x00\x4f\x00\x0d\x00\x0e\x00\x52\x00\x53\x00\x54\x00\x55\x00\x97\x00\x29\x00\x43\x00\x2b\x00\x45\x00\xc7\x00\x47\x00\xe2\x00\x49\x00\x4a\x00\x4b\x00\x4c\x00\x4d\x00\x4e\x00\x4f\x00\xdb\x00\xe2\x00\xdd\x00\xd4\x00\xdf\x00\xd6\x00\xd7\x00\xd8\x00\x43\x00\x44\x00\xa3\x00\x46\x00\x47\x00\x43\x00\x8d\x00\x45\x00\x4b\x00\x47\x00\x5a\x00\x49\x00\x4a\x00\xe6\x00\xe7\x00\x34\x00\x4e\x00\x4f\x00\x34\x00\x51\x00\x52\x00\x53\x00\xc7\x00\xa0\x00\xa1\x00\xa2\x00\x4e\x00\x4f\x00\x62\x00\x63\x00\x64\x00\x70\x00\x71\x00\x72\x00\x73\x00\xd4\x00\x39\x00\xd6\x00\xd7\x00\xd8\x00\x7c\x00\x25\x00\x97\x00\x27\x00\xd8\x00\x43\x00\x6c\x00\x45\x00\x25\x00\x47\x00\x27\x00\x49\x00\x4a\x00\xe6\x00\xe7\x00\xd4\x00\x4e\x00\x4f\x00\xd7\x00\xd8\x00\x52\x00\x53\x00\x54\x00\x55\x00\xd4\x00\x43\x00\x39\x00\xd7\x00\xd8\x00\x47\x00\x39\x00\x49\x00\x4a\x00\xe6\x00\xe7\x00\xe3\x00\x4e\x00\x4f\x00\x84\x00\x85\x00\x86\x00\xd4\x00\xe6\x00\xe7\x00\xd7\x00\xd8\x00\x3d\x00\x3e\x00\x3f\x00\x40\x00\xc7\x00\x84\x00\x85\x00\x86\x00\x43\x00\x44\x00\x45\x00\x46\x00\x47\x00\xe6\x00\xe7\x00\xd2\x00\x4b\x00\xd4\x00\x39\x00\xd6\x00\xd7\x00\xd8\x00\xd9\x00\xda\x00\xdb\x00\xc7\x00\xdd\x00\xc9\x00\xdf\x00\xa0\x00\xa1\x00\xa2\x00\x0d\x00\x0e\x00\xc9\x00\xe6\x00\xe7\x00\x25\x00\xd4\x00\x27\x00\xd6\x00\xd7\x00\xd8\x00\x97\x00\xd3\x00\xd4\x00\xd5\x00\x39\x00\xd7\x00\xd8\x00\x22\x00\xc7\x00\x48\x00\x49\x00\x4a\x00\xe5\x00\xe6\x00\xe7\x00\x4e\x00\x4f\x00\x56\x00\x25\x00\xe5\x00\x27\x00\xd4\x00\x9e\x00\xd6\x00\xd7\x00\xd8\x00\x43\x00\x25\x00\x45\x00\x27\x00\x47\x00\xd4\x00\x49\x00\x4a\x00\xd7\x00\xd8\x00\xe1\x00\x4e\x00\x4f\x00\xe6\x00\xe7\x00\x52\x00\x53\x00\x17\x00\x55\x00\xd2\x00\xb8\x00\xb9\x00\xc7\x00\xe6\x00\xe7\x00\x17\x00\xd9\x00\xda\x00\xdb\x00\x25\x00\xdd\x00\x27\x00\xdf\x00\xd2\x00\x25\x00\xd4\x00\x27\x00\xd6\x00\xd7\x00\xd8\x00\xd9\x00\xda\x00\xdb\x00\x6e\x00\xdd\x00\x43\x00\xdf\x00\x45\x00\x25\x00\x47\x00\x27\x00\x49\x00\x4a\x00\xe6\x00\xe7\x00\xd4\x00\x4e\x00\x4f\x00\xd7\x00\xd8\x00\x52\x00\x53\x00\x25\x00\x55\x00\x27\x00\x43\x00\xc4\x00\x45\x00\x16\x00\x47\x00\xc8\x00\x49\x00\x4a\x00\xe6\x00\xe7\x00\x20\x00\x4e\x00\x4f\x00\x17\x00\x51\x00\x52\x00\x53\x00\xd4\x00\xd5\x00\x16\x00\x43\x00\xd8\x00\x45\x00\x25\x00\x47\x00\x27\x00\x49\x00\x4a\x00\x3e\x00\x3f\x00\x40\x00\x4e\x00\x4f\x00\x30\x00\x51\x00\x52\x00\x53\x00\x86\x00\x87\x00\x88\x00\x43\x00\x25\x00\x45\x00\x27\x00\x47\x00\x5a\x00\x49\x00\x4a\x00\x25\x00\x48\x00\x27\x00\x4e\x00\x4f\x00\x01\x00\x51\x00\x52\x00\x53\x00\x60\x00\xd4\x00\x72\x00\x73\x00\xd7\x00\xd8\x00\x25\x00\x97\x00\x27\x00\x48\x00\xc7\x00\xd3\x00\xd4\x00\xd5\x00\x89\x00\xd7\x00\xd8\x00\x16\x00\x89\x00\xe6\x00\xe7\x00\x0c\x00\x25\x00\xd4\x00\x27\x00\xd6\x00\xd7\x00\xd8\x00\x25\x00\x25\x00\x27\x00\x27\x00\x43\x00\x25\x00\x45\x00\x27\x00\x47\x00\x1a\x00\x49\x00\x4a\x00\x26\x00\xe6\x00\xe7\x00\x4e\x00\x4f\x00\x27\x00\x51\x00\x52\x00\x53\x00\xd3\x00\xd4\x00\xd5\x00\x27\x00\xd7\x00\xd8\x00\xc7\x00\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x59\x00\x78\x00\x3d\x00\x3e\x00\x3f\x00\x40\x00\xd4\x00\x17\x00\xd6\x00\xd7\x00\xd8\x00\x4e\x00\x27\x00\xc7\x00\x29\x00\x60\x00\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x17\x00\x5a\x00\xe6\x00\xe7\x00\xd4\x00\xc9\x00\xd6\x00\xd7\x00\xd8\x00\x0f\x00\x29\x00\xc7\x00\x2b\x00\x66\x00\x17\x00\xd3\x00\xd4\x00\xd5\x00\x1a\x00\xd7\x00\xd8\x00\xc4\x00\xe6\x00\xe7\x00\xd4\x00\xc8\x00\xd6\x00\xd7\x00\xd8\x00\xa1\x00\xa2\x00\xc7\x00\x79\x00\xe5\x00\x27\x00\x27\x00\x29\x00\xd4\x00\xd5\x00\x0d\x00\x0e\x00\xd8\x00\xe6\x00\xe7\x00\xd4\x00\x30\x00\xd6\x00\xd7\x00\xd8\x00\x05\x00\x06\x00\x0d\x00\x0e\x00\x43\x00\x30\x00\x45\x00\x20\x00\x47\x00\x26\x00\x49\x00\x4a\x00\x77\x00\xe6\x00\xe7\x00\x4e\x00\x4f\x00\x25\x00\x51\x00\x52\x00\x53\x00\x0d\x00\x0e\x00\x08\x00\x09\x00\x45\x00\x43\x00\x1a\x00\x45\x00\x20\x00\x47\x00\xc7\x00\x49\x00\x4a\x00\x37\x00\x38\x00\x78\x00\x4e\x00\x4f\x00\x29\x00\x51\x00\x52\x00\x53\x00\x0c\x00\xd4\x00\x17\x00\xd6\x00\xd7\x00\xd8\x00\x05\x00\x06\x00\x26\x00\x27\x00\x43\x00\x89\x00\x45\x00\x62\x00\x47\x00\x17\x00\x49\x00\x4a\x00\x01\x00\xe6\x00\xe7\x00\x4e\x00\x4f\x00\xc4\x00\x51\x00\x52\x00\x53\x00\xc8\x00\x05\x00\x06\x00\x43\x00\x20\x00\x45\x00\x1c\x00\x47\x00\x89\x00\x49\x00\x4a\x00\x05\x00\xd4\x00\xd5\x00\x4e\x00\x4f\x00\xd8\x00\x51\x00\x52\x00\x53\x00\x05\x00\x06\x00\x1c\x00\x43\x00\x27\x00\x45\x00\x26\x00\x47\x00\x30\x00\x49\x00\x4a\x00\x48\x00\x49\x00\x16\x00\x4e\x00\x4f\x00\x20\x00\x51\x00\x52\x00\x53\x00\x03\x00\x04\x00\x17\x00\x43\x00\x1a\x00\x45\x00\x7e\x00\x47\x00\x89\x00\x49\x00\x4a\x00\x1f\x00\x20\x00\x79\x00\x4e\x00\x4f\x00\x27\x00\x51\x00\x52\x00\x53\x00\x70\x00\x71\x00\x72\x00\x73\x00\x48\x00\x49\x00\xc7\x00\x78\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\x89\x00\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xd4\x00\x5a\x00\xd6\x00\xd7\x00\xd8\x00\x83\x00\x84\x00\x26\x00\x27\x00\xc7\x00\x63\x00\x64\x00\x89\x00\x66\x00\x1f\x00\x20\x00\x26\x00\x27\x00\xe6\x00\xe7\x00\x05\x00\x06\x00\xd4\x00\x23\x00\xd6\x00\xd7\x00\xd8\x00\x89\x00\x20\x00\x89\x00\x1a\x00\x43\x00\x79\x00\x45\x00\x1c\x00\x47\x00\xc7\x00\x49\x00\x4a\x00\x26\x00\xe6\x00\xe7\x00\x4e\x00\x4f\x00\x17\x00\x51\x00\x52\x00\x53\x00\x69\x00\xd4\x00\x64\x00\xd6\x00\xd7\x00\xd8\x00\x1b\x00\x89\x00\xc7\x00\x20\x00\x17\x00\x17\x00\x2f\x00\x60\x00\x2b\x00\x1a\x00\x5b\x00\x05\x00\x20\x00\xe6\x00\xe7\x00\xd4\x00\xc9\x00\xd6\x00\xd7\x00\xd8\x00\x29\x00\x1c\x00\xc7\x00\x3f\x00\x1a\x00\x29\x00\xd3\x00\xd4\x00\xd5\x00\x27\x00\xd7\x00\xd8\x00\x4a\x00\xe6\x00\xe7\x00\xd4\x00\x0c\x00\xd6\x00\xd7\x00\xd8\x00\x20\x00\x17\x00\xc7\x00\x20\x00\xe5\x00\x26\x00\x22\x00\x22\x00\x78\x00\x60\x00\x1b\x00\x60\x00\x1b\x00\xe6\x00\xe7\x00\xd4\x00\x29\x00\xd6\x00\xd7\x00\xd8\x00\x1b\x00\x2a\x00\x1b\x00\x20\x00\x43\x00\x26\x00\x45\x00\x5c\x00\x47\x00\x27\x00\x49\x00\x4a\x00\x1f\x00\xe6\x00\xe7\x00\x4e\x00\x4f\x00\x0f\x00\x51\x00\x52\x00\x53\x00\x0f\x00\x44\x00\x16\x00\x43\x00\x1c\x00\x45\x00\x17\x00\x47\x00\x0f\x00\x49\x00\x4a\x00\x17\x00\x26\x00\x4a\x00\x4e\x00\x4f\x00\x0c\x00\x51\x00\x52\x00\x53\x00\x64\x00\x1f\x00\x26\x00\x43\x00\x89\x00\x45\x00\x89\x00\x47\x00\xc7\x00\x49\x00\x4a\x00\x27\x00\x89\x00\x26\x00\x4e\x00\x4f\x00\x1c\x00\x51\x00\x52\x00\x53\x00\x1c\x00\xd4\x00\x1b\x00\xd6\x00\xd7\x00\xd8\x00\x43\x00\x17\x00\x45\x00\x20\x00\x47\x00\x26\x00\x49\x00\x4a\x00\x22\x00\x78\x00\x17\x00\x4e\x00\x4f\x00\xe6\x00\xe7\x00\x52\x00\x53\x00\x29\x00\x64\x00\x78\x00\x43\x00\x78\x00\x45\x00\x60\x00\x47\x00\x78\x00\x49\x00\x4a\x00\x20\x00\x22\x00\x1a\x00\x4e\x00\x4f\x00\x1c\x00\x25\x00\x52\x00\x53\x00\x75\x00\x1c\x00\x1b\x00\x43\x00\x45\x00\x45\x00\x17\x00\x47\x00\x22\x00\x49\x00\x4a\x00\x22\x00\x20\x00\x89\x00\x4e\x00\x4f\x00\x05\x00\x27\x00\x52\x00\x53\x00\x1a\x00\x17\x00\x17\x00\x43\x00\x17\x00\x45\x00\x17\x00\x47\x00\x17\x00\x49\x00\x4a\x00\x17\x00\xc7\x00\x17\x00\x4e\x00\x4f\x00\x01\x00\x17\x00\x52\x00\x53\x00\x70\x00\x71\x00\x72\x00\x73\x00\x20\x00\xd4\x00\x1f\x00\xd6\x00\xd7\x00\xd8\x00\x0f\x00\x1b\x00\xc7\x00\x17\x00\x3f\x00\x3f\x00\x2b\x00\x16\x00\x29\x00\x3f\x00\x1f\x00\x4a\x00\x25\x00\xe6\x00\xe7\x00\xd4\x00\x72\x00\xd6\x00\xd7\x00\xd8\x00\x1a\x00\x69\x00\xc7\x00\x0c\x00\x2b\x00\x89\x00\x17\x00\x89\x00\x17\x00\x17\x00\x20\x00\x17\x00\x20\x00\xe6\x00\xe7\x00\xd4\x00\x20\x00\xd6\x00\xd7\x00\xd8\x00\x43\x00\x1f\x00\x45\x00\x1f\x00\x47\x00\xc7\x00\x49\x00\x4a\x00\x29\x00\x78\x00\x69\x00\x4e\x00\x4f\x00\xe6\x00\xe7\x00\x52\x00\x53\x00\x27\x00\xd4\x00\x1e\x00\xd6\x00\xd7\x00\xd8\x00\x4e\x00\x20\x00\xc7\x00\x0f\x00\x05\x00\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x1b\x00\x5a\x00\xe6\x00\xe7\x00\xd4\x00\xc9\x00\xd6\x00\xd7\x00\xd8\x00\x70\x00\x89\x00\xc7\x00\x0f\x00\x66\x00\x8a\x00\xd3\x00\xd4\x00\xd5\x00\x6d\x00\xd7\x00\xd8\x00\x1e\x00\xe6\x00\xe7\x00\xd4\x00\x1b\x00\xd6\x00\xd7\x00\xd8\x00\x16\x00\x8a\x00\xc7\x00\x79\x00\xe5\x00\x63\x00\x0c\x00\x8a\x00\x26\x00\x1b\x00\x0f\x00\x0f\x00\x0f\x00\xe6\x00\xe7\x00\xd4\x00\x1b\x00\xd6\x00\xd7\x00\xd8\x00\x43\x00\x8a\x00\x45\x00\x2a\x00\x47\x00\x1b\x00\x49\x00\x4a\x00\x1b\x00\x2a\x00\x8a\x00\x4e\x00\x4f\x00\xe6\x00\xe7\x00\x52\x00\x53\x00\x00\x00\xff\xff\x70\x00\x43\x00\xff\xff\x45\x00\xff\xff\x47\x00\xff\xff\x49\x00\x4a\x00\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\xff\xff\xff\xff\x52\x00\x53\x00\xff\xff\xff\xff\xff\xff\x43\x00\xff\xff\x45\x00\xff\xff\x47\x00\xff\xff\x49\x00\x4a\x00\xff\xff\xc7\x00\xff\xff\x4e\x00\x4f\x00\xff\xff\xff\xff\x52\x00\x53\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd4\x00\xff\xff\xd6\x00\xd7\x00\xd8\x00\x43\x00\xff\xff\x45\x00\xff\xff\x47\x00\xff\xff\x49\x00\x4a\x00\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\xe6\x00\xe7\x00\x52\x00\x53\x00\xff\xff\xff\xff\xff\xff\x43\x00\xff\xff\x45\x00\xff\xff\x47\x00\xff\xff\x49\x00\x4a\x00\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\xff\xff\xff\xff\x52\x00\x53\x00\xff\xff\xff\xff\xff\xff\x43\x00\xff\xff\x45\x00\xff\xff\x47\x00\xff\xff\x49\x00\x4a\x00\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\xff\xff\xff\xff\x52\x00\x53\x00\xff\xff\xff\xff\xff\xff\x43\x00\xff\xff\x45\x00\xff\xff\x47\x00\xc7\x00\x49\x00\x4a\x00\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\xff\xff\xff\xff\x52\x00\x53\x00\xff\xff\xd4\x00\xff\xff\xd6\x00\xd7\x00\xd8\x00\x43\x00\xff\xff\xc7\x00\xff\xff\x47\x00\xff\xff\x49\x00\x4a\x00\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\xe6\x00\xe7\x00\xd4\x00\xff\xff\xd6\x00\xd7\x00\xd8\x00\x43\x00\xff\xff\xc7\x00\xff\xff\x47\x00\xff\xff\x49\x00\x4a\x00\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\xe6\x00\xe7\x00\xd4\x00\xff\xff\xd6\x00\xd7\x00\xd8\x00\x43\x00\xff\xff\x45\x00\xff\xff\x47\x00\xff\xff\x49\x00\x4a\x00\xff\xff\xc7\x00\xff\xff\x4e\x00\x4f\x00\xe6\x00\xe7\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd4\x00\xff\xff\xd6\x00\xd7\x00\xd8\x00\x43\x00\xff\xff\xc7\x00\xff\xff\x47\x00\xff\xff\x49\x00\x4a\x00\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\xe6\x00\xe7\x00\xd4\x00\xff\xff\xd6\x00\xd7\x00\xd8\x00\xff\xff\x47\x00\xc7\x00\x49\x00\x4a\x00\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\xff\xff\xff\xff\xff\xff\xe6\x00\xe7\x00\xd4\x00\xff\xff\xd6\x00\xd7\x00\xd8\x00\x46\x00\x47\x00\xc7\x00\x49\x00\x4a\x00\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\xff\xff\xff\xff\xff\xff\xe6\x00\xe7\x00\xd4\x00\xff\xff\xd6\x00\xd7\x00\xd8\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xe6\x00\xe7\x00\x69\x00\x46\x00\x47\x00\xd4\x00\x49\x00\x4a\x00\xd7\x00\xd8\x00\xff\xff\x4e\x00\x4f\x00\x49\x00\x4a\x00\xff\xff\x53\x00\xff\xff\x4e\x00\x4f\x00\x50\x00\xff\xff\xff\xff\xe6\x00\xe7\x00\xd4\x00\xff\xff\xff\xff\xd7\x00\xd8\x00\xff\xff\xff\xff\xc7\x00\xff\xff\x65\x00\x66\x00\x67\x00\xff\xff\x69\x00\xff\xff\xff\xff\xff\xff\xff\xff\xe6\x00\xe7\x00\xd4\x00\xff\xff\xd6\x00\xd7\x00\xd8\x00\x46\x00\x47\x00\xff\xff\x49\x00\x4a\x00\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x48\x00\x49\x00\x4a\x00\xe6\x00\xe7\x00\xff\xff\x4e\x00\x4f\x00\xff\xff\xff\xff\xff\xff\x47\x00\xd4\x00\x49\x00\x4a\x00\xd7\x00\xd8\x00\xff\xff\x4e\x00\x4f\x00\xff\xff\x65\x00\x66\x00\x67\x00\xff\xff\x69\x00\xff\xff\xff\xff\xff\xff\xd4\x00\xe6\x00\xe7\x00\xd7\x00\xd8\x00\x47\x00\xff\xff\x49\x00\x4a\x00\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\xff\xff\x48\x00\x49\x00\x4a\x00\xe6\x00\xe7\x00\xd4\x00\x4e\x00\x4f\x00\xd7\x00\xd8\x00\xff\xff\x48\x00\x49\x00\x4a\x00\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\xff\xff\x48\x00\x49\x00\x4a\x00\xe6\x00\xe7\x00\xff\xff\x4e\x00\x4f\x00\xff\xff\xff\xff\xc9\x00\xff\xff\x68\x00\x48\x00\x49\x00\x4a\x00\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\xd4\x00\xce\x00\xff\xff\xd7\x00\xd8\x00\xd2\x00\xff\xff\xd4\x00\xff\xff\xff\xff\xd7\x00\xd8\x00\xd9\x00\xda\x00\x48\x00\x49\x00\x4a\x00\xff\xff\xe6\x00\xe7\x00\x4e\x00\x4f\x00\xff\xff\xff\xff\xff\xff\xe6\x00\xe7\x00\xe8\x00\xe9\x00\xff\xff\xff\xff\x48\x00\x49\x00\x4a\x00\xff\xff\xff\xff\xc9\x00\x4e\x00\x4f\x00\x49\x00\x4a\x00\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\xff\xff\xd4\x00\xff\xff\xff\xff\xd7\x00\xd8\x00\xff\xff\xff\xff\xff\xff\xd4\x00\xff\xff\xff\xff\xd7\x00\xd8\x00\xff\xff\x71\x00\x72\x00\x73\x00\xff\xff\xe6\x00\xe7\x00\xd4\x00\x72\x00\x73\x00\xd7\x00\xd8\x00\x8c\x00\xe6\x00\xe7\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xe6\x00\xe7\x00\xff\xff\xff\xff\x9d\x00\xd4\x00\xff\xff\xff\xff\xd7\x00\xd8\x00\xff\xff\xff\xff\xff\xff\xff\xff\xd4\x00\xff\xff\xff\xff\xd7\x00\xd8\x00\xff\xff\xff\xff\xff\xff\xff\xff\xe6\x00\xe7\x00\xd4\x00\xff\xff\xff\xff\xd7\x00\xd8\x00\xff\xff\xff\xff\xe6\x00\xe7\x00\xd4\x00\xff\xff\xff\xff\xd7\x00\xd8\x00\xff\xff\xff\xff\xff\xff\xc3\x00\xe6\x00\xe7\x00\xc6\x00\xc7\x00\xd4\x00\xc9\x00\xff\xff\xd7\x00\xd8\x00\xe6\x00\xe7\x00\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc9\x00\xe6\x00\xe7\x00\xff\xff\xff\xff\xd4\x00\xc9\x00\xe0\x00\xd7\x00\xd8\x00\xd3\x00\xd4\x00\xd5\x00\xff\xff\xd7\x00\xd8\x00\xd3\x00\xd4\x00\xd5\x00\xff\xff\xd7\x00\xd8\x00\xd4\x00\xe6\x00\xe7\x00\xd7\x00\xd8\x00\xff\xff\xe5\x00\xd4\x00\xff\xff\xff\xff\xd7\x00\xd8\x00\xe5\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xe6\x00\xe7\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xe6\x00\xe7\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\x1a\x00\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4d\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\x5d\x00\x5e\x00\x5f\x00\x60\x00\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\x68\x00\xff\xff\x6a\x00\x6b\x00\x6c\x00\x6d\x00\x6e\x00\x6f\x00\xff\xff\x71\x00\xff\xff\x73\x00\xff\xff\xff\xff\x76\x00\xff\xff\xff\xff\x79\x00\x7a\x00\x7b\x00\x7c\x00\x7d\x00\xff\xff\x7f\x00\x80\x00\x81\x00\x82\x00\x83\x00\x84\x00\xff\xff\xff\xff\xff\xff\x88\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4d\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\x5d\x00\x5e\x00\x5f\x00\x60\x00\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\x68\x00\xff\xff\x6a\x00\x6b\x00\x6c\x00\x6d\x00\x6e\x00\x6f\x00\xff\xff\x71\x00\xff\xff\x73\x00\xff\xff\xff\xff\x76\x00\xff\xff\xff\xff\x79\x00\x7a\x00\x7b\x00\x7c\x00\x7d\x00\xff\xff\x7f\x00\x80\x00\x81\x00\x82\x00\x83\x00\x84\x00\xff\xff\xff\xff\xff\xff\x88\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4d\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\x5d\x00\x5e\x00\x5f\x00\x60\x00\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\x6a\x00\x6b\x00\x6c\x00\x6d\x00\x6e\x00\x6f\x00\x83\x00\x71\x00\xff\xff\x73\x00\xff\xff\xff\xff\x76\x00\xff\xff\xff\xff\x79\x00\x7a\x00\x7b\x00\x7c\x00\x7d\x00\xff\xff\x7f\x00\x80\x00\x81\x00\x82\x00\x83\x00\x84\x00\x01\x00\xff\xff\xff\xff\x88\x00\x05\x00\x06\x00\xff\xff\xff\xff\x09\x00\xff\xff\xff\xff\x0c\x00\xff\xff\x01\x00\x0f\x00\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\xff\xff\x16\x00\xff\xff\x18\x00\x0c\x00\xff\xff\xff\xff\x0f\x00\xff\xff\x1e\x00\xff\xff\xff\xff\xff\xff\x22\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\x2d\x00\x2e\x00\xff\xff\x30\x00\x71\x00\x72\x00\x73\x00\xcb\x00\xff\xff\xff\xff\xce\x00\xff\xff\xd0\x00\xff\xff\xd2\x00\xff\xff\x30\x00\xff\xff\xff\xff\x40\x00\xff\xff\xd9\x00\xda\x00\xdb\x00\xff\xff\xdd\x00\xff\xff\xdf\x00\xff\xff\xff\xff\xff\xff\xff\xff\x40\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x66\x00\x5a\x00\xff\xff\xff\xff\x01\x00\xff\xff\xff\xff\xff\xff\x05\x00\x06\x00\xff\xff\x64\x00\x09\x00\x66\x00\xff\xff\x0c\x00\xff\xff\xff\xff\x0f\x00\x79\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\x85\x00\x79\x00\x1e\x00\xff\xff\xc9\x00\xff\xff\x22\x00\xff\xff\x08\x00\x09\x00\x0a\x00\x0b\x00\xff\xff\x85\x00\xd3\x00\xd4\x00\xd5\x00\x01\x00\xd7\x00\xd8\x00\x30\x00\x05\x00\x06\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x0c\x00\xff\xff\xff\xff\x0f\x00\xe5\x00\xff\xff\x22\x00\x23\x00\x40\x00\x25\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\x2f\x00\x30\x00\x31\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x01\x00\x5a\x00\xff\xff\x30\x00\x05\x00\x06\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x0c\x00\xff\xff\x66\x00\x0f\x00\xff\xff\xff\xff\xff\xff\xff\xff\x40\x00\xff\xff\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x79\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\x83\x00\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x85\x00\x5a\x00\xff\xff\x30\x00\xff\xff\x05\x00\x06\x00\xff\xff\xff\xff\x09\x00\xff\xff\x0b\x00\x0c\x00\x66\x00\xff\xff\x0f\x00\xff\xff\xff\xff\xff\xff\x40\x00\x01\x00\x02\x00\x16\x00\x17\x00\x05\x00\x06\x00\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\x79\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\x25\x00\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x85\x00\x5a\x00\xff\xff\xff\xff\x01\x00\xff\xff\xff\xff\xff\xff\x05\x00\x06\x00\xff\xff\xff\xff\xff\xff\x66\x00\xff\xff\x0c\x00\xff\xff\xff\xff\x0f\x00\xff\xff\x40\x00\xcb\x00\xff\xff\xff\xff\xce\x00\x16\x00\xd0\x00\x18\x00\xd2\x00\xff\xff\xff\xff\xff\xff\x79\x00\x1e\x00\xff\xff\xd9\x00\xda\x00\xdb\x00\xff\xff\xdd\x00\xff\xff\xdf\x00\xff\xff\xff\xff\x85\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x30\x00\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x40\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\xff\xff\x79\x00\xff\xff\xff\xff\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x66\x00\x08\x00\x09\x00\x0a\x00\x0b\x00\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\x19\x00\x79\x00\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\x20\x00\x21\x00\x22\x00\x23\x00\xff\xff\x25\x00\x85\x00\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\x31\x00\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x80\x00\x81\x00\x82\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\x08\x00\x09\x00\x0a\x00\x0b\x00\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\x17\x00\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\x20\x00\x21\x00\x22\x00\x23\x00\xff\xff\x25\x00\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\x31\x00\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x80\x00\x81\x00\x82\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\x08\x00\x09\x00\x0a\x00\x0b\x00\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\x17\x00\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\x20\x00\x21\x00\x22\x00\x23\x00\xff\xff\x25\x00\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\x31\x00\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x80\x00\x81\x00\x82\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\x08\x00\x09\x00\x0a\x00\x0b\x00\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\x19\x00\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\x20\x00\x21\x00\x22\x00\x23\x00\xff\xff\x25\x00\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\x31\x00\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x80\x00\x81\x00\x82\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\x08\x00\x09\x00\x0a\x00\x0b\x00\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\x19\x00\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\x20\x00\x21\x00\x22\x00\x23\x00\xff\xff\x25\x00\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\x31\x00\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x80\x00\x81\x00\x82\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\x08\x00\x09\x00\x0a\x00\x0b\x00\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\x17\x00\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\x20\x00\x21\x00\x22\x00\x23\x00\xff\xff\x25\x00\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\x31\x00\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x80\x00\x81\x00\x82\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\x08\x00\x09\x00\x0a\x00\x0b\x00\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\x1f\x00\xff\xff\x21\x00\x22\x00\x23\x00\xff\xff\x25\x00\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\x31\x00\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x80\x00\x81\x00\x82\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\x08\x00\x09\x00\x0a\x00\x0b\x00\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\x22\x00\x23\x00\xff\xff\x25\x00\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\x31\x00\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\xff\xff\x08\x00\x09\x00\x0a\x00\x0b\x00\x79\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x80\x00\x81\x00\x82\x00\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\x23\x00\x18\x00\x25\x00\x1a\x00\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\x2f\x00\x30\x00\x31\x00\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\x74\x00\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x80\x00\x81\x00\x82\x00\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\xff\xff\x75\x00\xff\xff\xff\xff\xff\xff\x79\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x80\x00\x81\x00\x82\x00\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\x1a\x00\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x80\x00\x81\x00\x82\x00\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\xff\xff\x75\x00\xff\xff\xff\xff\xff\xff\x79\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x80\x00\x81\x00\x82\x00\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\x74\x00\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x80\x00\x81\x00\x82\x00\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x80\x00\x81\x00\x82\x00\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x80\x00\x81\x00\x82\x00\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\x65\x00\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x80\x00\x81\x00\x82\x00\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x80\x00\x81\x00\x82\x00\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\x1b\x00\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x80\x00\x81\x00\x82\x00\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x80\x00\x81\x00\x82\x00\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x67\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6e\x00\x6f\x00\xff\xff\xff\xff\xff\xff\x73\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x80\x00\x81\x00\x82\x00\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\x1a\x00\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\x08\x00\x09\x00\x0a\x00\x0b\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x17\x00\x5a\x00\xff\xff\x5c\x00\xff\xff\x5e\x00\xff\xff\xff\xff\x61\x00\x20\x00\x63\x00\x64\x00\x23\x00\x66\x00\x25\x00\xff\xff\xff\xff\x6a\x00\x6b\x00\x6c\x00\x2b\x00\xff\xff\x6f\x00\xff\xff\x2f\x00\x30\x00\x31\x00\xff\xff\xff\xff\x76\x00\xff\xff\xff\xff\x79\x00\x7a\x00\x7b\x00\x7c\x00\x7d\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\x5e\x00\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\xff\xff\xff\xff\x6a\x00\x6b\x00\x6c\x00\xff\xff\xff\xff\x6f\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x76\x00\xff\xff\xff\xff\x79\x00\x7a\x00\x7b\x00\x7c\x00\x7d\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\x1a\x00\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\x5e\x00\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6f\x00\xff\xff\x71\x00\xff\xff\xff\xff\xff\xff\xff\xff\x76\x00\xff\xff\xff\xff\x79\x00\x7a\x00\x7b\x00\x7c\x00\x7d\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\x1a\x00\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\xff\xff\xff\xff\x6a\x00\x6b\x00\x6c\x00\xff\xff\xff\xff\x6f\x00\xff\xff\xff\xff\xff\xff\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x79\x00\x7a\x00\x7b\x00\x7c\x00\x7d\x00\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\x5e\x00\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x6f\x00\xff\xff\x71\x00\xff\xff\xff\xff\xff\xff\xff\xff\x76\x00\xff\xff\xff\xff\x79\x00\x7a\x00\x7b\x00\x7c\x00\x7d\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xff\xff\xff\xff\xff\xff\x61\x00\xcb\x00\x63\x00\x64\x00\xce\x00\x66\x00\xd0\x00\xff\xff\xd2\x00\x6a\x00\x6b\x00\x6c\x00\xff\xff\xff\xff\x6f\x00\xd9\x00\xda\x00\xdb\x00\xff\xff\xdd\x00\xff\xff\xdf\x00\xff\xff\xff\xff\x79\x00\x7a\x00\x7b\x00\x7c\x00\x7d\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x07\x00\xff\xff\xff\xff\xff\xff\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xcb\x00\x21\x00\xff\xff\xce\x00\xff\xff\xd0\x00\xff\xff\xd2\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xd9\x00\xda\x00\xdb\x00\xff\xff\xdd\x00\xff\xff\xdf\x00\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4d\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\x5b\x00\x5c\x00\x5d\x00\x5e\x00\x5f\x00\x60\x00\x61\x00\x62\x00\x63\x00\x64\x00\x65\x00\x66\x00\x67\x00\x68\x00\x69\x00\x6a\x00\x6b\x00\x6c\x00\x6d\x00\x6e\x00\x6f\x00\x70\x00\x71\x00\x72\x00\x73\x00\x74\x00\x75\x00\x76\x00\x77\x00\x78\x00\x79\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x72\x00\x73\x00\xff\xff\x75\x00\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x28\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\x2f\x00\x71\x00\x72\x00\x73\x00\x74\x00\x75\x00\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\xff\xff\xc9\x00\xff\xff\xff\xff\x61\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\xd3\x00\xd4\x00\xd5\x00\xff\xff\xd7\x00\xd8\x00\xff\xff\x6f\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\xff\xff\xff\xff\x79\x00\xe5\x00\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x21\x00\xd2\x00\xd3\x00\xd4\x00\xd5\x00\xff\xff\xd7\x00\xd8\x00\xd9\x00\xda\x00\x2b\x00\xff\xff\x2d\x00\xff\xff\xff\xff\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xe5\x00\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\xff\xff\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\xff\xff\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x5c\x00\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x2d\x00\xff\xff\xff\xff\x30\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\xff\xff\x01\x00\x02\x00\x03\x00\x04\x00\x05\x00\x06\x00\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\x0c\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x13\x00\x14\x00\x15\x00\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\x1e\x00\xff\xff\xff\xff\x21\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x01\x00\xff\xff\xff\xff\xff\xff\x05\x00\x2d\x00\x07\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x36\x00\xff\xff\xff\xff\x39\x00\x3a\x00\x3b\x00\x3c\x00\x3d\x00\x3e\x00\xff\xff\x40\x00\x41\x00\x42\x00\xff\xff\x44\x00\xff\xff\x46\x00\x47\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4d\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\x79\x00\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\x5b\x00\x5c\x00\x5d\x00\x5e\x00\x5f\x00\x60\x00\x61\x00\x62\x00\x63\x00\x64\x00\x65\x00\x66\x00\x67\x00\x68\x00\x69\x00\x6a\x00\x6b\x00\x6c\x00\x6d\x00\x6e\x00\x6f\x00\x70\x00\x71\x00\x72\x00\x73\x00\x74\x00\x75\x00\x76\x00\x77\x00\x78\x00\x79\x00\x01\x00\xff\xff\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\x08\x00\x09\x00\x0a\x00\x0b\x00\x0c\x00\xff\xff\xff\xff\x0f\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\x17\x00\x18\x00\xff\xff\x08\x00\x09\x00\x0a\x00\x0b\x00\x1e\x00\xff\xff\x20\x00\xff\xff\xff\xff\x23\x00\xff\xff\x25\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x2b\x00\xff\xff\xff\xff\xff\xff\x2f\x00\x30\x00\x31\x00\x20\x00\xff\xff\x22\x00\x23\x00\xff\xff\x25\x00\x26\x00\x27\x00\xff\xff\x29\x00\xff\xff\xff\xff\xff\xff\xff\xff\x40\x00\x2f\x00\x30\x00\x31\x00\xff\xff\xff\xff\x01\x00\xff\xff\xff\xff\xff\xff\x05\x00\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\x16\x00\x17\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x01\x00\x64\x00\xff\xff\x66\x00\x05\x00\x06\x00\x24\x00\x08\x00\x09\x00\xff\xff\x0b\x00\x0c\x00\xff\xff\xff\xff\x0f\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\x79\x00\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\xff\xff\x22\x00\xff\xff\xff\xff\x25\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x2b\x00\xff\xff\x2d\x00\x2e\x00\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\xff\xff\x40\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x01\x00\x5a\x00\xff\xff\x79\x00\x05\x00\x06\x00\xff\xff\x08\x00\x09\x00\xff\xff\x0b\x00\x0c\x00\xff\xff\x66\x00\x0f\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\x18\x00\xff\xff\x01\x00\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x79\x00\x22\x00\xff\xff\xff\xff\x25\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x2b\x00\xff\xff\x2d\x00\x01\x00\x02\x00\xff\xff\xff\xff\x05\x00\x06\x00\xff\xff\xff\xff\x09\x00\xff\xff\x0b\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x40\x00\xff\xff\xff\xff\x16\x00\x17\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\x25\x00\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\xff\xff\x30\x00\x31\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x66\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\x66\x00\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\xff\xff\x01\x00\xff\xff\x03\x00\x04\x00\x05\x00\x06\x00\x63\x00\x64\x00\x79\x00\x66\x00\xff\xff\x0c\x00\xff\xff\xff\xff\x0f\x00\xff\xff\x08\x00\x09\x00\x0a\x00\x0b\x00\xff\xff\x16\x00\xff\xff\x18\x00\x19\x00\xff\xff\xff\xff\xff\xff\x79\x00\x1e\x00\x01\x00\x20\x00\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\xff\xff\xff\xff\xff\xff\x20\x00\x0c\x00\x22\x00\x23\x00\x0f\x00\x25\x00\x26\x00\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\x17\x00\x18\x00\xff\xff\x2f\x00\x30\x00\x31\x00\xff\xff\x1e\x00\xff\xff\xff\xff\xff\xff\x40\x00\xff\xff\x08\x00\x09\x00\x0a\x00\x0b\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x22\x00\x23\x00\x40\x00\x25\x00\x26\x00\xff\xff\xff\xff\xff\xff\x64\x00\xff\xff\x66\x00\xff\xff\xff\xff\x2f\x00\x30\x00\x31\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\x79\x00\xff\xff\x01\x00\xff\xff\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\x64\x00\xff\xff\x66\x00\xff\xff\x0c\x00\xff\xff\xff\xff\x0f\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\x1e\x00\xff\xff\xff\xff\x01\x00\xff\xff\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x0c\x00\xff\xff\xff\xff\x0f\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\x40\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x40\x00\xff\xff\xff\xff\xff\xff\x64\x00\xff\xff\x66\x00\x01\x00\xff\xff\xff\xff\xff\xff\x05\x00\x06\x00\x6d\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x79\x00\x5a\x00\x01\x00\x02\x00\xff\xff\xff\xff\x05\x00\x06\x00\xff\xff\xff\xff\xff\xff\x64\x00\xff\xff\x66\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\x79\x00\x20\x00\x01\x00\xff\xff\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2b\x00\x0c\x00\xff\xff\xff\xff\x0f\x00\x30\x00\x31\x00\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\x18\x00\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\x1e\x00\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\x66\x00\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x40\x00\xff\xff\xff\xff\x63\x00\x64\x00\x79\x00\x66\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x79\x00\x5a\x00\xff\xff\xff\xff\x01\x00\xff\xff\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\x64\x00\xff\xff\x66\x00\xff\xff\x0c\x00\xff\xff\x01\x00\x0f\x00\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\xff\xff\x16\x00\xff\xff\x18\x00\x0c\x00\xff\xff\xff\xff\x0f\x00\x79\x00\x1e\x00\x1f\x00\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x01\x00\xff\xff\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x0c\x00\xff\xff\xff\xff\x0f\x00\xff\xff\xff\xff\xff\xff\x40\x00\xff\xff\xff\xff\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\x40\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x66\x00\x5a\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x40\x00\xff\xff\xff\xff\xff\xff\x64\x00\xff\xff\x66\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x79\x00\x5a\x00\x01\x00\x02\x00\xff\xff\xff\xff\x05\x00\x06\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x66\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\x79\x00\xff\xff\x01\x00\xff\xff\x03\x00\x04\x00\x05\x00\x06\x00\xff\xff\xff\xff\xff\xff\xff\xff\x2b\x00\x0c\x00\xff\xff\x01\x00\x0f\x00\x30\x00\x31\x00\x05\x00\x06\x00\xff\xff\xff\xff\x16\x00\xff\xff\x18\x00\x0c\x00\xff\xff\xff\xff\x0f\x00\xff\xff\x1e\x00\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\xff\xff\x01\x00\x02\x00\xff\xff\x40\x00\x05\x00\x06\x00\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x40\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\x16\x00\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x79\x00\x5a\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x66\x00\x5a\x00\xff\xff\xff\xff\x01\x00\xff\xff\xff\xff\xff\xff\x05\x00\x06\x00\x63\x00\xff\xff\xff\xff\x66\x00\xff\xff\x0c\x00\xff\xff\x01\x00\x0f\x00\x79\x00\x6d\x00\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\x01\x00\x02\x00\x79\x00\x1e\x00\x05\x00\x06\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\x2e\x00\xff\xff\x16\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x63\x00\x64\x00\x1e\x00\x66\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x40\x00\xff\xff\xff\xff\x70\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x30\x00\x31\x00\x79\x00\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x66\x00\x5a\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x66\x00\x5a\x00\xff\xff\xff\xff\x01\x00\xff\xff\x79\x00\xff\xff\x05\x00\x06\x00\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\x0c\x00\xff\xff\x01\x00\x0f\x00\x79\x00\xff\xff\x05\x00\x06\x00\xff\xff\xff\xff\x16\x00\xff\xff\x18\x00\x0c\x00\xff\xff\x01\x00\x0f\x00\x79\x00\x1e\x00\x05\x00\x06\x00\xff\xff\xff\xff\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\x01\x00\x02\x00\xff\xff\x1e\x00\x05\x00\x06\x00\x2e\x00\xff\xff\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x1e\x00\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\xff\xff\x40\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x20\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x40\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x66\x00\x5a\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x66\x00\x5a\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\x79\x00\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x66\x00\x5a\x00\xff\xff\xff\xff\x01\x00\xff\xff\x79\x00\xff\xff\x05\x00\x06\x00\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\x0c\x00\xff\xff\x01\x00\x0f\x00\x79\x00\xff\xff\x05\x00\xff\xff\x70\x00\xff\xff\x16\x00\xff\xff\x18\x00\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\x1e\x00\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\x17\x00\xff\xff\xff\xff\xff\xff\x01\x00\xff\xff\xff\xff\xff\xff\x05\x00\xff\xff\xff\xff\xff\xff\xff\xff\x24\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\xff\xff\x40\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x66\x00\x5a\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x01\x00\x02\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\x79\x00\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\x16\x00\xff\xff\xff\xff\xff\xff\x79\x00\xff\xff\x1c\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\xff\xff\x24\x00\xff\xff\xff\xff\xff\xff\x01\x00\xff\xff\xff\xff\x70\x00\x05\x00\xff\xff\xff\xff\xff\xff\xff\xff\x76\x00\xff\xff\xff\xff\x79\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\x01\x00\x02\x00\xff\xff\xff\xff\x05\x00\x06\x00\xff\xff\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\x08\x00\x09\x00\x0a\x00\x0b\x00\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\x79\x00\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\x22\x00\x23\x00\xff\xff\x25\x00\x26\x00\xff\xff\xff\xff\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x01\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\xff\xff\x70\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x76\x00\xff\xff\xff\xff\x79\x00\xff\xff\x01\x00\xff\xff\xff\xff\xff\xff\x05\x00\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\xff\xff\x16\x00\x2b\x00\xff\xff\xff\xff\x01\x00\x02\x00\x63\x00\x64\x00\xff\xff\x66\x00\x20\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\xff\xff\x79\x00\xff\xff\xff\xff\x1c\x00\x01\x00\x02\x00\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\x66\x00\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x24\x00\x5a\x00\xff\xff\xff\xff\x01\x00\xff\xff\xff\xff\xff\xff\x05\x00\xff\xff\x63\x00\x64\x00\x79\x00\x66\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x16\x00\x5a\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\xff\xff\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\x01\x00\x02\x00\x79\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\x79\x00\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\x01\x00\xff\xff\xff\xff\xff\xff\x05\x00\xff\xff\xff\xff\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x01\x00\x02\x00\x79\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x16\x00\x5a\x00\xff\xff\xff\xff\x01\x00\xff\xff\xff\xff\xff\xff\x05\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\x79\x00\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x01\x00\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\x16\x00\xff\xff\x79\x00\xff\xff\x01\x00\xff\xff\xff\xff\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\x16\x00\xff\xff\xff\xff\xff\xff\x01\x00\x79\x00\xff\xff\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x16\x00\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x01\x00\xff\xff\x63\x00\x64\x00\x05\x00\x66\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\x01\x00\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\xff\xff\x16\x00\xff\xff\xff\xff\x79\x00\xff\xff\xff\xff\x63\x00\x64\x00\x01\x00\x66\x00\xff\xff\xff\xff\xff\xff\x23\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\x16\x00\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\x23\x00\x5a\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x63\x00\x64\x00\xff\xff\x66\x00\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x79\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x66\x00\xff\xff\xff\xff\xff\xff\xff\xff\x4e\x00\x4f\x00\x50\x00\x51\x00\xff\xff\x53\x00\x54\x00\x55\x00\x56\x00\x57\x00\x58\x00\xff\xff\x5a\x00\xff\xff\x79\x00\xff\xff\xff\xff\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\x66\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\x79\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xa8\x00\xa9\x00\xaa\x00\xab\x00\xff\xff\xad\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xb5\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xa9\x00\xaa\x00\xab\x00\xff\xff\xad\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xb5\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xaa\x00\xab\x00\xff\xff\xad\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xb5\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xac\x00\xad\x00\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xb5\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\x9d\x00\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\xff\xff\xff\xff\xb5\x00\xff\xff\xff\xff\xff\xff\xe0\x00\xff\xff\xbb\x00\xff\xff\xbd\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xac\x00\xad\x00\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xb5\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\x9d\x00\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\xff\xff\xff\xff\xb5\x00\xff\xff\xff\xff\xff\xff\xe0\x00\xff\xff\xbb\x00\xff\xff\xbd\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xac\x00\xad\x00\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xb5\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\x9d\x00\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\xff\xff\xff\xff\xb5\x00\xff\xff\xff\xff\xff\xff\xe0\x00\xff\xff\xbb\x00\xff\xff\xbd\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xb5\x00\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xbb\x00\xff\xff\xbd\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xad\x00\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xb5\x00\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xe0\x00\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xb5\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xbd\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xb6\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xe0\x00\xe1\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xe0\x00\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xe0\x00\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xe0\x00\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xe0\x00\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xe0\x00\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xe0\x00\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xe0\x00\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xe0\x00\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\x87\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xe0\x00\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\x97\x00\x98\x00\xff\xff\xff\xff\xff\xff\x9c\x00\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\x98\x00\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xa6\x00\xa7\x00\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xcc\x00\xff\xff\xce\x00\xff\xff\xff\xff\xd1\x00\xd2\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xd9\x00\xda\x00\xdb\x00\xdc\x00\xdd\x00\xde\x00\xdf\x00\xe0\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xcc\x00\xff\xff\xce\x00\xff\xff\xff\xff\xd1\x00\xd2\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xd9\x00\xda\x00\xff\xff\xdc\x00\xff\xff\xde\x00\xdf\x00\xe0\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\x97\x00\x98\x00\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\x98\x00\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xcc\x00\xff\xff\xce\x00\xff\xff\xff\xff\xd1\x00\xd2\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xd9\x00\xda\x00\xff\xff\xdc\x00\xff\xff\xde\x00\xdf\x00\xe0\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xcc\x00\xff\xff\xce\x00\xff\xff\xff\xff\xd1\x00\xd2\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xd9\x00\xda\x00\xff\xff\xdc\x00\xff\xff\xde\x00\xdf\x00\xe0\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\x98\x00\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\x98\x00\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xcc\x00\xff\xff\xce\x00\xff\xff\xff\xff\xd1\x00\xd2\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xd9\x00\xda\x00\xff\xff\xdc\x00\xff\xff\xde\x00\xdf\x00\xe0\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xcc\x00\xff\xff\xce\x00\xff\xff\xff\xff\xd1\x00\xd2\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xd9\x00\xda\x00\xff\xff\xdc\x00\xff\xff\xde\x00\xdf\x00\xe0\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\x98\x00\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\x98\x00\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xcc\x00\xff\xff\xce\x00\xff\xff\xff\xff\xd1\x00\xd2\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xd9\x00\xda\x00\xff\xff\xdc\x00\xff\xff\xde\x00\xdf\x00\xe0\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xcc\x00\xff\xff\xce\x00\xff\xff\xff\xff\xd1\x00\xd2\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xd9\x00\xda\x00\xff\xff\xdc\x00\xff\xff\xde\x00\xdf\x00\xe0\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\x98\x00\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xcc\x00\xff\xff\xce\x00\xff\xff\xff\xff\xd1\x00\xd2\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xd9\x00\xda\x00\xff\xff\xdc\x00\xff\xff\xde\x00\xdf\x00\xe0\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xb0\x00\xb1\x00\xff\xff\xff\xff\xff\xff\xb5\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xb1\x00\x94\x00\x95\x00\x96\x00\xb5\x00\xff\xff\xff\xff\xff\xff\x9b\x00\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xe0\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9c\x00\x9d\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\x9b\x00\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xb5\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\x88\x00\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\xff\xff\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xff\xff\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\x8f\x00\xff\xff\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x9d\x00\xff\xff\x89\x00\x8a\x00\x8b\x00\x8c\x00\xff\xff\xe0\x00\x8f\x00\x90\x00\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\x91\x00\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xe0\x00\xff\xff\xff\xff\xff\xff\x92\x00\x93\x00\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\x93\x00\x94\x00\x95\x00\x96\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\x93\x00\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xe0\x00\xff\xff\x9d\x00\xff\xff\xff\xff\xa0\x00\xa1\x00\xa2\x00\xff\xff\xa4\x00\xa5\x00\xff\xff\xff\xff\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xa0\x00\xa1\x00\xa2\x00\xff\xff\xa4\x00\xa5\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x94\x00\x95\x00\x96\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x9d\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xc3\x00\xff\xff\xff\xff\xc6\x00\xc7\x00\xff\xff\xc9\x00\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xd3\x00\xd4\x00\xd5\x00\xd6\x00\xd7\x00\xd8\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xe0\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"#

happyTable :: HappyAddr
happyTable = HappyA# "\x00\x00\x0c\x00\x09\x00\x0b\x00\xa1\x00\x09\x00\xa2\x00\x64\x02\x9d\x03\x9e\x03\x9f\x03\x39\x02\x6d\x03\x24\x03\x8b\x04\x7b\x04\x0b\x04\x0c\x04\xc4\x00\x8d\x02\xa1\x01\xe9\x01\x9f\x01\x69\x03\x4b\x03\xc4\x00\x8c\x04\x8d\x04\x14\x01\x97\x02\x6c\x03\xe1\x02\xa1\x01\x01\x03\x62\x04\x8d\x02\x3c\x04\x65\x00\x3a\x02\x3b\x02\xba\xff\x67\x00\xcf\x01\x68\x00\x69\x00\xd5\x01\xa1\x01\xe9\x01\x6a\x00\x6b\x00\xa1\x01\x01\x03\xf8\x00\x6e\x00\x3c\x02\xff\x00\xc4\x00\xc4\x00\x36\x03\xa1\x01\x40\x02\xc5\x00\x1b\x01\xa1\x01\x76\x03\xa1\x01\x76\x03\x15\x01\xa1\x01\x40\x02\x53\x02\xbc\x00\xa7\x00\x1a\x01\xa8\x00\xa1\x01\xdf\x03\x55\x02\xa1\x01\xdf\x03\x55\x02\x37\x03\xd0\x01\xf8\x03\x20\x01\xeb\x00\xa9\x03\x86\x03\x41\x02\x3e\x03\x84\x04\x4f\x02\xdb\x00\xd0\x01\xdd\x00\xd5\x01\xdc\x00\x43\x02\xa6\x01\x24\x00\x25\x00\xa1\x01\xc0\x03\x95\x04\xa1\x01\xc0\x03\x85\x03\x86\x03\xce\x00\x97\x04\x3f\x04\xf9\x03\x15\x01\xec\x00\x3e\x04\x3f\x04\xa9\x00\x92\x04\xd6\x01\xe1\x00\x89\x04\xcf\x03\x0d\x04\x15\x01\x50\x02\x92\x03\x93\x03\xa7\x01\x4c\x03\x25\x00\x24\x03\xe4\x00\xe5\x00\xe6\x00\xcf\x00\xa9\x00\x24\x00\x25\x00\x7e\x00\x0d\x04\x7f\x00\x67\x03\x94\x03\xdc\x01\xdd\x01\x81\x00\x20\x03\x78\x04\x34\x04\x82\x00\xbe\x00\x83\x00\x84\x00\x24\x00\x25\x00\x98\x02\xe0\x03\x21\x00\xad\x00\xe1\x03\x7e\x00\xf5\xff\x7f\x00\xc0\xfd\x35\x04\xdc\x01\xdd\x01\x81\x00\xc0\xfd\x6f\x00\x99\x02\x82\x00\x98\x02\x83\x00\x84\x00\x77\x03\x7c\x04\x78\x03\x17\x02\x40\x04\x13\x02\x54\x02\x70\x00\x41\x04\x23\x00\x71\x00\x25\x00\x99\x02\x56\x02\xa0\x03\xea\x01\x59\x02\xeb\x00\x8e\x04\x02\x03\x21\x00\xad\x00\x8f\x04\xc6\x00\xae\x00\x72\x00\x73\x00\x20\x00\x21\x00\x22\x00\xc6\x00\x24\x00\x25\x00\xc1\x03\x21\x00\xad\x00\xc2\x03\x03\x03\xae\x00\xc1\xfd\xa0\x03\xeb\x01\x6c\x04\xeb\x00\xc1\xfd\xa1\x03\x65\x02\xa0\x01\x0a\x00\x0a\x00\x90\x04\x0a\x00\x0a\x00\x20\x00\x21\x00\x22\x00\xa0\x01\x24\x00\x25\x00\x85\x00\x25\x02\xc6\x00\xc6\x00\x42\x02\x21\x00\x87\x03\x17\x04\x86\x00\x10\x00\x87\x00\x12\x00\xa1\x03\x44\x02\x13\x00\x14\x00\x15\x00\x0d\x02\x44\x03\x16\x00\x17\x00\x18\x00\x85\x00\x17\x04\x6d\x04\x21\x00\x87\x03\x2b\x03\x19\x00\x70\x04\x86\x00\x10\x00\x87\x00\x12\x00\xfd\x00\x8e\x04\x13\x00\x14\x00\x15\x00\x8f\x04\x40\x04\x16\x00\x17\x00\x18\x00\x41\x04\x40\x04\xcd\x03\x0d\x02\xca\x02\x41\x04\x19\x00\x21\x00\xad\x00\x6f\x03\x71\x04\xae\x00\x21\x00\xad\x00\xac\x02\x1b\x02\xae\x00\x21\x00\xad\x00\x6a\x04\x54\x04\xae\x00\x1c\x00\x90\x04\x99\x02\x1d\x00\x1e\x00\x70\x00\x1f\x00\x0d\x02\xf1\x00\xf2\x00\xf3\x00\x45\x04\x18\x04\x19\x04\x2c\x00\x2d\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x14\x02\x73\x00\x1d\x00\x1e\x00\x70\x00\x1f\x00\x26\x00\x92\x02\x10\x02\x46\x04\x93\x02\x94\x02\x8d\x02\xc0\x02\x1f\x02\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x18\x02\x25\x04\x73\x00\x19\x02\x7e\x00\x1a\x02\x7f\x00\x26\x00\xdb\x01\xdc\x01\xdd\x01\x81\x00\x36\x03\xc1\x02\x20\x02\x82\x00\x1b\x02\x83\x00\x84\x00\x1c\x02\x8d\x02\x0d\x02\xf1\x00\x5d\x04\x11\x02\xaf\x03\x7e\x00\xdc\x00\x7f\x00\xde\x00\xb0\x03\x8c\x01\x38\x03\x81\x00\xeb\x00\xe1\x02\x77\x04\x82\x00\x26\x02\x83\x00\x84\x00\xa9\x00\x9e\x01\x8d\x01\x26\x02\x23\x00\x2b\x04\x81\x03\x24\x00\x25\x00\x39\x04\x20\x00\x21\x00\x22\x00\x9c\x04\xe2\x00\x01\x04\x20\x00\x21\x00\x22\x00\x26\x02\x75\x04\xf5\x00\x76\x04\xeb\x00\x65\x00\x70\x00\x66\x00\x70\x00\x67\x00\x76\x04\x68\x00\x69\x00\x20\x00\x21\x00\x22\x00\x6a\x00\x6b\x00\x24\x00\x25\x00\x82\x03\x6e\x00\x72\x02\xb5\x00\x74\x03\x73\x00\xc6\x02\x73\x00\x05\x02\xcd\x01\xeb\x00\x02\x04\xf5\x00\xce\x01\x4a\x04\x4b\x04\xc3\x02\x4c\x04\xcf\x01\xc1\x00\x85\x00\x68\x00\x69\x00\x06\x02\x24\x00\x25\x00\x6a\x00\x6b\x00\x86\x00\x10\x00\x87\x00\x12\x00\x4e\x04\x4f\x04\x13\x00\x14\x00\x15\x00\x2c\x03\xf5\x00\x16\x00\x17\x00\x18\x00\x85\x00\x09\x04\xc2\x01\x29\x03\x0d\x02\xf1\x00\x19\x00\x0e\x02\x86\x00\x10\x00\x87\x00\x12\x00\xfd\x00\x1d\x04\x13\x00\x14\x00\x15\x00\x36\x03\xdb\x00\x16\x00\x17\x00\x18\x00\xc8\x02\x7e\x00\x6c\x03\x7f\x00\xeb\x00\x2e\x04\x19\x00\x80\x00\x81\x00\xc9\x02\x2c\x00\x2d\x00\x82\x00\xde\x01\x83\x00\x84\x00\xa9\x00\xaa\x00\x24\x00\x25\x00\xd5\x00\xd6\x00\x1c\x00\xe1\x00\x12\x03\x1d\x00\x1e\x00\x10\x02\x1f\x00\xab\x00\x92\x02\xfd\x00\xf5\x00\xac\x00\xbf\x03\xe4\x00\xe5\x00\xe6\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x21\x00\xad\x00\x1d\x00\x1e\x00\xae\x00\x1f\x00\x26\x00\x37\x04\x6f\x00\xc2\x01\x2c\x00\x2d\x00\x29\x03\x19\x03\x03\x04\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x70\x00\xfd\x00\x23\x00\x71\x00\x25\x00\xab\x00\xee\x00\x26\x00\xe3\x01\xac\x00\xe4\x01\x2c\x00\x2d\x00\x7e\x00\xef\x00\x7f\x00\x30\x04\x6f\x00\x72\x00\x73\x00\xe2\x01\x21\x00\xad\x00\x86\x02\x87\x02\xae\x00\x83\x00\x84\x00\x4e\x03\xc5\x01\x70\x00\x85\x00\x23\x00\x71\x00\x25\x00\xe3\x01\xfd\x00\xe4\x01\x33\x04\x86\x00\x10\x00\x87\x00\x12\x00\xeb\x00\x97\x03\x13\x00\x14\x00\x15\x00\x72\x00\x73\x00\x16\x00\x17\x00\x18\x00\x7e\x00\x5c\x02\x7f\x00\xbd\x01\x24\x00\x25\x00\x19\x00\x4a\x02\x02\x02\x2c\x00\x2d\x00\x1f\x04\xfd\x00\x83\x00\x84\x00\xe5\x03\x75\x00\xc0\x01\xf5\x00\x76\x00\x20\x00\x21\x00\x22\x00\x1c\x04\x46\x02\xfc\x00\x10\x02\xb2\x03\xc1\x00\xab\x00\x68\x00\x69\x00\xfd\x00\xac\x00\x11\x02\x6a\x00\x6b\x00\xac\x02\x57\x04\x58\x04\x7e\x00\x00\x01\x7f\x00\xad\x02\x1c\x00\x21\x00\xad\x00\x1d\x00\x1e\x00\xae\x00\x1f\x00\x1f\x04\xc2\x01\x83\x00\x84\x00\x10\x03\x85\x00\xe9\x03\xdb\x00\xdc\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x88\x02\x07\x02\x7a\x00\xe8\x03\x13\x00\xde\x01\x15\x00\x26\x00\xc0\x01\x16\x00\x17\x00\x18\x00\xd5\x00\xd6\x00\x7e\x00\xec\x03\x7f\x00\x70\x00\x19\x00\xe1\x00\xc8\x01\x25\x00\x20\x04\x21\x04\x22\x04\xae\x03\xc2\x01\x83\x00\x84\x00\xc3\x01\x85\x00\xe4\x00\xe5\x00\xe6\x00\x01\x01\xfa\x01\x73\x00\x85\x04\x86\x04\x87\x04\x88\x02\xd5\x00\xd6\x00\x02\x01\x13\x00\xd8\x00\x15\x00\xd9\x00\x30\x03\x16\x00\x17\x00\x18\x00\x4b\x02\x52\x04\x89\x02\x8a\x02\x1c\x00\x53\x04\x19\x00\x1d\x00\x8b\x02\xb4\x03\x1f\x00\x6b\x04\x22\x04\x20\x00\x21\x00\x22\x00\xb5\x03\xe3\x01\x85\x00\xe4\x01\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x18\x02\x4a\x03\x88\x02\x19\x02\x49\x03\x46\x03\x13\x00\x26\x00\x15\x00\xa3\x00\x08\x02\x16\x00\x17\x00\x18\x00\xa4\x00\xa5\x00\x1b\x02\x09\x02\x1c\x00\x1c\x02\x19\x00\x1d\x00\x1e\x00\xb4\x03\x1f\x00\x70\x00\x4e\x03\x1d\x03\x71\x00\x25\x00\xb5\x03\x1e\x03\x85\x00\x50\x03\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x5a\x03\x4d\x02\x88\x02\x72\x00\x73\x00\x6e\x03\x13\x00\x26\x00\x15\x00\x28\x04\x29\x04\x16\x00\x17\x00\x18\x00\x20\x00\x21\x00\x22\x00\x71\x03\x1c\x00\x9d\x01\x19\x00\x1d\x00\x1e\x00\x8f\x03\x1f\x00\x5b\x02\x9c\x01\xa5\x00\x5c\x02\xb6\x03\xb7\x03\xb8\x03\xb9\x03\x6a\x02\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x91\x03\x50\x02\xba\x03\x60\x04\x25\x01\x26\x01\x13\x00\x26\x00\x15\x00\x99\x03\x9a\x03\x16\x00\x17\x00\x18\x00\x20\x00\x21\x00\x22\x00\x74\x02\x1c\x00\x9b\x01\x19\x00\x1d\x00\x1e\x00\xba\x01\x1f\x00\x27\x03\x9c\x01\xa5\x00\xdb\x00\xdc\x00\x47\x04\xb8\x03\xb9\x03\x6c\x02\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x21\x00\xa8\x03\xba\x03\x8a\x04\xcb\x03\x69\x00\x13\x00\x26\x00\x15\x00\x6a\x00\x6b\x00\x16\x00\x17\x00\x18\x00\x1e\x02\xe1\x00\x75\x02\xba\x01\x1c\x00\x17\x03\x19\x00\x1d\x00\x1e\x00\x80\x02\x1f\x00\x80\x03\x81\x03\xe4\x00\xe5\x00\xe6\x00\x9e\x02\x9f\x02\xa0\x02\xa1\x02\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x82\x02\x21\x00\xc2\x02\x83\x02\x65\x00\xc3\x02\x66\x00\x26\x00\x67\x00\x8e\x02\x68\x00\x69\x00\xbd\x03\xbe\x03\x5a\x01\x6a\x00\x6b\x00\x5b\x01\x1c\x00\x82\x03\x6e\x00\x1d\x00\x1e\x00\x91\x02\x1f\x00\x84\x03\x81\x03\x20\x00\x21\x00\x22\x00\xa9\x02\x24\x00\x25\x00\x95\x02\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x76\x04\x25\x01\x26\x01\x9c\x02\x65\x00\x25\x03\x66\x00\x26\x00\x67\x00\xa1\x02\x68\x00\x69\x00\xb0\x01\xc8\x03\xb0\x02\x6a\x00\x6b\x00\xf3\x03\xf4\x03\x82\x03\x6e\x00\xa1\x01\x3f\x03\x26\x04\x46\x02\x20\x00\x21\x00\x22\x00\xc1\x00\xaa\x02\x68\x00\x69\x00\xc3\x03\x46\x02\xbc\x02\x6a\x00\x6b\x00\xc1\x00\xbe\x02\x68\x00\x69\x00\xee\x03\xef\x03\xf0\x03\x6a\x00\x6b\x00\xd4\x02\xae\x01\xaf\x01\x45\x02\x46\x02\x67\x02\xa1\x01\x68\x02\xc1\x00\xb0\x01\x68\x00\x69\x00\x69\x02\xa1\x01\x68\x02\x6a\x00\x6b\x00\x70\x00\x8c\x03\x8d\x03\x71\x00\x25\x00\x20\x00\x21\x00\x22\x00\xd6\x02\x65\x00\xe6\x01\x66\x00\xba\x01\x67\x00\xbb\x01\x68\x00\x69\x00\xdc\x02\x72\x00\x73\x00\x6a\x00\x6b\x00\x7a\x03\x7b\x03\xfd\x00\x6e\x00\xfe\x00\xff\x00\xfa\x00\xf2\x03\x65\x00\xf3\x03\xef\x00\x6f\x00\xc1\x00\xfd\x02\x68\x00\xf0\x00\xf1\x00\xf2\x00\xf3\x00\x6a\x00\x6b\x00\x83\x01\xa4\x01\xd8\x00\x70\x00\xd9\x00\x23\x00\x71\x00\x25\x00\xe7\x01\x4a\x00\xe0\x01\x4b\x00\x4c\x00\x65\x00\xec\x01\x66\x00\xe8\x01\x67\x00\x36\x02\x68\x00\x69\x00\x72\x00\x73\x00\x2d\x02\x6a\x00\x6b\x00\x35\x02\x66\x04\x6d\x00\x6e\x00\x6f\x00\x63\x02\x25\x01\x26\x01\x55\x01\x6b\x00\xae\x02\xaf\x02\xb0\x02\xe6\x03\xb3\x02\xb4\x02\xb5\x02\x70\x00\x51\x02\x23\x00\x71\x00\x25\x00\x3e\x02\x16\xfe\x00\x01\x16\xfe\x59\x02\x65\x00\x67\x04\x66\x00\x28\xfe\x67\x00\x28\xfe\x68\x00\x69\x00\x72\x00\x73\x00\x70\x00\x6a\x00\x6b\x00\x71\x00\x25\x00\xfd\x00\x6e\x00\xfe\x00\xff\x00\x70\x00\xfb\x01\xba\x00\x71\x00\x25\x00\xc1\x00\xbc\x00\x68\x00\x69\x00\x72\x00\x73\x00\x5f\x02\x6a\x00\x6b\x00\x27\x02\x28\x02\x29\x02\x70\x00\x72\x00\x73\x00\x71\x00\x25\x00\xb2\x00\xb3\x00\xb4\x00\xb5\x00\x6f\x00\x2c\x02\x28\x02\x29\x02\xe7\x01\x4a\x00\x3c\x04\x4b\x00\x4c\x00\x72\x00\x73\x00\x09\x02\xe8\x01\x70\x00\xbe\x00\x23\x00\x71\x00\x25\x00\xd5\x00\xd6\x00\x02\x01\x6f\x00\xd8\x00\xeb\x00\xd9\x00\x24\x01\x25\x01\x26\x01\x7d\x03\x7e\x03\xeb\x00\x72\x00\x73\x00\x2a\xfe\x70\x00\x2a\xfe\x23\x00\xf4\x00\x25\x00\x00\x01\xb6\x02\x21\x00\x22\x00\xbf\x00\x24\x00\x25\x00\xe8\x00\x6f\x00\x7f\x04\xcb\x03\x69\x00\xf5\x00\x72\x00\x73\x00\x6a\x00\x6b\x00\xea\x00\x2b\xfe\xb7\x02\x2b\xfe\x70\x00\x22\x01\x23\x00\x71\x00\x25\x00\x65\x00\x32\xfe\x66\x00\x32\xfe\x67\x00\x70\x00\x68\x00\x69\x00\x71\x00\x25\x00\x5d\x01\x6a\x00\x6b\x00\x72\x00\x73\x00\xf8\x00\x6e\x00\x9b\x04\xe2\x03\xde\x01\xde\x02\xdf\x02\x6f\x00\x56\x01\x73\x00\x97\x04\xd5\x00\xd6\x00\x83\x01\x33\xfe\xd8\x00\x33\xfe\xd9\x00\x01\x01\x34\xfe\x70\x00\x34\xfe\x23\x00\x71\x00\x25\x00\xd5\x00\xd6\x00\x02\x01\xe3\x03\xd8\x00\x65\x00\xd9\x00\x66\x00\x35\xfe\x67\x00\x35\xfe\x68\x00\x69\x00\x72\x00\x73\x00\x70\x00\x6a\x00\x6b\x00\x71\x00\x25\x00\xf8\x00\x6e\x00\x36\xfe\xf9\x00\x36\xfe\x65\x00\xab\x00\x66\x00\x92\x04\x67\x00\xac\x00\x68\x00\x69\x00\x72\x00\x73\x00\x94\x04\x6a\x00\x6b\x00\x95\x04\x7d\x04\x6d\x00\x6e\x00\x21\x00\xad\x00\x8a\x04\x65\x00\xae\x00\x66\x00\x37\xfe\x67\x00\x37\xfe\x68\x00\x69\x00\x73\x02\xb4\x00\xb5\x00\x6a\x00\x6b\x00\x84\x04\x63\x04\x6d\x00\x6e\x00\x99\x01\x9a\x01\x9b\x01\x65\x00\x38\xfe\x66\x00\x38\xfe\x67\x00\x7a\x04\x68\x00\x69\x00\x39\xfe\x7b\x04\x39\xfe\x6a\x00\x6b\x00\x28\x00\x6f\x04\x6d\x00\x6e\x00\xc0\x02\x70\x00\x0e\x04\xb5\x02\x71\x00\x25\x00\x3a\xfe\xfa\x00\x3a\xfe\x82\x04\x6f\x00\xc7\x01\x21\x00\x22\x00\x7f\x04\xc8\x01\x25\x00\x9c\x02\x81\x04\x72\x00\x73\x00\x62\x04\x3b\xfe\x70\x00\x3b\xfe\x23\x00\x71\x00\x25\x00\x3c\xfe\x43\xfe\x3c\xfe\x43\xfe\x65\x00\x44\xfe\x66\x00\x44\xfe\x67\x00\x8d\x02\x68\x00\x69\x00\x22\x02\x72\x00\x73\x00\x6a\x00\x6b\x00\x49\xff\x44\x04\x6d\x00\x6e\x00\x24\x02\x21\x00\x22\x00\x6f\x04\xc8\x01\x25\x00\x6f\x00\x2f\x02\x30\x02\x31\x02\x32\x02\x33\x02\x34\x02\x35\x02\xbe\x02\xb6\x00\xb3\x00\xb4\x00\xb5\x00\x70\x00\x73\x04\x23\x00\x71\x00\x25\x00\x4d\x00\xb0\xfe\x6f\x00\xb0\xfe\xc0\x02\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x74\x04\x57\x00\x72\x00\x73\x00\x70\x00\xeb\x00\x23\x00\x71\x00\x25\x00\x3b\x04\xc1\x01\x6f\x00\xc2\x01\x5c\x00\x3e\x04\xb6\x02\x21\x00\x22\x00\x8d\x02\x24\x00\x25\x00\xab\x00\x72\x00\x73\x00\x70\x00\xac\x00\x23\x00\x71\x00\x25\x00\xfa\x02\x26\x01\x6f\x00\x62\x00\xb7\x02\x2b\x02\x4a\x04\x2c\x02\x21\x00\xad\x00\xfb\x02\xfc\x02\xae\x00\x72\x00\x73\x00\x70\x00\x55\x04\x23\x00\x71\x00\x25\x00\x1e\x03\x97\x01\xa1\x01\x21\x03\x65\x00\x56\x04\x66\x00\xd0\xfe\x67\x00\x57\x04\x68\x00\x69\x00\x5e\x04\x72\x00\x73\x00\x6a\x00\x6b\x00\x5a\x04\x48\x04\x6d\x00\x6e\x00\xa1\x01\xa2\x01\x62\x02\xa5\x00\x5c\x04\x65\x00\x8d\x02\x66\x00\x2f\x03\x67\x00\x6f\x00\x68\x00\x69\x00\xb7\x00\xb8\x00\xa3\x02\x6a\x00\x6b\x00\xf2\x03\x2a\x04\x6d\x00\x6e\x00\x07\x04\x70\x00\x08\x04\x23\x00\x71\x00\x25\x00\x96\x01\x97\x01\x22\x02\x23\x02\x65\x00\x11\x04\x66\x00\x06\x04\x67\x00\x09\x04\x68\x00\x69\x00\x28\x00\x72\x00\x73\x00\x6a\x00\x6b\x00\xab\x00\x66\x03\x6d\x00\x6e\x00\xac\x00\x61\x02\x62\x02\x65\x00\x09\x02\x66\x00\x0b\x04\x67\x00\x14\x04\x68\x00\x69\x00\x2c\x00\x21\x00\xad\x00\x6a\x00\x6b\x00\xae\x00\x7f\x03\x6d\x00\x6e\x00\x61\x02\x62\x02\x1f\x04\x65\x00\x7a\x03\x66\x00\x2d\x04\x67\x00\x1c\x04\x68\x00\x69\x00\x9c\x03\x9d\x03\x30\x04\x6a\x00\x6b\x00\x32\x04\x7e\x02\x6d\x00\x6e\x00\x2a\x00\x2b\x00\x33\x04\x65\x00\x8d\x02\x66\x00\x99\x03\x67\x00\x37\x04\x68\x00\x69\x00\xde\x03\xdf\x03\x39\x04\x6a\x00\x6b\x00\xa8\x03\xab\x01\x6d\x00\x6e\x00\xb2\x02\xb3\x02\xb4\x02\xb5\x02\x3a\x03\x3b\x03\x6f\x00\xa5\x03\x4d\x00\x4e\x00\x4f\x00\x50\x00\xa6\x03\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x70\x00\x57\x00\x23\x00\x71\x00\x25\x00\x6e\x02\x6f\x02\x22\x02\x23\x02\x6f\x00\x5a\x00\x5b\x00\xa7\x03\x5c\x00\xca\x02\xc9\x02\x22\x02\x23\x02\x72\x00\x73\x00\x61\x02\x62\x02\x70\x00\xab\x03\x23\x00\x71\x00\x25\x00\xac\x03\xad\x03\xae\x03\x8d\x02\x65\x00\x62\x00\x66\x00\xb4\x03\x67\x00\x6f\x00\x68\x00\x69\x00\xc5\x03\x72\x00\x73\x00\x6a\x00\x6b\x00\xc6\x03\x1e\x02\x6d\x00\x6e\x00\xef\x01\x70\x00\xb2\x02\x23\x00\x71\x00\x25\x00\xcd\x03\xca\x03\x6f\x00\xdb\x03\xdc\x03\xdd\x03\xe8\x03\xc0\x02\xeb\x03\xee\x01\xfa\x03\x2c\x00\x2f\x03\x72\x00\x73\x00\x70\x00\xeb\x00\x23\x00\x71\x00\x25\x00\xc1\x01\x32\x03\x6f\x00\x38\x03\x8d\x02\x30\x03\xb6\x02\x21\x00\x22\x00\x3c\x03\x24\x00\x25\x00\x3d\x03\x72\x00\x73\x00\x70\x00\x43\x03\x23\x00\x71\x00\x25\x00\x15\xff\x44\x03\x6f\x00\x29\xff\xb7\x02\x46\x03\x48\x03\x49\x03\xbe\x02\xc0\x02\x50\x03\xc0\x02\xc8\x00\x72\x00\x73\x00\x70\x00\x66\x03\x23\x00\x71\x00\x25\x00\xc8\x00\x6b\x03\x73\x03\x74\x03\x65\x00\x71\x03\x66\x00\x6b\x01\x67\x00\x7a\x03\x68\x00\x69\x00\x84\x03\x72\x00\x73\x00\x6a\x00\x6b\x00\x8b\x03\x23\x02\x6d\x00\x6e\x00\x8c\x03\xa7\x00\x8f\x03\x65\x00\x97\x03\x66\x00\x70\x02\x67\x00\xba\x00\x68\x00\x69\x00\x71\x02\x78\x02\x6c\x02\x6a\x00\x6b\x00\x79\x02\x5e\x01\x6d\x00\x6e\x00\x77\x02\x7a\x02\x7b\x02\x65\x00\x7c\x02\x66\x00\x7d\x02\x67\x00\x6f\x00\x68\x00\x69\x00\x82\x02\x7e\x02\x22\x02\x6a\x00\x6b\x00\x85\x02\x6c\x00\x6d\x00\x6e\x00\x86\x02\x70\x00\x90\x02\x23\x00\x71\x00\x25\x00\x65\x00\x91\x02\x66\x00\x1a\xff\x67\x00\x22\x02\x68\x00\x69\x00\xae\x02\x97\x02\xcc\x02\x6a\x00\x6b\x00\x72\x00\x73\x00\x4d\x04\x6e\x00\x2c\x02\xb2\x02\xa3\x02\x65\x00\xa3\x02\x66\x00\xc0\x02\x67\x00\xbe\x02\x68\x00\x69\x00\xc5\x02\xd0\x02\xd8\x02\x6a\x00\x6b\x00\xd9\x02\xd2\x02\xc5\x02\x6e\x00\xdc\x02\xde\x02\xd2\x01\x65\x00\xe5\x02\x66\x00\xb9\x01\x67\x00\x0f\x03\x68\x00\x69\x00\x10\x03\x21\x03\x23\x03\x6a\x00\x6b\x00\x2c\x00\x1c\x03\xcc\x02\x6e\x00\xa4\x01\x8a\xfd\x8c\xfd\x65\x00\x8b\xfd\x66\x00\x8e\xfd\x67\x00\x91\xfd\x68\x00\x69\x00\xb8\x01\x6f\x00\xb9\x01\x6a\x00\x6b\x00\x28\x00\xba\x01\xce\x02\x6e\x00\xcd\x02\xb3\x02\xb4\x02\xb5\x02\xcb\x01\x70\x00\xcc\x01\x23\x00\x71\x00\x25\x00\x9b\x02\xd2\x01\x6f\x00\xd8\x01\xd9\x01\xda\x01\xc2\x01\x9c\x02\xd7\x01\xdb\x01\xf8\x00\xe2\x01\xe0\x01\x72\x00\x73\x00\x70\x00\xe9\x01\x23\x00\x71\x00\x25\x00\xee\x01\xef\x01\x6f\x00\xf8\x01\xf1\x01\xf6\x01\xff\x01\xf7\x01\x00\x02\x01\x02\x03\x02\x04\x02\x0b\x02\x72\x00\x73\x00\x70\x00\x2d\xff\x23\x00\x71\x00\x25\x00\x65\x00\x0c\x02\x66\x00\x0d\x02\x67\x00\x6f\x00\x68\x00\x69\x00\x38\x02\x40\x02\xef\x01\x6a\x00\x6b\x00\x72\x00\x73\x00\xf9\x01\x6e\x00\x4a\x02\x70\x00\xbc\x00\x23\x00\x71\x00\x25\x00\x4d\x00\x57\x02\x6f\x00\x58\x02\x2c\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x67\x02\x57\x00\x72\x00\x73\x00\x70\x00\xeb\x00\x23\x00\x71\x00\x25\x00\xa9\x00\x5e\x02\x6f\x00\xba\x00\x5c\x00\xff\xff\xb6\x02\x21\x00\x22\x00\xca\x00\x24\x00\x25\x00\xbc\x00\x72\x00\x73\x00\x70\x00\xc8\x00\x23\x00\x71\x00\x25\x00\xcb\x00\xff\xff\x6f\x00\x62\x00\xb7\x02\xcc\x00\xea\x00\xff\xff\x06\x01\x17\x01\x12\x01\x13\x01\x14\x01\x72\x00\x73\x00\x70\x00\xc8\x00\x23\x00\x71\x00\x25\x00\x65\x00\xff\xff\x66\x00\x8e\x01\x67\x00\x17\x01\x68\x00\x69\x00\x8f\x01\xec\xfd\xff\xff\x6a\x00\x6b\x00\x72\x00\x73\x00\x38\x02\x6e\x00\xff\xff\x00\x00\xa9\x00\x65\x00\x00\x00\x66\x00\x00\x00\x67\x00\x00\x00\x68\x00\x69\x00\x00\x00\x00\x00\x00\x00\x6a\x00\x6b\x00\x00\x00\x00\x00\x3d\x02\x6e\x00\x00\x00\x00\x00\x00\x00\x65\x00\x00\x00\x66\x00\x00\x00\x67\x00\x00\x00\x68\x00\x69\x00\x00\x00\x6f\x00\x00\x00\x6a\x00\x6b\x00\x00\x00\x00\x00\x4c\x02\x6e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x70\x00\x00\x00\x23\x00\x71\x00\x25\x00\x65\x00\x00\x00\x66\x00\x00\x00\x67\x00\x00\x00\x68\x00\x69\x00\x00\x00\x00\x00\x00\x00\x6a\x00\x6b\x00\x72\x00\x73\x00\xc8\x00\x6e\x00\x00\x00\x00\x00\x00\x00\x65\x00\x00\x00\x66\x00\x00\x00\x67\x00\x00\x00\x68\x00\x69\x00\x00\x00\x00\x00\x00\x00\x6a\x00\x6b\x00\x00\x00\x00\x00\xcc\x00\x6e\x00\x00\x00\x00\x00\x00\x00\x65\x00\x00\x00\x66\x00\x00\x00\x67\x00\x00\x00\x68\x00\x69\x00\x00\x00\x00\x00\x00\x00\x6a\x00\x6b\x00\x00\x00\x00\x00\xe6\x00\x6e\x00\x00\x00\x00\x00\x00\x00\x65\x00\x00\x00\x66\x00\x00\x00\x67\x00\x6f\x00\x68\x00\x69\x00\x00\x00\x00\x00\x00\x00\x6a\x00\x6b\x00\x00\x00\x00\x00\x06\x01\x6e\x00\x00\x00\x70\x00\x00\x00\x23\x00\x71\x00\x25\x00\xfc\x01\x00\x00\x6f\x00\x00\x00\xc1\x00\x00\x00\x68\x00\x69\x00\x00\x00\x00\x00\x00\x00\x6a\x00\x6b\x00\x72\x00\x73\x00\x70\x00\x00\x00\x23\x00\x71\x00\x25\x00\xfd\x01\x00\x00\x6f\x00\x00\x00\xc1\x00\x00\x00\x68\x00\x69\x00\x00\x00\x00\x00\x00\x00\x6a\x00\x6b\x00\x72\x00\x73\x00\x70\x00\x00\x00\x23\x00\x71\x00\x25\x00\x65\x00\x00\x00\x47\x02\x00\x00\xc1\x00\x00\x00\x68\x00\x69\x00\x00\x00\x6f\x00\x00\x00\x6a\x00\x6b\x00\x72\x00\x73\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x70\x00\x00\x00\x23\x00\x71\x00\x25\x00\xc0\x00\x00\x00\x6f\x00\x00\x00\xc1\x00\x00\x00\x68\x00\x69\x00\x00\x00\x00\x00\x00\x00\x6a\x00\x6b\x00\x72\x00\x73\x00\x70\x00\x00\x00\x23\x00\x71\x00\x25\x00\x00\x00\x12\x04\x6f\x00\x68\x00\x69\x00\x00\x00\x00\x00\x00\x00\x6a\x00\x6b\x00\x00\x00\x00\x00\x00\x00\x72\x00\x73\x00\x70\x00\x00\x00\x23\x00\x71\x00\x25\x00\x5b\x03\x15\x04\x6f\x00\x68\x00\x69\x00\x00\x00\x00\x00\x00\x00\x6a\x00\x6b\x00\x00\x00\x00\x00\x00\x00\x72\x00\x73\x00\x70\x00\x00\x00\x23\x00\x71\x00\x25\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x00\x73\x00\x16\x04\x5b\x03\x5c\x03\x70\x00\x68\x00\x69\x00\x71\x00\x25\x00\x00\x00\x6a\x00\x6b\x00\x07\x01\x69\x00\x00\x00\x5d\x03\x00\x00\x6a\x00\x6b\x00\x08\x01\x00\x00\x00\x00\x72\x00\x73\x00\x70\x00\x00\x00\x00\x00\x71\x00\x25\x00\x00\x00\x00\x00\x6f\x00\x00\x00\x5e\x03\x5f\x03\x60\x03\x00\x00\x61\x03\x00\x00\x00\x00\x00\x00\x00\x00\x72\x00\x73\x00\x70\x00\x00\x00\x23\x00\x71\x00\x25\x00\x5b\x03\xd3\x03\x00\x00\x68\x00\x69\x00\x00\x00\x00\x00\x00\x00\x6a\x00\x6b\x00\x64\x04\xcb\x03\x69\x00\x72\x00\x73\x00\x00\x00\x6a\x00\x6b\x00\x00\x00\x00\x00\x00\x00\x7f\x02\x70\x00\x68\x00\x69\x00\x71\x00\x25\x00\x00\x00\x6a\x00\x6b\x00\x00\x00\xd4\x03\x5f\x03\x60\x03\x00\x00\x61\x03\x00\x00\x00\x00\x00\x00\x70\x00\x72\x00\x73\x00\x71\x00\x25\x00\xf8\x01\x00\x00\x68\x00\x69\x00\x00\x00\x00\x00\x00\x00\x6a\x00\x6b\x00\x00\x00\xcf\x03\xcb\x03\x69\x00\x72\x00\x73\x00\x70\x00\x6a\x00\x6b\x00\x71\x00\x25\x00\x00\x00\x65\x04\xcb\x03\x69\x00\x00\x00\x00\x00\x00\x00\x6a\x00\x6b\x00\x00\x00\x50\x04\xcb\x03\x69\x00\x72\x00\x73\x00\x00\x00\x6a\x00\x6b\x00\x00\x00\x00\x00\x62\x03\x00\x00\xd0\x03\x11\x04\xcb\x03\x69\x00\x00\x00\x00\x00\x00\x00\x6a\x00\x6b\x00\x70\x00\x09\x01\x00\x00\xf4\x00\x25\x00\xd4\x00\x00\x00\x70\x00\x00\x00\x00\x00\x71\x00\x25\x00\xd5\x00\xd6\x00\x14\x04\xcb\x03\x69\x00\x00\x00\x72\x00\x73\x00\x6a\x00\x6b\x00\x00\x00\x00\x00\x00\x00\x72\x00\x73\x00\x0a\x01\x0b\x01\x00\x00\x00\x00\xca\x03\xcb\x03\x69\x00\x00\x00\x00\x00\x62\x03\x6a\x00\x6b\x00\x07\x01\x69\x00\x00\x00\x00\x00\x00\x00\x6a\x00\x6b\x00\x00\x00\x70\x00\x00\x00\x00\x00\xf4\x00\x25\x00\x00\x00\x00\x00\x00\x00\x70\x00\x00\x00\x00\x00\x71\x00\x25\x00\x00\x00\xd9\x03\xb4\x02\xb5\x02\x00\x00\x72\x00\x73\x00\x70\x00\x0f\x04\xb5\x02\x71\x00\x25\x00\xac\x01\x72\x00\x73\x00\xad\x01\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x72\x00\x73\x00\x00\x00\x00\x00\x19\x00\x70\x00\x00\x00\x00\x00\x71\x00\x25\x00\x00\x00\x00\x00\x00\x00\x00\x00\x70\x00\x00\x00\x00\x00\x71\x00\x25\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x00\x73\x00\x70\x00\x00\x00\x00\x00\x71\x00\x25\x00\x00\x00\x00\x00\x72\x00\x73\x00\x70\x00\x00\x00\x00\x00\x71\x00\x25\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x72\x00\x73\x00\x1d\x00\x1e\x00\x70\x00\x1f\x00\x00\x00\x71\x00\x25\x00\x72\x00\x73\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xeb\x00\x72\x00\x73\x00\x00\x00\x00\x00\x70\x00\xeb\x00\x26\x00\x71\x00\x25\x00\xb6\x02\x21\x00\x22\x00\x00\x00\x24\x00\x25\x00\xb6\x02\x21\x00\x22\x00\x00\x00\x24\x00\x25\x00\x70\x00\x72\x00\x73\x00\x71\x00\x25\x00\x00\x00\xb7\x02\x70\x00\x00\x00\x00\x00\x71\x00\x25\x00\xb7\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x00\x73\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x00\x73\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x8d\x02\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x89\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x8a\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x8b\x00\x8c\x00\x8d\x00\x8e\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x96\x03\x00\x00\x8f\x00\x90\x00\x91\x00\x92\x00\x93\x00\x5f\x00\x00\x00\x94\x00\x00\x00\x60\x00\x00\x00\x00\x00\x95\x00\x00\x00\x00\x00\x62\x00\x96\x00\x97\x00\x98\x00\x99\x00\x00\x00\x9a\x00\x63\x00\x64\x00\x65\x00\x9b\x00\x9c\x00\x00\x00\x00\x00\x00\x00\x9d\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x89\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x8a\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x8b\x00\x8c\x00\x8d\x00\x8e\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x96\x03\x00\x00\x8f\x00\x90\x00\x91\x00\x92\x00\x93\x00\x5f\x00\x00\x00\x94\x00\x00\x00\x60\x00\x00\x00\x00\x00\x95\x00\x00\x00\x00\x00\x62\x00\x96\x00\x97\x00\x98\x00\x99\x00\x00\x00\x9a\x00\x63\x00\x64\x00\x65\x00\x9b\x00\x9c\x00\x00\x00\x00\x00\x00\x00\x9d\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x89\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x8a\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x8b\x00\x8c\x00\x8d\x00\x8e\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x8f\x00\x90\x00\x91\x00\x92\x00\x93\x00\x5f\x00\xd0\x00\x94\x00\x00\x00\x60\x00\x00\x00\x00\x00\x95\x00\x00\x00\x00\x00\x62\x00\x96\x00\x97\x00\x98\x00\x99\x00\x00\x00\x9a\x00\x63\x00\x64\x00\x65\x00\x9b\x00\x9c\x00\x28\x00\x00\x00\x00\x00\x9d\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x40\xff\x00\x00\x00\x00\x75\x00\x00\x00\x28\x00\x76\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x77\x00\x00\x00\x78\x00\x75\x00\x00\x00\x00\x00\x76\x00\x00\x00\x79\x00\x00\x00\x00\x00\x00\x00\x40\xff\x77\x00\x00\x00\x78\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00\x00\x00\xd6\x03\x11\x01\x00\x00\xd7\x03\x58\x03\xb4\x02\xb5\x02\xd1\x00\x00\x00\x00\x00\xd2\x00\x00\x00\x7c\x03\x00\x00\xd4\x00\x00\x00\x69\x04\x00\x00\x00\x00\x7a\x00\x00\x00\xd5\x00\xd6\x00\xd7\x00\x00\x00\xd8\x00\x00\x00\xd9\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x5c\x00\x57\x00\x00\x00\x00\x00\x28\x00\x00\x00\x00\x00\x00\x00\x2c\x00\x2d\x00\x00\x00\x7e\x00\x40\xff\x5c\x00\x00\x00\x75\x00\x00\x00\x00\x00\x76\x00\x62\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x77\x00\x00\x00\x78\x00\x00\x00\x00\x00\x00\x00\xd8\x03\x62\x00\x79\x00\x00\x00\xeb\x00\x00\x00\x40\xff\x00\x00\xdb\x00\xdc\x00\xdd\x00\xde\x00\x00\x00\x6a\x04\xb6\x02\x21\x00\x22\x00\x28\x00\x24\x00\x25\x00\xd7\x03\x2c\x00\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x00\x00\x00\x00\x76\x00\xb7\x02\x00\x00\xe0\x00\xe1\x00\x7a\x00\xe2\x00\x77\x00\x00\x00\x78\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00\xe4\x00\xe5\x00\xe6\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x28\x00\x57\x00\x00\x00\x64\x03\x2c\x00\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x00\x00\x5c\x00\x76\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x77\x00\x00\x00\x78\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00\x00\x00\x00\x00\x62\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\xd0\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\xd8\x03\x57\x00\x00\x00\xd2\x03\x00\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\xdc\x00\x00\x00\xde\x00\x75\x00\x5c\x00\x00\x00\x76\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x28\x00\x29\x00\x10\x02\x13\x02\x2c\x00\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x11\x02\x00\x00\x62\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\xe2\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x65\x03\x57\x00\x00\x00\x00\x00\x28\x00\x00\x00\x00\x00\x00\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x00\x00\x5c\x00\x00\x00\x75\x00\x00\x00\x00\x00\x76\x00\x00\x00\x7a\x00\xd1\x00\x00\x00\x00\x00\xd2\x00\xf7\x00\xd3\x00\x78\x00\xd4\x00\x00\x00\x00\x00\x00\x00\x62\x00\x79\x00\x00\x00\xd5\x00\xd6\x00\xd7\x00\x00\x00\xd8\x00\x00\x00\xd9\x00\x00\x00\x00\x00\xd3\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x7a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x00\x00\x62\x00\x00\x00\x00\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x5c\x00\x76\x01\xdc\x00\xdd\x00\xde\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x14\x03\x62\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\xc0\x01\x3b\x00\x78\x01\x79\x01\x00\x00\xe2\x00\x65\x03\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x7a\x01\x7b\x01\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x93\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x63\x00\x64\x00\x65\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x76\x01\xdc\x00\xdd\x00\xde\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x1b\x03\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\xc0\x01\x3b\x00\x78\x01\x79\x01\x00\x00\xe2\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x7a\x01\x7b\x01\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x93\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x63\x00\x64\x00\x65\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x76\x01\xdc\x00\xdd\x00\xde\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\xbf\x01\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\xc0\x01\x3b\x00\x78\x01\x79\x01\x00\x00\xe2\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x7a\x01\x7b\x01\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x93\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x63\x00\x64\x00\x65\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x76\x01\xdc\x00\xdd\x00\xde\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\xc7\x01\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\xc0\x01\x3b\x00\x78\x01\x79\x01\x00\x00\xe2\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x7a\x01\x7b\x01\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x93\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x63\x00\x64\x00\x65\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x76\x01\xdc\x00\xdd\x00\xde\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x7e\x01\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\xfd\x00\x3b\x00\x78\x01\x79\x01\x00\x00\xe2\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x7a\x01\x7b\x01\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x93\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x63\x00\x64\x00\x65\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x86\x01\xdc\x00\xdd\x00\xde\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x87\x01\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\xfd\x00\x3b\x00\x78\x01\x88\x01\x00\x00\xe2\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x89\x01\x8a\x01\x8b\x01\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x93\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x63\x00\x64\x00\x65\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x76\x01\xdc\x00\xdd\x00\xde\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x77\x01\x00\x00\x3b\x00\x78\x01\x79\x01\x00\x00\xe2\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x7a\x01\x7b\x01\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x93\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x63\x00\x64\x00\x65\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x76\x01\xdc\x00\xdd\x00\xde\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x78\x01\x79\x01\x00\x00\xe2\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x7a\x01\x7b\x01\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x93\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x00\x00\xdb\x00\xdc\x00\xdd\x00\xde\x00\x62\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x63\x00\x64\x00\x65\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\xe1\x00\x39\x00\xe2\x00\xf4\x01\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\xe4\x00\xe5\x00\xe6\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x3f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5e\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x61\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x63\x00\x64\x00\x65\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x3f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa8\x02\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x00\x00\x0a\x03\x00\x00\x00\x00\x00\x00\x62\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x63\x00\x64\x00\x65\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x8d\x02\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x3f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x93\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x63\x00\x64\x00\x65\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x3f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa8\x02\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x00\x00\x0a\x03\x00\x00\x00\x00\x00\x00\x62\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x63\x00\x64\x00\x65\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x3f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5e\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x61\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x63\x00\x64\x00\x65\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x3f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x93\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x63\x00\x64\x00\x65\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x3f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa8\x02\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x63\x00\x64\x00\x65\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x2e\x03\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x93\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x63\x00\x64\x00\x65\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x3f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa8\x02\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x63\x00\x64\x00\x65\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x20\x01\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x93\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x63\x00\x64\x00\x65\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x3f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x93\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x63\x00\x64\x00\x65\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x5d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x93\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x60\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x63\x00\x64\x00\x65\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x8d\x02\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x8e\x02\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x89\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\xdb\x00\xdc\x00\xdd\x00\xde\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x04\x01\x57\x00\x00\x00\x58\x00\x00\x00\x24\x04\x00\x00\x00\x00\x59\x00\xfd\x00\x5a\x00\x5b\x00\xe1\x00\x5c\x00\xe2\x00\x00\x00\x00\x00\x8f\x00\x90\x00\x91\x00\x05\x01\x00\x00\x5f\x00\x00\x00\xe4\x00\xe5\x00\xe6\x00\x00\x00\x00\x00\x25\x04\x00\x00\x00\x00\x62\x00\x96\x00\x97\x00\x98\x00\x99\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x8e\x02\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x89\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x24\x04\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x00\x00\x00\x00\x8f\x00\x90\x00\x91\x00\x00\x00\x00\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x25\x04\x00\x00\x00\x00\x62\x00\x96\x00\x97\x00\x98\x00\x99\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x8d\x02\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x8e\x02\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x89\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\xbc\x03\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5f\x00\x00\x00\x94\x00\x00\x00\x00\x00\x00\x00\x00\x00\xbd\x03\x00\x00\x00\x00\x62\x00\x96\x00\x97\x00\x98\x00\x99\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x8d\x02\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x8e\x02\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x89\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x00\x00\x00\x00\x8f\x00\x90\x00\x91\x00\x00\x00\x00\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x62\x00\x96\x00\x97\x00\x98\x00\x99\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x8e\x02\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x89\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\xbc\x03\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5f\x00\x00\x00\x94\x00\x00\x00\x00\x00\x00\x00\x00\x00\xbd\x03\x00\x00\x00\x00\x62\x00\x96\x00\x97\x00\x98\x00\x99\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x8e\x02\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x89\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\xd1\x00\x5a\x00\x5b\x00\xd2\x00\x5c\x00\xd3\x00\x00\x00\xd4\x00\x8f\x00\x90\x00\x91\x00\x00\x00\x00\x00\x5f\x00\xd5\x00\xd6\x00\xd7\x00\x00\x00\xd8\x00\x00\x00\xd9\x00\x00\x00\x00\x00\x62\x00\x96\x00\x97\x00\x98\x00\x99\x00\xea\x02\x29\x00\x2a\x00\x2b\x00\xeb\x02\x2d\x00\x2a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\xd1\x00\x3b\x00\x00\x00\xd2\x00\x00\x00\x90\x01\x00\x00\xd4\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\xd5\x00\xd6\x00\xd7\x00\x00\x00\xd8\x00\x00\x00\xd9\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x2b\x01\xec\x02\xed\x02\xee\x02\xef\x02\x00\x00\xf0\x02\xf1\x02\xf2\x02\xf3\x02\xf4\x02\xf5\x02\x00\x00\xf6\x02\x37\x01\x38\x01\x39\x01\x3a\x01\x3b\x01\x3c\x01\x3d\x01\x3e\x01\xf7\x02\xf8\x02\x41\x01\xf9\x02\x43\x01\x44\x01\x45\x01\x46\x01\x47\x01\x48\x01\x49\x01\x4a\x01\x4b\x01\x4c\x01\x4d\x01\x4e\x01\x4f\x01\x50\x01\x51\x01\x52\x01\x53\x01\x54\x01\xfa\x02\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x51\x03\xb5\x02\x00\x00\x52\x03\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x8e\x02\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x3e\x00\x53\x03\x54\x03\xb5\x02\x55\x03\x56\x03\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x58\x00\x00\x00\xeb\x00\x00\x00\x00\x00\x59\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\xb6\x02\x21\x00\x22\x00\x00\x00\x24\x00\x25\x00\x00\x00\x5f\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x00\x00\x62\x00\xb7\x02\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\xeb\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\xde\x01\xb6\x02\x21\x00\x22\x00\x00\x00\x24\x00\x25\x00\xd5\x00\xd6\x00\xd4\x01\x00\x00\x3d\x00\x00\x00\x00\x00\x1a\x01\x00\x00\x00\x00\x00\x00\x00\x00\xb7\x02\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x00\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x00\x00\x1a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x6b\x01\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x00\x00\x00\x00\x1a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x00\x00\x28\x00\x29\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x2e\x00\x2f\x00\x30\x00\x31\x00\x32\x00\x33\x00\x34\x00\x35\x00\x36\x00\x37\x00\x38\x00\x00\x00\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x3a\x00\x00\x00\x00\x00\x3b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x28\x01\x00\x00\x00\x00\x00\x00\x29\x01\x3d\x00\x2a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x41\x00\x42\x00\x43\x00\x44\x00\x45\x00\x46\x00\x00\x00\x47\x00\x48\x00\x49\x00\x00\x00\x4a\x00\x00\x00\x4b\x00\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x2b\x01\x2c\x01\x2d\x01\x2e\x01\x2f\x01\x62\x00\x30\x01\x31\x01\x32\x01\x33\x01\x34\x01\x35\x01\x00\x00\x36\x01\x37\x01\x38\x01\x39\x01\x3a\x01\x3b\x01\x3c\x01\x3d\x01\x3e\x01\x3f\x01\x40\x01\x41\x01\x42\x01\x43\x01\x44\x01\x45\x01\x46\x01\x47\x01\x48\x01\x49\x01\x4a\x01\x4b\x01\x4c\x01\x4d\x01\x4e\x01\x4f\x01\x50\x01\x51\x01\x52\x01\x53\x01\x54\x01\x55\x01\x28\x00\x00\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\xdb\x00\xdc\x00\xdd\x00\xde\x00\x75\x00\x00\x00\x00\x00\x76\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x77\x00\x04\x01\x78\x00\x00\x00\xdb\x00\xdc\x00\xdd\x00\xde\x00\x79\x00\x00\x00\xfd\x00\x00\x00\x00\x00\xe1\x00\x00\x00\xe2\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x05\x01\x00\x00\x00\x00\x00\x00\xe4\x00\xe5\x00\xe6\x00\xdf\x00\x00\x00\xe0\x00\xe1\x00\x00\x00\xe2\x00\xe3\x00\xaf\xfe\x00\x00\xaf\xfe\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\xe4\x00\xe5\x00\xe6\x00\x00\x00\x00\x00\x28\x00\x00\x00\x00\x00\x00\x00\x2c\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\xb0\x00\x99\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x28\x00\x7e\x00\x00\x00\x5c\x00\x2c\x00\x2d\x00\x9a\x04\x0d\x01\xdc\x00\x00\x00\xde\x00\x75\x00\x00\x00\x00\x00\x76\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x77\x00\x62\x00\x78\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00\x00\x00\x00\x00\x00\x00\x0e\x01\x00\x00\x00\x00\xe2\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0f\x01\x00\x00\x10\x01\x11\x01\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x28\x00\x57\x00\x00\x00\x62\x00\x2c\x00\x2d\x00\x00\x00\x0d\x01\xdc\x00\x00\x00\xde\x00\x75\x00\x00\x00\x5c\x00\x76\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x77\x00\x00\x00\x78\x00\x00\x00\x10\xff\x00\x00\x00\x00\x00\x00\x79\x00\x00\x00\x00\x00\x62\x00\x0e\x01\x00\x00\x00\x00\xe2\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0f\x01\x00\x00\x49\x02\x28\x00\x29\x00\x00\x00\x00\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\xdc\x00\x00\x00\xde\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x00\x00\xb9\x02\x58\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xba\x02\x00\x00\x00\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\xe2\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x00\x00\xbb\x02\xbc\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5c\x00\x10\xff\x10\xff\x10\xff\x10\xff\x00\x00\x10\xff\x10\xff\x10\xff\x10\xff\x10\xff\x10\xff\x00\x00\x10\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x10\xff\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x00\x00\x28\x00\x00\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x5a\x00\x5b\x00\x10\xff\x5c\x00\x00\x00\x75\x00\x00\x00\x00\x00\x76\x00\x00\x00\xdb\x00\xdc\x00\xdd\x00\xde\x00\x00\x00\x77\x00\x00\x00\x78\x00\xfc\x00\x00\x00\x00\x00\x00\x00\x62\x00\x79\x00\x28\x00\xfd\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\xdf\x00\x75\x00\xe0\x00\xe1\x00\x76\x00\xe2\x00\xe3\x00\x00\x00\x00\x00\x00\x00\x00\x00\x77\x00\xe5\x03\x78\x00\x00\x00\xe4\x00\xe5\x00\xe6\x00\x00\x00\x79\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\xdb\x00\xdc\x00\xdd\x00\xde\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\xe0\x00\xe1\x00\x7a\x00\xe2\x00\x28\x04\x00\x00\x00\x00\x00\x00\x7e\x00\x00\x00\x5c\x00\x00\x00\x00\x00\xe4\x00\xe5\x00\xe6\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x62\x00\x00\x00\x28\x00\x00\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x7e\x00\x00\x00\x5c\x00\x00\x00\x75\x00\x00\x00\x00\x00\x76\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x77\x00\x00\x00\x78\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x79\x00\x00\x00\x00\x00\x28\x00\x00\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x00\x00\x00\x00\x76\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x77\x00\x00\x00\x78\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00\x00\x00\x7a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x00\x00\x00\x00\x7e\x00\x00\x00\x5c\x00\x28\x00\x00\x00\x00\x00\x00\x00\x2c\x00\x2d\x00\xe8\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x62\x00\x57\x00\x28\x00\x29\x00\x00\x00\x00\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x00\x00\x7e\x00\x00\x00\x5c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb9\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xba\x02\x62\x00\xd1\xfe\x28\x00\x00\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5a\x03\x75\x00\x00\x00\x00\x00\x76\x00\xbb\x02\xbc\x02\x00\x00\x00\x00\x00\x00\x00\x00\x77\x00\x00\x00\x78\x00\x00\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x79\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x5c\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x00\x00\x5a\x00\x5b\x00\x62\x00\x5c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x62\x00\x57\x00\x00\x00\x00\x00\x28\x00\x00\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x7e\x00\x00\x00\x5c\x00\x00\x00\x75\x00\x00\x00\x28\x00\x76\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\xf7\x00\x00\x00\x78\x00\x75\x00\x00\x00\x00\x00\x76\x00\x62\x00\x79\x00\xf8\x00\x00\x00\x00\x00\x00\x00\x77\x00\x00\x00\x78\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00\x00\x00\x00\x00\x28\x00\x00\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x00\x00\x00\x00\x76\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x00\x00\x77\x00\x00\x00\x78\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00\x00\x00\x7a\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x5c\x00\x57\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x00\x00\x00\x00\x7e\x00\x00\x00\x5c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x00\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x62\x00\x57\x00\x28\x00\x29\x00\x00\x00\x00\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb9\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xba\x02\x62\x00\x00\x00\x28\x00\x00\x00\x2a\x00\x2b\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5a\x03\x75\x00\x00\x00\x28\x00\x76\x00\xbb\x02\xbc\x02\x2c\x00\x2d\x00\x00\x00\x00\x00\x77\x00\x00\x00\x78\x00\x75\x00\x00\x00\x00\x00\x76\x00\x00\x00\x79\x00\x00\x00\x00\x00\x00\x00\x00\x00\x77\x00\x00\x00\x78\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x00\x00\x28\x00\x29\x00\x00\x00\x7a\x00\x2c\x00\x2d\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x5d\x01\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x62\x00\x57\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x5c\x00\x57\x00\x00\x00\x00\x00\x28\x00\x00\x00\x00\x00\x00\x00\x2c\x00\x2d\x00\xc3\x00\x00\x00\x00\x00\x5c\x00\x00\x00\x75\x00\x00\x00\x28\x00\x76\x00\x62\x00\xc4\x00\x00\x00\x00\x00\x00\x00\x00\x00\x77\x00\x00\x00\x78\x00\x00\x00\x00\x00\x28\x00\x29\x00\x62\x00\x79\x00\x2c\x00\x2d\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\xd1\x02\x00\x00\xb9\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5a\x00\x5b\x00\xba\x02\x5c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x00\x00\xa4\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xbb\x02\xbc\x02\x62\x00\x00\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x5c\x00\x57\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x5c\x00\x57\x00\x00\x00\x00\x00\x28\x00\x00\x00\x62\x00\x00\x00\x2c\x00\x2d\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x75\x00\x00\x00\x28\x00\x76\x00\x62\x00\x00\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x77\x00\x00\x00\x78\x00\x75\x00\x00\x00\x28\x00\x76\x00\x62\x00\x79\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x77\x00\x00\x00\x78\x00\x00\x00\x00\x00\x28\x00\x29\x00\x00\x00\x79\x00\x2c\x00\x2d\x00\xd1\x02\x00\x00\x58\x01\x00\x00\x59\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x5d\x01\x00\x00\x00\x00\x7a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa3\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x5c\x00\x57\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x5c\x00\x57\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x62\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x5c\x00\x57\x00\x00\x00\x00\x00\x28\x00\x00\x00\x62\x00\x00\x00\x2c\x00\x2d\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x75\x00\x00\x00\x28\x00\x76\x00\x62\x00\x00\x00\x2c\x00\x00\x00\xa4\x03\x00\x00\x77\x00\x00\x00\x78\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x79\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb0\x00\x43\x04\x00\x00\x00\x00\x00\x00\x28\x00\x00\x00\x00\x00\x00\x00\x2c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x44\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb0\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x5c\x00\x57\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x28\x00\x29\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x62\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\xb2\x01\x00\x00\x00\x00\x00\x00\x62\x00\x00\x00\xb3\x01\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x00\x00\xb4\x01\x00\x00\x00\x00\x00\x00\x28\x00\x00\x00\x00\x00\xb1\x00\x2c\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb2\x00\x00\x00\x00\x00\x62\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x28\x00\x29\x00\x00\x00\x00\x00\x2c\x00\x2d\x00\x00\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\xdb\x00\xdc\x00\xdd\x00\xde\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5d\x01\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x62\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\xe0\x00\xe1\x00\x00\x00\xe2\x00\x92\x01\x00\x00\x00\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x28\x00\xe4\x00\xe5\x00\xe6\x00\x93\x01\x94\x01\x95\x01\x96\x01\x00\x00\xb1\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb2\x00\x00\x00\x00\x00\x62\x00\x00\x00\x28\x00\x00\x00\x00\x00\x00\x00\x2c\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x00\x00\xb0\x00\x76\x03\x00\x00\x00\x00\x28\x00\x29\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\xa3\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb2\x01\x00\x00\x00\x00\x62\x00\x00\x00\x00\x00\x1b\x04\x28\x00\x29\x00\x00\x00\x00\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb2\x01\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x5c\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\xb4\x01\x57\x00\x00\x00\x00\x00\x28\x00\x00\x00\x00\x00\x00\x00\x2c\x00\x00\x00\x5a\x00\x5b\x00\x62\x00\x5c\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\xb0\x00\x57\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x00\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x28\x00\x29\x00\x62\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb2\x01\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x62\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x28\x00\x00\x00\x00\x00\x00\x00\x2c\x00\x00\x00\x00\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x28\x00\x29\x00\x62\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\xb2\x01\x57\x00\x00\x00\x00\x00\x28\x00\x00\x00\x00\x00\x00\x00\x2c\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb0\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x62\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x28\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x89\x03\x00\x00\x62\x00\x00\x00\x28\x00\x00\x00\x00\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x69\x03\x00\x00\x00\x00\x00\x00\x28\x00\x62\x00\x00\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89\x03\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x28\x00\x00\x00\x5a\x00\x5b\x00\x2c\x00\x5c\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x28\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x00\x00\x16\x02\x00\x00\x00\x00\x62\x00\x00\x00\x00\x00\x5a\x00\x5b\x00\x28\x00\x5c\x00\x00\x00\x00\x00\x00\x00\xd9\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x16\x02\x00\x00\x4d\x00\x4e\x00\x4f\x00\x50\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x17\x02\x57\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5a\x00\x5b\x00\x00\x00\x5c\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x62\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x4d\x00\x7b\x00\x7c\x00\x7d\x00\x00\x00\x51\x00\x52\x00\x53\x00\x54\x00\x55\x00\x56\x00\x00\x00\x57\x00\x00\x00\x62\x00\x00\x00\x00\x00\xa3\x02\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x5c\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x62\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x04\x03\x05\x03\x06\x03\x07\x03\x00\x00\x08\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa6\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xa3\x02\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xf5\x03\x06\x03\x07\x03\x00\x00\x08\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa6\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xa3\x02\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xf6\x03\x07\x03\x00\x00\x08\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa6\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xa3\x02\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x04\xa5\x02\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\xa6\x02\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x19\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x00\x00\x00\x00\x1a\x00\x00\x00\x00\x00\x00\x00\x26\x00\x00\x00\x41\x03\x00\x00\xf2\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xa3\x02\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa4\x02\xa5\x02\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\xa6\x02\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x19\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x00\x00\x00\x00\x1a\x00\x00\x00\x00\x00\x00\x00\x26\x00\x00\x00\xd5\x02\x00\x00\xf2\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xa3\x02\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xfe\x02\xa5\x02\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\xa6\x02\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x19\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x00\x00\x00\x00\x1a\x00\x00\x00\x00\x00\x00\x00\x26\x00\x00\x00\xf1\x01\x00\x00\xf2\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa3\x02\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x1a\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\xf4\x01\x00\x00\xf2\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\xed\x03\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa6\x02\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x26\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x0d\x00\x0e\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x1c\x01\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x1d\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x5e\x04\xa0\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x26\x00\x1e\x01\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x5f\x04\xa0\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x26\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xfb\x03\xa0\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\xfc\x03\xa0\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x26\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xfd\x03\xa0\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\xff\x03\xa0\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x26\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x04\x04\xa0\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x2d\x04\xa0\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x26\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xb1\x03\xa0\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\xc6\x03\xa0\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x26\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xc7\x03\xa0\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x2c\x03\xa0\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x26\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x3e\x03\xa0\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\xa8\x02\xa0\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x26\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xb4\x01\xa0\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x61\x01\xa0\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x26\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x62\x01\xa0\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x9f\x00\xa0\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x26\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x7e\x01\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x7f\x01\x80\x01\x00\x00\x00\x00\x00\x00\x81\x01\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x01\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x6c\x01\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x6d\x01\x6e\x01\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x6f\x01\x00\x00\x70\x01\x00\x00\x00\x00\x71\x01\x82\x01\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xd5\x00\xd6\x00\x83\x01\x72\x01\xd8\x00\x73\x01\x84\x01\x26\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x6f\x01\x00\x00\x70\x01\x00\x00\x00\x00\x71\x01\xd4\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xd5\x00\xd6\x00\x00\x00\x72\x01\x00\x00\x73\x01\x74\x01\x26\x00\x6b\x01\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x7b\x01\x7c\x01\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0b\x03\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x0c\x03\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x6f\x01\x00\x00\x70\x01\x00\x00\x00\x00\x71\x01\xd4\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xd5\x00\xd6\x00\x00\x00\x72\x01\x00\x00\x73\x01\x74\x01\x26\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x6f\x01\x00\x00\x70\x01\x00\x00\x00\x00\x71\x01\xd4\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xd5\x00\xd6\x00\x00\x00\x72\x01\x00\x00\x73\x01\x74\x01\x26\x00\x6b\x01\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x0d\x03\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x01\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x12\x03\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x6f\x01\x00\x00\x70\x01\x00\x00\x00\x00\x71\x01\xd4\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xd5\x00\xd6\x00\x00\x00\x72\x01\x00\x00\x73\x01\x74\x01\x26\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x6f\x01\x00\x00\x70\x01\x00\x00\x00\x00\x71\x01\xd4\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xd5\x00\xd6\x00\x00\x00\x72\x01\x00\x00\x73\x01\x74\x01\x26\x00\x6b\x01\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x19\x03\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x01\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\xbd\x01\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x6f\x01\x00\x00\x70\x01\x00\x00\x00\x00\x71\x01\xd4\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xd5\x00\xd6\x00\x00\x00\x72\x01\x00\x00\x73\x01\x74\x01\x26\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x6f\x01\x00\x00\x70\x01\x00\x00\x00\x00\x71\x01\xd4\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xd5\x00\xd6\x00\x00\x00\x72\x01\x00\x00\x73\x01\x74\x01\x26\x00\x6b\x01\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\xc5\x01\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x6f\x01\x00\x00\x70\x01\x00\x00\x00\x00\x71\x01\xd4\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xd5\x00\xd6\x00\x00\x00\x72\x01\x00\x00\x73\x01\x74\x01\x26\x00\x9d\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x32\x03\x33\x03\x00\x00\x00\x00\x00\x00\x34\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x9d\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xe2\x02\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\xfe\x03\x16\x00\x17\x00\x18\x00\x34\x03\x00\x00\x00\x00\x00\x00\xe3\x02\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x26\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x15\x03\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x16\x03\x19\x00\x63\x01\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x9d\x00\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x82\x04\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x9e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x5a\x04\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\xfa\x03\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x26\x03\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x40\x03\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\xd2\x02\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\xd3\x02\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\xd9\x02\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\xda\x02\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\xff\x02\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x03\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x0a\x03\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x14\x03\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\xa7\x01\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\xa8\x01\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\xa9\x01\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\xaa\x01\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x21\x01\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x23\x01\x0f\x00\x10\x00\x11\x00\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x00\x00\x89\x03\x10\x00\x60\x01\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\xc9\x01\x10\x00\x60\x01\x12\x00\x00\x00\x00\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\xad\x01\x00\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x19\x00\x00\x00\x5f\x01\x10\x00\x60\x01\x12\x00\x00\x00\x26\x00\x13\x00\x14\x00\x15\x00\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x66\x01\x00\x00\x00\x00\x16\x00\x17\x00\x18\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x26\x00\x00\x00\x00\x00\x00\x00\x68\x01\x69\x01\x18\x01\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\xd2\x01\x18\x01\x17\x00\x18\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x17\x01\x18\x01\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\xe5\x02\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x26\x00\x00\x00\x19\x00\x00\x00\x00\x00\xe6\x02\x25\x01\x26\x01\x00\x00\xe7\x02\x90\x03\x00\x00\x00\x00\x00\x00\x00\x00\xe5\x02\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\xe6\x02\x25\x01\x26\x01\x00\x00\xe7\x02\xe8\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xeb\x03\x17\x00\x18\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x71\x02\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb5\x01\x17\x00\x18\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb6\x01\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x8f\x01\x17\x00\x18\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xef\x01\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5a\x02\x17\x00\x18\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5e\x02\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xcf\x00\x17\x00\x18\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x01\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x67\x01\x17\x00\x18\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x8f\x01\x17\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x1c\x00\x00\x00\x00\x00\x1d\x00\x1e\x00\x00\x00\x1f\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x21\x00\x22\x00\x23\x00\x24\x00\x25\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"#

happyReduceArr = Happy_Data_Array.array (8, 659) [
	(8 , happyReduce_8),
	(9 , happyReduce_9),
	(10 , happyReduce_10),
	(11 , happyReduce_11),
	(12 , happyReduce_12),
	(13 , happyReduce_13),
	(14 , happyReduce_14),
	(15 , happyReduce_15),
	(16 , happyReduce_16),
	(17 , happyReduce_17),
	(18 , happyReduce_18),
	(19 , happyReduce_19),
	(20 , happyReduce_20),
	(21 , happyReduce_21),
	(22 , happyReduce_22),
	(23 , happyReduce_23),
	(24 , happyReduce_24),
	(25 , happyReduce_25),
	(26 , happyReduce_26),
	(27 , happyReduce_27),
	(28 , happyReduce_28),
	(29 , happyReduce_29),
	(30 , happyReduce_30),
	(31 , happyReduce_31),
	(32 , happyReduce_32),
	(33 , happyReduce_33),
	(34 , happyReduce_34),
	(35 , happyReduce_35),
	(36 , happyReduce_36),
	(37 , happyReduce_37),
	(38 , happyReduce_38),
	(39 , happyReduce_39),
	(40 , happyReduce_40),
	(41 , happyReduce_41),
	(42 , happyReduce_42),
	(43 , happyReduce_43),
	(44 , happyReduce_44),
	(45 , happyReduce_45),
	(46 , happyReduce_46),
	(47 , happyReduce_47),
	(48 , happyReduce_48),
	(49 , happyReduce_49),
	(50 , happyReduce_50),
	(51 , happyReduce_51),
	(52 , happyReduce_52),
	(53 , happyReduce_53),
	(54 , happyReduce_54),
	(55 , happyReduce_55),
	(56 , happyReduce_56),
	(57 , happyReduce_57),
	(58 , happyReduce_58),
	(59 , happyReduce_59),
	(60 , happyReduce_60),
	(61 , happyReduce_61),
	(62 , happyReduce_62),
	(63 , happyReduce_63),
	(64 , happyReduce_64),
	(65 , happyReduce_65),
	(66 , happyReduce_66),
	(67 , happyReduce_67),
	(68 , happyReduce_68),
	(69 , happyReduce_69),
	(70 , happyReduce_70),
	(71 , happyReduce_71),
	(72 , happyReduce_72),
	(73 , happyReduce_73),
	(74 , happyReduce_74),
	(75 , happyReduce_75),
	(76 , happyReduce_76),
	(77 , happyReduce_77),
	(78 , happyReduce_78),
	(79 , happyReduce_79),
	(80 , happyReduce_80),
	(81 , happyReduce_81),
	(82 , happyReduce_82),
	(83 , happyReduce_83),
	(84 , happyReduce_84),
	(85 , happyReduce_85),
	(86 , happyReduce_86),
	(87 , happyReduce_87),
	(88 , happyReduce_88),
	(89 , happyReduce_89),
	(90 , happyReduce_90),
	(91 , happyReduce_91),
	(92 , happyReduce_92),
	(93 , happyReduce_93),
	(94 , happyReduce_94),
	(95 , happyReduce_95),
	(96 , happyReduce_96),
	(97 , happyReduce_97),
	(98 , happyReduce_98),
	(99 , happyReduce_99),
	(100 , happyReduce_100),
	(101 , happyReduce_101),
	(102 , happyReduce_102),
	(103 , happyReduce_103),
	(104 , happyReduce_104),
	(105 , happyReduce_105),
	(106 , happyReduce_106),
	(107 , happyReduce_107),
	(108 , happyReduce_108),
	(109 , happyReduce_109),
	(110 , happyReduce_110),
	(111 , happyReduce_111),
	(112 , happyReduce_112),
	(113 , happyReduce_113),
	(114 , happyReduce_114),
	(115 , happyReduce_115),
	(116 , happyReduce_116),
	(117 , happyReduce_117),
	(118 , happyReduce_118),
	(119 , happyReduce_119),
	(120 , happyReduce_120),
	(121 , happyReduce_121),
	(122 , happyReduce_122),
	(123 , happyReduce_123),
	(124 , happyReduce_124),
	(125 , happyReduce_125),
	(126 , happyReduce_126),
	(127 , happyReduce_127),
	(128 , happyReduce_128),
	(129 , happyReduce_129),
	(130 , happyReduce_130),
	(131 , happyReduce_131),
	(132 , happyReduce_132),
	(133 , happyReduce_133),
	(134 , happyReduce_134),
	(135 , happyReduce_135),
	(136 , happyReduce_136),
	(137 , happyReduce_137),
	(138 , happyReduce_138),
	(139 , happyReduce_139),
	(140 , happyReduce_140),
	(141 , happyReduce_141),
	(142 , happyReduce_142),
	(143 , happyReduce_143),
	(144 , happyReduce_144),
	(145 , happyReduce_145),
	(146 , happyReduce_146),
	(147 , happyReduce_147),
	(148 , happyReduce_148),
	(149 , happyReduce_149),
	(150 , happyReduce_150),
	(151 , happyReduce_151),
	(152 , happyReduce_152),
	(153 , happyReduce_153),
	(154 , happyReduce_154),
	(155 , happyReduce_155),
	(156 , happyReduce_156),
	(157 , happyReduce_157),
	(158 , happyReduce_158),
	(159 , happyReduce_159),
	(160 , happyReduce_160),
	(161 , happyReduce_161),
	(162 , happyReduce_162),
	(163 , happyReduce_163),
	(164 , happyReduce_164),
	(165 , happyReduce_165),
	(166 , happyReduce_166),
	(167 , happyReduce_167),
	(168 , happyReduce_168),
	(169 , happyReduce_169),
	(170 , happyReduce_170),
	(171 , happyReduce_171),
	(172 , happyReduce_172),
	(173 , happyReduce_173),
	(174 , happyReduce_174),
	(175 , happyReduce_175),
	(176 , happyReduce_176),
	(177 , happyReduce_177),
	(178 , happyReduce_178),
	(179 , happyReduce_179),
	(180 , happyReduce_180),
	(181 , happyReduce_181),
	(182 , happyReduce_182),
	(183 , happyReduce_183),
	(184 , happyReduce_184),
	(185 , happyReduce_185),
	(186 , happyReduce_186),
	(187 , happyReduce_187),
	(188 , happyReduce_188),
	(189 , happyReduce_189),
	(190 , happyReduce_190),
	(191 , happyReduce_191),
	(192 , happyReduce_192),
	(193 , happyReduce_193),
	(194 , happyReduce_194),
	(195 , happyReduce_195),
	(196 , happyReduce_196),
	(197 , happyReduce_197),
	(198 , happyReduce_198),
	(199 , happyReduce_199),
	(200 , happyReduce_200),
	(201 , happyReduce_201),
	(202 , happyReduce_202),
	(203 , happyReduce_203),
	(204 , happyReduce_204),
	(205 , happyReduce_205),
	(206 , happyReduce_206),
	(207 , happyReduce_207),
	(208 , happyReduce_208),
	(209 , happyReduce_209),
	(210 , happyReduce_210),
	(211 , happyReduce_211),
	(212 , happyReduce_212),
	(213 , happyReduce_213),
	(214 , happyReduce_214),
	(215 , happyReduce_215),
	(216 , happyReduce_216),
	(217 , happyReduce_217),
	(218 , happyReduce_218),
	(219 , happyReduce_219),
	(220 , happyReduce_220),
	(221 , happyReduce_221),
	(222 , happyReduce_222),
	(223 , happyReduce_223),
	(224 , happyReduce_224),
	(225 , happyReduce_225),
	(226 , happyReduce_226),
	(227 , happyReduce_227),
	(228 , happyReduce_228),
	(229 , happyReduce_229),
	(230 , happyReduce_230),
	(231 , happyReduce_231),
	(232 , happyReduce_232),
	(233 , happyReduce_233),
	(234 , happyReduce_234),
	(235 , happyReduce_235),
	(236 , happyReduce_236),
	(237 , happyReduce_237),
	(238 , happyReduce_238),
	(239 , happyReduce_239),
	(240 , happyReduce_240),
	(241 , happyReduce_241),
	(242 , happyReduce_242),
	(243 , happyReduce_243),
	(244 , happyReduce_244),
	(245 , happyReduce_245),
	(246 , happyReduce_246),
	(247 , happyReduce_247),
	(248 , happyReduce_248),
	(249 , happyReduce_249),
	(250 , happyReduce_250),
	(251 , happyReduce_251),
	(252 , happyReduce_252),
	(253 , happyReduce_253),
	(254 , happyReduce_254),
	(255 , happyReduce_255),
	(256 , happyReduce_256),
	(257 , happyReduce_257),
	(258 , happyReduce_258),
	(259 , happyReduce_259),
	(260 , happyReduce_260),
	(261 , happyReduce_261),
	(262 , happyReduce_262),
	(263 , happyReduce_263),
	(264 , happyReduce_264),
	(265 , happyReduce_265),
	(266 , happyReduce_266),
	(267 , happyReduce_267),
	(268 , happyReduce_268),
	(269 , happyReduce_269),
	(270 , happyReduce_270),
	(271 , happyReduce_271),
	(272 , happyReduce_272),
	(273 , happyReduce_273),
	(274 , happyReduce_274),
	(275 , happyReduce_275),
	(276 , happyReduce_276),
	(277 , happyReduce_277),
	(278 , happyReduce_278),
	(279 , happyReduce_279),
	(280 , happyReduce_280),
	(281 , happyReduce_281),
	(282 , happyReduce_282),
	(283 , happyReduce_283),
	(284 , happyReduce_284),
	(285 , happyReduce_285),
	(286 , happyReduce_286),
	(287 , happyReduce_287),
	(288 , happyReduce_288),
	(289 , happyReduce_289),
	(290 , happyReduce_290),
	(291 , happyReduce_291),
	(292 , happyReduce_292),
	(293 , happyReduce_293),
	(294 , happyReduce_294),
	(295 , happyReduce_295),
	(296 , happyReduce_296),
	(297 , happyReduce_297),
	(298 , happyReduce_298),
	(299 , happyReduce_299),
	(300 , happyReduce_300),
	(301 , happyReduce_301),
	(302 , happyReduce_302),
	(303 , happyReduce_303),
	(304 , happyReduce_304),
	(305 , happyReduce_305),
	(306 , happyReduce_306),
	(307 , happyReduce_307),
	(308 , happyReduce_308),
	(309 , happyReduce_309),
	(310 , happyReduce_310),
	(311 , happyReduce_311),
	(312 , happyReduce_312),
	(313 , happyReduce_313),
	(314 , happyReduce_314),
	(315 , happyReduce_315),
	(316 , happyReduce_316),
	(317 , happyReduce_317),
	(318 , happyReduce_318),
	(319 , happyReduce_319),
	(320 , happyReduce_320),
	(321 , happyReduce_321),
	(322 , happyReduce_322),
	(323 , happyReduce_323),
	(324 , happyReduce_324),
	(325 , happyReduce_325),
	(326 , happyReduce_326),
	(327 , happyReduce_327),
	(328 , happyReduce_328),
	(329 , happyReduce_329),
	(330 , happyReduce_330),
	(331 , happyReduce_331),
	(332 , happyReduce_332),
	(333 , happyReduce_333),
	(334 , happyReduce_334),
	(335 , happyReduce_335),
	(336 , happyReduce_336),
	(337 , happyReduce_337),
	(338 , happyReduce_338),
	(339 , happyReduce_339),
	(340 , happyReduce_340),
	(341 , happyReduce_341),
	(342 , happyReduce_342),
	(343 , happyReduce_343),
	(344 , happyReduce_344),
	(345 , happyReduce_345),
	(346 , happyReduce_346),
	(347 , happyReduce_347),
	(348 , happyReduce_348),
	(349 , happyReduce_349),
	(350 , happyReduce_350),
	(351 , happyReduce_351),
	(352 , happyReduce_352),
	(353 , happyReduce_353),
	(354 , happyReduce_354),
	(355 , happyReduce_355),
	(356 , happyReduce_356),
	(357 , happyReduce_357),
	(358 , happyReduce_358),
	(359 , happyReduce_359),
	(360 , happyReduce_360),
	(361 , happyReduce_361),
	(362 , happyReduce_362),
	(363 , happyReduce_363),
	(364 , happyReduce_364),
	(365 , happyReduce_365),
	(366 , happyReduce_366),
	(367 , happyReduce_367),
	(368 , happyReduce_368),
	(369 , happyReduce_369),
	(370 , happyReduce_370),
	(371 , happyReduce_371),
	(372 , happyReduce_372),
	(373 , happyReduce_373),
	(374 , happyReduce_374),
	(375 , happyReduce_375),
	(376 , happyReduce_376),
	(377 , happyReduce_377),
	(378 , happyReduce_378),
	(379 , happyReduce_379),
	(380 , happyReduce_380),
	(381 , happyReduce_381),
	(382 , happyReduce_382),
	(383 , happyReduce_383),
	(384 , happyReduce_384),
	(385 , happyReduce_385),
	(386 , happyReduce_386),
	(387 , happyReduce_387),
	(388 , happyReduce_388),
	(389 , happyReduce_389),
	(390 , happyReduce_390),
	(391 , happyReduce_391),
	(392 , happyReduce_392),
	(393 , happyReduce_393),
	(394 , happyReduce_394),
	(395 , happyReduce_395),
	(396 , happyReduce_396),
	(397 , happyReduce_397),
	(398 , happyReduce_398),
	(399 , happyReduce_399),
	(400 , happyReduce_400),
	(401 , happyReduce_401),
	(402 , happyReduce_402),
	(403 , happyReduce_403),
	(404 , happyReduce_404),
	(405 , happyReduce_405),
	(406 , happyReduce_406),
	(407 , happyReduce_407),
	(408 , happyReduce_408),
	(409 , happyReduce_409),
	(410 , happyReduce_410),
	(411 , happyReduce_411),
	(412 , happyReduce_412),
	(413 , happyReduce_413),
	(414 , happyReduce_414),
	(415 , happyReduce_415),
	(416 , happyReduce_416),
	(417 , happyReduce_417),
	(418 , happyReduce_418),
	(419 , happyReduce_419),
	(420 , happyReduce_420),
	(421 , happyReduce_421),
	(422 , happyReduce_422),
	(423 , happyReduce_423),
	(424 , happyReduce_424),
	(425 , happyReduce_425),
	(426 , happyReduce_426),
	(427 , happyReduce_427),
	(428 , happyReduce_428),
	(429 , happyReduce_429),
	(430 , happyReduce_430),
	(431 , happyReduce_431),
	(432 , happyReduce_432),
	(433 , happyReduce_433),
	(434 , happyReduce_434),
	(435 , happyReduce_435),
	(436 , happyReduce_436),
	(437 , happyReduce_437),
	(438 , happyReduce_438),
	(439 , happyReduce_439),
	(440 , happyReduce_440),
	(441 , happyReduce_441),
	(442 , happyReduce_442),
	(443 , happyReduce_443),
	(444 , happyReduce_444),
	(445 , happyReduce_445),
	(446 , happyReduce_446),
	(447 , happyReduce_447),
	(448 , happyReduce_448),
	(449 , happyReduce_449),
	(450 , happyReduce_450),
	(451 , happyReduce_451),
	(452 , happyReduce_452),
	(453 , happyReduce_453),
	(454 , happyReduce_454),
	(455 , happyReduce_455),
	(456 , happyReduce_456),
	(457 , happyReduce_457),
	(458 , happyReduce_458),
	(459 , happyReduce_459),
	(460 , happyReduce_460),
	(461 , happyReduce_461),
	(462 , happyReduce_462),
	(463 , happyReduce_463),
	(464 , happyReduce_464),
	(465 , happyReduce_465),
	(466 , happyReduce_466),
	(467 , happyReduce_467),
	(468 , happyReduce_468),
	(469 , happyReduce_469),
	(470 , happyReduce_470),
	(471 , happyReduce_471),
	(472 , happyReduce_472),
	(473 , happyReduce_473),
	(474 , happyReduce_474),
	(475 , happyReduce_475),
	(476 , happyReduce_476),
	(477 , happyReduce_477),
	(478 , happyReduce_478),
	(479 , happyReduce_479),
	(480 , happyReduce_480),
	(481 , happyReduce_481),
	(482 , happyReduce_482),
	(483 , happyReduce_483),
	(484 , happyReduce_484),
	(485 , happyReduce_485),
	(486 , happyReduce_486),
	(487 , happyReduce_487),
	(488 , happyReduce_488),
	(489 , happyReduce_489),
	(490 , happyReduce_490),
	(491 , happyReduce_491),
	(492 , happyReduce_492),
	(493 , happyReduce_493),
	(494 , happyReduce_494),
	(495 , happyReduce_495),
	(496 , happyReduce_496),
	(497 , happyReduce_497),
	(498 , happyReduce_498),
	(499 , happyReduce_499),
	(500 , happyReduce_500),
	(501 , happyReduce_501),
	(502 , happyReduce_502),
	(503 , happyReduce_503),
	(504 , happyReduce_504),
	(505 , happyReduce_505),
	(506 , happyReduce_506),
	(507 , happyReduce_507),
	(508 , happyReduce_508),
	(509 , happyReduce_509),
	(510 , happyReduce_510),
	(511 , happyReduce_511),
	(512 , happyReduce_512),
	(513 , happyReduce_513),
	(514 , happyReduce_514),
	(515 , happyReduce_515),
	(516 , happyReduce_516),
	(517 , happyReduce_517),
	(518 , happyReduce_518),
	(519 , happyReduce_519),
	(520 , happyReduce_520),
	(521 , happyReduce_521),
	(522 , happyReduce_522),
	(523 , happyReduce_523),
	(524 , happyReduce_524),
	(525 , happyReduce_525),
	(526 , happyReduce_526),
	(527 , happyReduce_527),
	(528 , happyReduce_528),
	(529 , happyReduce_529),
	(530 , happyReduce_530),
	(531 , happyReduce_531),
	(532 , happyReduce_532),
	(533 , happyReduce_533),
	(534 , happyReduce_534),
	(535 , happyReduce_535),
	(536 , happyReduce_536),
	(537 , happyReduce_537),
	(538 , happyReduce_538),
	(539 , happyReduce_539),
	(540 , happyReduce_540),
	(541 , happyReduce_541),
	(542 , happyReduce_542),
	(543 , happyReduce_543),
	(544 , happyReduce_544),
	(545 , happyReduce_545),
	(546 , happyReduce_546),
	(547 , happyReduce_547),
	(548 , happyReduce_548),
	(549 , happyReduce_549),
	(550 , happyReduce_550),
	(551 , happyReduce_551),
	(552 , happyReduce_552),
	(553 , happyReduce_553),
	(554 , happyReduce_554),
	(555 , happyReduce_555),
	(556 , happyReduce_556),
	(557 , happyReduce_557),
	(558 , happyReduce_558),
	(559 , happyReduce_559),
	(560 , happyReduce_560),
	(561 , happyReduce_561),
	(562 , happyReduce_562),
	(563 , happyReduce_563),
	(564 , happyReduce_564),
	(565 , happyReduce_565),
	(566 , happyReduce_566),
	(567 , happyReduce_567),
	(568 , happyReduce_568),
	(569 , happyReduce_569),
	(570 , happyReduce_570),
	(571 , happyReduce_571),
	(572 , happyReduce_572),
	(573 , happyReduce_573),
	(574 , happyReduce_574),
	(575 , happyReduce_575),
	(576 , happyReduce_576),
	(577 , happyReduce_577),
	(578 , happyReduce_578),
	(579 , happyReduce_579),
	(580 , happyReduce_580),
	(581 , happyReduce_581),
	(582 , happyReduce_582),
	(583 , happyReduce_583),
	(584 , happyReduce_584),
	(585 , happyReduce_585),
	(586 , happyReduce_586),
	(587 , happyReduce_587),
	(588 , happyReduce_588),
	(589 , happyReduce_589),
	(590 , happyReduce_590),
	(591 , happyReduce_591),
	(592 , happyReduce_592),
	(593 , happyReduce_593),
	(594 , happyReduce_594),
	(595 , happyReduce_595),
	(596 , happyReduce_596),
	(597 , happyReduce_597),
	(598 , happyReduce_598),
	(599 , happyReduce_599),
	(600 , happyReduce_600),
	(601 , happyReduce_601),
	(602 , happyReduce_602),
	(603 , happyReduce_603),
	(604 , happyReduce_604),
	(605 , happyReduce_605),
	(606 , happyReduce_606),
	(607 , happyReduce_607),
	(608 , happyReduce_608),
	(609 , happyReduce_609),
	(610 , happyReduce_610),
	(611 , happyReduce_611),
	(612 , happyReduce_612),
	(613 , happyReduce_613),
	(614 , happyReduce_614),
	(615 , happyReduce_615),
	(616 , happyReduce_616),
	(617 , happyReduce_617),
	(618 , happyReduce_618),
	(619 , happyReduce_619),
	(620 , happyReduce_620),
	(621 , happyReduce_621),
	(622 , happyReduce_622),
	(623 , happyReduce_623),
	(624 , happyReduce_624),
	(625 , happyReduce_625),
	(626 , happyReduce_626),
	(627 , happyReduce_627),
	(628 , happyReduce_628),
	(629 , happyReduce_629),
	(630 , happyReduce_630),
	(631 , happyReduce_631),
	(632 , happyReduce_632),
	(633 , happyReduce_633),
	(634 , happyReduce_634),
	(635 , happyReduce_635),
	(636 , happyReduce_636),
	(637 , happyReduce_637),
	(638 , happyReduce_638),
	(639 , happyReduce_639),
	(640 , happyReduce_640),
	(641 , happyReduce_641),
	(642 , happyReduce_642),
	(643 , happyReduce_643),
	(644 , happyReduce_644),
	(645 , happyReduce_645),
	(646 , happyReduce_646),
	(647 , happyReduce_647),
	(648 , happyReduce_648),
	(649 , happyReduce_649),
	(650 , happyReduce_650),
	(651 , happyReduce_651),
	(652 , happyReduce_652),
	(653 , happyReduce_653),
	(654 , happyReduce_654),
	(655 , happyReduce_655),
	(656 , happyReduce_656),
	(657 , happyReduce_657),
	(658 , happyReduce_658),
	(659 , happyReduce_659)
	]

happy_n_terms = 139 :: Int
happy_n_nonterms = 234 :: Int

happyReduce_8 = happySpecReduce_2  0# happyReduction_8
happyReduction_8 happy_x_2
	happy_x_1
	 =  case happyOut15 happy_x_1 of { happy_var_1 -> 
	case happyOut12 happy_x_2 of { happy_var_2 -> 
	happyIn11
		 (let (os,ss,l) = happy_var_1 in map (\x -> x os ss l) happy_var_2
	)}}

happyReduce_9 = happySpecReduce_2  1# happyReduction_9
happyReduction_9 happy_x_2
	happy_x_1
	 =  case happyOut19 happy_x_1 of { happy_var_1 -> 
	case happyOut12 happy_x_2 of { happy_var_2 -> 
	happyIn12
		 (happy_var_1 : happy_var_2
	)}}

happyReduce_10 = happySpecReduce_1  1# happyReduction_10
happyReduction_10 happy_x_1
	 =  case happyOut19 happy_x_1 of { happy_var_1 -> 
	happyIn12
		 ([happy_var_1]
	)}

happyReduce_11 = happyMonadReduce 2# 2# happyReduction_11
happyReduction_11 (happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut15 happy_x_1 of { happy_var_1 -> 
	case happyOut14 happy_x_2 of { happy_var_2 -> 
	( checkPageModule happy_var_2 happy_var_1)}}
	) (\r -> happyReturn (happyIn13 r))

happyReduce_12 = happyMonadReduce 5# 2# happyReduction_12
happyReduction_12 (happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut15 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 XCodeTagOpen) -> 
	case happyOut19 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 XCodeTagClose) -> 
	case happyOut14 happy_x_5 of { happy_var_5 -> 
	( let (os,ss,l) = happy_var_1 in checkHybridModule happy_var_5 (happy_var_3 os ss l) happy_var_2 happy_var_4)}}}}}
	) (\r -> happyReturn (happyIn13 r))

happyReduce_13 = happySpecReduce_2  2# happyReduction_13
happyReduction_13 happy_x_2
	happy_x_1
	 =  case happyOut15 happy_x_1 of { happy_var_1 -> 
	case happyOut19 happy_x_2 of { happy_var_2 -> 
	happyIn13
		 (let (os,ss,l) = happy_var_1 in happy_var_2 os ss l
	)}}

happyReduce_14 = happyMonadReduce 9# 3# happyReduction_14
happyReduction_14 (happy_x_9 `HappyStk`
	happy_x_8 `HappyStk`
	happy_x_7 `HappyStk`
	happy_x_6 `HappyStk`
	happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 XStdTagOpen) -> 
	case happyOut171 happy_x_2 of { happy_var_2 -> 
	case happyOut174 happy_x_3 of { happy_var_3 -> 
	case happyOut176 happy_x_4 of { happy_var_4 -> 
	case happyOutTok happy_x_5 of { (Loc happy_var_5 XStdTagClose) -> 
	case happyOut169 happy_x_6 of { happy_var_6 -> 
	case happyOutTok happy_x_7 of { (Loc happy_var_7 XCloseTagOpen) -> 
	case happyOut171 happy_x_8 of { happy_var_8 -> 
	case happyOutTok happy_x_9 of { (Loc happy_var_9 XStdTagClose) -> 
	( do { n <- checkEqNames happy_var_2 happy_var_8;
                                                                       let { cn = reverse happy_var_6;
                                                                             as = reverse happy_var_3; };
                                                                       return $ XTag (happy_var_1 <^^> happy_var_9 <** [happy_var_1,happy_var_5,happy_var_7,happy_var_9]) n as happy_var_4 cn })}}}}}}}}}
	) (\r -> happyReturn (happyIn14 r))

happyReduce_15 = happyReduce 5# 3# happyReduction_15
happyReduction_15 (happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 XStdTagOpen) -> 
	case happyOut171 happy_x_2 of { happy_var_2 -> 
	case happyOut174 happy_x_3 of { happy_var_3 -> 
	case happyOut176 happy_x_4 of { happy_var_4 -> 
	case happyOutTok happy_x_5 of { (Loc happy_var_5 XEmptyTagClose) -> 
	happyIn14
		 (XETag (happy_var_1 <^^> happy_var_5 <** [happy_var_1,happy_var_5]) happy_var_2 (reverse happy_var_3) happy_var_4
	) `HappyStk` happyRest}}}}}

happyReduce_16 = happySpecReduce_3  4# happyReduction_16
happyReduction_16 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut236 happy_x_1 of { happy_var_1 -> 
	case happyOut16 happy_x_2 of { happy_var_2 -> 
	case happyOut237 happy_x_3 of { happy_var_3 -> 
	happyIn15
		 (let (os,ss,ml) = happy_var_2 in (os,happy_var_1:ss++[happy_var_3],happy_var_1 <^^> happy_var_3)
	)}}}

happyReduce_17 = happySpecReduce_3  5# happyReduction_17
happyReduction_17 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut17 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 SemiColon) -> 
	case happyOut16 happy_x_3 of { happy_var_3 -> 
	happyIn16
		 (let (os,ss,ml) = happy_var_3 in (happy_var_1 : os, happy_var_2 : ss, Just $ ann happy_var_1 <++> nIS happy_var_2 <+?> ml)
	)}}}

happyReduce_18 = happySpecReduce_0  5# happyReduction_18
happyReduction_18  =  happyIn16
		 (([],[],Nothing)
	)

happyReduce_19 = happyReduce 4# 6# happyReduction_19
happyReduction_19 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 LANGUAGE) -> 
	case happyOut18 happy_x_2 of { happy_var_2 -> 
	case happyOut25 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 PragmaEnd) -> 
	happyIn17
		 (LanguagePragma (happy_var_1 <^^> happy_var_4 <** (happy_var_1:snd happy_var_2 ++ reverse happy_var_3 ++ [happy_var_4])) (fst happy_var_2)
	) `HappyStk` happyRest}}}}

happyReduce_20 = happySpecReduce_3  6# happyReduction_20
happyReduction_20 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	case happyOut25 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 PragmaEnd) -> 
	happyIn17
		 (let Loc l (OPTIONS (mc, s)) = happy_var_1
                                                      in OptionsPragma (l <^^> happy_var_3 <** (l:reverse happy_var_2 ++ [happy_var_3])) (readTool mc) s
	)}}}

happyReduce_21 = happySpecReduce_3  6# happyReduction_21
happyReduction_21 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 ANN) -> 
	case happyOut76 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 PragmaEnd) -> 
	happyIn17
		 (AnnModulePragma (happy_var_1 <^^> happy_var_3 <** [happy_var_1,happy_var_3]) happy_var_2
	)}}}

happyReduce_22 = happySpecReduce_3  7# happyReduction_22
happyReduction_22 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut227 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut18 happy_x_3 of { happy_var_3 -> 
	happyIn18
		 ((happy_var_1 : fst happy_var_3, happy_var_2 : snd happy_var_3)
	)}}}

happyReduce_23 = happySpecReduce_1  7# happyReduction_23
happyReduction_23 happy_x_1
	 =  case happyOut227 happy_x_1 of { happy_var_1 -> 
	happyIn18
		 (([happy_var_1],[])
	)}

happyReduce_24 = happySpecReduce_2  8# happyReduction_24
happyReduction_24 happy_x_2
	happy_x_1
	 =  case happyOut20 happy_x_1 of { happy_var_1 -> 
	case happyOut22 happy_x_2 of { happy_var_2 -> 
	happyIn19
		 (let (is,ds,ss1,inf) = happy_var_2
                 in \os ss l -> Module (l <++> inf <** (ss ++ ss1)) happy_var_1 os is ds
	)}}

happyReduce_25 = happyReduce 5# 9# happyReduction_25
happyReduction_25 (happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Module) -> 
	case happyOut238 happy_x_2 of { happy_var_2 -> 
	case happyOut21 happy_x_3 of { happy_var_3 -> 
	case happyOut26 happy_x_4 of { happy_var_4 -> 
	case happyOutTok happy_x_5 of { (Loc happy_var_5 KW_Where) -> 
	happyIn20
		 (Just $ ModuleHead (happy_var_1 <^^> happy_var_5 <** [happy_var_1,happy_var_5]) happy_var_2 happy_var_3 happy_var_4
	) `HappyStk` happyRest}}}}}

happyReduce_26 = happySpecReduce_0  9# happyReduction_26
happyReduction_26  =  happyIn20
		 (Nothing
	)

happyReduce_27 = happySpecReduce_3  10# happyReduction_27
happyReduction_27 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 DEPRECATED) -> 
	case happyOutTok happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 PragmaEnd) -> 
	happyIn21
		 (let Loc l (StringTok (s,_)) = happy_var_2 in Just $ DeprText (happy_var_1 <^^> happy_var_3 <** [happy_var_1,l,happy_var_3]) s
	)}}}

happyReduce_28 = happySpecReduce_3  10# happyReduction_28
happyReduction_28 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 WARNING) -> 
	case happyOutTok happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 PragmaEnd) -> 
	happyIn21
		 (let Loc l (StringTok (s,_)) = happy_var_2 in Just $ WarnText (happy_var_1 <^^> happy_var_3 <** [happy_var_1,l,happy_var_3]) s
	)}}}

happyReduce_29 = happySpecReduce_0  10# happyReduction_29
happyReduction_29  =  happyIn21
		 (Nothing
	)

happyReduce_30 = happySpecReduce_3  11# happyReduction_30
happyReduction_30 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftCurly) -> 
	case happyOut23 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightCurly) -> 
	happyIn22
		 (let (is,ds,ss) = happy_var_2 in (is,ds,happy_var_1:ss ++ [happy_var_3], happy_var_1 <^^> happy_var_3)
	)}}}

happyReduce_31 = happySpecReduce_3  11# happyReduction_31
happyReduction_31 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut236 happy_x_1 of { happy_var_1 -> 
	case happyOut23 happy_x_2 of { happy_var_2 -> 
	case happyOut237 happy_x_3 of { happy_var_3 -> 
	happyIn22
		 (let (is,ds,ss) = happy_var_2 in (is,ds,happy_var_1:ss ++ [happy_var_3], happy_var_1 <^^> happy_var_3)
	)}}}

happyReduce_32 = happyReduce 4# 12# happyReduction_32
happyReduction_32 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOut25 happy_x_1 of { happy_var_1 -> 
	case happyOut31 happy_x_2 of { happy_var_2 -> 
	case happyOut24 happy_x_3 of { happy_var_3 -> 
	case happyOut48 happy_x_4 of { happy_var_4 -> 
	happyIn23
		 ((reverse (fst happy_var_2), fst happy_var_4, reverse happy_var_1 ++ snd happy_var_2 ++ reverse happy_var_3 ++ snd happy_var_4)
	) `HappyStk` happyRest}}}}

happyReduce_33 = happySpecReduce_2  12# happyReduction_33
happyReduction_33 happy_x_2
	happy_x_1
	 =  case happyOut25 happy_x_1 of { happy_var_1 -> 
	case happyOut48 happy_x_2 of { happy_var_2 -> 
	happyIn23
		 (([], fst happy_var_2, reverse happy_var_1 ++ snd happy_var_2)
	)}}

happyReduce_34 = happySpecReduce_3  12# happyReduction_34
happyReduction_34 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut25 happy_x_1 of { happy_var_1 -> 
	case happyOut31 happy_x_2 of { happy_var_2 -> 
	case happyOut25 happy_x_3 of { happy_var_3 -> 
	happyIn23
		 ((reverse (fst happy_var_2), [], reverse happy_var_1 ++ snd happy_var_2 ++ reverse happy_var_3)
	)}}}

happyReduce_35 = happySpecReduce_1  12# happyReduction_35
happyReduction_35 happy_x_1
	 =  case happyOut25 happy_x_1 of { happy_var_1 -> 
	happyIn23
		 (([], [], reverse happy_var_1)
	)}

happyReduce_36 = happySpecReduce_2  13# happyReduction_36
happyReduction_36 happy_x_2
	happy_x_1
	 =  case happyOut25 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 SemiColon) -> 
	happyIn24
		 (happy_var_2 : happy_var_1
	)}}

happyReduce_37 = happySpecReduce_1  14# happyReduction_37
happyReduction_37 happy_x_1
	 =  case happyOut24 happy_x_1 of { happy_var_1 -> 
	happyIn25
		 (happy_var_1
	)}

happyReduce_38 = happySpecReduce_0  14# happyReduction_38
happyReduction_38  =  happyIn25
		 ([]
	)

happyReduce_39 = happySpecReduce_1  15# happyReduction_39
happyReduction_39 happy_x_1
	 =  case happyOut27 happy_x_1 of { happy_var_1 -> 
	happyIn26
		 (Just happy_var_1
	)}

happyReduce_40 = happySpecReduce_0  15# happyReduction_40
happyReduction_40  =  happyIn26
		 (Nothing
	)

happyReduce_41 = happyReduce 4# 16# happyReduction_41
happyReduction_41 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut29 happy_x_2 of { happy_var_2 -> 
	case happyOut28 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 RightParen) -> 
	happyIn27
		 (ExportSpecList (happy_var_1 <^^> happy_var_4 <** (happy_var_1:reverse (snd happy_var_2) ++ happy_var_3 ++ [happy_var_4])) (reverse (fst happy_var_2))
	) `HappyStk` happyRest}}}}

happyReduce_42 = happySpecReduce_3  16# happyReduction_42
happyReduction_42 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut28 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn27
		 (ExportSpecList (happy_var_1 <^^> happy_var_3 <** (happy_var_1:happy_var_2++[happy_var_3])) []
	)}}}

happyReduce_43 = happySpecReduce_1  17# happyReduction_43
happyReduction_43 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Comma) -> 
	happyIn28
		 ([happy_var_1]
	)}

happyReduce_44 = happySpecReduce_0  17# happyReduction_44
happyReduction_44  =  happyIn28
		 ([  ]
	)

happyReduce_45 = happySpecReduce_3  18# happyReduction_45
happyReduction_45 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut29 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut30 happy_x_3 of { happy_var_3 -> 
	happyIn29
		 ((happy_var_3 : fst happy_var_1, happy_var_2 : snd happy_var_1)
	)}}}

happyReduce_46 = happySpecReduce_1  18# happyReduction_46
happyReduction_46 happy_x_1
	 =  case happyOut30 happy_x_1 of { happy_var_1 -> 
	happyIn29
		 (([happy_var_1],[])
	)}

happyReduce_47 = happySpecReduce_1  19# happyReduction_47
happyReduction_47 happy_x_1
	 =  case happyOut209 happy_x_1 of { happy_var_1 -> 
	happyIn30
		 (EVar (ann happy_var_1) happy_var_1
	)}

happyReduce_48 = happySpecReduce_1  19# happyReduction_48
happyReduction_48 happy_x_1
	 =  case happyOut240 happy_x_1 of { happy_var_1 -> 
	happyIn30
		 (EAbs (ann happy_var_1) happy_var_1
	)}

happyReduce_49 = happyReduce 4# 19# happyReduction_49
happyReduction_49 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOut240 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftParen) -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 DotDot) -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 RightParen) -> 
	happyIn30
		 (EThingAll  (ann happy_var_1 <++> nIS happy_var_4 <** [happy_var_2,happy_var_3,happy_var_4]) happy_var_1
	) `HappyStk` happyRest}}}}

happyReduce_50 = happySpecReduce_3  19# happyReduction_50
happyReduction_50 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut240 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftParen) -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn30
		 (EThingWith (ann happy_var_1 <++> nIS happy_var_3 <** [happy_var_2,happy_var_3])    happy_var_1 []
	)}}}

happyReduce_51 = happyReduce 4# 19# happyReduction_51
happyReduction_51 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOut240 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftParen) -> 
	case happyOut42 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 RightParen) -> 
	happyIn30
		 (EThingWith (ann happy_var_1 <++> nIS happy_var_4 <** (happy_var_2:reverse (snd happy_var_3) ++ [happy_var_4])) happy_var_1 (reverse (fst happy_var_3))
	) `HappyStk` happyRest}}}}

happyReduce_52 = happySpecReduce_2  19# happyReduction_52
happyReduction_52 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Module) -> 
	case happyOut238 happy_x_2 of { happy_var_2 -> 
	happyIn30
		 (EModuleContents (nIS happy_var_1 <++> ann happy_var_2 <** [happy_var_1]) happy_var_2
	)}}

happyReduce_53 = happySpecReduce_3  20# happyReduction_53
happyReduction_53 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut31 happy_x_1 of { happy_var_1 -> 
	case happyOut24 happy_x_2 of { happy_var_2 -> 
	case happyOut32 happy_x_3 of { happy_var_3 -> 
	happyIn31
		 ((happy_var_3 : fst happy_var_1, snd happy_var_1 ++ reverse happy_var_2)
	)}}}

happyReduce_54 = happySpecReduce_1  20# happyReduction_54
happyReduction_54 happy_x_1
	 =  case happyOut32 happy_x_1 of { happy_var_1 -> 
	happyIn31
		 (([happy_var_1],[])
	)}

happyReduce_55 = happyReduce 7# 21# happyReduction_55
happyReduction_55 (happy_x_7 `HappyStk`
	happy_x_6 `HappyStk`
	happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Import) -> 
	case happyOut33 happy_x_2 of { happy_var_2 -> 
	case happyOut34 happy_x_3 of { happy_var_3 -> 
	case happyOut35 happy_x_4 of { happy_var_4 -> 
	case happyOut238 happy_x_5 of { happy_var_5 -> 
	case happyOut36 happy_x_6 of { happy_var_6 -> 
	case happyOut37 happy_x_7 of { happy_var_7 -> 
	happyIn32
		 (let { (mmn,ss,ml) = happy_var_6 ;
                                      l = nIS happy_var_1 <++> ann happy_var_5 <+?> ml <+?> (fmap ann) happy_var_7 <** (happy_var_1:snd happy_var_2 ++ snd happy_var_3 ++ snd happy_var_4 ++ ss)}
                                 in ImportDecl l happy_var_5 (fst happy_var_3) (fst happy_var_2) (fst happy_var_4) mmn happy_var_7
	) `HappyStk` happyRest}}}}}}}

happyReduce_56 = happySpecReduce_2  22# happyReduction_56
happyReduction_56 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 SOURCE) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 PragmaEnd) -> 
	happyIn33
		 ((True,[happy_var_1,happy_var_2])
	)}}

happyReduce_57 = happySpecReduce_0  22# happyReduction_57
happyReduction_57  =  happyIn33
		 ((False,[])
	)

happyReduce_58 = happySpecReduce_1  23# happyReduction_58
happyReduction_58 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Qualified) -> 
	happyIn34
		 ((True,[happy_var_1])
	)}

happyReduce_59 = happySpecReduce_0  23# happyReduction_59
happyReduction_59  =  happyIn34
		 ((False, [])
	)

happyReduce_60 = happyMonadReduce 1# 24# happyReduction_60
happyReduction_60 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { happy_var_1 -> 
	( do { checkEnabled PackageImports ;
                                                      let { Loc l (StringTok (s,_)) = happy_var_1 } ;
                                                      return $ (Just s,[l]) })}
	) (\r -> happyReturn (happyIn35 r))

happyReduce_61 = happySpecReduce_0  24# happyReduction_61
happyReduction_61  =  happyIn35
		 ((Nothing,[])
	)

happyReduce_62 = happySpecReduce_2  25# happyReduction_62
happyReduction_62 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_As) -> 
	case happyOut238 happy_x_2 of { happy_var_2 -> 
	happyIn36
		 ((Just happy_var_2,[happy_var_1],Just (nIS happy_var_1 <++> ann happy_var_2))
	)}}

happyReduce_63 = happySpecReduce_0  25# happyReduction_63
happyReduction_63  =  happyIn36
		 ((Nothing,[],Nothing)
	)

happyReduce_64 = happySpecReduce_1  26# happyReduction_64
happyReduction_64 happy_x_1
	 =  case happyOut38 happy_x_1 of { happy_var_1 -> 
	happyIn37
		 (Just happy_var_1
	)}

happyReduce_65 = happySpecReduce_0  26# happyReduction_65
happyReduction_65  =  happyIn37
		 (Nothing
	)

happyReduce_66 = happyReduce 5# 27# happyReduction_66
happyReduction_66 (happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOut39 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftParen) -> 
	case happyOut40 happy_x_3 of { happy_var_3 -> 
	case happyOut28 happy_x_4 of { happy_var_4 -> 
	case happyOutTok happy_x_5 of { (Loc happy_var_5 RightParen) -> 
	happyIn38
		 (let {(b,ml,s) = happy_var_1 ;
                                                      l = (ml <?+> (happy_var_2 <^^> happy_var_5)) <** (s ++ happy_var_2:reverse (snd happy_var_3) ++ happy_var_4 ++ [happy_var_5])}
                                                 in ImportSpecList l b (reverse (fst happy_var_3))
	) `HappyStk` happyRest}}}}}

happyReduce_67 = happyReduce 4# 27# happyReduction_67
happyReduction_67 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOut39 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftParen) -> 
	case happyOut28 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 RightParen) -> 
	happyIn38
		 (let {(b,ml,s) = happy_var_1 ; l = (ml <?+> (happy_var_2 <^^> happy_var_4)) <** (s ++ happy_var_2:happy_var_3 ++ [happy_var_4])}
                                                 in ImportSpecList l b []
	) `HappyStk` happyRest}}}}

happyReduce_68 = happySpecReduce_1  28# happyReduction_68
happyReduction_68 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Hiding) -> 
	happyIn39
		 ((True,Just (nIS happy_var_1),[happy_var_1])
	)}

happyReduce_69 = happySpecReduce_0  28# happyReduction_69
happyReduction_69  =  happyIn39
		 ((False,Nothing,[])
	)

happyReduce_70 = happySpecReduce_3  29# happyReduction_70
happyReduction_70 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut40 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut41 happy_x_3 of { happy_var_3 -> 
	happyIn40
		 ((happy_var_3 : fst happy_var_1, happy_var_2 : snd happy_var_1)
	)}}}

happyReduce_71 = happySpecReduce_1  29# happyReduction_71
happyReduction_71 happy_x_1
	 =  case happyOut41 happy_x_1 of { happy_var_1 -> 
	happyIn40
		 (([happy_var_1],[])
	)}

happyReduce_72 = happySpecReduce_1  30# happyReduction_72
happyReduction_72 happy_x_1
	 =  case happyOut207 happy_x_1 of { happy_var_1 -> 
	happyIn41
		 (IVar (ann happy_var_1) happy_var_1
	)}

happyReduce_73 = happySpecReduce_1  30# happyReduction_73
happyReduction_73 happy_x_1
	 =  case happyOut239 happy_x_1 of { happy_var_1 -> 
	happyIn41
		 (IAbs (ann happy_var_1) happy_var_1
	)}

happyReduce_74 = happyReduce 4# 30# happyReduction_74
happyReduction_74 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOut239 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftParen) -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 DotDot) -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 RightParen) -> 
	happyIn41
		 (IThingAll  (ann happy_var_1 <++> nIS happy_var_4 <** [happy_var_2,happy_var_3,happy_var_4]) happy_var_1
	) `HappyStk` happyRest}}}}

happyReduce_75 = happySpecReduce_3  30# happyReduction_75
happyReduction_75 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut239 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftParen) -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn41
		 (IThingWith (ann happy_var_1 <++> nIS happy_var_3 <** [happy_var_2,happy_var_3])    happy_var_1 []
	)}}}

happyReduce_76 = happyReduce 4# 30# happyReduction_76
happyReduction_76 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOut239 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftParen) -> 
	case happyOut42 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 RightParen) -> 
	happyIn41
		 (IThingWith (ann happy_var_1 <++> nIS happy_var_4 <** (happy_var_2:reverse (snd happy_var_3) ++ [happy_var_4])) happy_var_1 (reverse (fst happy_var_3))
	) `HappyStk` happyRest}}}}

happyReduce_77 = happySpecReduce_3  31# happyReduction_77
happyReduction_77 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut42 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut43 happy_x_3 of { happy_var_3 -> 
	happyIn42
		 ((happy_var_3 : fst happy_var_1, happy_var_2 : snd happy_var_1)
	)}}}

happyReduce_78 = happySpecReduce_1  31# happyReduction_78
happyReduction_78 happy_x_1
	 =  case happyOut43 happy_x_1 of { happy_var_1 -> 
	happyIn42
		 (([happy_var_1],[])
	)}

happyReduce_79 = happySpecReduce_1  32# happyReduction_79
happyReduction_79 happy_x_1
	 =  case happyOut207 happy_x_1 of { happy_var_1 -> 
	happyIn43
		 (VarName (ann happy_var_1) happy_var_1
	)}

happyReduce_80 = happySpecReduce_1  32# happyReduction_80
happyReduction_80 happy_x_1
	 =  case happyOut211 happy_x_1 of { happy_var_1 -> 
	happyIn43
		 (ConName (ann happy_var_1) happy_var_1
	)}

happyReduce_81 = happySpecReduce_3  33# happyReduction_81
happyReduction_81 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut46 happy_x_1 of { happy_var_1 -> 
	case happyOut45 happy_x_2 of { happy_var_2 -> 
	case happyOut47 happy_x_3 of { happy_var_3 -> 
	happyIn44
		 (let (ops,ss,l) = happy_var_3
                                                 in InfixDecl (ann happy_var_1 <++> l <** (snd happy_var_2 ++ reverse ss)) happy_var_1 (fst happy_var_2) (reverse ops)
	)}}}

happyReduce_82 = happySpecReduce_0  34# happyReduction_82
happyReduction_82  =  happyIn45
		 ((Nothing, [])
	)

happyReduce_83 = happyMonadReduce 1# 34# happyReduction_83
happyReduction_83 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { happy_var_1 -> 
	( let Loc l (IntTok (i,_)) = happy_var_1 in checkPrec i >>= \i -> return (Just i, [l]))}
	) (\r -> happyReturn (happyIn45 r))

happyReduce_84 = happySpecReduce_1  35# happyReduction_84
happyReduction_84 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Infix) -> 
	happyIn46
		 (AssocNone  $ nIS happy_var_1
	)}

happyReduce_85 = happySpecReduce_1  35# happyReduction_85
happyReduction_85 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_InfixL) -> 
	happyIn46
		 (AssocLeft  $ nIS happy_var_1
	)}

happyReduce_86 = happySpecReduce_1  35# happyReduction_86
happyReduction_86 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_InfixR) -> 
	happyIn46
		 (AssocRight $ nIS happy_var_1
	)}

happyReduce_87 = happySpecReduce_3  36# happyReduction_87
happyReduction_87 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut47 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut218 happy_x_3 of { happy_var_3 -> 
	happyIn47
		 (let (ops,ss,l) = happy_var_1 in (happy_var_3 : ops, happy_var_2 : ss, l <++> ann happy_var_3)
	)}}}

happyReduce_88 = happySpecReduce_1  36# happyReduction_88
happyReduction_88 happy_x_1
	 =  case happyOut218 happy_x_1 of { happy_var_1 -> 
	happyIn47
		 (([happy_var_1],[],ann happy_var_1)
	)}

happyReduce_89 = happyMonadReduce 2# 37# happyReduction_89
happyReduction_89 (happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut49 happy_x_1 of { happy_var_1 -> 
	case happyOut25 happy_x_2 of { happy_var_2 -> 
	( checkRevDecls (fst happy_var_1) >>= \ds -> return (ds, snd happy_var_1 ++ reverse happy_var_2))}}
	) (\r -> happyReturn (happyIn48 r))

happyReduce_90 = happySpecReduce_3  38# happyReduction_90
happyReduction_90 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut49 happy_x_1 of { happy_var_1 -> 
	case happyOut24 happy_x_2 of { happy_var_2 -> 
	case happyOut50 happy_x_3 of { happy_var_3 -> 
	happyIn49
		 ((happy_var_3 : fst happy_var_1, snd happy_var_1 ++ reverse happy_var_2)
	)}}}

happyReduce_91 = happySpecReduce_1  38# happyReduction_91
happyReduction_91 happy_x_1
	 =  case happyOut50 happy_x_1 of { happy_var_1 -> 
	happyIn49
		 (([happy_var_1],[])
	)}

happyReduce_92 = happyMonadReduce 4# 39# happyReduction_92
happyReduction_92 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Type) -> 
	case happyOut78 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 Equals) -> 
	case happyOut92 happy_x_4 of { happy_var_4 -> 
	( do { dh <- checkSimpleType happy_var_2;
                       let {l = nIS happy_var_1 <++> ann happy_var_4 <** [happy_var_1,happy_var_3]};
                       return (TypeDecl l dh happy_var_4) })}}}}
	) (\r -> happyReturn (happyIn50 r))

happyReduce_93 = happyMonadReduce 4# 39# happyReduction_93
happyReduction_93 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Type) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 KW_Family) -> 
	case happyOut80 happy_x_3 of { happy_var_3 -> 
	case happyOut129 happy_x_4 of { happy_var_4 -> 
	( do { dh <- checkSimpleType happy_var_3;
                       let {l = nIS happy_var_1 <++> ann happy_var_3 <+?> (fmap ann) (fst happy_var_4) <** (happy_var_1:happy_var_2:snd happy_var_4)};
                       return (TypeFamDecl l dh (fst happy_var_4)) })}}}}
	) (\r -> happyReturn (happyIn50 r))

happyReduce_94 = happyMonadReduce 5# 39# happyReduction_94
happyReduction_94 (happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Type) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 KW_Instance) -> 
	case happyOut77 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 Equals) -> 
	case happyOut92 happy_x_5 of { happy_var_5 -> 
	( do { -- no checkSimpleType happy_var_4 since dtype may contain type patterns
                       checkEnabled TypeFamilies ;
                       let {l = nIS happy_var_1 <++> ann happy_var_5 <** [happy_var_1,happy_var_2,happy_var_4]};
                       return (TypeInsDecl l happy_var_3 happy_var_5) })}}}}}
	) (\r -> happyReturn (happyIn50 r))

happyReduce_95 = happyMonadReduce 4# 39# happyReduction_95
happyReduction_95 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut51 happy_x_1 of { happy_var_1 -> 
	case happyOut93 happy_x_2 of { happy_var_2 -> 
	case happyOut108 happy_x_3 of { happy_var_3 -> 
	case happyOut120 happy_x_4 of { happy_var_4 -> 
	( do { (cs,dh) <- checkDataHeader happy_var_2;
                       let { (qds,ss,minf) = happy_var_3;
                             l = happy_var_1 <> happy_var_2 <+?> minf <+?> fmap ann happy_var_4 <** ss};
                       checkDataOrNew happy_var_1 qds;
                       return (DataDecl l happy_var_1 cs dh (reverse qds) happy_var_4) })}}}}
	) (\r -> happyReturn (happyIn50 r))

happyReduce_96 = happyMonadReduce 5# 39# happyReduction_96
happyReduction_96 (happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut51 happy_x_1 of { happy_var_1 -> 
	case happyOut93 happy_x_2 of { happy_var_2 -> 
	case happyOut129 happy_x_3 of { happy_var_3 -> 
	case happyOut104 happy_x_4 of { happy_var_4 -> 
	case happyOut120 happy_x_5 of { happy_var_5 -> 
	( do { (cs,dh) <- checkDataHeader happy_var_2;
                       let { (gs,ss,minf) = happy_var_4;
                             l = ann happy_var_1 <+?> minf <+?> fmap ann happy_var_5 <** (snd happy_var_3 ++ ss)};
                       checkDataOrNewG happy_var_1 gs;
                       case (gs, fst happy_var_3) of
                        ([], Nothing) -> return (DataDecl l happy_var_1 cs dh [] happy_var_5)
                        _ -> checkEnabled GADTs >> return (GDataDecl l happy_var_1 cs dh (fst happy_var_3) (reverse gs) happy_var_5) })}}}}}
	) (\r -> happyReturn (happyIn50 r))

happyReduce_97 = happyMonadReduce 4# 39# happyReduction_97
happyReduction_97 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Data) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 KW_Family) -> 
	case happyOut93 happy_x_3 of { happy_var_3 -> 
	case happyOut129 happy_x_4 of { happy_var_4 -> 
	( do { (cs,dh) <- checkDataHeader happy_var_3;
                       let {l = nIS happy_var_1 <++> ann happy_var_3 <+?> (fmap ann) (fst happy_var_4) <** (happy_var_1:happy_var_2:snd happy_var_4)};
                       return (DataFamDecl l cs dh (fst happy_var_4)) })}}}}
	) (\r -> happyReturn (happyIn50 r))

happyReduce_98 = happyMonadReduce 5# 39# happyReduction_98
happyReduction_98 (happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut51 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 KW_Instance) -> 
	case happyOut92 happy_x_3 of { happy_var_3 -> 
	case happyOut108 happy_x_4 of { happy_var_4 -> 
	case happyOut120 happy_x_5 of { happy_var_5 -> 
	( do { -- (cs,c,t) <- checkDataHeader happy_var_4;
                       checkEnabled TypeFamilies ;
                       let { (qds,ss,minf) = happy_var_4 ;
                             l = happy_var_1 <> happy_var_3 <+?> minf <+?> fmap ann happy_var_5 <** happy_var_2:ss };
                       checkDataOrNew happy_var_1 qds;
                       return (DataInsDecl l happy_var_1 happy_var_3 (reverse qds) happy_var_5) })}}}}}
	) (\r -> happyReturn (happyIn50 r))

happyReduce_99 = happyMonadReduce 6# 39# happyReduction_99
happyReduction_99 (happy_x_6 `HappyStk`
	happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut51 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 KW_Instance) -> 
	case happyOut92 happy_x_3 of { happy_var_3 -> 
	case happyOut129 happy_x_4 of { happy_var_4 -> 
	case happyOut104 happy_x_5 of { happy_var_5 -> 
	case happyOut120 happy_x_6 of { happy_var_6 -> 
	( do { -- (cs,c,t) <- checkDataHeader happy_var_4;
                       checkEnabled TypeFamilies ;
                       let {(gs,ss,minf) = happy_var_5;
                            l = ann happy_var_1 <+?> minf <+?> fmap ann happy_var_6 <** (happy_var_2:snd happy_var_4 ++ ss)};
                       checkDataOrNewG happy_var_1 gs;
                       return (GDataInsDecl l happy_var_1 happy_var_3 (fst happy_var_4) (reverse gs) happy_var_6) })}}}}}}
	) (\r -> happyReturn (happyIn50 r))

happyReduce_100 = happyMonadReduce 4# 39# happyReduction_100
happyReduction_100 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Class) -> 
	case happyOut93 happy_x_2 of { happy_var_2 -> 
	case happyOut101 happy_x_3 of { happy_var_3 -> 
	case happyOut130 happy_x_4 of { happy_var_4 -> 
	( do { (cs,dh) <- checkClassHeader happy_var_2;
                       let {(fds,ss1,minf1) = happy_var_3;(mcs,ss2,minf2) = happy_var_4} ;
                       let { l = nIS happy_var_1 <++> ann happy_var_2 <+?> minf1 <+?> minf2 <** (happy_var_1:ss1 ++ ss2)} ;
                       return (ClassDecl l cs dh fds mcs) })}}}}
	) (\r -> happyReturn (happyIn50 r))

happyReduce_101 = happyMonadReduce 3# 39# happyReduction_101
happyReduction_101 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Instance) -> 
	case happyOut93 happy_x_2 of { happy_var_2 -> 
	case happyOut135 happy_x_3 of { happy_var_3 -> 
	( do { (cs,ih) <- checkInstHeader happy_var_2;
                       let {(mis,ss,minf) = happy_var_3};
                       return (InstDecl (nIS happy_var_1 <++> ann happy_var_2 <+?> minf <** (happy_var_1:ss)) cs ih mis) })}}}
	) (\r -> happyReturn (happyIn50 r))

happyReduce_102 = happyMonadReduce 3# 39# happyReduction_102
happyReduction_102 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Deriving) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 KW_Instance) -> 
	case happyOut93 happy_x_3 of { happy_var_3 -> 
	( do { checkEnabled StandaloneDeriving ;
                       (cs, ih) <- checkInstHeader happy_var_3;
                       let {l = nIS happy_var_1 <++> ann happy_var_3 <** [happy_var_1,happy_var_2]};
                       return (DerivDecl l cs ih) })}}}
	) (\r -> happyReturn (happyIn50 r))

happyReduce_103 = happyReduce 4# 39# happyReduction_103
happyReduction_103 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Default) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftParen) -> 
	case happyOut52 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 RightParen) -> 
	happyIn50
		 (DefaultDecl (happy_var_1 <^^> happy_var_4 <** (happy_var_1:happy_var_2 : snd happy_var_3 ++ [happy_var_4])) (fst happy_var_3)
	) `HappyStk` happyRest}}}}

happyReduce_104 = happyMonadReduce 1# 39# happyReduction_104
happyReduction_104 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut148 happy_x_1 of { happy_var_1 -> 
	( do
              checkToplevel happy_var_1
              checkExpr happy_var_1 >>= \e -> return (SpliceDecl (ann e) e))}
	) (\r -> happyReturn (happyIn50 r))

happyReduce_105 = happyReduce 5# 39# happyReduction_105
happyReduction_105 (happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Foreign) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 KW_Import) -> 
	case happyOut63 happy_x_3 of { happy_var_3 -> 
	case happyOut64 happy_x_4 of { happy_var_4 -> 
	case happyOut65 happy_x_5 of { happy_var_5 -> 
	happyIn50
		 (let (s,n,t,ss) = happy_var_5 in ForImp (nIS happy_var_1 <++> ann t <** (happy_var_1:happy_var_2:ss)) happy_var_3 happy_var_4 s n t
	) `HappyStk` happyRest}}}}}

happyReduce_106 = happyReduce 4# 39# happyReduction_106
happyReduction_106 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Foreign) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 KW_Export) -> 
	case happyOut63 happy_x_3 of { happy_var_3 -> 
	case happyOut65 happy_x_4 of { happy_var_4 -> 
	happyIn50
		 (let (s,n,t,ss) = happy_var_4 in ForExp (nIS happy_var_1 <++> ann t <** (happy_var_1:happy_var_2:ss)) happy_var_3    s n t
	) `HappyStk` happyRest}}}}

happyReduce_107 = happySpecReduce_3  39# happyReduction_107
happyReduction_107 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 RULES) -> 
	case happyOut66 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 PragmaEnd) -> 
	happyIn50
		 (RulePragmaDecl (happy_var_1 <^^> happy_var_3 <** [happy_var_1,happy_var_3]) $ reverse happy_var_2
	)}}}

happyReduce_108 = happySpecReduce_3  39# happyReduction_108
happyReduction_108 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 DEPRECATED) -> 
	case happyOut72 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 PragmaEnd) -> 
	happyIn50
		 (DeprPragmaDecl (happy_var_1 <^^> happy_var_3 <** (happy_var_1:snd happy_var_2++[happy_var_3])) $ reverse (fst happy_var_2)
	)}}}

happyReduce_109 = happySpecReduce_3  39# happyReduction_109
happyReduction_109 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 WARNING) -> 
	case happyOut72 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 PragmaEnd) -> 
	happyIn50
		 (WarnPragmaDecl (happy_var_1 <^^> happy_var_3 <** (happy_var_1:snd happy_var_2++[happy_var_3])) $ reverse (fst happy_var_2)
	)}}}

happyReduce_110 = happySpecReduce_3  39# happyReduction_110
happyReduction_110 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 ANN) -> 
	case happyOut76 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 PragmaEnd) -> 
	happyIn50
		 (AnnPragma      (happy_var_1 <^^> happy_var_3 <** [happy_var_1,happy_var_3]) happy_var_2
	)}}}

happyReduce_111 = happySpecReduce_1  39# happyReduction_111
happyReduction_111 happy_x_1
	 =  case happyOut55 happy_x_1 of { happy_var_1 -> 
	happyIn50
		 (happy_var_1
	)}

happyReduce_112 = happySpecReduce_1  40# happyReduction_112
happyReduction_112 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Data) -> 
	happyIn51
		 (DataType $ nIS happy_var_1
	)}

happyReduce_113 = happySpecReduce_1  40# happyReduction_113
happyReduction_113 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_NewType) -> 
	happyIn51
		 (NewType  $ nIS happy_var_1
	)}

happyReduce_114 = happyMonadReduce 1# 41# happyReduction_114
happyReduction_114 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut95 happy_x_1 of { happy_var_1 -> 
	( do { ts <- mapM checkType (fst happy_var_1);
                                              return $ (reverse ts, reverse (snd happy_var_1)) })}
	) (\r -> happyReturn (happyIn52 r))

happyReduce_115 = happySpecReduce_1  41# happyReduction_115
happyReduction_115 happy_x_1
	 =  case happyOut79 happy_x_1 of { happy_var_1 -> 
	happyIn52
		 (([happy_var_1],[])
	)}

happyReduce_116 = happySpecReduce_0  41# happyReduction_116
happyReduction_116  =  happyIn52
		 (([],[])
	)

happyReduce_117 = happyMonadReduce 3# 42# happyReduction_117
happyReduction_117 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut25 happy_x_1 of { happy_var_1 -> 
	case happyOut54 happy_x_2 of { happy_var_2 -> 
	case happyOut25 happy_x_3 of { happy_var_3 -> 
	( checkRevDecls (fst happy_var_2) >>= \ds -> return (ds, reverse happy_var_1 ++ snd happy_var_2 ++ reverse happy_var_3))}}}
	) (\r -> happyReturn (happyIn53 r))

happyReduce_118 = happySpecReduce_1  42# happyReduction_118
happyReduction_118 happy_x_1
	 =  case happyOut25 happy_x_1 of { happy_var_1 -> 
	happyIn53
		 (([],reverse happy_var_1)
	)}

happyReduce_119 = happySpecReduce_3  43# happyReduction_119
happyReduction_119 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut54 happy_x_1 of { happy_var_1 -> 
	case happyOut24 happy_x_2 of { happy_var_2 -> 
	case happyOut55 happy_x_3 of { happy_var_3 -> 
	happyIn54
		 ((happy_var_3 : fst happy_var_1, snd happy_var_1 ++ reverse happy_var_2)
	)}}}

happyReduce_120 = happySpecReduce_1  43# happyReduction_120
happyReduction_120 happy_x_1
	 =  case happyOut55 happy_x_1 of { happy_var_1 -> 
	happyIn54
		 (([happy_var_1],[])
	)}

happyReduce_121 = happySpecReduce_1  44# happyReduction_121
happyReduction_121 happy_x_1
	 =  case happyOut57 happy_x_1 of { happy_var_1 -> 
	happyIn55
		 (happy_var_1
	)}

happyReduce_122 = happySpecReduce_1  44# happyReduction_122
happyReduction_122 happy_x_1
	 =  case happyOut44 happy_x_1 of { happy_var_1 -> 
	happyIn55
		 (happy_var_1
	)}

happyReduce_123 = happySpecReduce_1  44# happyReduction_123
happyReduction_123 happy_x_1
	 =  case happyOut140 happy_x_1 of { happy_var_1 -> 
	happyIn55
		 (happy_var_1
	)}

happyReduce_124 = happySpecReduce_3  45# happyReduction_124
happyReduction_124 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftCurly) -> 
	case happyOut53 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightCurly) -> 
	happyIn56
		 (BDecls (happy_var_1 <^^> happy_var_3 <** (happy_var_1:snd happy_var_2++[happy_var_3])) (fst happy_var_2)
	)}}}

happyReduce_125 = happySpecReduce_3  45# happyReduction_125
happyReduction_125 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut236 happy_x_1 of { happy_var_1 -> 
	case happyOut53 happy_x_2 of { happy_var_2 -> 
	case happyOut237 happy_x_3 of { happy_var_3 -> 
	happyIn56
		 (BDecls (happy_var_1 <^^> happy_var_3 <** (happy_var_1:snd happy_var_2++[happy_var_3])) (fst happy_var_2)
	)}}}

happyReduce_126 = happyMonadReduce 3# 46# happyReduction_126
happyReduction_126 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut150 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 DoubleColon) -> 
	case happyOut92 happy_x_3 of { happy_var_3 -> 
	( do { v <- checkSigVar happy_var_1;
                                                               return $ TypeSig (happy_var_1 <> happy_var_3 <** [happy_var_2]) [v] happy_var_3 })}}}
	) (\r -> happyReturn (happyIn57 r))

happyReduce_127 = happyMonadReduce 5# 46# happyReduction_127
happyReduction_127 (happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut150 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut62 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 DoubleColon) -> 
	case happyOut92 happy_x_5 of { happy_var_5 -> 
	( do { v <- checkSigVar happy_var_1;
                                                               let {(vs,ss,_) = happy_var_3 ; l = happy_var_1 <> happy_var_5 <** (happy_var_2 : reverse ss ++ [happy_var_4]) } ;
                                                               return $ TypeSig l (v : reverse vs) happy_var_5 })}}}}}
	) (\r -> happyReturn (happyIn57 r))

happyReduce_128 = happySpecReduce_1  46# happyReduction_128
happyReduction_128 happy_x_1
	 =  case happyOut58 happy_x_1 of { happy_var_1 -> 
	happyIn57
		 (happy_var_1
	)}

happyReduce_129 = happyReduce 4# 47# happyReduction_129
happyReduction_129 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { happy_var_1 -> 
	case happyOut68 happy_x_2 of { happy_var_2 -> 
	case happyOut209 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 PragmaEnd) -> 
	happyIn58
		 (let Loc l (INLINE s) = happy_var_1 in InlineSig (l <^^> happy_var_4 <** [l,happy_var_4]) s happy_var_2 happy_var_3
	) `HappyStk` happyRest}}}}

happyReduce_130 = happyReduce 4# 47# happyReduction_130
happyReduction_130 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 INLINE_CONLIKE) -> 
	case happyOut68 happy_x_2 of { happy_var_2 -> 
	case happyOut209 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 PragmaEnd) -> 
	happyIn58
		 (InlineConlikeSig (happy_var_1 <^^> happy_var_4 <** [happy_var_1,happy_var_4]) happy_var_2 happy_var_3
	) `HappyStk` happyRest}}}}

happyReduce_131 = happyReduce 6# 47# happyReduction_131
happyReduction_131 (happy_x_6 `HappyStk`
	happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 SPECIALISE) -> 
	case happyOut68 happy_x_2 of { happy_var_2 -> 
	case happyOut209 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 DoubleColon) -> 
	case happyOut59 happy_x_5 of { happy_var_5 -> 
	case happyOutTok happy_x_6 of { (Loc happy_var_6 PragmaEnd) -> 
	happyIn58
		 (SpecSig (happy_var_1 <^^> happy_var_6 <** (happy_var_1: happy_var_4 : snd happy_var_5 ++ [happy_var_6])) happy_var_2 happy_var_3 (fst happy_var_5)
	) `HappyStk` happyRest}}}}}}

happyReduce_132 = happyReduce 6# 47# happyReduction_132
happyReduction_132 (happy_x_6 `HappyStk`
	happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { happy_var_1 -> 
	case happyOut68 happy_x_2 of { happy_var_2 -> 
	case happyOut209 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 DoubleColon) -> 
	case happyOut59 happy_x_5 of { happy_var_5 -> 
	case happyOutTok happy_x_6 of { (Loc happy_var_6 PragmaEnd) -> 
	happyIn58
		 (let Loc l (SPECIALISE_INLINE s) = happy_var_1
               in SpecInlineSig (l <^^> happy_var_6 <** (l:happy_var_4:snd happy_var_5++[happy_var_6])) s happy_var_2 happy_var_3 (fst happy_var_5)
	) `HappyStk` happyRest}}}}}}

happyReduce_133 = happyMonadReduce 4# 47# happyReduction_133
happyReduction_133 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 SPECIALISE) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 KW_Instance) -> 
	case happyOut93 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 PragmaEnd) -> 
	( do { (cs,ih) <- checkInstHeader happy_var_3;
                                                               let {l = happy_var_1 <^^> happy_var_4 <** [happy_var_1,happy_var_2,happy_var_4]};
                                                               return $ InstSig l cs ih })}}}}
	) (\r -> happyReturn (happyIn58 r))

happyReduce_134 = happySpecReduce_1  48# happyReduction_134
happyReduction_134 happy_x_1
	 =  case happyOut60 happy_x_1 of { happy_var_1 -> 
	happyIn59
		 (([happy_var_1],[])
	)}

happyReduce_135 = happySpecReduce_3  48# happyReduction_135
happyReduction_135 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut60 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut59 happy_x_3 of { happy_var_3 -> 
	happyIn59
		 ((happy_var_1 : fst happy_var_3, happy_var_2 : snd happy_var_3)
	)}}}

happyReduce_136 = happyMonadReduce 1# 49# happyReduction_136
happyReduction_136 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut93 happy_x_1 of { happy_var_1 -> 
	( checkType $ mkTyForall (ann happy_var_1) Nothing Nothing happy_var_1)}
	) (\r -> happyReturn (happyIn60 r))

happyReduce_137 = happySpecReduce_1  50# happyReduction_137
happyReduction_137 happy_x_1
	 =  case happyOut56 happy_x_1 of { happy_var_1 -> 
	happyIn61
		 (happy_var_1
	)}

happyReduce_138 = happySpecReduce_3  50# happyReduction_138
happyReduction_138 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftCurly) -> 
	case happyOut203 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightCurly) -> 
	happyIn61
		 (IPBinds (happy_var_1 <^^> happy_var_3 <** snd happy_var_2) (fst happy_var_2)
	)}}}

happyReduce_139 = happySpecReduce_3  50# happyReduction_139
happyReduction_139 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut236 happy_x_1 of { happy_var_1 -> 
	case happyOut203 happy_x_2 of { happy_var_2 -> 
	case happyOut237 happy_x_3 of { happy_var_3 -> 
	happyIn61
		 (IPBinds (happy_var_1 <^^> happy_var_3 <** snd happy_var_2) (fst happy_var_2)
	)}}}

happyReduce_140 = happySpecReduce_3  51# happyReduction_140
happyReduction_140 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut62 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut207 happy_x_3 of { happy_var_3 -> 
	happyIn62
		 (let (ns,ss,l) = happy_var_1 in (happy_var_3 : ns, happy_var_2 : ss, l <++> ann happy_var_3)
	)}}}

happyReduce_141 = happyMonadReduce 1# 51# happyReduction_141
happyReduction_141 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut209 happy_x_1 of { happy_var_1 -> 
	( do { n <- checkUnQual happy_var_1;
                                              return ([n],[],ann n) })}
	) (\r -> happyReturn (happyIn62 r))

happyReduce_142 = happySpecReduce_1  52# happyReduction_142
happyReduction_142 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_StdCall) -> 
	happyIn63
		 (StdCall   (nIS happy_var_1)
	)}

happyReduce_143 = happySpecReduce_1  52# happyReduction_143
happyReduction_143 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_CCall) -> 
	happyIn63
		 (CCall     (nIS happy_var_1)
	)}

happyReduce_144 = happySpecReduce_1  52# happyReduction_144
happyReduction_144 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_CPlusPlus) -> 
	happyIn63
		 (CPlusPlus (nIS happy_var_1)
	)}

happyReduce_145 = happySpecReduce_1  52# happyReduction_145
happyReduction_145 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_DotNet) -> 
	happyIn63
		 (DotNet    (nIS happy_var_1)
	)}

happyReduce_146 = happySpecReduce_1  52# happyReduction_146
happyReduction_146 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Jvm) -> 
	happyIn63
		 (Jvm       (nIS happy_var_1)
	)}

happyReduce_147 = happySpecReduce_1  52# happyReduction_147
happyReduction_147 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Js) -> 
	happyIn63
		 (Js        (nIS happy_var_1)
	)}

happyReduce_148 = happySpecReduce_1  52# happyReduction_148
happyReduction_148 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_CApi) -> 
	happyIn63
		 (CApi      (nIS happy_var_1)
	)}

happyReduce_149 = happySpecReduce_1  53# happyReduction_149
happyReduction_149 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Safe) -> 
	happyIn64
		 (Just $ PlaySafe  (nIS happy_var_1) False
	)}

happyReduce_150 = happySpecReduce_1  53# happyReduction_150
happyReduction_150 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Unsafe) -> 
	happyIn64
		 (Just $ PlayRisky (nIS happy_var_1)
	)}

happyReduce_151 = happySpecReduce_1  53# happyReduction_151
happyReduction_151 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Threadsafe) -> 
	happyIn64
		 (Just $ PlaySafe  (nIS happy_var_1) True
	)}

happyReduce_152 = happySpecReduce_1  53# happyReduction_152
happyReduction_152 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Interruptible) -> 
	happyIn64
		 (Just $ PlayInterruptible (nIS happy_var_1)
	)}

happyReduce_153 = happySpecReduce_0  53# happyReduction_153
happyReduction_153  =  happyIn64
		 (Nothing
	)

happyReduce_154 = happyReduce 4# 54# happyReduction_154
happyReduction_154 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { happy_var_1 -> 
	case happyOut208 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 DoubleColon) -> 
	case happyOut77 happy_x_4 of { happy_var_4 -> 
	happyIn65
		 (let Loc l (StringTok (s,_)) = happy_var_1 in (Just s, happy_var_2, happy_var_4, [l,happy_var_3])
	) `HappyStk` happyRest}}}}

happyReduce_155 = happySpecReduce_3  54# happyReduction_155
happyReduction_155 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut208 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 DoubleColon) -> 
	case happyOut77 happy_x_3 of { happy_var_3 -> 
	happyIn65
		 ((Nothing, happy_var_1, happy_var_3, [happy_var_2])
	)}}}

happyReduce_156 = happySpecReduce_3  55# happyReduction_156
happyReduction_156 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut66 happy_x_1 of { happy_var_1 -> 
	case happyOut67 happy_x_3 of { happy_var_3 -> 
	happyIn66
		 (happy_var_3 : happy_var_1
	)}}

happyReduce_157 = happySpecReduce_2  55# happyReduction_157
happyReduction_157 happy_x_2
	happy_x_1
	 =  case happyOut66 happy_x_1 of { happy_var_1 -> 
	happyIn66
		 (happy_var_1
	)}

happyReduce_158 = happySpecReduce_1  55# happyReduction_158
happyReduction_158 happy_x_1
	 =  case happyOut67 happy_x_1 of { happy_var_1 -> 
	happyIn66
		 ([happy_var_1]
	)}

happyReduce_159 = happySpecReduce_0  55# happyReduction_159
happyReduction_159  =  happyIn66
		 ([]
	)

happyReduce_160 = happyMonadReduce 6# 56# happyReduction_160
happyReduction_160 (happy_x_6 `HappyStk`
	happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { happy_var_1 -> 
	case happyOut68 happy_x_2 of { happy_var_2 -> 
	case happyOut69 happy_x_3 of { happy_var_3 -> 
	case happyOut148 happy_x_4 of { happy_var_4 -> 
	case happyOutTok happy_x_5 of { (Loc happy_var_5 Equals) -> 
	case happyOut146 happy_x_6 of { happy_var_6 -> 
	( do { let {Loc l (StringTok (s,_)) = happy_var_1};
                                                                  e <- checkRuleExpr happy_var_4;
                                                                  return $ Rule (nIS l <++> ann happy_var_6 <** l:snd happy_var_3 ++ [happy_var_5]) s happy_var_2 (fst happy_var_3) e happy_var_6 })}}}}}}
	) (\r -> happyReturn (happyIn67 r))

happyReduce_161 = happySpecReduce_0  57# happyReduction_161
happyReduction_161  =  happyIn68
		 (Nothing
	)

happyReduce_162 = happySpecReduce_3  57# happyReduction_162
happyReduction_162 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftSquare) -> 
	case happyOutTok happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightSquare) -> 
	happyIn68
		 (let Loc l (IntTok (i,_)) = happy_var_2 in Just $ ActiveFrom  (happy_var_1 <^^> happy_var_3 <** [happy_var_1,l,happy_var_3])    (fromInteger i)
	)}}}

happyReduce_163 = happyReduce 4# 57# happyReduction_163
happyReduction_163 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftSquare) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Tilde) -> 
	case happyOutTok happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 RightSquare) -> 
	happyIn68
		 (let Loc l (IntTok (i,_)) = happy_var_3 in Just $ ActiveUntil (happy_var_1 <^^> happy_var_4 <** [happy_var_1,happy_var_2,l,happy_var_4]) (fromInteger i)
	) `HappyStk` happyRest}}}}

happyReduce_164 = happySpecReduce_0  58# happyReduction_164
happyReduction_164  =  happyIn69
		 ((Nothing,[])
	)

happyReduce_165 = happySpecReduce_3  58# happyReduction_165
happyReduction_165 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Forall) -> 
	case happyOut70 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 Dot) -> 
	happyIn69
		 ((Just happy_var_2,[happy_var_1,happy_var_3])
	)}}}

happyReduce_166 = happySpecReduce_1  59# happyReduction_166
happyReduction_166 happy_x_1
	 =  case happyOut71 happy_x_1 of { happy_var_1 -> 
	happyIn70
		 ([happy_var_1]
	)}

happyReduce_167 = happySpecReduce_2  59# happyReduction_167
happyReduction_167 happy_x_2
	happy_x_1
	 =  case happyOut71 happy_x_1 of { happy_var_1 -> 
	case happyOut70 happy_x_2 of { happy_var_2 -> 
	happyIn70
		 (happy_var_1 : happy_var_2
	)}}

happyReduce_168 = happySpecReduce_1  60# happyReduction_168
happyReduction_168 happy_x_1
	 =  case happyOut224 happy_x_1 of { happy_var_1 -> 
	happyIn71
		 (RuleVar (ann happy_var_1) happy_var_1
	)}

happyReduce_169 = happyReduce 5# 60# happyReduction_169
happyReduction_169 (happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut224 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 DoubleColon) -> 
	case happyOut92 happy_x_4 of { happy_var_4 -> 
	case happyOutTok happy_x_5 of { (Loc happy_var_5 RightParen) -> 
	happyIn71
		 (TypedRuleVar (happy_var_1 <^^> happy_var_5 <** [happy_var_1,happy_var_3,happy_var_5]) happy_var_2 happy_var_4
	) `HappyStk` happyRest}}}}}

happyReduce_170 = happySpecReduce_3  61# happyReduction_170
happyReduction_170 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut72 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 SemiColon) -> 
	case happyOut73 happy_x_3 of { happy_var_3 -> 
	happyIn72
		 ((fst happy_var_3 : fst happy_var_1, snd happy_var_1 ++ (happy_var_2:snd happy_var_3))
	)}}}

happyReduce_171 = happySpecReduce_2  61# happyReduction_171
happyReduction_171 happy_x_2
	happy_x_1
	 =  case happyOut72 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 SemiColon) -> 
	happyIn72
		 ((fst happy_var_1, snd happy_var_1 ++ [happy_var_2])
	)}}

happyReduce_172 = happySpecReduce_1  61# happyReduction_172
happyReduction_172 happy_x_1
	 =  case happyOut73 happy_x_1 of { happy_var_1 -> 
	happyIn72
		 (([fst happy_var_1],snd happy_var_1)
	)}

happyReduce_173 = happySpecReduce_0  61# happyReduction_173
happyReduction_173  =  happyIn72
		 (([],[])
	)

happyReduce_174 = happySpecReduce_2  62# happyReduction_174
happyReduction_174 happy_x_2
	happy_x_1
	 =  case happyOut74 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { happy_var_2 -> 
	happyIn73
		 (let Loc l (StringTok (s,_)) = happy_var_2 in ((fst happy_var_1,s),snd happy_var_1 ++ [l])
	)}}

happyReduce_175 = happySpecReduce_1  63# happyReduction_175
happyReduction_175 happy_x_1
	 =  case happyOut75 happy_x_1 of { happy_var_1 -> 
	happyIn74
		 (([happy_var_1],[])
	)}

happyReduce_176 = happySpecReduce_3  63# happyReduction_176
happyReduction_176 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut75 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut74 happy_x_3 of { happy_var_3 -> 
	happyIn74
		 ((happy_var_1 : fst happy_var_3, happy_var_2 : snd happy_var_3)
	)}}}

happyReduce_177 = happySpecReduce_1  64# happyReduction_177
happyReduction_177 happy_x_1
	 =  case happyOut211 happy_x_1 of { happy_var_1 -> 
	happyIn75
		 (happy_var_1
	)}

happyReduce_178 = happySpecReduce_1  64# happyReduction_178
happyReduction_178 happy_x_1
	 =  case happyOut207 happy_x_1 of { happy_var_1 -> 
	happyIn75
		 (happy_var_1
	)}

happyReduce_179 = happyMonadReduce 3# 65# happyReduction_179
happyReduction_179 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Type) -> 
	case happyOut227 happy_x_2 of { happy_var_2 -> 
	case happyOut159 happy_x_3 of { happy_var_3 -> 
	( checkExpr happy_var_3 >>= \e -> return (TypeAnn   (nIS happy_var_1 <++> ann e <** [happy_var_1]) happy_var_2 e))}}}
	) (\r -> happyReturn (happyIn76 r))

happyReduce_180 = happyMonadReduce 2# 65# happyReduction_180
happyReduction_180 (happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Module) -> 
	case happyOut159 happy_x_2 of { happy_var_2 -> 
	( checkExpr happy_var_2 >>= \e -> return (ModuleAnn (nIS happy_var_1 <++> ann e <** [happy_var_1])    e))}}
	) (\r -> happyReturn (happyIn76 r))

happyReduce_181 = happyMonadReduce 2# 65# happyReduction_181
happyReduction_181 (happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut75 happy_x_1 of { happy_var_1 -> 
	case happyOut159 happy_x_2 of { happy_var_2 -> 
	( checkExpr happy_var_2 >>= \e -> return (Ann (happy_var_1 <> e) happy_var_1 e))}}
	) (\r -> happyReturn (happyIn76 r))

happyReduce_182 = happyMonadReduce 1# 66# happyReduction_182
happyReduction_182 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut78 happy_x_1 of { happy_var_1 -> 
	( checkType happy_var_1)}
	) (\r -> happyReturn (happyIn77 r))

happyReduce_183 = happySpecReduce_1  67# happyReduction_183
happyReduction_183 happy_x_1
	 =  case happyOut82 happy_x_1 of { happy_var_1 -> 
	happyIn78
		 (happy_var_1
	)}

happyReduce_184 = happySpecReduce_3  67# happyReduction_184
happyReduction_184 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut82 happy_x_1 of { happy_var_1 -> 
	case happyOut91 happy_x_2 of { happy_var_2 -> 
	case happyOut78 happy_x_3 of { happy_var_3 -> 
	happyIn78
		 (TyInfix (happy_var_1 <> happy_var_3) happy_var_1 happy_var_2 happy_var_3
	)}}}

happyReduce_185 = happySpecReduce_3  67# happyReduction_185
happyReduction_185 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut82 happy_x_1 of { happy_var_1 -> 
	case happyOut243 happy_x_2 of { happy_var_2 -> 
	case happyOut78 happy_x_3 of { happy_var_3 -> 
	happyIn78
		 (TyInfix (happy_var_1 <> happy_var_3) happy_var_1 happy_var_2 happy_var_3
	)}}}

happyReduce_186 = happySpecReduce_3  67# happyReduction_186
happyReduction_186 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut82 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 RightArrow) -> 
	case happyOut93 happy_x_3 of { happy_var_3 -> 
	happyIn78
		 (TyFun (happy_var_1 <> happy_var_3 <** [happy_var_2]) happy_var_1 happy_var_3
	)}}}

happyReduce_187 = happyMonadReduce 3# 67# happyReduction_187
happyReduction_187 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut82 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Tilde) -> 
	case happyOut82 happy_x_3 of { happy_var_3 -> 
	( do { checkEnabled TypeFamilies ;
                                              let {l = happy_var_1 <> happy_var_3 <** [happy_var_2]};
                                              return $ TyPred l $ EqualP l happy_var_1 happy_var_3 })}}}
	) (\r -> happyReturn (happyIn78 r))

happyReduce_188 = happyMonadReduce 1# 68# happyReduction_188
happyReduction_188 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut80 happy_x_1 of { happy_var_1 -> 
	( checkType happy_var_1)}
	) (\r -> happyReturn (happyIn79 r))

happyReduce_189 = happySpecReduce_3  69# happyReduction_189
happyReduction_189 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut210 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 DoubleColon) -> 
	case happyOut78 happy_x_3 of { happy_var_3 -> 
	happyIn80
		 (let l = (happy_var_1 <> happy_var_3 <** [happy_var_2]) in TyPred l $ IParam l happy_var_1 happy_var_3
	)}}}

happyReduce_190 = happySpecReduce_1  69# happyReduction_190
happyReduction_190 happy_x_1
	 =  case happyOut78 happy_x_1 of { happy_var_1 -> 
	happyIn80
		 (happy_var_1
	)}

happyReduce_191 = happyMonadReduce 1# 70# happyReduction_191
happyReduction_191 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut82 happy_x_1 of { happy_var_1 -> 
	( checkType happy_var_1)}
	) (\r -> happyReturn (happyIn81 r))

happyReduce_192 = happySpecReduce_2  71# happyReduction_192
happyReduction_192 happy_x_2
	happy_x_1
	 =  case happyOut82 happy_x_1 of { happy_var_1 -> 
	case happyOut84 happy_x_2 of { happy_var_2 -> 
	happyIn82
		 (TyApp (happy_var_1 <> happy_var_2) happy_var_1 happy_var_2
	)}}

happyReduce_193 = happySpecReduce_1  71# happyReduction_193
happyReduction_193 happy_x_1
	 =  case happyOut84 happy_x_1 of { happy_var_1 -> 
	happyIn82
		 (happy_var_1
	)}

happyReduce_194 = happyMonadReduce 1# 72# happyReduction_194
happyReduction_194 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut84 happy_x_1 of { happy_var_1 -> 
	( checkType happy_var_1)}
	) (\r -> happyReturn (happyIn83 r))

happyReduce_195 = happySpecReduce_1  73# happyReduction_195
happyReduction_195 happy_x_1
	 =  case happyOut89 happy_x_1 of { happy_var_1 -> 
	happyIn84
		 (TyCon   (ann happy_var_1) happy_var_1
	)}

happyReduce_196 = happySpecReduce_1  73# happyReduction_196
happyReduction_196 happy_x_1
	 =  case happyOut241 happy_x_1 of { happy_var_1 -> 
	happyIn84
		 (TyVar   (ann happy_var_1) happy_var_1
	)}

happyReduce_197 = happySpecReduce_3  73# happyReduction_197
happyReduction_197 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut95 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn84
		 (TyTuple (happy_var_1 <^^> happy_var_3 <** (happy_var_1:reverse (happy_var_3:snd happy_var_2))) Boxed   (reverse (fst happy_var_2))
	)}}}

happyReduce_198 = happySpecReduce_3  73# happyReduction_198
happyReduction_198 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftHashParen) -> 
	case happyOut96 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightHashParen) -> 
	happyIn84
		 (TyTuple (happy_var_1 <^^> happy_var_3 <** (happy_var_1:reverse (happy_var_3:snd happy_var_2))) Unboxed (reverse (fst happy_var_2))
	)}}}

happyReduce_199 = happySpecReduce_3  73# happyReduction_199
happyReduction_199 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftSquare) -> 
	case happyOut80 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightSquare) -> 
	happyIn84
		 (TyList  (happy_var_1 <^^> happy_var_3 <** [happy_var_1,happy_var_3]) happy_var_2
	)}}}

happyReduce_200 = happySpecReduce_3  73# happyReduction_200
happyReduction_200 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut93 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn84
		 (TyParen (happy_var_1 <^^> happy_var_3 <** [happy_var_1,happy_var_3]) happy_var_2
	)}}}

happyReduce_201 = happyReduce 5# 73# happyReduction_201
happyReduction_201 (happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut93 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 DoubleColon) -> 
	case happyOut123 happy_x_4 of { happy_var_4 -> 
	case happyOutTok happy_x_5 of { (Loc happy_var_5 RightParen) -> 
	happyIn84
		 (TyKind  (happy_var_1 <^^> happy_var_5 <** [happy_var_1,happy_var_3,happy_var_5]) happy_var_2 happy_var_4
	) `HappyStk` happyRest}}}}}

happyReduce_202 = happyMonadReduce 1# 73# happyReduction_202
happyReduction_202 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut85 happy_x_1 of { happy_var_1 -> 
	( checkEnabled DataKinds >> return (TyPromoted (ann happy_var_1) happy_var_1))}
	) (\r -> happyReturn (happyIn84 r))

happyReduce_203 = happyReduce 4# 74# happyReduction_203
happyReduction_203 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 THVarQuote) -> 
	case happyOut88 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 RightSquare) -> 
	happyIn85
		 (PromotedList  (happy_var_1 <^^> happy_var_4 <** (happy_var_1: reverse(happy_var_4:snd happy_var_3))) True  (reverse (fst happy_var_3))
	) `HappyStk` happyRest}}}

happyReduce_204 = happySpecReduce_3  74# happyReduction_204
happyReduction_204 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftSquare) -> 
	case happyOut87 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightSquare) -> 
	happyIn85
		 (PromotedList  (happy_var_1 <^^> happy_var_3 <** (happy_var_1: reverse(happy_var_3:snd happy_var_2))) False (reverse (fst happy_var_2))
	)}}}

happyReduce_205 = happyReduce 4# 74# happyReduction_205
happyReduction_205 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 THVarQuote) -> 
	case happyOut88 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 RightParen) -> 
	happyIn85
		 (PromotedTuple (happy_var_1 <^^> happy_var_4 <** (happy_var_1: reverse(happy_var_4:snd happy_var_3)))       (reverse (fst happy_var_3))
	) `HappyStk` happyRest}}}

happyReduce_206 = happySpecReduce_3  74# happyReduction_206
happyReduction_206 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 THVarQuote) -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn85
		 (PromotedUnit  (happy_var_1 <^^> happy_var_3 )
	)}}

happyReduce_207 = happySpecReduce_2  74# happyReduction_207
happyReduction_207 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 THVarQuote) -> 
	case happyOut240 happy_x_2 of { happy_var_2 -> 
	happyIn85
		 (PromotedCon ((noInfoSpan happy_var_1 <++> ann happy_var_2) <** [happy_var_1]) True  happy_var_2
	)}}

happyReduce_208 = happySpecReduce_1  74# happyReduction_208
happyReduction_208 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn85
		 (let Loc l (IntTok  (i,raw)) = happy_var_1 in PromotedInteger (nIS l) i raw
	)}

happyReduce_209 = happySpecReduce_1  74# happyReduction_209
happyReduction_209 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn85
		 (let Loc l (StringTok (s,raw)) = happy_var_1 in PromotedString (nIS l) s raw
	)}

happyReduce_210 = happySpecReduce_1  75# happyReduction_210
happyReduction_210 happy_x_1
	 =  case happyOut85 happy_x_1 of { happy_var_1 -> 
	happyIn86
		 (happy_var_1
	)}

happyReduce_211 = happySpecReduce_1  75# happyReduction_211
happyReduction_211 happy_x_1
	 =  case happyOut240 happy_x_1 of { happy_var_1 -> 
	happyIn86
		 (PromotedCon (ann happy_var_1) False happy_var_1
	)}

happyReduce_212 = happySpecReduce_3  76# happyReduction_212
happyReduction_212 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut88 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut86 happy_x_3 of { happy_var_3 -> 
	happyIn87
		 ((happy_var_3 : fst happy_var_1, happy_var_2 : snd happy_var_1)
	)}}}

happyReduce_213 = happySpecReduce_1  77# happyReduction_213
happyReduction_213 happy_x_1
	 =  case happyOut86 happy_x_1 of { happy_var_1 -> 
	happyIn88
		 (([happy_var_1],[])
	)}

happyReduce_214 = happySpecReduce_3  77# happyReduction_214
happyReduction_214 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut88 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut86 happy_x_3 of { happy_var_3 -> 
	happyIn88
		 ((happy_var_3 : fst happy_var_1, happy_var_2 : snd happy_var_1)
	)}}}

happyReduce_215 = happySpecReduce_1  78# happyReduction_215
happyReduction_215 happy_x_1
	 =  case happyOut90 happy_x_1 of { happy_var_1 -> 
	happyIn89
		 (happy_var_1
	)}

happyReduce_216 = happySpecReduce_2  78# happyReduction_216
happyReduction_216 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 RightParen) -> 
	happyIn89
		 (unit_tycon_name              (happy_var_1 <^^> happy_var_2 <** [happy_var_1,happy_var_2])
	)}}

happyReduce_217 = happySpecReduce_3  78# happyReduction_217
happyReduction_217 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 RightArrow) -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn89
		 (fun_tycon_name               (happy_var_1 <^^> happy_var_3 <** [happy_var_1,happy_var_2,happy_var_3])
	)}}}

happyReduce_218 = happySpecReduce_2  78# happyReduction_218
happyReduction_218 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftSquare) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 RightSquare) -> 
	happyIn89
		 (list_tycon_name              (happy_var_1 <^^> happy_var_2 <** [happy_var_1,happy_var_2])
	)}}

happyReduce_219 = happySpecReduce_3  78# happyReduction_219
happyReduction_219 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut162 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn89
		 (tuple_tycon_name             (happy_var_1 <^^> happy_var_3 <** (happy_var_1:reverse happy_var_2 ++ [happy_var_3])) Boxed (length happy_var_2)
	)}}}

happyReduce_220 = happySpecReduce_2  78# happyReduction_220
happyReduction_220 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftHashParen) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 RightHashParen) -> 
	happyIn89
		 (unboxed_singleton_tycon_name (happy_var_1 <^^> happy_var_2 <** [happy_var_1,happy_var_2])
	)}}

happyReduce_221 = happySpecReduce_3  78# happyReduction_221
happyReduction_221 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftHashParen) -> 
	case happyOut162 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightHashParen) -> 
	happyIn89
		 (tuple_tycon_name             (happy_var_1 <^^> happy_var_3 <** (happy_var_1:reverse happy_var_2 ++ [happy_var_3])) Unboxed (length happy_var_2)
	)}}}

happyReduce_222 = happySpecReduce_1  79# happyReduction_222
happyReduction_222 happy_x_1
	 =  case happyOut226 happy_x_1 of { happy_var_1 -> 
	happyIn90
		 (happy_var_1
	)}

happyReduce_223 = happySpecReduce_3  79# happyReduction_223
happyReduction_223 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut221 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn90
		 (fmap (const (happy_var_1 <^^> happy_var_3 <** [happy_var_1, srcInfoSpan (ann happy_var_2), happy_var_3])) happy_var_2
	)}}}

happyReduce_224 = happySpecReduce_3  79# happyReduction_224
happyReduction_224 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut230 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn90
		 (fmap (const (happy_var_1 <^^> happy_var_3 <** [happy_var_1, srcInfoSpan (ann happy_var_2), happy_var_3])) happy_var_2
	)}}}

happyReduce_225 = happySpecReduce_1  80# happyReduction_225
happyReduction_225 happy_x_1
	 =  case happyOut217 happy_x_1 of { happy_var_1 -> 
	happyIn91
		 (happy_var_1
	)}

happyReduce_226 = happyMonadReduce 1# 81# happyReduction_226
happyReduction_226 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut93 happy_x_1 of { happy_var_1 -> 
	( checkType happy_var_1)}
	) (\r -> happyReturn (happyIn92 r))

happyReduce_227 = happyReduce 4# 82# happyReduction_227
happyReduction_227 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Forall) -> 
	case happyOut97 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 Dot) -> 
	case happyOut93 happy_x_4 of { happy_var_4 -> 
	happyIn93
		 (TyForall (nIS happy_var_1 <++> ann happy_var_4 <** [happy_var_1,happy_var_3]) (Just (reverse (fst happy_var_2))) Nothing happy_var_4
	) `HappyStk` happyRest}}}}

happyReduce_228 = happySpecReduce_2  82# happyReduction_228
happyReduction_228 happy_x_2
	happy_x_1
	 =  case happyOut94 happy_x_1 of { happy_var_1 -> 
	case happyOut93 happy_x_2 of { happy_var_2 -> 
	happyIn93
		 (TyForall (happy_var_1 <> happy_var_2) Nothing (Just happy_var_1) happy_var_2
	)}}

happyReduce_229 = happySpecReduce_1  82# happyReduction_229
happyReduction_229 happy_x_1
	 =  case happyOut80 happy_x_1 of { happy_var_1 -> 
	happyIn93
		 (happy_var_1
	)}

happyReduce_230 = happyMonadReduce 2# 83# happyReduction_230
happyReduction_230 (happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut82 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 DoubleArrow) -> 
	( checkPContext $ (amap (\l -> l <++> nIS happy_var_2 <** (srcInfoPoints l ++ [happy_var_2]))) happy_var_1)}}
	) (\r -> happyReturn (happyIn94 r))

happyReduce_231 = happyMonadReduce 4# 83# happyReduction_231
happyReduction_231 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut82 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Tilde) -> 
	case happyOut82 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 DoubleArrow) -> 
	( do { checkEnabled TypeFamilies;
                                              let {l = happy_var_1 <> happy_var_3 <** [happy_var_2,happy_var_4]};
                                              checkPContext (TyPred l $ EqualP l happy_var_1 happy_var_3) })}}}}
	) (\r -> happyReturn (happyIn94 r))

happyReduce_232 = happySpecReduce_3  84# happyReduction_232
happyReduction_232 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut96 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut93 happy_x_3 of { happy_var_3 -> 
	happyIn95
		 ((happy_var_3 : fst happy_var_1, happy_var_2 : snd happy_var_1)
	)}}}

happyReduce_233 = happySpecReduce_1  85# happyReduction_233
happyReduction_233 happy_x_1
	 =  case happyOut93 happy_x_1 of { happy_var_1 -> 
	happyIn96
		 (([happy_var_1],[])
	)}

happyReduce_234 = happySpecReduce_3  85# happyReduction_234
happyReduction_234 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut96 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut93 happy_x_3 of { happy_var_3 -> 
	happyIn96
		 ((happy_var_3 : fst happy_var_1, happy_var_2 : snd happy_var_1)
	)}}}

happyReduce_235 = happySpecReduce_2  86# happyReduction_235
happyReduction_235 happy_x_2
	happy_x_1
	 =  case happyOut97 happy_x_1 of { happy_var_1 -> 
	case happyOut98 happy_x_2 of { happy_var_2 -> 
	happyIn97
		 ((happy_var_2 : fst happy_var_1, Just (snd happy_var_1 <?+> ann happy_var_2))
	)}}

happyReduce_236 = happySpecReduce_0  86# happyReduction_236
happyReduction_236  =  happyIn97
		 (([],Nothing)
	)

happyReduce_237 = happySpecReduce_1  87# happyReduction_237
happyReduction_237 happy_x_1
	 =  case happyOut241 happy_x_1 of { happy_var_1 -> 
	happyIn98
		 (UnkindedVar (ann happy_var_1) happy_var_1
	)}

happyReduce_238 = happyReduce 5# 87# happyReduction_238
happyReduction_238 (happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut241 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 DoubleColon) -> 
	case happyOut123 happy_x_4 of { happy_var_4 -> 
	case happyOutTok happy_x_5 of { (Loc happy_var_5 RightParen) -> 
	happyIn98
		 (KindedVar (happy_var_1 <^^> happy_var_5 <** [happy_var_1,happy_var_3,happy_var_5]) happy_var_2 happy_var_4
	) `HappyStk` happyRest}}}}}

happyReduce_239 = happySpecReduce_2  88# happyReduction_239
happyReduction_239 happy_x_2
	happy_x_1
	 =  case happyOut99 happy_x_1 of { happy_var_1 -> 
	case happyOut241 happy_x_2 of { happy_var_2 -> 
	happyIn99
		 ((happy_var_2 : fst happy_var_1, Just (snd happy_var_1 <?+> ann happy_var_2))
	)}}

happyReduce_240 = happySpecReduce_0  88# happyReduction_240
happyReduction_240  =  happyIn99
		 (([], Nothing)
	)

happyReduce_241 = happySpecReduce_2  89# happyReduction_241
happyReduction_241 happy_x_2
	happy_x_1
	 =  case happyOut99 happy_x_1 of { happy_var_1 -> 
	case happyOut241 happy_x_2 of { happy_var_2 -> 
	happyIn100
		 ((happy_var_2 : fst happy_var_1, snd happy_var_1 <?+> ann happy_var_2)
	)}}

happyReduce_242 = happySpecReduce_0  90# happyReduction_242
happyReduction_242  =  happyIn101
		 (([],[], Nothing)
	)

happyReduce_243 = happyMonadReduce 2# 90# happyReduction_243
happyReduction_243 (happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 Bar) -> 
	case happyOut102 happy_x_2 of { happy_var_2 -> 
	( do { checkEnabled FunctionalDependencies ;
                                              let {(fds,ss,l) = happy_var_2} ;
                                              return (reverse fds, happy_var_1 : reverse ss, Just (nIS happy_var_1 <++> l)) })}}
	) (\r -> happyReturn (happyIn101 r))

happyReduce_244 = happySpecReduce_3  91# happyReduction_244
happyReduction_244 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut102 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut103 happy_x_3 of { happy_var_3 -> 
	happyIn102
		 (let (fds,ss,l) = happy_var_1 in (happy_var_3 : fds, happy_var_2 : ss, l <++> ann happy_var_3)
	)}}}

happyReduce_245 = happySpecReduce_1  91# happyReduction_245
happyReduction_245 happy_x_1
	 =  case happyOut103 happy_x_1 of { happy_var_1 -> 
	happyIn102
		 (([happy_var_1],[],ann happy_var_1)
	)}

happyReduce_246 = happySpecReduce_3  92# happyReduction_246
happyReduction_246 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut99 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 RightArrow) -> 
	case happyOut100 happy_x_3 of { happy_var_3 -> 
	happyIn103
		 (FunDep (snd happy_var_1 <?+> nIS happy_var_2 <++> snd happy_var_3 <** [happy_var_2]) (reverse (fst happy_var_1)) (reverse (fst happy_var_3))
	)}}}

happyReduce_247 = happyMonadReduce 4# 93# happyReduction_247
happyReduction_247 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Where) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftCurly) -> 
	case happyOut105 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 RightCurly) -> 
	( return (fst happy_var_3, happy_var_1 : happy_var_2 : snd happy_var_3 ++ [happy_var_4], Just $ happy_var_1 <^^> happy_var_4))}}}}
	) (\r -> happyReturn (happyIn104 r))

happyReduce_248 = happyMonadReduce 4# 93# happyReduction_248
happyReduction_248 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Where) -> 
	case happyOut236 happy_x_2 of { happy_var_2 -> 
	case happyOut105 happy_x_3 of { happy_var_3 -> 
	case happyOut237 happy_x_4 of { happy_var_4 -> 
	( return (fst happy_var_3, happy_var_1 : happy_var_2 : snd happy_var_3 ++ [happy_var_4], Just $ happy_var_1 <^^> happy_var_4))}}}}
	) (\r -> happyReturn (happyIn104 r))

happyReduce_249 = happyMonadReduce 0# 93# happyReduction_249
happyReduction_249 (happyRest) tk
	 = happyThen (( checkEnabled EmptyDataDecls >> return ([],[],Nothing))
	) (\r -> happyReturn (happyIn104 r))

happyReduce_250 = happySpecReduce_3  94# happyReduction_250
happyReduction_250 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut25 happy_x_1 of { happy_var_1 -> 
	case happyOut106 happy_x_2 of { happy_var_2 -> 
	case happyOut25 happy_x_3 of { happy_var_3 -> 
	happyIn105
		 ((fst happy_var_2, reverse happy_var_1 ++ snd happy_var_2 ++ reverse happy_var_3)
	)}}}

happyReduce_251 = happySpecReduce_3  95# happyReduction_251
happyReduction_251 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut106 happy_x_1 of { happy_var_1 -> 
	case happyOut24 happy_x_2 of { happy_var_2 -> 
	case happyOut107 happy_x_3 of { happy_var_3 -> 
	happyIn106
		 ((happy_var_3 : fst happy_var_1, snd happy_var_1 ++ reverse happy_var_2)
	)}}}

happyReduce_252 = happySpecReduce_1  95# happyReduction_252
happyReduction_252 happy_x_1
	 =  case happyOut107 happy_x_1 of { happy_var_1 -> 
	happyIn106
		 (([happy_var_1],[])
	)}

happyReduce_253 = happyMonadReduce 3# 96# happyReduction_253
happyReduction_253 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut212 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 DoubleColon) -> 
	case happyOut92 happy_x_3 of { happy_var_3 -> 
	( do { c <- checkUnQual happy_var_1;
                                               return $ GadtDecl (happy_var_1 <> happy_var_3 <** [happy_var_2]) c happy_var_3 })}}}
	) (\r -> happyReturn (happyIn107 r))

happyReduce_254 = happySpecReduce_2  97# happyReduction_254
happyReduction_254 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Equals) -> 
	case happyOut109 happy_x_2 of { happy_var_2 -> 
	happyIn108
		 (let (ds,ss,l) = happy_var_2 in (ds, happy_var_1 : reverse ss, Just $ nIS happy_var_1 <++> l)
	)}}

happyReduce_255 = happySpecReduce_3  98# happyReduction_255
happyReduction_255 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut109 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Bar) -> 
	case happyOut110 happy_x_3 of { happy_var_3 -> 
	happyIn109
		 (let (ds,ss,l) = happy_var_1 in (happy_var_3 : ds, happy_var_2 : ss, l <++> ann happy_var_3)
	)}}}

happyReduce_256 = happySpecReduce_1  98# happyReduction_256
happyReduction_256 happy_x_1
	 =  case happyOut110 happy_x_1 of { happy_var_1 -> 
	happyIn109
		 (([happy_var_1],[],ann happy_var_1)
	)}

happyReduce_257 = happyMonadReduce 3# 99# happyReduction_257
happyReduction_257 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut111 happy_x_1 of { happy_var_1 -> 
	case happyOut94 happy_x_2 of { happy_var_2 -> 
	case happyOut112 happy_x_3 of { happy_var_3 -> 
	( do { checkEnabled ExistentialQuantification ;
                                               ctxt <- checkContext (Just happy_var_2) ;
                                               let {(mtvs,ss,ml) = happy_var_1} ;
                                               return $ QualConDecl (ml <?+> ann happy_var_3 <** ss) mtvs ctxt happy_var_3 })}}}
	) (\r -> happyReturn (happyIn110 r))

happyReduce_258 = happySpecReduce_2  99# happyReduction_258
happyReduction_258 happy_x_2
	happy_x_1
	 =  case happyOut111 happy_x_1 of { happy_var_1 -> 
	case happyOut112 happy_x_2 of { happy_var_2 -> 
	happyIn110
		 (let (mtvs, ss, ml) = happy_var_1 in QualConDecl (ml <?+> ann happy_var_2 <** ss) mtvs Nothing happy_var_2
	)}}

happyReduce_259 = happyMonadReduce 3# 100# happyReduction_259
happyReduction_259 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Forall) -> 
	case happyOut97 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 Dot) -> 
	( checkEnabled ExistentialQuantification >> return (Just (fst happy_var_2), [happy_var_1,happy_var_3], Just $ happy_var_1 <^^> happy_var_3))}}}
	) (\r -> happyReturn (happyIn111 r))

happyReduce_260 = happySpecReduce_0  100# happyReduction_260
happyReduction_260  =  happyIn111
		 ((Nothing, [], Nothing)
	)

happyReduce_261 = happySpecReduce_1  101# happyReduction_261
happyReduction_261 happy_x_1
	 =  case happyOut113 happy_x_1 of { happy_var_1 -> 
	happyIn112
		 (let (n,ts,l) = happy_var_1 in ConDecl l n ts
	)}

happyReduce_262 = happySpecReduce_3  101# happyReduction_262
happyReduction_262 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut116 happy_x_1 of { happy_var_1 -> 
	case happyOut216 happy_x_2 of { happy_var_2 -> 
	case happyOut116 happy_x_3 of { happy_var_3 -> 
	happyIn112
		 (InfixConDecl (happy_var_1 <> happy_var_3) happy_var_1 happy_var_2 happy_var_3
	)}}}

happyReduce_263 = happyMonadReduce 3# 101# happyReduction_263
happyReduction_263 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut212 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftCurly) -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightCurly) -> 
	( do { c <- checkUnQual happy_var_1; return $ RecDecl (ann happy_var_1 <++> nIS happy_var_3 <** [happy_var_2,happy_var_3]) c [] })}}}
	) (\r -> happyReturn (happyIn112 r))

happyReduce_264 = happyMonadReduce 4# 101# happyReduction_264
happyReduction_264 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut212 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftCurly) -> 
	case happyOut117 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 RightCurly) -> 
	( do { c <- checkUnQual happy_var_1;
                                              return $ RecDecl (ann happy_var_1 <++> nIS happy_var_4 <** (happy_var_2:reverse (snd happy_var_3) ++ [happy_var_4])) c (reverse (fst happy_var_3)) })}}}}
	) (\r -> happyReturn (happyIn112 r))

happyReduce_265 = happyMonadReduce 1# 102# happyReduction_265
happyReduction_265 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut82 happy_x_1 of { happy_var_1 -> 
	( do { (c,ts) <- splitTyConApp happy_var_1;
                                              return (c,map (\t -> UnBangedTy (ann t) t) ts,ann happy_var_1) })}
	) (\r -> happyReturn (happyIn113 r))

happyReduce_266 = happySpecReduce_1  102# happyReduction_266
happyReduction_266 happy_x_1
	 =  case happyOut114 happy_x_1 of { happy_var_1 -> 
	happyIn113
		 (happy_var_1
	)}

happyReduce_267 = happyMonadReduce 3# 103# happyReduction_267
happyReduction_267 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut82 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Exclamation) -> 
	case happyOut83 happy_x_3 of { happy_var_3 -> 
	( do { (c,ts) <- splitTyConApp happy_var_1;
                                                          return (c,map (\t -> UnBangedTy (ann t) t) ts++
                                                                  [BangedTy (nIS happy_var_2 <++> ann happy_var_3 <** [happy_var_2]) happy_var_3], happy_var_1 <> happy_var_3) })}}}
	) (\r -> happyReturn (happyIn114 r))

happyReduce_268 = happyMonadReduce 5# 103# happyReduction_268
happyReduction_268 (happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut82 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 UNPACK) -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 PragmaEnd) -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 Exclamation) -> 
	case happyOut83 happy_x_5 of { happy_var_5 -> 
	( do { (c,ts) <- splitTyConApp happy_var_1;
                                                          return (c,map (\t -> UnBangedTy (ann t) t) ts++
                                                                  [UnpackedTy (nIS happy_var_2 <++> ann happy_var_5 <** [happy_var_2,happy_var_3,happy_var_4]) happy_var_5], happy_var_1 <> happy_var_5) })}}}}}
	) (\r -> happyReturn (happyIn114 r))

happyReduce_269 = happySpecReduce_2  103# happyReduction_269
happyReduction_269 happy_x_2
	happy_x_1
	 =  case happyOut114 happy_x_1 of { happy_var_1 -> 
	case happyOut115 happy_x_2 of { happy_var_2 -> 
	happyIn114
		 (let (n,ts,l) = happy_var_1 in (n, ts ++ [happy_var_2],l <++> ann happy_var_2)
	)}}

happyReduce_270 = happySpecReduce_1  104# happyReduction_270
happyReduction_270 happy_x_1
	 =  case happyOut83 happy_x_1 of { happy_var_1 -> 
	happyIn115
		 (UnBangedTy (ann happy_var_1) happy_var_1
	)}

happyReduce_271 = happySpecReduce_2  104# happyReduction_271
happyReduction_271 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Exclamation) -> 
	case happyOut83 happy_x_2 of { happy_var_2 -> 
	happyIn115
		 (BangedTy   (nIS happy_var_1 <++> ann happy_var_2 <** [happy_var_1]) happy_var_2
	)}}

happyReduce_272 = happyReduce 4# 104# happyReduction_272
happyReduction_272 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 UNPACK) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 PragmaEnd) -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 Exclamation) -> 
	case happyOut83 happy_x_4 of { happy_var_4 -> 
	happyIn115
		 (UnpackedTy (nIS happy_var_1 <++> ann happy_var_4 <** [happy_var_1,happy_var_2,happy_var_3]) happy_var_4
	) `HappyStk` happyRest}}}}

happyReduce_273 = happySpecReduce_1  105# happyReduction_273
happyReduction_273 happy_x_1
	 =  case happyOut81 happy_x_1 of { happy_var_1 -> 
	happyIn116
		 (UnBangedTy (ann happy_var_1) happy_var_1
	)}

happyReduce_274 = happySpecReduce_2  105# happyReduction_274
happyReduction_274 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Exclamation) -> 
	case happyOut83 happy_x_2 of { happy_var_2 -> 
	happyIn116
		 (BangedTy   (nIS happy_var_1 <++> ann happy_var_2 <** [happy_var_1]) happy_var_2
	)}}

happyReduce_275 = happyReduce 4# 105# happyReduction_275
happyReduction_275 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 UNPACK) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 PragmaEnd) -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 Exclamation) -> 
	case happyOut83 happy_x_4 of { happy_var_4 -> 
	happyIn116
		 (UnpackedTy (nIS happy_var_1 <++> ann happy_var_4 <** [happy_var_1,happy_var_2,happy_var_3]) happy_var_4
	) `HappyStk` happyRest}}}}

happyReduce_276 = happySpecReduce_3  106# happyReduction_276
happyReduction_276 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut117 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut118 happy_x_3 of { happy_var_3 -> 
	happyIn117
		 ((happy_var_3 : fst happy_var_1, happy_var_2 : snd happy_var_1)
	)}}}

happyReduce_277 = happySpecReduce_1  106# happyReduction_277
happyReduction_277 happy_x_1
	 =  case happyOut118 happy_x_1 of { happy_var_1 -> 
	happyIn117
		 (([happy_var_1],[])
	)}

happyReduce_278 = happySpecReduce_3  107# happyReduction_278
happyReduction_278 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut62 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 DoubleColon) -> 
	case happyOut119 happy_x_3 of { happy_var_3 -> 
	happyIn118
		 (let (ns,ss,l) = happy_var_1 in FieldDecl (l <++> ann happy_var_3 <** (reverse ss ++ [happy_var_2])) (reverse ns) happy_var_3
	)}}}

happyReduce_279 = happySpecReduce_1  108# happyReduction_279
happyReduction_279 happy_x_1
	 =  case happyOut92 happy_x_1 of { happy_var_1 -> 
	happyIn119
		 (UnBangedTy (ann happy_var_1) happy_var_1
	)}

happyReduce_280 = happySpecReduce_2  108# happyReduction_280
happyReduction_280 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Exclamation) -> 
	case happyOut83 happy_x_2 of { happy_var_2 -> 
	happyIn119
		 (BangedTy   (nIS happy_var_1 <++> ann happy_var_2 <** [happy_var_1]) happy_var_2
	)}}

happyReduce_281 = happyReduce 4# 108# happyReduction_281
happyReduction_281 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 UNPACK) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 PragmaEnd) -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 Exclamation) -> 
	case happyOut83 happy_x_4 of { happy_var_4 -> 
	happyIn119
		 (UnpackedTy (nIS happy_var_1 <++> ann happy_var_4 <** [happy_var_1,happy_var_2,happy_var_3]) happy_var_4
	) `HappyStk` happyRest}}}}

happyReduce_282 = happySpecReduce_0  109# happyReduction_282
happyReduction_282  =  happyIn120
		 (Nothing
	)

happyReduce_283 = happySpecReduce_2  109# happyReduction_283
happyReduction_283 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Deriving) -> 
	case happyOut122 happy_x_2 of { happy_var_2 -> 
	happyIn120
		 (let l = nIS happy_var_1 <++> ann happy_var_2 <** [happy_var_1] in Just $ Deriving l [IHead (ann happy_var_2) happy_var_2 []]
	)}}

happyReduce_284 = happySpecReduce_3  109# happyReduction_284
happyReduction_284 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Deriving) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftParen) -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn120
		 (Just $ Deriving (happy_var_1 <^^> happy_var_3 <** [happy_var_1,happy_var_2,happy_var_3]) []
	)}}}

happyReduce_285 = happyReduce 4# 109# happyReduction_285
happyReduction_285 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Deriving) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftParen) -> 
	case happyOut121 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 RightParen) -> 
	happyIn120
		 (Just $ Deriving (happy_var_1 <^^> happy_var_4 <** happy_var_1:happy_var_2: reverse (snd happy_var_3) ++ [happy_var_4]) (reverse (fst happy_var_3))
	) `HappyStk` happyRest}}}}

happyReduce_286 = happyMonadReduce 1# 110# happyReduction_286
happyReduction_286 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut96 happy_x_1 of { happy_var_1 -> 
	( checkDeriving (fst happy_var_1) >>= \ds -> return (ds, snd happy_var_1))}
	) (\r -> happyReturn (happyIn121 r))

happyReduce_287 = happySpecReduce_1  111# happyReduction_287
happyReduction_287 happy_x_1
	 =  case happyOut226 happy_x_1 of { happy_var_1 -> 
	happyIn122
		 (happy_var_1
	)}

happyReduce_288 = happyMonadReduce 1# 112# happyReduction_288
happyReduction_288 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut124 happy_x_1 of { happy_var_1 -> 
	( checkEnabled KindSignatures >> return happy_var_1)}
	) (\r -> happyReturn (happyIn123 r))

happyReduce_289 = happySpecReduce_1  113# happyReduction_289
happyReduction_289 happy_x_1
	 =  case happyOut125 happy_x_1 of { happy_var_1 -> 
	happyIn124
		 (happy_var_1
	)}

happyReduce_290 = happySpecReduce_2  113# happyReduction_290
happyReduction_290 happy_x_2
	happy_x_1
	 =  case happyOut125 happy_x_1 of { happy_var_1 -> 
	case happyOut124 happy_x_2 of { happy_var_2 -> 
	happyIn124
		 (KindApp (happy_var_1 <> happy_var_2) happy_var_1 happy_var_2
	)}}

happyReduce_291 = happySpecReduce_3  113# happyReduction_291
happyReduction_291 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut125 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 RightArrow) -> 
	case happyOut124 happy_x_3 of { happy_var_3 -> 
	happyIn124
		 (KindFn (happy_var_1 <> happy_var_3 <** [happy_var_2]) happy_var_1 happy_var_3
	)}}}

happyReduce_292 = happySpecReduce_1  114# happyReduction_292
happyReduction_292 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Star) -> 
	happyIn125
		 (KindStar  (nIS happy_var_1)
	)}

happyReduce_293 = happySpecReduce_1  114# happyReduction_293
happyReduction_293 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Exclamation) -> 
	happyIn125
		 (KindBang  (nIS happy_var_1)
	)}

happyReduce_294 = happySpecReduce_3  114# happyReduction_294
happyReduction_294 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut124 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn125
		 (KindParen (happy_var_1 <^^> happy_var_3 <** [happy_var_1,happy_var_3]) happy_var_2
	)}}}

happyReduce_295 = happyMonadReduce 1# 114# happyReduction_295
happyReduction_295 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut126 happy_x_1 of { happy_var_1 -> 
	( checkEnabled DataKinds >> return happy_var_1)}
	) (\r -> happyReturn (happyIn125 r))

happyReduce_296 = happyMonadReduce 1# 114# happyReduction_296
happyReduction_296 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut222 happy_x_1 of { happy_var_1 -> 
	( checkEnabled PolyKinds >> return (KindVar (ann happy_var_1) happy_var_1))}
	) (\r -> happyReturn (happyIn125 r))

happyReduce_297 = happySpecReduce_1  115# happyReduction_297
happyReduction_297 happy_x_1
	 =  case happyOut240 happy_x_1 of { happy_var_1 -> 
	happyIn126
		 (KindVar (ann happy_var_1) happy_var_1
	)}

happyReduce_298 = happySpecReduce_2  115# happyReduction_298
happyReduction_298 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 RightParen) -> 
	happyIn126
		 (let l = happy_var_1 <^^> happy_var_2 in KindVar l (unit_tycon_name l)
	)}}

happyReduce_299 = happySpecReduce_3  115# happyReduction_299
happyReduction_299 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut127 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn126
		 (KindTuple (happy_var_1 <^^> happy_var_3 <** (happy_var_1:reverse (happy_var_3:snd happy_var_2))) (reverse (fst happy_var_2))
	)}}}

happyReduce_300 = happySpecReduce_3  115# happyReduction_300
happyReduction_300 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftSquare) -> 
	case happyOut128 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightSquare) -> 
	happyIn126
		 (KindList  (happy_var_1 <^^> happy_var_3 <** (happy_var_1:reverse (happy_var_3:snd happy_var_2))) (reverse (fst happy_var_2))
	)}}}

happyReduce_301 = happySpecReduce_3  116# happyReduction_301
happyReduction_301 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut128 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut125 happy_x_3 of { happy_var_3 -> 
	happyIn127
		 ((happy_var_3 : fst happy_var_1, happy_var_2 : snd happy_var_1)
	)}}}

happyReduce_302 = happySpecReduce_1  117# happyReduction_302
happyReduction_302 happy_x_1
	 =  case happyOut125 happy_x_1 of { happy_var_1 -> 
	happyIn128
		 (([happy_var_1],[])
	)}

happyReduce_303 = happySpecReduce_3  117# happyReduction_303
happyReduction_303 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut128 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut125 happy_x_3 of { happy_var_3 -> 
	happyIn128
		 ((happy_var_3 : fst happy_var_1, happy_var_2 : snd happy_var_1)
	)}}}

happyReduce_304 = happySpecReduce_0  118# happyReduction_304
happyReduction_304  =  happyIn129
		 ((Nothing,[])
	)

happyReduce_305 = happySpecReduce_2  118# happyReduction_305
happyReduction_305 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 DoubleColon) -> 
	case happyOut123 happy_x_2 of { happy_var_2 -> 
	happyIn129
		 ((Just happy_var_2,[happy_var_1])
	)}}

happyReduce_306 = happyMonadReduce 4# 119# happyReduction_306
happyReduction_306 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Where) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftCurly) -> 
	case happyOut131 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 RightCurly) -> 
	( checkClassBody (fst happy_var_3) >>= \vs -> return (Just vs, happy_var_1:happy_var_2: snd happy_var_3 ++ [happy_var_4], Just (happy_var_1 <^^> happy_var_4)))}}}}
	) (\r -> happyReturn (happyIn130 r))

happyReduce_307 = happyMonadReduce 4# 119# happyReduction_307
happyReduction_307 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Where) -> 
	case happyOut236 happy_x_2 of { happy_var_2 -> 
	case happyOut131 happy_x_3 of { happy_var_3 -> 
	case happyOut237 happy_x_4 of { happy_var_4 -> 
	( checkClassBody (fst happy_var_3) >>= \vs -> return (Just vs, happy_var_1:happy_var_2: snd happy_var_3 ++ [happy_var_4], Just (happy_var_1 <^^> happy_var_4)))}}}}
	) (\r -> happyReturn (happyIn130 r))

happyReduce_308 = happySpecReduce_0  119# happyReduction_308
happyReduction_308  =  happyIn130
		 ((Nothing,[],Nothing)
	)

happyReduce_309 = happyMonadReduce 3# 120# happyReduction_309
happyReduction_309 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut25 happy_x_1 of { happy_var_1 -> 
	case happyOut132 happy_x_2 of { happy_var_2 -> 
	case happyOut25 happy_x_3 of { happy_var_3 -> 
	( checkRevClsDecls (fst happy_var_2) >>= \cs -> return (cs, reverse happy_var_1 ++ snd happy_var_2 ++ reverse happy_var_3))}}}
	) (\r -> happyReturn (happyIn131 r))

happyReduce_310 = happySpecReduce_1  120# happyReduction_310
happyReduction_310 happy_x_1
	 =  case happyOut25 happy_x_1 of { happy_var_1 -> 
	happyIn131
		 (([],reverse happy_var_1)
	)}

happyReduce_311 = happySpecReduce_3  121# happyReduction_311
happyReduction_311 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut132 happy_x_1 of { happy_var_1 -> 
	case happyOut24 happy_x_2 of { happy_var_2 -> 
	case happyOut133 happy_x_3 of { happy_var_3 -> 
	happyIn132
		 ((happy_var_3 : fst happy_var_1, snd happy_var_1 ++ reverse happy_var_2)
	)}}}

happyReduce_312 = happySpecReduce_1  121# happyReduction_312
happyReduction_312 happy_x_1
	 =  case happyOut133 happy_x_1 of { happy_var_1 -> 
	happyIn132
		 (([happy_var_1],[])
	)}

happyReduce_313 = happySpecReduce_1  122# happyReduction_313
happyReduction_313 happy_x_1
	 =  case happyOut55 happy_x_1 of { happy_var_1 -> 
	happyIn133
		 (ClsDecl (ann happy_var_1) happy_var_1
	)}

happyReduce_314 = happyMonadReduce 1# 122# happyReduction_314
happyReduction_314 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut134 happy_x_1 of { happy_var_1 -> 
	( checkEnabled TypeFamilies >> return happy_var_1)}
	) (\r -> happyReturn (happyIn133 r))

happyReduce_315 = happyMonadReduce 3# 123# happyReduction_315
happyReduction_315 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Type) -> 
	case happyOut80 happy_x_2 of { happy_var_2 -> 
	case happyOut129 happy_x_3 of { happy_var_3 -> 
	( do { dh <- checkSimpleType happy_var_2;
                    return (ClsTyFam  (nIS happy_var_1 <++> ann happy_var_2 <+?> (fmap ann) (fst happy_var_3) <** happy_var_1:snd happy_var_3) dh (fst happy_var_3)) })}}}
	) (\r -> happyReturn (happyIn134 r))

happyReduce_316 = happyReduce 4# 123# happyReduction_316
happyReduction_316 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Type) -> 
	case happyOut77 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 Equals) -> 
	case happyOut92 happy_x_4 of { happy_var_4 -> 
	happyIn134
		 (ClsTyDef (nIS happy_var_1 <++> ann happy_var_4 <** [happy_var_1,happy_var_3]) happy_var_2 happy_var_4
	) `HappyStk` happyRest}}}}

happyReduce_317 = happyMonadReduce 3# 123# happyReduction_317
happyReduction_317 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Data) -> 
	case happyOut93 happy_x_2 of { happy_var_2 -> 
	case happyOut129 happy_x_3 of { happy_var_3 -> 
	( do { (cs,dh) <- checkDataHeader happy_var_2;
                    return (ClsDataFam (nIS happy_var_1 <++> ann happy_var_2 <+?> (fmap ann) (fst happy_var_3) <** happy_var_1:snd happy_var_3) cs dh (fst happy_var_3)) })}}}
	) (\r -> happyReturn (happyIn134 r))

happyReduce_318 = happyMonadReduce 4# 124# happyReduction_318
happyReduction_318 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Where) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftCurly) -> 
	case happyOut136 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 RightCurly) -> 
	( checkInstBody (fst happy_var_3) >>= \vs -> return (Just vs, happy_var_1:happy_var_2: snd happy_var_3 ++ [happy_var_4], Just (happy_var_1 <^^> happy_var_4)))}}}}
	) (\r -> happyReturn (happyIn135 r))

happyReduce_319 = happyMonadReduce 4# 124# happyReduction_319
happyReduction_319 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Where) -> 
	case happyOut236 happy_x_2 of { happy_var_2 -> 
	case happyOut136 happy_x_3 of { happy_var_3 -> 
	case happyOut237 happy_x_4 of { happy_var_4 -> 
	( checkInstBody (fst happy_var_3) >>= \vs -> return (Just vs, happy_var_1:happy_var_2: snd happy_var_3 ++ [happy_var_4], Just (happy_var_1 <^^> happy_var_4)))}}}}
	) (\r -> happyReturn (happyIn135 r))

happyReduce_320 = happySpecReduce_0  124# happyReduction_320
happyReduction_320  =  happyIn135
		 ((Nothing, [], Nothing)
	)

happyReduce_321 = happyMonadReduce 3# 125# happyReduction_321
happyReduction_321 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut25 happy_x_1 of { happy_var_1 -> 
	case happyOut137 happy_x_2 of { happy_var_2 -> 
	case happyOut25 happy_x_3 of { happy_var_3 -> 
	( checkRevInstDecls (fst happy_var_2) >>= \is -> return (is, reverse happy_var_1 ++ snd happy_var_2 ++ reverse happy_var_3))}}}
	) (\r -> happyReturn (happyIn136 r))

happyReduce_322 = happySpecReduce_1  125# happyReduction_322
happyReduction_322 happy_x_1
	 =  case happyOut25 happy_x_1 of { happy_var_1 -> 
	happyIn136
		 (([],reverse happy_var_1)
	)}

happyReduce_323 = happySpecReduce_3  126# happyReduction_323
happyReduction_323 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut137 happy_x_1 of { happy_var_1 -> 
	case happyOut24 happy_x_2 of { happy_var_2 -> 
	case happyOut138 happy_x_3 of { happy_var_3 -> 
	happyIn137
		 ((happy_var_3 : fst happy_var_1, snd happy_var_1 ++ reverse happy_var_2)
	)}}}

happyReduce_324 = happySpecReduce_1  126# happyReduction_324
happyReduction_324 happy_x_1
	 =  case happyOut138 happy_x_1 of { happy_var_1 -> 
	happyIn137
		 (([happy_var_1],[])
	)}

happyReduce_325 = happySpecReduce_1  127# happyReduction_325
happyReduction_325 happy_x_1
	 =  case happyOut140 happy_x_1 of { happy_var_1 -> 
	happyIn138
		 (InsDecl (ann happy_var_1) happy_var_1
	)}

happyReduce_326 = happyMonadReduce 1# 127# happyReduction_326
happyReduction_326 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut139 happy_x_1 of { happy_var_1 -> 
	( checkEnabled TypeFamilies >> return happy_var_1)}
	) (\r -> happyReturn (happyIn138 r))

happyReduce_327 = happySpecReduce_1  127# happyReduction_327
happyReduction_327 happy_x_1
	 =  case happyOut58 happy_x_1 of { happy_var_1 -> 
	happyIn138
		 (InsDecl (ann happy_var_1) happy_var_1
	)}

happyReduce_328 = happyMonadReduce 4# 128# happyReduction_328
happyReduction_328 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Type) -> 
	case happyOut77 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 Equals) -> 
	case happyOut92 happy_x_4 of { happy_var_4 -> 
	( do { -- no checkSimpleType happy_var_4 since dtype may contain type patterns
                       return (InsType (nIS happy_var_1 <++> ann happy_var_4 <** [happy_var_1,happy_var_3]) happy_var_2 happy_var_4) })}}}}
	) (\r -> happyReturn (happyIn139 r))

happyReduce_329 = happyMonadReduce 4# 128# happyReduction_329
happyReduction_329 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut51 happy_x_1 of { happy_var_1 -> 
	case happyOut92 happy_x_2 of { happy_var_2 -> 
	case happyOut108 happy_x_3 of { happy_var_3 -> 
	case happyOut120 happy_x_4 of { happy_var_4 -> 
	( do { -- (cs,c,t) <- checkDataHeader happy_var_4;
                       let {(ds,ss,minf) = happy_var_3};
                       checkDataOrNew happy_var_1 ds;
                       return (InsData (happy_var_1 <> happy_var_2 <+?> minf <+?> fmap ann happy_var_4 <** ss ) happy_var_1 happy_var_2 (reverse ds) happy_var_4) })}}}}
	) (\r -> happyReturn (happyIn139 r))

happyReduce_330 = happyMonadReduce 5# 128# happyReduction_330
happyReduction_330 (happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut51 happy_x_1 of { happy_var_1 -> 
	case happyOut92 happy_x_2 of { happy_var_2 -> 
	case happyOut129 happy_x_3 of { happy_var_3 -> 
	case happyOut104 happy_x_4 of { happy_var_4 -> 
	case happyOut120 happy_x_5 of { happy_var_5 -> 
	( do { -- (cs,c,t) <- checkDataHeader happy_var_4;
                       let { (gs,ss,minf) = happy_var_4 } ;
                       checkDataOrNewG happy_var_1 gs;
                       return $ InsGData (ann happy_var_1 <+?> minf <+?> fmap ann happy_var_5 <** (snd happy_var_3 ++ ss)) happy_var_1 happy_var_2 (fst happy_var_3) (reverse gs) happy_var_5 })}}}}}
	) (\r -> happyReturn (happyIn139 r))

happyReduce_331 = happyMonadReduce 4# 129# happyReduction_331
happyReduction_331 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut150 happy_x_1 of { happy_var_1 -> 
	case happyOut142 happy_x_2 of { happy_var_2 -> 
	case happyOut143 happy_x_3 of { happy_var_3 -> 
	case happyOut141 happy_x_4 of { happy_var_4 -> 
	( checkValDef ((happy_var_1 <> happy_var_3 <+?> (fmap ann) (fst happy_var_4)) <** (snd happy_var_2 ++ snd happy_var_4)) happy_var_1 (fst happy_var_2) happy_var_3 (fst happy_var_4))}}}}
	) (\r -> happyReturn (happyIn140 r))

happyReduce_332 = happyMonadReduce 4# 129# happyReduction_332
happyReduction_332 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 Exclamation) -> 
	case happyOut159 happy_x_2 of { happy_var_2 -> 
	case happyOut143 happy_x_3 of { happy_var_3 -> 
	case happyOut141 happy_x_4 of { happy_var_4 -> 
	( do { checkEnabled BangPatterns ;
                                              let { l = nIS happy_var_1 <++> ann happy_var_2 <** [happy_var_1] };
                                              p <- checkPattern (BangPat l happy_var_2);
                                              return $ PatBind (p <> happy_var_3 <+?> (fmap ann) (fst happy_var_4) <** snd happy_var_4)
                                                          p Nothing happy_var_3 (fst happy_var_4) })}}}}
	) (\r -> happyReturn (happyIn140 r))

happyReduce_333 = happySpecReduce_2  130# happyReduction_333
happyReduction_333 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Where) -> 
	case happyOut61 happy_x_2 of { happy_var_2 -> 
	happyIn141
		 ((Just happy_var_2, [happy_var_1])
	)}}

happyReduce_334 = happySpecReduce_0  130# happyReduction_334
happyReduction_334  =  happyIn141
		 ((Nothing, [])
	)

happyReduce_335 = happyMonadReduce 2# 131# happyReduction_335
happyReduction_335 (happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 DoubleColon) -> 
	case happyOut92 happy_x_2 of { happy_var_2 -> 
	( checkEnabled ScopedTypeVariables >> return (Just happy_var_2, [happy_var_1]))}}
	) (\r -> happyReturn (happyIn142 r))

happyReduce_336 = happySpecReduce_0  131# happyReduction_336
happyReduction_336  =  happyIn142
		 ((Nothing,[])
	)

happyReduce_337 = happySpecReduce_2  132# happyReduction_337
happyReduction_337 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Equals) -> 
	case happyOut146 happy_x_2 of { happy_var_2 -> 
	happyIn143
		 (UnGuardedRhs (nIS happy_var_1 <++> ann happy_var_2 <** [happy_var_1]) happy_var_2
	)}}

happyReduce_338 = happySpecReduce_1  132# happyReduction_338
happyReduction_338 happy_x_1
	 =  case happyOut144 happy_x_1 of { happy_var_1 -> 
	happyIn143
		 (GuardedRhss (snd happy_var_1) (reverse $ fst happy_var_1)
	)}

happyReduce_339 = happySpecReduce_2  133# happyReduction_339
happyReduction_339 happy_x_2
	happy_x_1
	 =  case happyOut144 happy_x_1 of { happy_var_1 -> 
	case happyOut145 happy_x_2 of { happy_var_2 -> 
	happyIn144
		 ((happy_var_2 : fst happy_var_1, snd happy_var_1 <++> ann happy_var_2)
	)}}

happyReduce_340 = happySpecReduce_1  133# happyReduction_340
happyReduction_340 happy_x_1
	 =  case happyOut145 happy_x_1 of { happy_var_1 -> 
	happyIn144
		 (([happy_var_1],ann happy_var_1)
	)}

happyReduce_341 = happyMonadReduce 4# 134# happyReduction_341
happyReduction_341 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 Bar) -> 
	case happyOut183 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 Equals) -> 
	case happyOut146 happy_x_4 of { happy_var_4 -> 
	( do { checkPatternGuards (fst happy_var_2);
                                       return $ GuardedRhs (nIS happy_var_1 <++> ann happy_var_4 <** (happy_var_1:snd happy_var_2 ++ [happy_var_3])) (reverse (fst happy_var_2)) happy_var_4 })}}}}
	) (\r -> happyReturn (happyIn145 r))

happyReduce_342 = happyMonadReduce 1# 135# happyReduction_342
happyReduction_342 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut147 happy_x_1 of { happy_var_1 -> 
	( checkExpr happy_var_1)}
	) (\r -> happyReturn (happyIn146 r))

happyReduce_343 = happySpecReduce_3  136# happyReduction_343
happyReduction_343 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut150 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 DoubleColon) -> 
	case happyOut92 happy_x_3 of { happy_var_3 -> 
	happyIn147
		 (ExpTypeSig      (happy_var_1 <> happy_var_3 <** [happy_var_2]) happy_var_1 happy_var_3
	)}}}

happyReduce_344 = happySpecReduce_1  136# happyReduction_344
happyReduction_344 happy_x_1
	 =  case happyOut148 happy_x_1 of { happy_var_1 -> 
	happyIn147
		 (happy_var_1
	)}

happyReduce_345 = happySpecReduce_2  136# happyReduction_345
happyReduction_345 happy_x_2
	happy_x_1
	 =  case happyOut150 happy_x_1 of { happy_var_1 -> 
	case happyOut219 happy_x_2 of { happy_var_2 -> 
	happyIn147
		 (PostOp          (happy_var_1 <> happy_var_2)          happy_var_1 happy_var_2
	)}}

happyReduce_346 = happySpecReduce_3  136# happyReduction_346
happyReduction_346 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut150 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftArrowTail) -> 
	case happyOut147 happy_x_3 of { happy_var_3 -> 
	happyIn147
		 (LeftArrApp      (happy_var_1 <> happy_var_3 <** [happy_var_2]) happy_var_1 happy_var_3
	)}}}

happyReduce_347 = happySpecReduce_3  136# happyReduction_347
happyReduction_347 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut150 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 RightArrowTail) -> 
	case happyOut147 happy_x_3 of { happy_var_3 -> 
	happyIn147
		 (RightArrApp     (happy_var_1 <> happy_var_3 <** [happy_var_2]) happy_var_1 happy_var_3
	)}}}

happyReduce_348 = happySpecReduce_3  136# happyReduction_348
happyReduction_348 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut150 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftDblArrowTail) -> 
	case happyOut147 happy_x_3 of { happy_var_3 -> 
	happyIn147
		 (LeftArrHighApp  (happy_var_1 <> happy_var_3 <** [happy_var_2]) happy_var_1 happy_var_3
	)}}}

happyReduce_349 = happySpecReduce_3  136# happyReduction_349
happyReduction_349 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut150 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 RightDblArrowTail) -> 
	case happyOut147 happy_x_3 of { happy_var_3 -> 
	happyIn147
		 (RightArrHighApp (happy_var_1 <> happy_var_3 <** [happy_var_2]) happy_var_1 happy_var_3
	)}}}

happyReduce_350 = happySpecReduce_1  137# happyReduction_350
happyReduction_350 happy_x_1
	 =  case happyOut149 happy_x_1 of { happy_var_1 -> 
	happyIn148
		 (happy_var_1
	)}

happyReduce_351 = happySpecReduce_1  137# happyReduction_351
happyReduction_351 happy_x_1
	 =  case happyOut150 happy_x_1 of { happy_var_1 -> 
	happyIn148
		 (happy_var_1
	)}

happyReduce_352 = happySpecReduce_3  138# happyReduction_352
happyReduction_352 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut150 happy_x_1 of { happy_var_1 -> 
	case happyOut219 happy_x_2 of { happy_var_2 -> 
	case happyOut151 happy_x_3 of { happy_var_3 -> 
	happyIn149
		 (InfixApp (happy_var_1 <> happy_var_3) happy_var_1 happy_var_2 happy_var_3
	)}}}

happyReduce_353 = happySpecReduce_1  138# happyReduction_353
happyReduction_353 happy_x_1
	 =  case happyOut151 happy_x_1 of { happy_var_1 -> 
	happyIn149
		 (happy_var_1
	)}

happyReduce_354 = happySpecReduce_3  139# happyReduction_354
happyReduction_354 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut150 happy_x_1 of { happy_var_1 -> 
	case happyOut219 happy_x_2 of { happy_var_2 -> 
	case happyOut154 happy_x_3 of { happy_var_3 -> 
	happyIn150
		 (InfixApp (happy_var_1 <> happy_var_3) happy_var_1 happy_var_2 happy_var_3
	)}}}

happyReduce_355 = happySpecReduce_1  139# happyReduction_355
happyReduction_355 happy_x_1
	 =  case happyOut154 happy_x_1 of { happy_var_1 -> 
	happyIn150
		 (happy_var_1
	)}

happyReduce_356 = happyReduce 4# 140# happyReduction_356
happyReduction_356 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 Backslash) -> 
	case happyOut157 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightArrow) -> 
	case happyOut147 happy_x_4 of { happy_var_4 -> 
	happyIn151
		 (Lambda (nIS happy_var_1 <++> ann happy_var_4 <** [happy_var_1,happy_var_3]) (reverse happy_var_2) happy_var_4
	) `HappyStk` happyRest}}}}

happyReduce_357 = happyReduce 4# 140# happyReduction_357
happyReduction_357 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Let) -> 
	case happyOut61 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 KW_In) -> 
	case happyOut147 happy_x_4 of { happy_var_4 -> 
	happyIn151
		 (Let    (nIS happy_var_1 <++> ann happy_var_4 <** [happy_var_1,happy_var_3])    happy_var_2 happy_var_4
	) `HappyStk` happyRest}}}}

happyReduce_358 = happyReduce 8# 140# happyReduction_358
happyReduction_358 (happy_x_8 `HappyStk`
	happy_x_7 `HappyStk`
	happy_x_6 `HappyStk`
	happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_If) -> 
	case happyOut147 happy_x_2 of { happy_var_2 -> 
	case happyOut152 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 KW_Then) -> 
	case happyOut147 happy_x_5 of { happy_var_5 -> 
	case happyOut152 happy_x_6 of { happy_var_6 -> 
	case happyOutTok happy_x_7 of { (Loc happy_var_7 KW_Else) -> 
	case happyOut147 happy_x_8 of { happy_var_8 -> 
	happyIn151
		 (If     (nIS happy_var_1 <++> ann happy_var_8 <** (happy_var_1:happy_var_3 ++ happy_var_4:happy_var_6 ++ [happy_var_7])) happy_var_2 happy_var_5 happy_var_8
	) `HappyStk` happyRest}}}}}}}}

happyReduce_359 = happyMonadReduce 2# 140# happyReduction_359
happyReduction_359 (happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_If) -> 
	case happyOut193 happy_x_2 of { happy_var_2 -> 
	( checkEnabled MultiWayIf >>
                                           let (alts, inf, ss) = happy_var_2
                                           in return (MultiIf (nIS happy_var_1 <++> inf <** (happy_var_1:ss)) alts))}}
	) (\r -> happyReturn (happyIn151 r))

happyReduce_360 = happyReduce 4# 140# happyReduction_360
happyReduction_360 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Proc) -> 
	case happyOut158 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightArrow) -> 
	case happyOut147 happy_x_4 of { happy_var_4 -> 
	happyIn151
		 (Proc   (nIS happy_var_1 <++> ann happy_var_4 <** [happy_var_1,happy_var_3])    happy_var_2 happy_var_4
	) `HappyStk` happyRest}}}}

happyReduce_361 = happySpecReduce_1  140# happyReduction_361
happyReduction_361 happy_x_1
	 =  case happyOut155 happy_x_1 of { happy_var_1 -> 
	happyIn151
		 (happy_var_1
	)}

happyReduce_362 = happyMonadReduce 1# 141# happyReduction_362
happyReduction_362 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 SemiColon) -> 
	( checkEnabled DoAndIfThenElse >> return [happy_var_1])}
	) (\r -> happyReturn (happyIn152 r))

happyReduce_363 = happySpecReduce_0  141# happyReduction_363
happyReduction_363  =  happyIn152
		 ([]
	)

happyReduce_364 = happySpecReduce_1  142# happyReduction_364
happyReduction_364 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 SemiColon) -> 
	happyIn153
		 ([happy_var_1]
	)}

happyReduce_365 = happySpecReduce_0  142# happyReduction_365
happyReduction_365  =  happyIn153
		 ([]
	)

happyReduce_366 = happyReduce 4# 143# happyReduction_366
happyReduction_366 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Case) -> 
	case happyOut147 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 KW_Of) -> 
	case happyOut185 happy_x_4 of { happy_var_4 -> 
	happyIn154
		 (let (als, inf, ss) = happy_var_4 in Case (nIS happy_var_1 <++> inf <** (happy_var_1:happy_var_3:ss)) happy_var_2 als
	) `HappyStk` happyRest}}}}

happyReduce_367 = happyMonadReduce 3# 143# happyReduction_367
happyReduction_367 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 Backslash) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 KW_Case) -> 
	case happyOut185 happy_x_3 of { happy_var_3 -> 
	( do { checkEnabled LambdaCase ;
                                              let { (als, inf, ss) = happy_var_3 } ;
                                              return (LCase (nIS happy_var_1 <++> inf <** (happy_var_1:happy_var_2:ss)) als) })}}}
	) (\r -> happyReturn (happyIn154 r))

happyReduce_368 = happySpecReduce_2  143# happyReduction_368
happyReduction_368 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Minus) -> 
	case happyOut156 happy_x_2 of { happy_var_2 -> 
	happyIn154
		 (NegApp (nIS happy_var_1 <++> ann happy_var_2 <** [happy_var_1]) happy_var_2
	)}}

happyReduce_369 = happySpecReduce_2  143# happyReduction_369
happyReduction_369 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Do) -> 
	case happyOut197 happy_x_2 of { happy_var_2 -> 
	happyIn154
		 (let (sts, inf, ss) = happy_var_2 in Do   (nIS happy_var_1 <++> inf <** happy_var_1:ss) sts
	)}}

happyReduce_370 = happySpecReduce_2  143# happyReduction_370
happyReduction_370 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_MDo) -> 
	case happyOut197 happy_x_2 of { happy_var_2 -> 
	happyIn154
		 (let (sts, inf, ss) = happy_var_2 in MDo  (nIS happy_var_1 <++> inf <** happy_var_1:ss) sts
	)}}

happyReduce_371 = happySpecReduce_1  143# happyReduction_371
happyReduction_371 happy_x_1
	 =  case happyOut156 happy_x_1 of { happy_var_1 -> 
	happyIn154
		 (happy_var_1
	)}

happyReduce_372 = happyReduce 4# 144# happyReduction_372
happyReduction_372 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 CORE) -> 
	case happyOutTok happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 PragmaEnd) -> 
	case happyOut147 happy_x_4 of { happy_var_4 -> 
	happyIn155
		 (let Loc l (StringTok (s,_)) = happy_var_2 in CorePragma (nIS happy_var_1 <++> ann happy_var_4 <** [l,happy_var_3]) s happy_var_4
	) `HappyStk` happyRest}}}}

happyReduce_373 = happyReduce 4# 144# happyReduction_373
happyReduction_373 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 SCC) -> 
	case happyOutTok happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 PragmaEnd) -> 
	case happyOut147 happy_x_4 of { happy_var_4 -> 
	happyIn155
		 (let Loc l (StringTok (s,_)) = happy_var_2 in SCCPragma  (nIS happy_var_1 <++> ann happy_var_4 <** [l,happy_var_3]) s happy_var_4
	) `HappyStk` happyRest}}}}

happyReduce_374 = happyReduce 11# 144# happyReduction_374
happyReduction_374 (happy_x_11 `HappyStk`
	happy_x_10 `HappyStk`
	happy_x_9 `HappyStk`
	happy_x_8 `HappyStk`
	happy_x_7 `HappyStk`
	happy_x_6 `HappyStk`
	happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 GENERATED) -> 
	case happyOutTok happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 Colon) -> 
	case happyOutTok happy_x_5 of { happy_var_5 -> 
	case happyOutTok happy_x_6 of { (Loc happy_var_6 Minus) -> 
	case happyOutTok happy_x_7 of { happy_var_7 -> 
	case happyOutTok happy_x_8 of { (Loc happy_var_8 Colon) -> 
	case happyOutTok happy_x_9 of { happy_var_9 -> 
	case happyOutTok happy_x_10 of { (Loc happy_var_10 PragmaEnd) -> 
	case happyOut147 happy_x_11 of { happy_var_11 -> 
	happyIn155
		 (let { Loc l0 (StringTok (s,_)) = happy_var_2;
                                                  Loc l1 (IntTok (i1,_))   = happy_var_3;
                                                  Loc l2 (IntTok (i2,_))   = happy_var_5;
                                                  Loc l3 (IntTok (i3,_))   = happy_var_7;
                                                  Loc l4 (IntTok (i4,_))   = happy_var_9}
                                             in GenPragma (nIS happy_var_1 <++> ann happy_var_11 <** [happy_var_1,l0,l1,happy_var_4,l2,happy_var_6,l3,happy_var_8,l4,happy_var_10])
                                                      s (fromInteger i1, fromInteger i2)
                                                        (fromInteger i3, fromInteger i4) happy_var_11
	) `HappyStk` happyRest}}}}}}}}}}}

happyReduce_375 = happySpecReduce_2  145# happyReduction_375
happyReduction_375 happy_x_2
	happy_x_1
	 =  case happyOut156 happy_x_1 of { happy_var_1 -> 
	case happyOut159 happy_x_2 of { happy_var_2 -> 
	happyIn156
		 (App (happy_var_1 <> happy_var_2) happy_var_1 happy_var_2
	)}}

happyReduce_376 = happySpecReduce_1  145# happyReduction_376
happyReduction_376 happy_x_1
	 =  case happyOut159 happy_x_1 of { happy_var_1 -> 
	happyIn156
		 (happy_var_1
	)}

happyReduce_377 = happySpecReduce_2  146# happyReduction_377
happyReduction_377 happy_x_2
	happy_x_1
	 =  case happyOut157 happy_x_1 of { happy_var_1 -> 
	case happyOut158 happy_x_2 of { happy_var_2 -> 
	happyIn157
		 (happy_var_2 : happy_var_1
	)}}

happyReduce_378 = happySpecReduce_1  146# happyReduction_378
happyReduction_378 happy_x_1
	 =  case happyOut158 happy_x_1 of { happy_var_1 -> 
	happyIn157
		 ([happy_var_1]
	)}

happyReduce_379 = happyMonadReduce 1# 147# happyReduction_379
happyReduction_379 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut159 happy_x_1 of { happy_var_1 -> 
	( checkPattern happy_var_1)}
	) (\r -> happyReturn (happyIn158 r))

happyReduce_380 = happyMonadReduce 2# 147# happyReduction_380
happyReduction_380 (happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 Exclamation) -> 
	case happyOut159 happy_x_2 of { happy_var_2 -> 
	( checkPattern (BangPat (nIS happy_var_1 <++> ann happy_var_2 <** [happy_var_1]) happy_var_2))}}
	) (\r -> happyReturn (happyIn158 r))

happyReduce_381 = happyMonadReduce 3# 148# happyReduction_381
happyReduction_381 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut209 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 At) -> 
	case happyOut159 happy_x_3 of { happy_var_3 -> 
	( do { n <- checkUnQual happy_var_1;
                                              return (AsPat (happy_var_1 <> happy_var_3 <** [happy_var_2]) n happy_var_3) })}}}
	) (\r -> happyReturn (happyIn159 r))

happyReduce_382 = happyMonadReduce 3# 148# happyReduction_382
happyReduction_382 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut209 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 RPCAt) -> 
	case happyOut159 happy_x_3 of { happy_var_3 -> 
	( do { n <- checkUnQual happy_var_1;
                                              return (CAsRP (happy_var_1 <> happy_var_3 <** [happy_var_2]) n happy_var_3) })}}}
	) (\r -> happyReturn (happyIn159 r))

happyReduce_383 = happySpecReduce_2  148# happyReduction_383
happyReduction_383 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Tilde) -> 
	case happyOut159 happy_x_2 of { happy_var_2 -> 
	happyIn159
		 (IrrPat (nIS happy_var_1 <++> ann happy_var_2 <** [happy_var_1]) happy_var_2
	)}}

happyReduce_384 = happySpecReduce_1  148# happyReduction_384
happyReduction_384 happy_x_1
	 =  case happyOut160 happy_x_1 of { happy_var_1 -> 
	happyIn159
		 (happy_var_1
	)}

happyReduce_385 = happyMonadReduce 3# 149# happyReduction_385
happyReduction_385 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut160 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftCurly) -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightCurly) -> 
	( liftM (amap (const (ann happy_var_1 <++> nIS happy_var_3 <** [happy_var_2,happy_var_3]))) $ mkRecConstrOrUpdate happy_var_1 [])}}}
	) (\r -> happyReturn (happyIn160 r))

happyReduce_386 = happyMonadReduce 4# 149# happyReduction_386
happyReduction_386 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut160 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftCurly) -> 
	case happyOut201 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 RightCurly) -> 
	( liftM (amap (const (ann happy_var_1 <++> nIS happy_var_4 <** (happy_var_2:reverse (snd happy_var_3) ++ [happy_var_4]))))
                                              $ mkRecConstrOrUpdate happy_var_1 (reverse (fst happy_var_3)))}}}}
	) (\r -> happyReturn (happyIn160 r))

happyReduce_387 = happySpecReduce_1  149# happyReduction_387
happyReduction_387 happy_x_1
	 =  case happyOut161 happy_x_1 of { happy_var_1 -> 
	happyIn160
		 (happy_var_1
	)}

happyReduce_388 = happySpecReduce_1  150# happyReduction_388
happyReduction_388 happy_x_1
	 =  case happyOut210 happy_x_1 of { happy_var_1 -> 
	happyIn161
		 (IPVar (ann happy_var_1) happy_var_1
	)}

happyReduce_389 = happySpecReduce_1  150# happyReduction_389
happyReduction_389 happy_x_1
	 =  case happyOut209 happy_x_1 of { happy_var_1 -> 
	happyIn161
		 (Var (ann happy_var_1) happy_var_1
	)}

happyReduce_390 = happySpecReduce_1  150# happyReduction_390
happyReduction_390 happy_x_1
	 =  case happyOut206 happy_x_1 of { happy_var_1 -> 
	happyIn161
		 (happy_var_1
	)}

happyReduce_391 = happySpecReduce_1  150# happyReduction_391
happyReduction_391 happy_x_1
	 =  case happyOut235 happy_x_1 of { happy_var_1 -> 
	happyIn161
		 (Lit (ann happy_var_1) happy_var_1
	)}

happyReduce_392 = happySpecReduce_3  150# happyReduction_392
happyReduction_392 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut163 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn161
		 (Paren (happy_var_1 <^^> happy_var_3 <** [happy_var_1,happy_var_3]) happy_var_2
	)}}}

happyReduce_393 = happySpecReduce_3  150# happyReduction_393
happyReduction_393 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut163 happy_x_2 of { happy_var_2 -> 
	case happyOut164 happy_x_3 of { happy_var_3 -> 
	happyIn161
		 (TupleSection (happy_var_1 <^^> head (snd happy_var_3) <** happy_var_1:reverse (snd happy_var_3)) Boxed (Just happy_var_2 : fst happy_var_3)
	)}}}

happyReduce_394 = happyReduce 4# 150# happyReduction_394
happyReduction_394 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut162 happy_x_2 of { happy_var_2 -> 
	case happyOut163 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 RightParen) -> 
	happyIn161
		 (TupleSection (happy_var_1 <^^> happy_var_4 <** happy_var_1:reverse (happy_var_4:happy_var_2)) Boxed
                                                      (replicate (length happy_var_2) Nothing ++ [Just happy_var_3])
	) `HappyStk` happyRest}}}}

happyReduce_395 = happyReduce 4# 150# happyReduction_395
happyReduction_395 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut162 happy_x_2 of { happy_var_2 -> 
	case happyOut163 happy_x_3 of { happy_var_3 -> 
	case happyOut164 happy_x_4 of { happy_var_4 -> 
	happyIn161
		 (TupleSection (happy_var_1 <^^> head (snd happy_var_4) <** happy_var_1:reverse (snd happy_var_4 ++ happy_var_2)) Boxed
                                                      (replicate (length happy_var_2) Nothing ++ Just happy_var_3 : fst happy_var_4)
	) `HappyStk` happyRest}}}}

happyReduce_396 = happySpecReduce_3  150# happyReduction_396
happyReduction_396 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftHashParen) -> 
	case happyOut163 happy_x_2 of { happy_var_2 -> 
	case happyOut165 happy_x_3 of { happy_var_3 -> 
	happyIn161
		 (TupleSection (happy_var_1 <^^> head (snd happy_var_3) <** happy_var_1:reverse (snd happy_var_3)) Unboxed (Just happy_var_2 : fst happy_var_3)
	)}}}

happyReduce_397 = happySpecReduce_3  150# happyReduction_397
happyReduction_397 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftHashParen) -> 
	case happyOut163 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightHashParen) -> 
	happyIn161
		 (TupleSection (happy_var_1 <^^> happy_var_3 <** [happy_var_1,happy_var_3]) Unboxed [Just happy_var_2]
	)}}}

happyReduce_398 = happyReduce 4# 150# happyReduction_398
happyReduction_398 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftHashParen) -> 
	case happyOut162 happy_x_2 of { happy_var_2 -> 
	case happyOut163 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 RightHashParen) -> 
	happyIn161
		 (TupleSection (happy_var_1 <^^> happy_var_4 <** happy_var_1:reverse (happy_var_4:happy_var_2)) Unboxed
                                                      (replicate (length happy_var_2) Nothing ++ [Just happy_var_3])
	) `HappyStk` happyRest}}}}

happyReduce_399 = happyReduce 4# 150# happyReduction_399
happyReduction_399 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftHashParen) -> 
	case happyOut162 happy_x_2 of { happy_var_2 -> 
	case happyOut163 happy_x_3 of { happy_var_3 -> 
	case happyOut165 happy_x_4 of { happy_var_4 -> 
	happyIn161
		 (TupleSection (happy_var_1 <^^> head (snd happy_var_4) <** happy_var_1:reverse (snd happy_var_4 ++ happy_var_2)) Unboxed
                                                      (replicate (length happy_var_2) Nothing ++ Just happy_var_3 : fst happy_var_4)
	) `HappyStk` happyRest}}}}

happyReduce_400 = happySpecReduce_3  150# happyReduction_400
happyReduction_400 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftSquare) -> 
	case happyOut177 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightSquare) -> 
	happyIn161
		 (amap (\l -> l <** [happy_var_3]) $ happy_var_2 (happy_var_1 <^^> happy_var_3 <** [happy_var_1])
	)}}}

happyReduce_401 = happySpecReduce_1  150# happyReduction_401
happyReduction_401 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Underscore) -> 
	happyIn161
		 (WildCard (nIS happy_var_1)
	)}

happyReduce_402 = happyMonadReduce 3# 150# happyReduction_402
happyReduction_402 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut167 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	( checkEnabled RegularPatterns >> return (Paren (happy_var_1 <^^> happy_var_3 <** [happy_var_1,happy_var_3]) happy_var_2))}}}
	) (\r -> happyReturn (happyIn161 r))

happyReduce_403 = happySpecReduce_3  150# happyReduction_403
happyReduction_403 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 RPGuardOpen) -> 
	case happyOut166 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RPGuardClose) -> 
	happyIn161
		 (SeqRP (happy_var_1 <^^> happy_var_3 <** (happy_var_1:reverse (snd happy_var_2) ++ [happy_var_3])) $ reverse (fst happy_var_2)
	)}}}

happyReduce_404 = happyReduce 5# 150# happyReduction_404
happyReduction_404 (happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 RPGuardOpen) -> 
	case happyOut147 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 Bar) -> 
	case happyOut183 happy_x_4 of { happy_var_4 -> 
	case happyOutTok happy_x_5 of { (Loc happy_var_5 RPGuardClose) -> 
	happyIn161
		 (GuardRP (happy_var_1 <^^> happy_var_5 <** (happy_var_1:happy_var_3 : snd happy_var_4 ++ [happy_var_5])) happy_var_2 $ (reverse $ fst happy_var_4)
	) `HappyStk` happyRest}}}}}

happyReduce_405 = happySpecReduce_1  150# happyReduction_405
happyReduction_405 happy_x_1
	 =  case happyOut168 happy_x_1 of { happy_var_1 -> 
	happyIn161
		 (happy_var_1
	)}

happyReduce_406 = happySpecReduce_1  150# happyReduction_406
happyReduction_406 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn161
		 (let Loc l (THIdEscape s) = happy_var_1 in SpliceExp (nIS l) $ IdSplice (nIS l) s
	)}

happyReduce_407 = happySpecReduce_3  150# happyReduction_407
happyReduction_407 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 THParenEscape) -> 
	case happyOut146 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn161
		 (SpliceExp  (happy_var_1 <^^> happy_var_3 <** [happy_var_1,happy_var_3]) $ ParenSplice (ann happy_var_2) happy_var_2
	)}}}

happyReduce_408 = happySpecReduce_3  150# happyReduction_408
happyReduction_408 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 THExpQuote) -> 
	case happyOut146 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 THCloseQuote) -> 
	happyIn161
		 (BracketExp (happy_var_1 <^^> happy_var_3 <** [happy_var_1,happy_var_3]) $ ExpBracket  (ann happy_var_2) happy_var_2
	)}}}

happyReduce_409 = happyMonadReduce 3# 150# happyReduction_409
happyReduction_409 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 THPatQuote) -> 
	case happyOut148 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 THCloseQuote) -> 
	( do { p <- checkPattern happy_var_2;
                                              return $ BracketExp (happy_var_1 <^^> happy_var_3 <** [happy_var_1,happy_var_3]) $ PatBracket (ann p) p })}}}
	) (\r -> happyReturn (happyIn161 r))

happyReduce_410 = happySpecReduce_3  150# happyReduction_410
happyReduction_410 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 THTypQuote) -> 
	case happyOut92 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 THCloseQuote) -> 
	happyIn161
		 (let l = happy_var_1 <^^> happy_var_3 <** [happy_var_1,happy_var_3] in BracketExp l $ TypeBracket l happy_var_2
	)}}}

happyReduce_411 = happyReduce 5# 150# happyReduction_411
happyReduction_411 (happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 THDecQuote) -> 
	case happyOut48 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_5 of { (Loc happy_var_5 THCloseQuote) -> 
	happyIn161
		 (let l = happy_var_1 <^^> happy_var_5 <** (happy_var_1:snd happy_var_3 ++ [happy_var_5]) in BracketExp l $ DeclBracket l (fst happy_var_3)
	) `HappyStk` happyRest}}}

happyReduce_412 = happySpecReduce_2  150# happyReduction_412
happyReduction_412 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 THVarQuote) -> 
	case happyOut209 happy_x_2 of { happy_var_2 -> 
	happyIn161
		 (VarQuote (nIS happy_var_1 <++> ann happy_var_2 <** [happy_var_1]) happy_var_2
	)}}

happyReduce_413 = happySpecReduce_2  150# happyReduction_413
happyReduction_413 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 THVarQuote) -> 
	case happyOut212 happy_x_2 of { happy_var_2 -> 
	happyIn161
		 (VarQuote (nIS happy_var_1 <++> ann happy_var_2 <** [happy_var_1]) happy_var_2
	)}}

happyReduce_414 = happySpecReduce_2  150# happyReduction_414
happyReduction_414 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 THTyQuote) -> 
	case happyOut241 happy_x_2 of { happy_var_2 -> 
	happyIn161
		 (TypQuote (nIS happy_var_1 <++> ann happy_var_2 <** [happy_var_1]) (UnQual (ann happy_var_2) happy_var_2)
	)}}

happyReduce_415 = happySpecReduce_2  150# happyReduction_415
happyReduction_415 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 THTyQuote) -> 
	case happyOut89 happy_x_2 of { happy_var_2 -> 
	happyIn161
		 (TypQuote (nIS happy_var_1 <++> ann happy_var_2 <** [happy_var_1]) happy_var_2
	)}}

happyReduce_416 = happySpecReduce_1  150# happyReduction_416
happyReduction_416 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn161
		 (let Loc l (THQuasiQuote (n,q)) = happy_var_1 in QuasiQuote (nIS l) n q
	)}

happyReduce_417 = happySpecReduce_2  151# happyReduction_417
happyReduction_417 happy_x_2
	happy_x_1
	 =  case happyOut162 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	happyIn162
		 (happy_var_2 : happy_var_1
	)}}

happyReduce_418 = happySpecReduce_1  151# happyReduction_418
happyReduction_418 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Comma) -> 
	happyIn162
		 ([happy_var_1]
	)}

happyReduce_419 = happySpecReduce_1  152# happyReduction_419
happyReduction_419 happy_x_1
	 =  case happyOut147 happy_x_1 of { happy_var_1 -> 
	happyIn163
		 (happy_var_1
	)}

happyReduce_420 = happySpecReduce_2  152# happyReduction_420
happyReduction_420 happy_x_2
	happy_x_1
	 =  case happyOut220 happy_x_1 of { happy_var_1 -> 
	case happyOut148 happy_x_2 of { happy_var_2 -> 
	happyIn163
		 (PreOp (happy_var_1 <> happy_var_2) happy_var_1 happy_var_2
	)}}

happyReduce_421 = happyMonadReduce 3# 152# happyReduction_421
happyReduction_421 (happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut147 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 RightArrow) -> 
	case happyOut147 happy_x_3 of { happy_var_3 -> 
	( do {checkEnabled ViewPatterns;
                                             return $ ViewPat (happy_var_1 <> happy_var_3 <** [happy_var_2]) happy_var_1 happy_var_3})}}}
	) (\r -> happyReturn (happyIn163 r))

happyReduce_422 = happySpecReduce_3  153# happyReduction_422
happyReduction_422 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut162 happy_x_1 of { happy_var_1 -> 
	case happyOut163 happy_x_2 of { happy_var_2 -> 
	case happyOut164 happy_x_3 of { happy_var_3 -> 
	happyIn164
		 (let (mes, ss) = happy_var_3 in (replicate (length happy_var_1 - 1) Nothing ++ Just happy_var_2 : mes, ss ++ happy_var_1)
	)}}}

happyReduce_423 = happySpecReduce_3  153# happyReduction_423
happyReduction_423 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut162 happy_x_1 of { happy_var_1 -> 
	case happyOut163 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn164
		 ((replicate (length happy_var_1 - 1) Nothing ++ [Just happy_var_2], happy_var_3 : happy_var_1)
	)}}}

happyReduce_424 = happySpecReduce_2  153# happyReduction_424
happyReduction_424 happy_x_2
	happy_x_1
	 =  case happyOut162 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 RightParen) -> 
	happyIn164
		 ((replicate (length happy_var_1) Nothing, happy_var_2 : happy_var_1)
	)}}

happyReduce_425 = happySpecReduce_3  154# happyReduction_425
happyReduction_425 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut162 happy_x_1 of { happy_var_1 -> 
	case happyOut163 happy_x_2 of { happy_var_2 -> 
	case happyOut165 happy_x_3 of { happy_var_3 -> 
	happyIn165
		 (let (mes, ss) = happy_var_3 in (replicate (length happy_var_1 - 1) Nothing ++ Just happy_var_2 : mes, ss ++ happy_var_1)
	)}}}

happyReduce_426 = happySpecReduce_3  154# happyReduction_426
happyReduction_426 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut162 happy_x_1 of { happy_var_1 -> 
	case happyOut163 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightHashParen) -> 
	happyIn165
		 ((replicate (length happy_var_1 - 1) Nothing ++ [Just happy_var_2], happy_var_3 : happy_var_1)
	)}}}

happyReduce_427 = happySpecReduce_2  154# happyReduction_427
happyReduction_427 happy_x_2
	happy_x_1
	 =  case happyOut162 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 RightHashParen) -> 
	happyIn165
		 ((replicate (length happy_var_1) Nothing, happy_var_2 : happy_var_1)
	)}}

happyReduce_428 = happySpecReduce_3  155# happyReduction_428
happyReduction_428 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut166 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut147 happy_x_3 of { happy_var_3 -> 
	happyIn166
		 ((happy_var_3 : fst happy_var_1, happy_var_2 : snd happy_var_1)
	)}}}

happyReduce_429 = happySpecReduce_1  155# happyReduction_429
happyReduction_429 happy_x_1
	 =  case happyOut147 happy_x_1 of { happy_var_1 -> 
	happyIn166
		 (([happy_var_1],[])
	)}

happyReduce_430 = happySpecReduce_3  156# happyReduction_430
happyReduction_430 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut147 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Bar) -> 
	case happyOut167 happy_x_3 of { happy_var_3 -> 
	happyIn167
		 (EitherRP (happy_var_1 <> happy_var_3 <** [happy_var_2]) happy_var_1 happy_var_3
	)}}}

happyReduce_431 = happySpecReduce_3  156# happyReduction_431
happyReduction_431 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut147 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Bar) -> 
	case happyOut147 happy_x_3 of { happy_var_3 -> 
	happyIn167
		 (EitherRP (happy_var_1 <> happy_var_3 <** [happy_var_2]) happy_var_1 happy_var_3
	)}}}

happyReduce_432 = happyMonadReduce 10# 157# happyReduction_432
happyReduction_432 (happy_x_10 `HappyStk`
	happy_x_9 `HappyStk`
	happy_x_8 `HappyStk`
	happy_x_7 `HappyStk`
	happy_x_6 `HappyStk`
	happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 XStdTagOpen) -> 
	case happyOut171 happy_x_2 of { happy_var_2 -> 
	case happyOut174 happy_x_3 of { happy_var_3 -> 
	case happyOut176 happy_x_4 of { happy_var_4 -> 
	case happyOutTok happy_x_5 of { (Loc happy_var_5 XStdTagClose) -> 
	case happyOut169 happy_x_6 of { happy_var_6 -> 
	case happyOut153 happy_x_7 of { happy_var_7 -> 
	case happyOutTok happy_x_8 of { (Loc happy_var_8 XCloseTagOpen) -> 
	case happyOut171 happy_x_9 of { happy_var_9 -> 
	case happyOutTok happy_x_10 of { (Loc happy_var_10 XStdTagClose) -> 
	( do { n <- checkEqNames happy_var_2 happy_var_9;
                                                                       let { cn = reverse happy_var_6;
                                                                             as = reverse happy_var_3;
                                                                             l  = happy_var_1 <^^> happy_var_10 <** [happy_var_1,happy_var_5] ++ happy_var_7 ++ [happy_var_8,srcInfoSpan (ann happy_var_9),happy_var_10] };
                                                                       return $ XTag l n as happy_var_4 cn })}}}}}}}}}}
	) (\r -> happyReturn (happyIn168 r))

happyReduce_433 = happyReduce 5# 157# happyReduction_433
happyReduction_433 (happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 XStdTagOpen) -> 
	case happyOut171 happy_x_2 of { happy_var_2 -> 
	case happyOut174 happy_x_3 of { happy_var_3 -> 
	case happyOut176 happy_x_4 of { happy_var_4 -> 
	case happyOutTok happy_x_5 of { (Loc happy_var_5 XEmptyTagClose) -> 
	happyIn168
		 (XETag   (happy_var_1 <^^> happy_var_5 <** [happy_var_1,happy_var_5]) happy_var_2 (reverse happy_var_3) happy_var_4
	) `HappyStk` happyRest}}}}}

happyReduce_434 = happySpecReduce_3  157# happyReduction_434
happyReduction_434 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 XCodeTagOpen) -> 
	case happyOut147 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 XCodeTagClose) -> 
	happyIn168
		 (XExpTag (happy_var_1 <^^> happy_var_3 <** [happy_var_1,happy_var_3]) happy_var_2
	)}}}

happyReduce_435 = happyReduce 5# 157# happyReduction_435
happyReduction_435 (happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 XChildTagOpen) -> 
	case happyOut169 happy_x_2 of { happy_var_2 -> 
	case happyOut153 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 XCloseTagOpen) -> 
	case happyOutTok happy_x_5 of { (Loc happy_var_5 XCodeTagClose) -> 
	happyIn168
		 (XChildTag (happy_var_1 <^^> happy_var_5 <** (happy_var_1:happy_var_3++[happy_var_4,happy_var_5])) (reverse happy_var_2)
	) `HappyStk` happyRest}}}}}

happyReduce_436 = happySpecReduce_2  158# happyReduction_436
happyReduction_436 happy_x_2
	happy_x_1
	 =  case happyOut169 happy_x_1 of { happy_var_1 -> 
	case happyOut170 happy_x_2 of { happy_var_2 -> 
	happyIn169
		 (happy_var_2 : happy_var_1
	)}}

happyReduce_437 = happySpecReduce_0  158# happyReduction_437
happyReduction_437  =  happyIn169
		 ([]
	)

happyReduce_438 = happySpecReduce_1  159# happyReduction_438
happyReduction_438 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn170
		 (let Loc l (XPCDATA pcd) = happy_var_1 in XPcdata (nIS l) pcd
	)}

happyReduce_439 = happySpecReduce_3  159# happyReduction_439
happyReduction_439 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 XRPatOpen) -> 
	case happyOut166 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 XRPatClose) -> 
	happyIn170
		 (XRPats (happy_var_1 <^^> happy_var_3 <** (snd happy_var_2 ++ [happy_var_1,happy_var_3])) $ reverse (fst happy_var_2)
	)}}}

happyReduce_440 = happySpecReduce_1  159# happyReduction_440
happyReduction_440 happy_x_1
	 =  case happyOut168 happy_x_1 of { happy_var_1 -> 
	happyIn170
		 (happy_var_1
	)}

happyReduce_441 = happySpecReduce_3  160# happyReduction_441
happyReduction_441 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut172 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Colon) -> 
	case happyOut172 happy_x_3 of { happy_var_3 -> 
	happyIn171
		 (let {Loc l1 s1 = happy_var_1; Loc l2 s2 = happy_var_3}
                                         in XDomName (nIS l1 <++> nIS l2 <** [l1,happy_var_2,l2]) s1 s2
	)}}}

happyReduce_442 = happySpecReduce_1  160# happyReduction_442
happyReduction_442 happy_x_1
	 =  case happyOut172 happy_x_1 of { happy_var_1 -> 
	happyIn171
		 (let Loc l str = happy_var_1 in XName (nIS l) str
	)}

happyReduce_443 = happySpecReduce_1  161# happyReduction_443
happyReduction_443 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn172
		 (let Loc l (VarId  s) = happy_var_1 in Loc l s
	)}

happyReduce_444 = happySpecReduce_1  161# happyReduction_444
happyReduction_444 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn172
		 (let Loc l (ConId  s) = happy_var_1 in Loc l s
	)}

happyReduce_445 = happySpecReduce_1  161# happyReduction_445
happyReduction_445 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn172
		 (let Loc l (DVarId s) = happy_var_1 in Loc l $ mkDVar s
	)}

happyReduce_446 = happySpecReduce_1  161# happyReduction_446
happyReduction_446 happy_x_1
	 =  case happyOut173 happy_x_1 of { happy_var_1 -> 
	happyIn172
		 (happy_var_1
	)}

happyReduce_447 = happySpecReduce_1  162# happyReduction_447
happyReduction_447 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Type) -> 
	happyIn173
		 (Loc happy_var_1 "type"
	)}

happyReduce_448 = happySpecReduce_1  162# happyReduction_448
happyReduction_448 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Class) -> 
	happyIn173
		 (Loc happy_var_1 "class"
	)}

happyReduce_449 = happySpecReduce_1  162# happyReduction_449
happyReduction_449 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Data) -> 
	happyIn173
		 (Loc happy_var_1 "data"
	)}

happyReduce_450 = happySpecReduce_1  162# happyReduction_450
happyReduction_450 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Foreign) -> 
	happyIn173
		 (Loc happy_var_1 "foreign"
	)}

happyReduce_451 = happySpecReduce_1  162# happyReduction_451
happyReduction_451 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Export) -> 
	happyIn173
		 (Loc happy_var_1 "export"
	)}

happyReduce_452 = happySpecReduce_1  162# happyReduction_452
happyReduction_452 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Safe) -> 
	happyIn173
		 (Loc happy_var_1 "safe"
	)}

happyReduce_453 = happySpecReduce_1  162# happyReduction_453
happyReduction_453 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Unsafe) -> 
	happyIn173
		 (Loc happy_var_1 "unsafe"
	)}

happyReduce_454 = happySpecReduce_1  162# happyReduction_454
happyReduction_454 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Threadsafe) -> 
	happyIn173
		 (Loc happy_var_1 "threadsafe"
	)}

happyReduce_455 = happySpecReduce_1  162# happyReduction_455
happyReduction_455 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_StdCall) -> 
	happyIn173
		 (Loc happy_var_1 "stdcall"
	)}

happyReduce_456 = happySpecReduce_1  162# happyReduction_456
happyReduction_456 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_CCall) -> 
	happyIn173
		 (Loc happy_var_1 "ccall"
	)}

happyReduce_457 = happySpecReduce_1  162# happyReduction_457
happyReduction_457 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_CPlusPlus) -> 
	happyIn173
		 (Loc happy_var_1 "cplusplus"
	)}

happyReduce_458 = happySpecReduce_1  162# happyReduction_458
happyReduction_458 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_DotNet) -> 
	happyIn173
		 (Loc happy_var_1 "dotnet"
	)}

happyReduce_459 = happySpecReduce_1  162# happyReduction_459
happyReduction_459 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Jvm) -> 
	happyIn173
		 (Loc happy_var_1 "jvm"
	)}

happyReduce_460 = happySpecReduce_1  162# happyReduction_460
happyReduction_460 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Js) -> 
	happyIn173
		 (Loc happy_var_1 "js"
	)}

happyReduce_461 = happySpecReduce_1  162# happyReduction_461
happyReduction_461 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_As) -> 
	happyIn173
		 (Loc happy_var_1 "as"
	)}

happyReduce_462 = happySpecReduce_1  162# happyReduction_462
happyReduction_462 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_By) -> 
	happyIn173
		 (Loc happy_var_1 "by"
	)}

happyReduce_463 = happySpecReduce_1  162# happyReduction_463
happyReduction_463 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Case) -> 
	happyIn173
		 (Loc happy_var_1 "case"
	)}

happyReduce_464 = happySpecReduce_1  162# happyReduction_464
happyReduction_464 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Default) -> 
	happyIn173
		 (Loc happy_var_1 "default"
	)}

happyReduce_465 = happySpecReduce_1  162# happyReduction_465
happyReduction_465 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Deriving) -> 
	happyIn173
		 (Loc happy_var_1 "deriving"
	)}

happyReduce_466 = happySpecReduce_1  162# happyReduction_466
happyReduction_466 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Do) -> 
	happyIn173
		 (Loc happy_var_1 "do"
	)}

happyReduce_467 = happySpecReduce_1  162# happyReduction_467
happyReduction_467 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Else) -> 
	happyIn173
		 (Loc happy_var_1 "else"
	)}

happyReduce_468 = happySpecReduce_1  162# happyReduction_468
happyReduction_468 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Family) -> 
	happyIn173
		 (Loc happy_var_1 "family"
	)}

happyReduce_469 = happySpecReduce_1  162# happyReduction_469
happyReduction_469 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Forall) -> 
	happyIn173
		 (Loc happy_var_1 "forall"
	)}

happyReduce_470 = happySpecReduce_1  162# happyReduction_470
happyReduction_470 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Group) -> 
	happyIn173
		 (Loc happy_var_1 "group"
	)}

happyReduce_471 = happySpecReduce_1  162# happyReduction_471
happyReduction_471 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Hiding) -> 
	happyIn173
		 (Loc happy_var_1 "hiding"
	)}

happyReduce_472 = happySpecReduce_1  162# happyReduction_472
happyReduction_472 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_If) -> 
	happyIn173
		 (Loc happy_var_1 "if"
	)}

happyReduce_473 = happySpecReduce_1  162# happyReduction_473
happyReduction_473 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Import) -> 
	happyIn173
		 (Loc happy_var_1 "import"
	)}

happyReduce_474 = happySpecReduce_1  162# happyReduction_474
happyReduction_474 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_In) -> 
	happyIn173
		 (Loc happy_var_1 "in"
	)}

happyReduce_475 = happySpecReduce_1  162# happyReduction_475
happyReduction_475 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Infix) -> 
	happyIn173
		 (Loc happy_var_1 "infix"
	)}

happyReduce_476 = happySpecReduce_1  162# happyReduction_476
happyReduction_476 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_InfixL) -> 
	happyIn173
		 (Loc happy_var_1 "infixl"
	)}

happyReduce_477 = happySpecReduce_1  162# happyReduction_477
happyReduction_477 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_InfixR) -> 
	happyIn173
		 (Loc happy_var_1 "infixr"
	)}

happyReduce_478 = happySpecReduce_1  162# happyReduction_478
happyReduction_478 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Instance) -> 
	happyIn173
		 (Loc happy_var_1 "instance"
	)}

happyReduce_479 = happySpecReduce_1  162# happyReduction_479
happyReduction_479 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Let) -> 
	happyIn173
		 (Loc happy_var_1 "let"
	)}

happyReduce_480 = happySpecReduce_1  162# happyReduction_480
happyReduction_480 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_MDo) -> 
	happyIn173
		 (Loc happy_var_1 "mdo"
	)}

happyReduce_481 = happySpecReduce_1  162# happyReduction_481
happyReduction_481 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Module) -> 
	happyIn173
		 (Loc happy_var_1 "module"
	)}

happyReduce_482 = happySpecReduce_1  162# happyReduction_482
happyReduction_482 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_NewType) -> 
	happyIn173
		 (Loc happy_var_1 "newtype"
	)}

happyReduce_483 = happySpecReduce_1  162# happyReduction_483
happyReduction_483 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Of) -> 
	happyIn173
		 (Loc happy_var_1 "of"
	)}

happyReduce_484 = happySpecReduce_1  162# happyReduction_484
happyReduction_484 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Proc) -> 
	happyIn173
		 (Loc happy_var_1 "proc"
	)}

happyReduce_485 = happySpecReduce_1  162# happyReduction_485
happyReduction_485 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Rec) -> 
	happyIn173
		 (Loc happy_var_1 "rec"
	)}

happyReduce_486 = happySpecReduce_1  162# happyReduction_486
happyReduction_486 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Then) -> 
	happyIn173
		 (Loc happy_var_1 "then"
	)}

happyReduce_487 = happySpecReduce_1  162# happyReduction_487
happyReduction_487 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Using) -> 
	happyIn173
		 (Loc happy_var_1 "using"
	)}

happyReduce_488 = happySpecReduce_1  162# happyReduction_488
happyReduction_488 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Where) -> 
	happyIn173
		 (Loc happy_var_1 "where"
	)}

happyReduce_489 = happySpecReduce_1  162# happyReduction_489
happyReduction_489 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Qualified) -> 
	happyIn173
		 (Loc happy_var_1 "qualified"
	)}

happyReduce_490 = happySpecReduce_2  163# happyReduction_490
happyReduction_490 happy_x_2
	happy_x_1
	 =  case happyOut174 happy_x_1 of { happy_var_1 -> 
	case happyOut175 happy_x_2 of { happy_var_2 -> 
	happyIn174
		 (happy_var_2 : happy_var_1
	)}}

happyReduce_491 = happySpecReduce_0  163# happyReduction_491
happyReduction_491  =  happyIn174
		 ([]
	)

happyReduce_492 = happySpecReduce_3  164# happyReduction_492
happyReduction_492 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut171 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Equals) -> 
	case happyOut159 happy_x_3 of { happy_var_3 -> 
	happyIn175
		 (XAttr (happy_var_1 <> happy_var_3 <** [happy_var_2]) happy_var_1 happy_var_3
	)}}}

happyReduce_493 = happySpecReduce_1  165# happyReduction_493
happyReduction_493 happy_x_1
	 =  case happyOut159 happy_x_1 of { happy_var_1 -> 
	happyIn176
		 (Just happy_var_1
	)}

happyReduce_494 = happySpecReduce_0  165# happyReduction_494
happyReduction_494  =  happyIn176
		 (Nothing
	)

happyReduce_495 = happySpecReduce_1  166# happyReduction_495
happyReduction_495 happy_x_1
	 =  case happyOut163 happy_x_1 of { happy_var_1 -> 
	happyIn177
		 (\l -> List l [happy_var_1]
	)}

happyReduce_496 = happySpecReduce_1  166# happyReduction_496
happyReduction_496 happy_x_1
	 =  case happyOut178 happy_x_1 of { happy_var_1 -> 
	happyIn177
		 (\l -> let (ps,ss) = happy_var_1 in List (l <** reverse ss) (reverse ps)
	)}

happyReduce_497 = happySpecReduce_2  166# happyReduction_497
happyReduction_497 happy_x_2
	happy_x_1
	 =  case happyOut163 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 DotDot) -> 
	happyIn177
		 (\l -> EnumFrom       (l <** [happy_var_2]) happy_var_1
	)}}

happyReduce_498 = happyReduce 4# 166# happyReduction_498
happyReduction_498 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOut163 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut147 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 DotDot) -> 
	happyIn177
		 (\l -> EnumFromThen   (l <** [happy_var_2,happy_var_4]) happy_var_1 happy_var_3
	) `HappyStk` happyRest}}}}

happyReduce_499 = happySpecReduce_3  166# happyReduction_499
happyReduction_499 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut163 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 DotDot) -> 
	case happyOut147 happy_x_3 of { happy_var_3 -> 
	happyIn177
		 (\l -> EnumFromTo     (l <** [happy_var_2]) happy_var_1 happy_var_3
	)}}}

happyReduce_500 = happyReduce 5# 166# happyReduction_500
happyReduction_500 (happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOut163 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut147 happy_x_3 of { happy_var_3 -> 
	case happyOutTok happy_x_4 of { (Loc happy_var_4 DotDot) -> 
	case happyOut147 happy_x_5 of { happy_var_5 -> 
	happyIn177
		 (\l -> EnumFromThenTo (l <** [happy_var_2,happy_var_4]) happy_var_1 happy_var_3 happy_var_5
	) `HappyStk` happyRest}}}}}

happyReduce_501 = happySpecReduce_3  166# happyReduction_501
happyReduction_501 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut163 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Bar) -> 
	case happyOut179 happy_x_3 of { happy_var_3 -> 
	happyIn177
		 (\l -> let (stss, ss) = happy_var_3 in ParComp (l <** (happy_var_2:ss)) happy_var_1 (reverse stss)
	)}}}

happyReduce_502 = happySpecReduce_3  167# happyReduction_502
happyReduction_502 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut178 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut163 happy_x_3 of { happy_var_3 -> 
	happyIn178
		 (let (es, ss) = happy_var_1 in (happy_var_3 : es, happy_var_2 : ss)
	)}}}

happyReduce_503 = happySpecReduce_3  167# happyReduction_503
happyReduction_503 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut163 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut163 happy_x_3 of { happy_var_3 -> 
	happyIn178
		 (([happy_var_3,happy_var_1], [happy_var_2])
	)}}}

happyReduce_504 = happySpecReduce_3  168# happyReduction_504
happyReduction_504 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut179 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Bar) -> 
	case happyOut180 happy_x_3 of { happy_var_3 -> 
	happyIn179
		 (let { (stss, ss1) = happy_var_1;
                                              (sts, ss2) = happy_var_3 }
                                         in (reverse sts : stss, ss1 ++ [happy_var_2] ++ reverse ss2)
	)}}}

happyReduce_505 = happySpecReduce_1  168# happyReduction_505
happyReduction_505 happy_x_1
	 =  case happyOut180 happy_x_1 of { happy_var_1 -> 
	happyIn179
		 (let (sts, ss) = happy_var_1 in ([reverse sts], reverse ss)
	)}

happyReduce_506 = happySpecReduce_3  169# happyReduction_506
happyReduction_506 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut180 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut181 happy_x_3 of { happy_var_3 -> 
	happyIn180
		 (let (sts, ss) = happy_var_1 in (happy_var_3 : sts, happy_var_2 : ss)
	)}}}

happyReduce_507 = happySpecReduce_1  169# happyReduction_507
happyReduction_507 happy_x_1
	 =  case happyOut181 happy_x_1 of { happy_var_1 -> 
	happyIn180
		 (([happy_var_1],[])
	)}

happyReduce_508 = happySpecReduce_1  170# happyReduction_508
happyReduction_508 happy_x_1
	 =  case happyOut182 happy_x_1 of { happy_var_1 -> 
	happyIn181
		 (happy_var_1
	)}

happyReduce_509 = happySpecReduce_1  170# happyReduction_509
happyReduction_509 happy_x_1
	 =  case happyOut184 happy_x_1 of { happy_var_1 -> 
	happyIn181
		 (QualStmt (ann happy_var_1) happy_var_1
	)}

happyReduce_510 = happySpecReduce_2  171# happyReduction_510
happyReduction_510 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Then) -> 
	case happyOut146 happy_x_2 of { happy_var_2 -> 
	happyIn182
		 (ThenTrans    (nIS happy_var_1 <++> ann happy_var_2 <** [happy_var_1]) happy_var_2
	)}}

happyReduce_511 = happyReduce 4# 171# happyReduction_511
happyReduction_511 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Then) -> 
	case happyOut146 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 KW_By) -> 
	case happyOut146 happy_x_4 of { happy_var_4 -> 
	happyIn182
		 (ThenBy       (nIS happy_var_1 <++> ann happy_var_4 <** [happy_var_1,happy_var_3]) happy_var_2 happy_var_4
	) `HappyStk` happyRest}}}}

happyReduce_512 = happyReduce 4# 171# happyReduction_512
happyReduction_512 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Then) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 KW_Group) -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 KW_By) -> 
	case happyOut146 happy_x_4 of { happy_var_4 -> 
	happyIn182
		 (GroupBy      (nIS happy_var_1 <++> ann happy_var_4 <** [happy_var_1,happy_var_2,happy_var_3]) happy_var_4
	) `HappyStk` happyRest}}}}

happyReduce_513 = happyReduce 4# 171# happyReduction_513
happyReduction_513 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Then) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 KW_Group) -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 KW_Using) -> 
	case happyOut146 happy_x_4 of { happy_var_4 -> 
	happyIn182
		 (GroupUsing   (nIS happy_var_1 <++> ann happy_var_4 <** [happy_var_1,happy_var_2,happy_var_3]) happy_var_4
	) `HappyStk` happyRest}}}}

happyReduce_514 = happyReduce 6# 171# happyReduction_514
happyReduction_514 (happy_x_6 `HappyStk`
	happy_x_5 `HappyStk`
	happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Then) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 KW_Group) -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 KW_By) -> 
	case happyOut146 happy_x_4 of { happy_var_4 -> 
	case happyOutTok happy_x_5 of { (Loc happy_var_5 KW_Using) -> 
	case happyOut146 happy_x_6 of { happy_var_6 -> 
	happyIn182
		 (GroupByUsing (nIS happy_var_1 <++> ann happy_var_6 <** [happy_var_1,happy_var_2,happy_var_3,happy_var_5]) happy_var_4 happy_var_6
	) `HappyStk` happyRest}}}}}}

happyReduce_515 = happySpecReduce_3  172# happyReduction_515
happyReduction_515 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut183 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut184 happy_x_3 of { happy_var_3 -> 
	happyIn183
		 (let (sts, ss) = happy_var_1 in (happy_var_3 : sts, happy_var_2 : ss)
	)}}}

happyReduce_516 = happySpecReduce_1  172# happyReduction_516
happyReduction_516 happy_x_1
	 =  case happyOut184 happy_x_1 of { happy_var_1 -> 
	happyIn183
		 (([happy_var_1],[])
	)}

happyReduce_517 = happySpecReduce_3  173# happyReduction_517
happyReduction_517 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut192 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftArrow) -> 
	case happyOut146 happy_x_3 of { happy_var_3 -> 
	happyIn184
		 (Generator (happy_var_1 <> happy_var_3 <** [happy_var_2]) happy_var_1 happy_var_3
	)}}}

happyReduce_518 = happySpecReduce_1  173# happyReduction_518
happyReduction_518 happy_x_1
	 =  case happyOut146 happy_x_1 of { happy_var_1 -> 
	happyIn184
		 (Qualifier (ann happy_var_1) happy_var_1
	)}

happyReduce_519 = happySpecReduce_2  173# happyReduction_519
happyReduction_519 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Let) -> 
	case happyOut61 happy_x_2 of { happy_var_2 -> 
	happyIn184
		 (LetStmt   (nIS happy_var_1 <++> ann happy_var_2 <** [happy_var_1]) happy_var_2
	)}}

happyReduce_520 = happySpecReduce_3  174# happyReduction_520
happyReduction_520 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftCurly) -> 
	case happyOut186 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightCurly) -> 
	happyIn185
		 ((fst happy_var_2, happy_var_1 <^^> happy_var_3, happy_var_1:snd happy_var_2 ++ [happy_var_3])
	)}}}

happyReduce_521 = happySpecReduce_3  174# happyReduction_521
happyReduction_521 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut236 happy_x_1 of { happy_var_1 -> 
	case happyOut186 happy_x_2 of { happy_var_2 -> 
	case happyOut237 happy_x_3 of { happy_var_3 -> 
	happyIn185
		 ((fst happy_var_2, happy_var_1 <^^> happy_var_3, happy_var_1:snd happy_var_2 ++ [happy_var_3])
	)}}}

happyReduce_522 = happySpecReduce_3  175# happyReduction_522
happyReduction_522 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut25 happy_x_1 of { happy_var_1 -> 
	case happyOut187 happy_x_2 of { happy_var_2 -> 
	case happyOut25 happy_x_3 of { happy_var_3 -> 
	happyIn186
		 ((reverse $ fst happy_var_2, happy_var_1 ++ snd happy_var_2 ++ happy_var_3)
	)}}}

happyReduce_523 = happySpecReduce_3  176# happyReduction_523
happyReduction_523 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut187 happy_x_1 of { happy_var_1 -> 
	case happyOut24 happy_x_2 of { happy_var_2 -> 
	case happyOut188 happy_x_3 of { happy_var_3 -> 
	happyIn187
		 ((happy_var_3 : fst happy_var_1, snd happy_var_1 ++ happy_var_2)
	)}}}

happyReduce_524 = happySpecReduce_1  176# happyReduction_524
happyReduction_524 happy_x_1
	 =  case happyOut188 happy_x_1 of { happy_var_1 -> 
	happyIn187
		 (([happy_var_1],[])
	)}

happyReduce_525 = happySpecReduce_3  177# happyReduction_525
happyReduction_525 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut192 happy_x_1 of { happy_var_1 -> 
	case happyOut189 happy_x_2 of { happy_var_2 -> 
	case happyOut141 happy_x_3 of { happy_var_3 -> 
	happyIn188
		 (Alt (happy_var_1 <> happy_var_2 <+?> (fmap ann) (fst happy_var_3) <** snd happy_var_3) happy_var_1 happy_var_2 (fst happy_var_3)
	)}}}

happyReduce_526 = happySpecReduce_2  178# happyReduction_526
happyReduction_526 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 RightArrow) -> 
	case happyOut146 happy_x_2 of { happy_var_2 -> 
	happyIn189
		 (UnGuardedAlt (nIS happy_var_1 <++> ann happy_var_2 <** [happy_var_1]) happy_var_2
	)}}

happyReduce_527 = happySpecReduce_1  178# happyReduction_527
happyReduction_527 happy_x_1
	 =  case happyOut190 happy_x_1 of { happy_var_1 -> 
	happyIn189
		 (GuardedAlts  (snd happy_var_1) (reverse $ fst happy_var_1)
	)}

happyReduce_528 = happySpecReduce_2  179# happyReduction_528
happyReduction_528 happy_x_2
	happy_x_1
	 =  case happyOut190 happy_x_1 of { happy_var_1 -> 
	case happyOut191 happy_x_2 of { happy_var_2 -> 
	happyIn190
		 ((happy_var_2 : fst happy_var_1, snd happy_var_1 <++> ann happy_var_2)
	)}}

happyReduce_529 = happySpecReduce_1  179# happyReduction_529
happyReduction_529 happy_x_1
	 =  case happyOut191 happy_x_1 of { happy_var_1 -> 
	happyIn190
		 (([happy_var_1], ann happy_var_1)
	)}

happyReduce_530 = happyMonadReduce 4# 180# happyReduction_530
happyReduction_530 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 Bar) -> 
	case happyOut183 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightArrow) -> 
	case happyOut146 happy_x_4 of { happy_var_4 -> 
	( do { checkPatternGuards (fst happy_var_2);
                                       let {l = nIS happy_var_1 <++> ann happy_var_4 <** (happy_var_1:snd happy_var_2 ++ [happy_var_3])};
                                       return (GuardedAlt l (reverse (fst happy_var_2)) happy_var_4) })}}}}
	) (\r -> happyReturn (happyIn191 r))

happyReduce_531 = happyMonadReduce 1# 181# happyReduction_531
happyReduction_531 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut147 happy_x_1 of { happy_var_1 -> 
	( checkPattern happy_var_1)}
	) (\r -> happyReturn (happyIn192 r))

happyReduce_532 = happyMonadReduce 2# 181# happyReduction_532
happyReduction_532 (happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 Exclamation) -> 
	case happyOut159 happy_x_2 of { happy_var_2 -> 
	( checkPattern (BangPat (nIS happy_var_1 <++> ann happy_var_2 <** [happy_var_1]) happy_var_2))}}
	) (\r -> happyReturn (happyIn192 r))

happyReduce_533 = happySpecReduce_3  182# happyReduction_533
happyReduction_533 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftCurly) -> 
	case happyOut194 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightCurly) -> 
	happyIn193
		 ((fst happy_var_2, happy_var_1 <^^> happy_var_3, happy_var_1:snd happy_var_2 ++ [happy_var_3])
	)}}}

happyReduce_534 = happySpecReduce_3  182# happyReduction_534
happyReduction_534 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut236 happy_x_1 of { happy_var_1 -> 
	case happyOut194 happy_x_2 of { happy_var_2 -> 
	case happyOut237 happy_x_3 of { happy_var_3 -> 
	happyIn193
		 ((fst happy_var_2, happy_var_1 <^^> happy_var_3, happy_var_1:snd happy_var_2 ++ [happy_var_3])
	)}}}

happyReduce_535 = happySpecReduce_3  183# happyReduction_535
happyReduction_535 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut25 happy_x_1 of { happy_var_1 -> 
	case happyOut195 happy_x_2 of { happy_var_2 -> 
	case happyOut25 happy_x_3 of { happy_var_3 -> 
	happyIn194
		 ((reverse $ fst happy_var_2, happy_var_1 ++ snd happy_var_2 ++ happy_var_3)
	)}}}

happyReduce_536 = happySpecReduce_3  184# happyReduction_536
happyReduction_536 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut195 happy_x_1 of { happy_var_1 -> 
	case happyOut25 happy_x_2 of { happy_var_2 -> 
	case happyOut196 happy_x_3 of { happy_var_3 -> 
	happyIn195
		 ((happy_var_3 : fst happy_var_1, snd happy_var_1 ++ happy_var_2)
	)}}}

happyReduce_537 = happySpecReduce_1  184# happyReduction_537
happyReduction_537 happy_x_1
	 =  case happyOut196 happy_x_1 of { happy_var_1 -> 
	happyIn195
		 (([happy_var_1], [])
	)}

happyReduce_538 = happyReduce 4# 185# happyReduction_538
happyReduction_538 (happy_x_4 `HappyStk`
	happy_x_3 `HappyStk`
	happy_x_2 `HappyStk`
	happy_x_1 `HappyStk`
	happyRest)
	 = case happyOutTok happy_x_1 of { (Loc happy_var_1 Bar) -> 
	case happyOut146 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightArrow) -> 
	case happyOut146 happy_x_4 of { happy_var_4 -> 
	happyIn196
		 (let l = nIS happy_var_1 <++> ann happy_var_2 <++> ann happy_var_4 <** [happy_var_1,happy_var_3] in (IfAlt l happy_var_2 happy_var_4)
	) `HappyStk` happyRest}}}}

happyReduce_539 = happySpecReduce_3  186# happyReduction_539
happyReduction_539 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftCurly) -> 
	case happyOut198 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightCurly) -> 
	happyIn197
		 ((fst happy_var_2, happy_var_1 <^^> happy_var_3, happy_var_1:snd happy_var_2 ++ [happy_var_3])
	)}}}

happyReduce_540 = happySpecReduce_3  186# happyReduction_540
happyReduction_540 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut236 happy_x_1 of { happy_var_1 -> 
	case happyOut198 happy_x_2 of { happy_var_2 -> 
	case happyOut237 happy_x_3 of { happy_var_3 -> 
	happyIn197
		 ((fst happy_var_2, happy_var_1 <^^> happy_var_3, happy_var_1:snd happy_var_2 ++ [happy_var_3])
	)}}}

happyReduce_541 = happySpecReduce_2  187# happyReduction_541
happyReduction_541 happy_x_2
	happy_x_1
	 =  case happyOut200 happy_x_1 of { happy_var_1 -> 
	case happyOut199 happy_x_2 of { happy_var_2 -> 
	happyIn198
		 ((happy_var_1 : fst happy_var_2, snd happy_var_2)
	)}}

happyReduce_542 = happySpecReduce_2  187# happyReduction_542
happyReduction_542 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 SemiColon) -> 
	case happyOut198 happy_x_2 of { happy_var_2 -> 
	happyIn198
		 ((fst happy_var_2, happy_var_1 : snd happy_var_2)
	)}}

happyReduce_543 = happySpecReduce_0  187# happyReduction_543
happyReduction_543  =  happyIn198
		 (([],[])
	)

happyReduce_544 = happySpecReduce_2  188# happyReduction_544
happyReduction_544 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 SemiColon) -> 
	case happyOut198 happy_x_2 of { happy_var_2 -> 
	happyIn199
		 ((fst happy_var_2, happy_var_1 : snd happy_var_2)
	)}}

happyReduce_545 = happySpecReduce_0  188# happyReduction_545
happyReduction_545  =  happyIn199
		 (([],[])
	)

happyReduce_546 = happySpecReduce_2  189# happyReduction_546
happyReduction_546 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Let) -> 
	case happyOut61 happy_x_2 of { happy_var_2 -> 
	happyIn200
		 (LetStmt (nIS happy_var_1 <++> ann happy_var_2 <** [happy_var_1]) happy_var_2
	)}}

happyReduce_547 = happySpecReduce_3  189# happyReduction_547
happyReduction_547 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut192 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 LeftArrow) -> 
	case happyOut146 happy_x_3 of { happy_var_3 -> 
	happyIn200
		 (Generator (happy_var_1 <> happy_var_3 <** [happy_var_2]) happy_var_1 happy_var_3
	)}}}

happyReduce_548 = happySpecReduce_1  189# happyReduction_548
happyReduction_548 happy_x_1
	 =  case happyOut146 happy_x_1 of { happy_var_1 -> 
	happyIn200
		 (Qualifier (ann happy_var_1) happy_var_1
	)}

happyReduce_549 = happySpecReduce_2  189# happyReduction_549
happyReduction_549 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Rec) -> 
	case happyOut197 happy_x_2 of { happy_var_2 -> 
	happyIn200
		 (let (stms,inf,ss) = happy_var_2 in RecStmt (nIS happy_var_1 <++> inf <** happy_var_1:ss) stms
	)}}

happyReduce_550 = happySpecReduce_3  190# happyReduction_550
happyReduction_550 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut201 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Comma) -> 
	case happyOut202 happy_x_3 of { happy_var_3 -> 
	happyIn201
		 (let (fbs, ss) = happy_var_1 in (happy_var_3 : fbs, happy_var_2 : ss)
	)}}}

happyReduce_551 = happySpecReduce_1  190# happyReduction_551
happyReduction_551 happy_x_1
	 =  case happyOut202 happy_x_1 of { happy_var_1 -> 
	happyIn201
		 (([happy_var_1],[])
	)}

happyReduce_552 = happySpecReduce_3  191# happyReduction_552
happyReduction_552 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut209 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Equals) -> 
	case happyOut147 happy_x_3 of { happy_var_3 -> 
	happyIn202
		 (FieldUpdate (happy_var_1 <>happy_var_3 <** [happy_var_2]) happy_var_1 happy_var_3
	)}}}

happyReduce_553 = happyMonadReduce 1# 191# happyReduction_553
happyReduction_553 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOut209 happy_x_1 of { happy_var_1 -> 
	( checkEnabled NamedFieldPuns >> checkUnQual happy_var_1 >>= return . FieldPun (ann happy_var_1))}
	) (\r -> happyReturn (happyIn202 r))

happyReduce_554 = happyMonadReduce 1# 191# happyReduction_554
happyReduction_554 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (case happyOutTok happy_x_1 of { (Loc happy_var_1 DotDot) -> 
	( checkEnabled RecordWildCards >> return (FieldWildcard (nIS happy_var_1)))}
	) (\r -> happyReturn (happyIn202 r))

happyReduce_555 = happySpecReduce_3  192# happyReduction_555
happyReduction_555 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut25 happy_x_1 of { happy_var_1 -> 
	case happyOut204 happy_x_2 of { happy_var_2 -> 
	case happyOut25 happy_x_3 of { happy_var_3 -> 
	happyIn203
		 ((reverse (fst happy_var_2), reverse happy_var_1 ++ snd happy_var_2 ++ reverse happy_var_3)
	)}}}

happyReduce_556 = happySpecReduce_3  193# happyReduction_556
happyReduction_556 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut204 happy_x_1 of { happy_var_1 -> 
	case happyOut24 happy_x_2 of { happy_var_2 -> 
	case happyOut205 happy_x_3 of { happy_var_3 -> 
	happyIn204
		 ((happy_var_3 : fst happy_var_1, snd happy_var_1 ++ reverse happy_var_2)
	)}}}

happyReduce_557 = happySpecReduce_1  193# happyReduction_557
happyReduction_557 happy_x_1
	 =  case happyOut205 happy_x_1 of { happy_var_1 -> 
	happyIn204
		 (([happy_var_1],[])
	)}

happyReduce_558 = happySpecReduce_3  194# happyReduction_558
happyReduction_558 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOut210 happy_x_1 of { happy_var_1 -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 Equals) -> 
	case happyOut146 happy_x_3 of { happy_var_3 -> 
	happyIn205
		 (IPBind (happy_var_1 <> happy_var_3 <** [happy_var_2]) happy_var_1 happy_var_3
	)}}}

happyReduce_559 = happySpecReduce_2  195# happyReduction_559
happyReduction_559 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 RightParen) -> 
	happyIn206
		 (p_unit_con              (happy_var_1 <^^> happy_var_2 <** [happy_var_1,happy_var_2])
	)}}

happyReduce_560 = happySpecReduce_2  195# happyReduction_560
happyReduction_560 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftSquare) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 RightSquare) -> 
	happyIn206
		 (List                    (happy_var_1 <^^> happy_var_2 <** [happy_var_1,happy_var_2]) []
	)}}

happyReduce_561 = happySpecReduce_3  195# happyReduction_561
happyReduction_561 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut162 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn206
		 (p_tuple_con             (happy_var_1 <^^> happy_var_3 <** happy_var_1:reverse (happy_var_3:happy_var_2)) Boxed (length happy_var_2)
	)}}}

happyReduce_562 = happySpecReduce_2  195# happyReduction_562
happyReduction_562 happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftHashParen) -> 
	case happyOutTok happy_x_2 of { (Loc happy_var_2 RightHashParen) -> 
	happyIn206
		 (p_unboxed_singleton_con (happy_var_1 <^^> happy_var_2 <** [happy_var_1,happy_var_2])
	)}}

happyReduce_563 = happySpecReduce_3  195# happyReduction_563
happyReduction_563 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftHashParen) -> 
	case happyOut162 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightHashParen) -> 
	happyIn206
		 (p_tuple_con             (happy_var_1 <^^> happy_var_3 <** happy_var_1:reverse (happy_var_3:happy_var_2)) Unboxed (length happy_var_2)
	)}}}

happyReduce_564 = happySpecReduce_1  195# happyReduction_564
happyReduction_564 happy_x_1
	 =  case happyOut212 happy_x_1 of { happy_var_1 -> 
	happyIn206
		 (Con (ann happy_var_1) happy_var_1
	)}

happyReduce_565 = happySpecReduce_1  196# happyReduction_565
happyReduction_565 happy_x_1
	 =  case happyOut224 happy_x_1 of { happy_var_1 -> 
	happyIn207
		 (happy_var_1
	)}

happyReduce_566 = happySpecReduce_3  196# happyReduction_566
happyReduction_566 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut232 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn207
		 (fmap (const (happy_var_1 <^^> happy_var_3 <** [happy_var_1, srcInfoSpan (ann happy_var_2), happy_var_3])) happy_var_2
	)}}}

happyReduce_567 = happySpecReduce_1  197# happyReduction_567
happyReduction_567 happy_x_1
	 =  case happyOut223 happy_x_1 of { happy_var_1 -> 
	happyIn208
		 (happy_var_1
	)}

happyReduce_568 = happySpecReduce_3  197# happyReduction_568
happyReduction_568 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut232 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn208
		 (fmap (const (happy_var_1 <^^> happy_var_3 <** [happy_var_1, srcInfoSpan (ann happy_var_2), happy_var_3])) happy_var_2
	)}}}

happyReduce_569 = happySpecReduce_1  198# happyReduction_569
happyReduction_569 happy_x_1
	 =  case happyOut222 happy_x_1 of { happy_var_1 -> 
	happyIn209
		 (happy_var_1
	)}

happyReduce_570 = happySpecReduce_3  198# happyReduction_570
happyReduction_570 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut230 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn209
		 (fmap (const (happy_var_1 <^^> happy_var_3 <** [happy_var_1, srcInfoSpan (ann happy_var_2), happy_var_3])) happy_var_2
	)}}}

happyReduce_571 = happySpecReduce_1  199# happyReduction_571
happyReduction_571 happy_x_1
	 =  case happyOut225 happy_x_1 of { happy_var_1 -> 
	happyIn210
		 (happy_var_1
	)}

happyReduce_572 = happySpecReduce_1  200# happyReduction_572
happyReduction_572 happy_x_1
	 =  case happyOut227 happy_x_1 of { happy_var_1 -> 
	happyIn211
		 (happy_var_1
	)}

happyReduce_573 = happySpecReduce_3  200# happyReduction_573
happyReduction_573 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut229 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn211
		 (fmap (const (happy_var_1 <^^> happy_var_3 <** [happy_var_1, srcInfoSpan (ann happy_var_2), happy_var_3])) happy_var_2
	)}}}

happyReduce_574 = happySpecReduce_1  201# happyReduction_574
happyReduction_574 happy_x_1
	 =  case happyOut226 happy_x_1 of { happy_var_1 -> 
	happyIn212
		 (happy_var_1
	)}

happyReduce_575 = happySpecReduce_3  201# happyReduction_575
happyReduction_575 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 LeftParen) -> 
	case happyOut221 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 RightParen) -> 
	happyIn212
		 (fmap (const (happy_var_1 <^^> happy_var_3 <** [happy_var_1, srcInfoSpan (ann happy_var_2), happy_var_3])) happy_var_2
	)}}}

happyReduce_576 = happySpecReduce_1  202# happyReduction_576
happyReduction_576 happy_x_1
	 =  case happyOut232 happy_x_1 of { happy_var_1 -> 
	happyIn213
		 (happy_var_1
	)}

happyReduce_577 = happySpecReduce_3  202# happyReduction_577
happyReduction_577 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 BackQuote) -> 
	case happyOut224 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 BackQuote) -> 
	happyIn213
		 (fmap (const (happy_var_1 <^^> happy_var_3 <** [happy_var_1, srcInfoSpan (ann happy_var_2), happy_var_3])) happy_var_2
	)}}}

happyReduce_578 = happySpecReduce_1  203# happyReduction_578
happyReduction_578 happy_x_1
	 =  case happyOut230 happy_x_1 of { happy_var_1 -> 
	happyIn214
		 (happy_var_1
	)}

happyReduce_579 = happySpecReduce_3  203# happyReduction_579
happyReduction_579 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 BackQuote) -> 
	case happyOut222 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 BackQuote) -> 
	happyIn214
		 (fmap (const (happy_var_1 <^^> happy_var_3 <** [happy_var_1, srcInfoSpan (ann happy_var_2), happy_var_3])) happy_var_2
	)}}}

happyReduce_580 = happySpecReduce_1  204# happyReduction_580
happyReduction_580 happy_x_1
	 =  case happyOut231 happy_x_1 of { happy_var_1 -> 
	happyIn215
		 (happy_var_1
	)}

happyReduce_581 = happySpecReduce_3  204# happyReduction_581
happyReduction_581 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 BackQuote) -> 
	case happyOut222 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 BackQuote) -> 
	happyIn215
		 (fmap (const (happy_var_1 <^^> happy_var_3 <** [happy_var_1, srcInfoSpan (ann happy_var_2), happy_var_3])) happy_var_2
	)}}}

happyReduce_582 = happySpecReduce_1  205# happyReduction_582
happyReduction_582 happy_x_1
	 =  case happyOut229 happy_x_1 of { happy_var_1 -> 
	happyIn216
		 (happy_var_1
	)}

happyReduce_583 = happySpecReduce_3  205# happyReduction_583
happyReduction_583 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 BackQuote) -> 
	case happyOut227 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 BackQuote) -> 
	happyIn216
		 (fmap (const (happy_var_1 <^^> happy_var_3 <** [happy_var_1, srcInfoSpan (ann happy_var_2), happy_var_3])) happy_var_2
	)}}}

happyReduce_584 = happySpecReduce_1  206# happyReduction_584
happyReduction_584 happy_x_1
	 =  case happyOut221 happy_x_1 of { happy_var_1 -> 
	happyIn217
		 (happy_var_1
	)}

happyReduce_585 = happySpecReduce_3  206# happyReduction_585
happyReduction_585 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 BackQuote) -> 
	case happyOut226 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 BackQuote) -> 
	happyIn217
		 (fmap (const (happy_var_1 <^^> happy_var_3 <** [happy_var_1, srcInfoSpan (ann happy_var_2), happy_var_3])) happy_var_2
	)}}}

happyReduce_586 = happySpecReduce_1  207# happyReduction_586
happyReduction_586 happy_x_1
	 =  case happyOut213 happy_x_1 of { happy_var_1 -> 
	happyIn218
		 (VarOp (ann happy_var_1) happy_var_1
	)}

happyReduce_587 = happySpecReduce_1  207# happyReduction_587
happyReduction_587 happy_x_1
	 =  case happyOut216 happy_x_1 of { happy_var_1 -> 
	happyIn218
		 (ConOp (ann happy_var_1) happy_var_1
	)}

happyReduce_588 = happySpecReduce_1  208# happyReduction_588
happyReduction_588 happy_x_1
	 =  case happyOut214 happy_x_1 of { happy_var_1 -> 
	happyIn219
		 (QVarOp (ann happy_var_1) happy_var_1
	)}

happyReduce_589 = happySpecReduce_1  208# happyReduction_589
happyReduction_589 happy_x_1
	 =  case happyOut217 happy_x_1 of { happy_var_1 -> 
	happyIn219
		 (QConOp (ann happy_var_1) happy_var_1
	)}

happyReduce_590 = happySpecReduce_1  209# happyReduction_590
happyReduction_590 happy_x_1
	 =  case happyOut215 happy_x_1 of { happy_var_1 -> 
	happyIn220
		 (QVarOp (ann happy_var_1) happy_var_1
	)}

happyReduce_591 = happySpecReduce_1  209# happyReduction_591
happyReduction_591 happy_x_1
	 =  case happyOut217 happy_x_1 of { happy_var_1 -> 
	happyIn220
		 (QConOp (ann happy_var_1) happy_var_1
	)}

happyReduce_592 = happySpecReduce_1  210# happyReduction_592
happyReduction_592 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Colon) -> 
	happyIn221
		 (list_cons_name (nIS happy_var_1)
	)}

happyReduce_593 = happySpecReduce_1  210# happyReduction_593
happyReduction_593 happy_x_1
	 =  case happyOut228 happy_x_1 of { happy_var_1 -> 
	happyIn221
		 (happy_var_1
	)}

happyReduce_594 = happySpecReduce_1  211# happyReduction_594
happyReduction_594 happy_x_1
	 =  case happyOut224 happy_x_1 of { happy_var_1 -> 
	happyIn222
		 (UnQual (ann happy_var_1) happy_var_1
	)}

happyReduce_595 = happySpecReduce_1  211# happyReduction_595
happyReduction_595 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn222
		 (let {Loc l (QVarId q) = happy_var_1; nis = nIS l}
                                 in Qual nis (ModuleName nis (fst q)) (Ident nis (snd q))
	)}

happyReduce_596 = happySpecReduce_1  212# happyReduction_596
happyReduction_596 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn223
		 (let Loc l (VarId v) = happy_var_1 in Ident (nIS l) v
	)}

happyReduce_597 = happySpecReduce_1  212# happyReduction_597
happyReduction_597 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_As) -> 
	happyIn223
		 (as_name        (nIS happy_var_1)
	)}

happyReduce_598 = happySpecReduce_1  212# happyReduction_598
happyReduction_598 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Qualified) -> 
	happyIn223
		 (qualified_name (nIS happy_var_1)
	)}

happyReduce_599 = happySpecReduce_1  212# happyReduction_599
happyReduction_599 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Hiding) -> 
	happyIn223
		 (hiding_name    (nIS happy_var_1)
	)}

happyReduce_600 = happySpecReduce_1  212# happyReduction_600
happyReduction_600 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Export) -> 
	happyIn223
		 (export_name    (nIS happy_var_1)
	)}

happyReduce_601 = happySpecReduce_1  212# happyReduction_601
happyReduction_601 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_StdCall) -> 
	happyIn223
		 (stdcall_name   (nIS happy_var_1)
	)}

happyReduce_602 = happySpecReduce_1  212# happyReduction_602
happyReduction_602 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_CCall) -> 
	happyIn223
		 (ccall_name     (nIS happy_var_1)
	)}

happyReduce_603 = happySpecReduce_1  212# happyReduction_603
happyReduction_603 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_CPlusPlus) -> 
	happyIn223
		 (cplusplus_name (nIS happy_var_1)
	)}

happyReduce_604 = happySpecReduce_1  212# happyReduction_604
happyReduction_604 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_DotNet) -> 
	happyIn223
		 (dotnet_name    (nIS happy_var_1)
	)}

happyReduce_605 = happySpecReduce_1  212# happyReduction_605
happyReduction_605 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Jvm) -> 
	happyIn223
		 (jvm_name       (nIS happy_var_1)
	)}

happyReduce_606 = happySpecReduce_1  212# happyReduction_606
happyReduction_606 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Js) -> 
	happyIn223
		 (js_name        (nIS happy_var_1)
	)}

happyReduce_607 = happySpecReduce_1  213# happyReduction_607
happyReduction_607 happy_x_1
	 =  case happyOut223 happy_x_1 of { happy_var_1 -> 
	happyIn224
		 (happy_var_1
	)}

happyReduce_608 = happySpecReduce_1  213# happyReduction_608
happyReduction_608 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Safe) -> 
	happyIn224
		 (safe_name       (nIS happy_var_1)
	)}

happyReduce_609 = happySpecReduce_1  213# happyReduction_609
happyReduction_609 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Unsafe) -> 
	happyIn224
		 (unsafe_name     (nIS happy_var_1)
	)}

happyReduce_610 = happySpecReduce_1  213# happyReduction_610
happyReduction_610 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Threadsafe) -> 
	happyIn224
		 (threadsafe_name (nIS happy_var_1)
	)}

happyReduce_611 = happySpecReduce_1  213# happyReduction_611
happyReduction_611 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Forall) -> 
	happyIn224
		 (forall_name	  (nIS happy_var_1)
	)}

happyReduce_612 = happySpecReduce_1  213# happyReduction_612
happyReduction_612 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Family) -> 
	happyIn224
		 (family_name     (nIS happy_var_1)
	)}

happyReduce_613 = happySpecReduce_1  214# happyReduction_613
happyReduction_613 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn225
		 (let Loc l (IDupVarId i) = happy_var_1 in IPDup (nIS l) i
	)}

happyReduce_614 = happySpecReduce_1  214# happyReduction_614
happyReduction_614 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn225
		 (let Loc l (ILinVarId i) = happy_var_1 in IPLin (nIS l) i
	)}

happyReduce_615 = happySpecReduce_1  215# happyReduction_615
happyReduction_615 happy_x_1
	 =  case happyOut227 happy_x_1 of { happy_var_1 -> 
	happyIn226
		 (UnQual (ann happy_var_1) happy_var_1
	)}

happyReduce_616 = happySpecReduce_1  215# happyReduction_616
happyReduction_616 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn226
		 (let {Loc l (QConId q) = happy_var_1; nis = nIS l} in Qual nis (ModuleName nis (fst q)) (Ident nis (snd q))
	)}

happyReduce_617 = happySpecReduce_1  216# happyReduction_617
happyReduction_617 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn227
		 (let Loc l (ConId c) = happy_var_1 in Ident (nIS l) c
	)}

happyReduce_618 = happySpecReduce_1  217# happyReduction_618
happyReduction_618 happy_x_1
	 =  case happyOut229 happy_x_1 of { happy_var_1 -> 
	happyIn228
		 (UnQual (ann happy_var_1) happy_var_1
	)}

happyReduce_619 = happySpecReduce_1  217# happyReduction_619
happyReduction_619 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn228
		 (let {Loc l (QConSym q) = happy_var_1; nis = nIS l} in Qual nis (ModuleName nis (fst q)) (Symbol nis (snd q))
	)}

happyReduce_620 = happySpecReduce_1  218# happyReduction_620
happyReduction_620 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn229
		 (let Loc l (ConSym c) = happy_var_1 in Symbol (nIS l) c
	)}

happyReduce_621 = happySpecReduce_1  219# happyReduction_621
happyReduction_621 happy_x_1
	 =  case happyOut232 happy_x_1 of { happy_var_1 -> 
	happyIn230
		 (UnQual (ann happy_var_1) happy_var_1
	)}

happyReduce_622 = happySpecReduce_1  219# happyReduction_622
happyReduction_622 happy_x_1
	 =  case happyOut234 happy_x_1 of { happy_var_1 -> 
	happyIn230
		 (happy_var_1
	)}

happyReduce_623 = happySpecReduce_1  220# happyReduction_623
happyReduction_623 happy_x_1
	 =  case happyOut233 happy_x_1 of { happy_var_1 -> 
	happyIn231
		 (UnQual (ann happy_var_1) happy_var_1
	)}

happyReduce_624 = happySpecReduce_1  220# happyReduction_624
happyReduction_624 happy_x_1
	 =  case happyOut234 happy_x_1 of { happy_var_1 -> 
	happyIn231
		 (happy_var_1
	)}

happyReduce_625 = happySpecReduce_1  221# happyReduction_625
happyReduction_625 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn232
		 (let Loc l (VarSym v) = happy_var_1 in Symbol (nIS l) v
	)}

happyReduce_626 = happySpecReduce_1  221# happyReduction_626
happyReduction_626 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Minus) -> 
	happyIn232
		 (minus_name (nIS happy_var_1)
	)}

happyReduce_627 = happySpecReduce_1  221# happyReduction_627
happyReduction_627 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Exclamation) -> 
	happyIn232
		 (bang_name  (nIS happy_var_1)
	)}

happyReduce_628 = happySpecReduce_1  221# happyReduction_628
happyReduction_628 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Dot) -> 
	happyIn232
		 (dot_name   (nIS happy_var_1)
	)}

happyReduce_629 = happySpecReduce_1  221# happyReduction_629
happyReduction_629 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Star) -> 
	happyIn232
		 (star_name  (nIS happy_var_1)
	)}

happyReduce_630 = happySpecReduce_1  222# happyReduction_630
happyReduction_630 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn233
		 (let Loc l (VarSym v) = happy_var_1 in Symbol (nIS l) v
	)}

happyReduce_631 = happySpecReduce_1  222# happyReduction_631
happyReduction_631 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Exclamation) -> 
	happyIn233
		 (bang_name (nIS happy_var_1)
	)}

happyReduce_632 = happySpecReduce_1  222# happyReduction_632
happyReduction_632 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Dot) -> 
	happyIn233
		 (dot_name  (nIS happy_var_1)
	)}

happyReduce_633 = happySpecReduce_1  222# happyReduction_633
happyReduction_633 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 Star) -> 
	happyIn233
		 (star_name (nIS happy_var_1)
	)}

happyReduce_634 = happySpecReduce_1  223# happyReduction_634
happyReduction_634 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn234
		 (let {Loc l (QVarSym q) = happy_var_1; nis = nIS l} in Qual nis (ModuleName nis (fst q)) (Symbol nis (snd q))
	)}

happyReduce_635 = happySpecReduce_1  224# happyReduction_635
happyReduction_635 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn235
		 (let Loc l (IntTok        (i,raw)) = happy_var_1 in Int        (nIS l) i raw
	)}

happyReduce_636 = happySpecReduce_1  224# happyReduction_636
happyReduction_636 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn235
		 (let Loc l (Character     (c,raw)) = happy_var_1 in Char       (nIS l) c raw
	)}

happyReduce_637 = happySpecReduce_1  224# happyReduction_637
happyReduction_637 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn235
		 (let Loc l (FloatTok      (r,raw)) = happy_var_1 in Frac       (nIS l) r raw
	)}

happyReduce_638 = happySpecReduce_1  224# happyReduction_638
happyReduction_638 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn235
		 (let Loc l (StringTok     (s,raw)) = happy_var_1 in String     (nIS l) s raw
	)}

happyReduce_639 = happySpecReduce_1  224# happyReduction_639
happyReduction_639 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn235
		 (let Loc l (IntTokHash    (i,raw)) = happy_var_1 in PrimInt    (nIS l) i raw
	)}

happyReduce_640 = happySpecReduce_1  224# happyReduction_640
happyReduction_640 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn235
		 (let Loc l (WordTokHash   (w,raw)) = happy_var_1 in PrimWord   (nIS l) w raw
	)}

happyReduce_641 = happySpecReduce_1  224# happyReduction_641
happyReduction_641 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn235
		 (let Loc l (FloatTokHash  (f,raw)) = happy_var_1 in PrimFloat  (nIS l) f raw
	)}

happyReduce_642 = happySpecReduce_1  224# happyReduction_642
happyReduction_642 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn235
		 (let Loc l (DoubleTokHash (d,raw)) = happy_var_1 in PrimDouble (nIS l) d raw
	)}

happyReduce_643 = happySpecReduce_1  224# happyReduction_643
happyReduction_643 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn235
		 (let Loc l (CharacterHash (c,raw)) = happy_var_1 in PrimChar   (nIS l) c raw
	)}

happyReduce_644 = happySpecReduce_1  224# happyReduction_644
happyReduction_644 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn235
		 (let Loc l (StringHash    (s,raw)) = happy_var_1 in PrimString (nIS l) s raw
	)}

happyReduce_645 = happyMonadReduce 0# 225# happyReduction_645
happyReduction_645 (happyRest) tk
	 = happyThen (( pushCurrentContext >> getSrcLoc >>= \s -> return $ mkSrcSpan s s {- >>= \x -> trace (show x) (return x) -})
	) (\r -> happyReturn (happyIn236 r))

happyReduce_646 = happySpecReduce_1  226# happyReduction_646
happyReduction_646 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 VRightCurly) -> 
	happyIn237
		 (happy_var_1 {- >>= \x -> trace (show x ++ show x ++ show x) (return x) -}
	)}

happyReduce_647 = happyMonadReduce 1# 226# happyReduction_647
happyReduction_647 (happy_x_1 `HappyStk`
	happyRest) tk
	 = happyThen (( popContext >> getSrcLoc >>= \s -> return $ mkSrcSpan s s {- >>= \x -> trace (show x ++ show x) (return x) -})
	) (\r -> happyReturn (happyIn237 r))

happyReduce_648 = happySpecReduce_1  227# happyReduction_648
happyReduction_648 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn238
		 (let Loc l (ConId  n) = happy_var_1 in ModuleName (nIS l) n
	)}

happyReduce_649 = happySpecReduce_1  227# happyReduction_649
happyReduction_649 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn238
		 (let Loc l (QConId n) = happy_var_1 in ModuleName (nIS l) (fst n ++ '.':snd n)
	)}

happyReduce_650 = happySpecReduce_1  228# happyReduction_650
happyReduction_650 happy_x_1
	 =  case happyOut211 happy_x_1 of { happy_var_1 -> 
	happyIn239
		 (happy_var_1
	)}

happyReduce_651 = happySpecReduce_1  229# happyReduction_651
happyReduction_651 happy_x_1
	 =  case happyOut212 happy_x_1 of { happy_var_1 -> 
	happyIn240
		 (happy_var_1
	)}

happyReduce_652 = happySpecReduce_1  230# happyReduction_652
happyReduction_652 happy_x_1
	 =  case happyOut242 happy_x_1 of { happy_var_1 -> 
	happyIn241
		 (happy_var_1
	)}

happyReduce_653 = happySpecReduce_1  231# happyReduction_653
happyReduction_653 happy_x_1
	 =  case happyOut223 happy_x_1 of { happy_var_1 -> 
	happyIn242
		 (happy_var_1
	)}

happyReduce_654 = happySpecReduce_1  231# happyReduction_654
happyReduction_654 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Safe) -> 
	happyIn242
		 (safe_name       (nIS happy_var_1)
	)}

happyReduce_655 = happySpecReduce_1  231# happyReduction_655
happyReduction_655 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Unsafe) -> 
	happyIn242
		 (unsafe_name     (nIS happy_var_1)
	)}

happyReduce_656 = happySpecReduce_1  231# happyReduction_656
happyReduction_656 happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 KW_Threadsafe) -> 
	happyIn242
		 (threadsafe_name (nIS happy_var_1)
	)}

happyReduce_657 = happySpecReduce_3  232# happyReduction_657
happyReduction_657 happy_x_3
	happy_x_2
	happy_x_1
	 =  case happyOutTok happy_x_1 of { (Loc happy_var_1 BackQuote) -> 
	case happyOut241 happy_x_2 of { happy_var_2 -> 
	case happyOutTok happy_x_3 of { (Loc happy_var_3 BackQuote) -> 
	happyIn243
		 (UnQual (happy_var_1 <^^> happy_var_3 <** [happy_var_1, srcInfoSpan (ann happy_var_2), happy_var_3]) happy_var_2
	)}}}

happyReduce_658 = happySpecReduce_1  232# happyReduction_658
happyReduction_658 happy_x_1
	 =  case happyOut244 happy_x_1 of { happy_var_1 -> 
	happyIn243
		 (UnQual (ann happy_var_1) happy_var_1
	)}

happyReduce_659 = happySpecReduce_1  233# happyReduction_659
happyReduction_659 happy_x_1
	 =  case happyOutTok happy_x_1 of { happy_var_1 -> 
	happyIn244
		 (let Loc l (VarSym x) = happy_var_1 in Symbol (nIS l) x
	)}

happyNewToken action sts stk
	= lexer(\tk -> 
	let cont i = happyDoAction i tk action sts stk in
	case tk of {
	Loc _ EOF -> happyDoAction 138# tk action sts stk;
	Loc _ (VarId _) -> cont 1#;
	Loc _ (QVarId _) -> cont 2#;
	Loc _ (IDupVarId _) -> cont 3#;
	Loc _ (ILinVarId _) -> cont 4#;
	Loc _ (ConId _) -> cont 5#;
	Loc _ (QConId _) -> cont 6#;
	Loc _ (DVarId _) -> cont 7#;
	Loc _ (VarSym _) -> cont 8#;
	Loc _ (ConSym _) -> cont 9#;
	Loc _ (QVarSym _) -> cont 10#;
	Loc _ (QConSym _) -> cont 11#;
	Loc _ (IntTok _) -> cont 12#;
	Loc _ (FloatTok _) -> cont 13#;
	Loc _ (Character _) -> cont 14#;
	Loc _ (StringTok _) -> cont 15#;
	Loc _ (IntTokHash _) -> cont 16#;
	Loc _ (WordTokHash _) -> cont 17#;
	Loc _ (FloatTokHash _) -> cont 18#;
	Loc _ (DoubleTokHash _) -> cont 19#;
	Loc _ (CharacterHash _) -> cont 20#;
	Loc _ (StringHash _) -> cont 21#;
	Loc happy_dollar_dollar LeftParen -> cont 22#;
	Loc happy_dollar_dollar RightParen -> cont 23#;
	Loc happy_dollar_dollar LeftHashParen -> cont 24#;
	Loc happy_dollar_dollar RightHashParen -> cont 25#;
	Loc happy_dollar_dollar SemiColon -> cont 26#;
	Loc happy_dollar_dollar LeftCurly -> cont 27#;
	Loc happy_dollar_dollar RightCurly -> cont 28#;
	Loc happy_dollar_dollar VRightCurly -> cont 29#;
	Loc happy_dollar_dollar LeftSquare -> cont 30#;
	Loc happy_dollar_dollar RightSquare -> cont 31#;
	Loc happy_dollar_dollar Comma -> cont 32#;
	Loc happy_dollar_dollar Underscore -> cont 33#;
	Loc happy_dollar_dollar BackQuote -> cont 34#;
	Loc happy_dollar_dollar Dot -> cont 35#;
	Loc happy_dollar_dollar DotDot -> cont 36#;
	Loc happy_dollar_dollar Colon -> cont 37#;
	Loc happy_dollar_dollar DoubleColon -> cont 38#;
	Loc happy_dollar_dollar Equals -> cont 39#;
	Loc happy_dollar_dollar Backslash -> cont 40#;
	Loc happy_dollar_dollar Bar -> cont 41#;
	Loc happy_dollar_dollar LeftArrow -> cont 42#;
	Loc happy_dollar_dollar RightArrow -> cont 43#;
	Loc happy_dollar_dollar At -> cont 44#;
	Loc happy_dollar_dollar Tilde -> cont 45#;
	Loc happy_dollar_dollar DoubleArrow -> cont 46#;
	Loc happy_dollar_dollar Minus -> cont 47#;
	Loc happy_dollar_dollar Exclamation -> cont 48#;
	Loc happy_dollar_dollar Star -> cont 49#;
	Loc happy_dollar_dollar LeftArrowTail -> cont 50#;
	Loc happy_dollar_dollar RightArrowTail -> cont 51#;
	Loc happy_dollar_dollar LeftDblArrowTail -> cont 52#;
	Loc happy_dollar_dollar RightDblArrowTail -> cont 53#;
	Loc happy_dollar_dollar RPGuardOpen -> cont 54#;
	Loc happy_dollar_dollar RPGuardClose -> cont 55#;
	Loc happy_dollar_dollar RPCAt -> cont 56#;
	Loc _ (THIdEscape _) -> cont 57#;
	Loc happy_dollar_dollar THParenEscape -> cont 58#;
	Loc happy_dollar_dollar THExpQuote -> cont 59#;
	Loc happy_dollar_dollar THPatQuote -> cont 60#;
	Loc happy_dollar_dollar THTypQuote -> cont 61#;
	Loc happy_dollar_dollar THDecQuote -> cont 62#;
	Loc happy_dollar_dollar THCloseQuote -> cont 63#;
	Loc happy_dollar_dollar THVarQuote -> cont 64#;
	Loc happy_dollar_dollar THTyQuote -> cont 65#;
	Loc _ (THQuasiQuote _) -> cont 66#;
	Loc _ (XPCDATA _) -> cont 67#;
	Loc happy_dollar_dollar XStdTagOpen -> cont 68#;
	Loc happy_dollar_dollar XCloseTagOpen -> cont 69#;
	Loc happy_dollar_dollar XCodeTagOpen -> cont 70#;
	Loc happy_dollar_dollar XChildTagOpen -> cont 71#;
	Loc happy_dollar_dollar XStdTagClose -> cont 72#;
	Loc happy_dollar_dollar XEmptyTagClose -> cont 73#;
	Loc happy_dollar_dollar XCodeTagClose -> cont 74#;
	Loc happy_dollar_dollar XRPatOpen -> cont 75#;
	Loc happy_dollar_dollar XRPatClose -> cont 76#;
	Loc happy_dollar_dollar KW_Foreign -> cont 77#;
	Loc happy_dollar_dollar KW_Export -> cont 78#;
	Loc happy_dollar_dollar KW_Safe -> cont 79#;
	Loc happy_dollar_dollar KW_Unsafe -> cont 80#;
	Loc happy_dollar_dollar KW_Threadsafe -> cont 81#;
	Loc happy_dollar_dollar KW_Interruptible -> cont 82#;
	Loc happy_dollar_dollar KW_StdCall -> cont 83#;
	Loc happy_dollar_dollar KW_CCall -> cont 84#;
	Loc happy_dollar_dollar KW_CPlusPlus -> cont 85#;
	Loc happy_dollar_dollar KW_DotNet -> cont 86#;
	Loc happy_dollar_dollar KW_Jvm -> cont 87#;
	Loc happy_dollar_dollar KW_Js -> cont 88#;
	Loc happy_dollar_dollar KW_CApi -> cont 89#;
	Loc happy_dollar_dollar KW_As -> cont 90#;
	Loc happy_dollar_dollar KW_By -> cont 91#;
	Loc happy_dollar_dollar KW_Case -> cont 92#;
	Loc happy_dollar_dollar KW_Class -> cont 93#;
	Loc happy_dollar_dollar KW_Data -> cont 94#;
	Loc happy_dollar_dollar KW_Default -> cont 95#;
	Loc happy_dollar_dollar KW_Deriving -> cont 96#;
	Loc happy_dollar_dollar KW_Do -> cont 97#;
	Loc happy_dollar_dollar KW_Else -> cont 98#;
	Loc happy_dollar_dollar KW_Family -> cont 99#;
	Loc happy_dollar_dollar KW_Forall -> cont 100#;
	Loc happy_dollar_dollar KW_Group -> cont 101#;
	Loc happy_dollar_dollar KW_Hiding -> cont 102#;
	Loc happy_dollar_dollar KW_If -> cont 103#;
	Loc happy_dollar_dollar KW_Import -> cont 104#;
	Loc happy_dollar_dollar KW_In -> cont 105#;
	Loc happy_dollar_dollar KW_Infix -> cont 106#;
	Loc happy_dollar_dollar KW_InfixL -> cont 107#;
	Loc happy_dollar_dollar KW_InfixR -> cont 108#;
	Loc happy_dollar_dollar KW_Instance -> cont 109#;
	Loc happy_dollar_dollar KW_Let -> cont 110#;
	Loc happy_dollar_dollar KW_MDo -> cont 111#;
	Loc happy_dollar_dollar KW_Module -> cont 112#;
	Loc happy_dollar_dollar KW_NewType -> cont 113#;
	Loc happy_dollar_dollar KW_Of -> cont 114#;
	Loc happy_dollar_dollar KW_Proc -> cont 115#;
	Loc happy_dollar_dollar KW_Rec -> cont 116#;
	Loc happy_dollar_dollar KW_Then -> cont 117#;
	Loc happy_dollar_dollar KW_Type -> cont 118#;
	Loc happy_dollar_dollar KW_Using -> cont 119#;
	Loc happy_dollar_dollar KW_Where -> cont 120#;
	Loc happy_dollar_dollar KW_Qualified -> cont 121#;
	Loc _ (INLINE _) -> cont 122#;
	Loc happy_dollar_dollar INLINE_CONLIKE -> cont 123#;
	Loc happy_dollar_dollar SPECIALISE -> cont 124#;
	Loc _ (SPECIALISE_INLINE _) -> cont 125#;
	Loc happy_dollar_dollar SOURCE -> cont 126#;
	Loc happy_dollar_dollar RULES -> cont 127#;
	Loc happy_dollar_dollar CORE -> cont 128#;
	Loc happy_dollar_dollar SCC -> cont 129#;
	Loc happy_dollar_dollar GENERATED -> cont 130#;
	Loc happy_dollar_dollar DEPRECATED -> cont 131#;
	Loc happy_dollar_dollar WARNING -> cont 132#;
	Loc happy_dollar_dollar UNPACK -> cont 133#;
	Loc _ (OPTIONS _) -> cont 134#;
	Loc happy_dollar_dollar LANGUAGE -> cont 135#;
	Loc happy_dollar_dollar ANN -> cont 136#;
	Loc happy_dollar_dollar PragmaEnd -> cont 137#;
	_ -> happyError' tk
	})

happyError_ 138# tk = happyError' tk
happyError_ _ tk = happyError' tk

happyThen :: () => P a -> (a -> P b) -> P b
happyThen = (>>=)
happyReturn :: () => a -> P a
happyReturn = (return)
happyThen1 = happyThen
happyReturn1 :: () => a -> P a
happyReturn1 = happyReturn
happyError' :: () => (Loc Token) -> P a
happyError' tk = parseError tk

mparseModule = happySomeParser where
  happySomeParser = happyThen (happyParse 0#) (\x -> happyReturn (happyOut13 x))

mparseExp = happySomeParser where
  happySomeParser = happyThen (happyParse 1#) (\x -> happyReturn (happyOut146 x))

mparsePat = happySomeParser where
  happySomeParser = happyThen (happyParse 2#) (\x -> happyReturn (happyOut192 x))

mparseDecl = happySomeParser where
  happySomeParser = happyThen (happyParse 3#) (\x -> happyReturn (happyOut50 x))

mparseType = happySomeParser where
  happySomeParser = happyThen (happyParse 4#) (\x -> happyReturn (happyOut92 x))

mparseStmt = happySomeParser where
  happySomeParser = happyThen (happyParse 5#) (\x -> happyReturn (happyOut200 x))

mparseModules = happySomeParser where
  happySomeParser = happyThen (happyParse 6#) (\x -> happyReturn (happyOut11 x))

mfindOptPragmas = happySomeParser where
  happySomeParser = happyThen (happyParse 7#) (\x -> happyReturn (happyOut15 x))

happySeq = happyDontSeq


type L = SrcSpanInfo -- just for convenience
type S = SrcSpan

parseError :: Loc Token -> P a
parseError t = fail $ "Parse error: " ++ showToken (unLoc t)

(<>) :: (Annotated a, Annotated b) => a SrcSpanInfo -> b SrcSpanInfo -> SrcSpanInfo
a <> b = ann a <++> ann b
infixl 6 <>

nIS = noInfoSpan
iS = infoSpan


-- | Parse of a string, which should contain a complete Haskell module.
parseModule :: String -> ParseResult (Module SrcSpanInfo)
parseModule = simpleParse mparseModule

-- | Parse of a string containing a complete Haskell module, using an explicit mode.
parseModuleWithMode :: ParseMode -> String -> ParseResult (Module SrcSpanInfo)
parseModuleWithMode = modeParse mparseModule

-- | Parse of a string containing a complete Haskell module, using an explicit mode, retaining comments.
parseModuleWithComments :: ParseMode -> String -> ParseResult (Module SrcSpanInfo, [Comment])
parseModuleWithComments = commentParse mparseModule

-- | Parse of a string containing a Haskell expression.
parseExp :: String -> ParseResult (Exp SrcSpanInfo)
parseExp = simpleParse mparseExp

-- | Parse of a string containing a Haskell expression, using an explicit mode.
parseExpWithMode :: ParseMode -> String -> ParseResult (Exp SrcSpanInfo)
parseExpWithMode = modeParse mparseExp

-- | Parse of a string containing a complete Haskell module, using an explicit mode, retaining comments.
parseExpWithComments :: ParseMode -> String -> ParseResult (Exp SrcSpanInfo, [Comment])
parseExpWithComments = commentParse mparseExp

-- | Parse of a string containing a Haskell pattern.
parsePat :: String -> ParseResult (Pat SrcSpanInfo)
parsePat = simpleParse mparsePat

-- | Parse of a string containing a Haskell pattern, using an explicit mode.
parsePatWithMode :: ParseMode -> String -> ParseResult (Pat SrcSpanInfo)
parsePatWithMode = modeParse mparsePat

-- | Parse of a string containing a complete Haskell module, using an explicit mode, retaining comments.
parsePatWithComments :: ParseMode -> String -> ParseResult (Pat SrcSpanInfo, [Comment])
parsePatWithComments = commentParse mparsePat

-- | Parse of a string containing a Haskell top-level declaration.
parseDecl :: String -> ParseResult (Decl SrcSpanInfo)
parseDecl = simpleParse mparseDecl

-- | Parse of a string containing a Haskell top-level declaration, using an explicit mode.
parseDeclWithMode :: ParseMode -> String -> ParseResult (Decl SrcSpanInfo)
parseDeclWithMode = modeParse mparseDecl

-- | Parse of a string containing a complete Haskell module, using an explicit mode, retaining comments.
parseDeclWithComments :: ParseMode -> String -> ParseResult (Decl SrcSpanInfo, [Comment])
parseDeclWithComments = commentParse mparseDecl

-- | Parse of a string containing a Haskell type.
parseType :: String -> ParseResult (Type SrcSpanInfo)
parseType = runParser mparseType

-- | Parse of a string containing a Haskell type, using an explicit mode.
parseTypeWithMode :: ParseMode -> String -> ParseResult (Type SrcSpanInfo)
parseTypeWithMode mode = runParserWithMode mode mparseType

-- | Parse of a string containing a complete Haskell module, using an explicit mode, retaining comments.
parseTypeWithComments :: ParseMode -> String -> ParseResult (Type SrcSpanInfo, [Comment])
parseTypeWithComments mode str = runParserWithModeComments mode mparseType str

-- | Parse of a string containing a Haskell statement.
parseStmt :: String -> ParseResult (Stmt SrcSpanInfo)
parseStmt = simpleParse mparseStmt

-- | Parse of a string containing a Haskell type, using an explicit mode.
parseStmtWithMode :: ParseMode -> String -> ParseResult (Stmt SrcSpanInfo)
parseStmtWithMode = modeParse mparseStmt

-- | Parse of a string containing a complete Haskell module, using an explicit mode, retaining comments.
parseStmtWithComments :: ParseMode -> String -> ParseResult (Stmt SrcSpanInfo, [Comment])
parseStmtWithComments = commentParse mparseStmt


simpleParse :: AppFixity a => P (a L) -> String -> ParseResult (a L)
simpleParse p = applyFixities preludeFixities <=< runParser p

modeParse :: AppFixity a => P (a L) -> ParseMode -> String -> ParseResult (a L)
modeParse p mode = applyFixities' (fixities mode) <=< runParserWithMode mode p

commentParse :: AppFixity a => P (a L) -> ParseMode -> String -> ParseResult (a L, [Comment])
commentParse p mode str = do (ast, cs) <- runParserWithModeComments mode p str
                             ast' <- applyFixities' (fixities mode) ast
                             return (ast', cs)

-- | Partial parse of a string starting with a series of top-level option pragmas.
getTopPragmas :: String -> ParseResult [ModulePragma SrcSpanInfo]
getTopPragmas = runParser (mfindOptPragmas >>= \(ps,_,_) -> return ps)

-- | Parse of a string, which should contain a complete Haskell module.
parseModules :: String -> ParseResult [Module SrcSpanInfo]
parseModules = mapM (applyFixities preludeFixities) <=< runParser mparseModules

-- | Parse of a string containing a complete Haskell module, using an explicit mode.
parseModulesWithMode :: ParseMode -> String -> ParseResult [Module SrcSpanInfo]
parseModulesWithMode mode = mapM (applyFixities' (fixities mode)) <=< runParserWithMode mode mparseModules

-- | Parse of a string containing a complete Haskell module, using an explicit mode, retaining comments.
parseModulesWithComments :: ParseMode -> String -> ParseResult ([Module SrcSpanInfo], [Comment])
parseModulesWithComments mode str = do (ast,cs) <- runParserWithModeComments mode mparseModules str
                                       ast' <- mapM (applyFixities' (fixities mode)) ast
                                       return (ast', cs)
applyFixities' :: (AppFixity a) => Maybe [Fixity] -> a L -> ParseResult (a L)
applyFixities' Nothing ast = return ast
applyFixities' (Just fixs) ast = applyFixities fixs ast
{-# LINE 1 "templates/GenericTemplate.hs" #-}
{-# LINE 1 "templates/GenericTemplate.hs" #-}
{-# LINE 1 "<command-line>" #-}
{-# LINE 8 "<command-line>" #-}
# 1 "/usr/include/stdc-predef.h" 1 3 4

# 17 "/usr/include/stdc-predef.h" 3 4



















# 47 "/usr/include/stdc-predef.h" 3 4

# 59 "/usr/include/stdc-predef.h" 3 4








{-# LINE 8 "<command-line>" #-}
{-# LINE 1 "templates/GenericTemplate.hs" #-}
-- Id: GenericTemplate.hs,v 1.26 2005/01/14 14:47:22 simonmar Exp 

{-# LINE 13 "templates/GenericTemplate.hs" #-}





-- Do not remove this comment. Required to fix CPP parsing when using GCC and a clang-compiled alex.
#if __GLASGOW_HASKELL__ > 706
#define LT(n,m) ((Happy_GHC_Exts.tagToEnum# (n Happy_GHC_Exts.<# m)) :: Bool)
#define GTE(n,m) ((Happy_GHC_Exts.tagToEnum# (n Happy_GHC_Exts.>=# m)) :: Bool)
#define EQ(n,m) ((Happy_GHC_Exts.tagToEnum# (n Happy_GHC_Exts.==# m)) :: Bool)
#else
#define LT(n,m) (n Happy_GHC_Exts.<# m)
#define GTE(n,m) (n Happy_GHC_Exts.>=# m)
#define EQ(n,m) (n Happy_GHC_Exts.==# m)
#endif
{-# LINE 46 "templates/GenericTemplate.hs" #-}


data Happy_IntList = HappyCons Happy_GHC_Exts.Int# Happy_IntList





{-# LINE 67 "templates/GenericTemplate.hs" #-}

{-# LINE 77 "templates/GenericTemplate.hs" #-}

{-# LINE 86 "templates/GenericTemplate.hs" #-}

infixr 9 `HappyStk`
data HappyStk a = HappyStk a (HappyStk a)

-----------------------------------------------------------------------------
-- starting the parse

happyParse start_state = happyNewToken start_state notHappyAtAll notHappyAtAll

-----------------------------------------------------------------------------
-- Accepting the parse

-- If the current token is 0#, it means we've just accepted a partial
-- parse (a %partial parser).  We must ignore the saved token on the top of
-- the stack in this case.
happyAccept 0# tk st sts (_ `HappyStk` ans `HappyStk` _) =
        happyReturn1 ans
happyAccept j tk st sts (HappyStk ans _) = 
        (happyTcHack j (happyTcHack st)) (happyReturn1 ans)

-----------------------------------------------------------------------------
-- Arrays only: do the next action



happyDoAction i tk st
        = {- nothing -}


          case action of
                0#           -> {- nothing -}
                                     happyFail i tk st
                -1#          -> {- nothing -}
                                     happyAccept i tk st
                n | LT(n,(0# :: Happy_GHC_Exts.Int#)) -> {- nothing -}

                                                   (happyReduceArr Happy_Data_Array.! rule) i tk st
                                                   where rule = (Happy_GHC_Exts.I# ((Happy_GHC_Exts.negateInt# ((n Happy_GHC_Exts.+# (1# :: Happy_GHC_Exts.Int#))))))
                n                 -> {- nothing -}


                                     happyShift new_state i tk st
                                     where new_state = (n Happy_GHC_Exts.-# (1# :: Happy_GHC_Exts.Int#))
   where off    = indexShortOffAddr happyActOffsets st
         off_i  = (off Happy_GHC_Exts.+# i)
         check  = if GTE(off_i,(0# :: Happy_GHC_Exts.Int#))
                  then EQ(indexShortOffAddr happyCheck off_i, i)
                  else False
         action
          | check     = indexShortOffAddr happyTable off_i
          | otherwise = indexShortOffAddr happyDefActions st


indexShortOffAddr (HappyA# arr) off =
        Happy_GHC_Exts.narrow16Int# i
  where
        i = Happy_GHC_Exts.word2Int# (Happy_GHC_Exts.or# (Happy_GHC_Exts.uncheckedShiftL# high 8#) low)
        high = Happy_GHC_Exts.int2Word# (Happy_GHC_Exts.ord# (Happy_GHC_Exts.indexCharOffAddr# arr (off' Happy_GHC_Exts.+# 1#)))
        low  = Happy_GHC_Exts.int2Word# (Happy_GHC_Exts.ord# (Happy_GHC_Exts.indexCharOffAddr# arr off'))
        off' = off Happy_GHC_Exts.*# 2#





data HappyAddr = HappyA# Happy_GHC_Exts.Addr#




-----------------------------------------------------------------------------
-- HappyState data type (not arrays)

{-# LINE 170 "templates/GenericTemplate.hs" #-}

-----------------------------------------------------------------------------
-- Shifting a token

happyShift new_state 0# tk st sts stk@(x `HappyStk` _) =
     let i = (case Happy_GHC_Exts.unsafeCoerce# x of { (Happy_GHC_Exts.I# (i)) -> i }) in
--     trace "shifting the error token" $
     happyDoAction i tk new_state (HappyCons (st) (sts)) (stk)

happyShift new_state i tk st sts stk =
     happyNewToken new_state (HappyCons (st) (sts)) ((happyInTok (tk))`HappyStk`stk)

-- happyReduce is specialised for the common cases.

happySpecReduce_0 i fn 0# tk st sts stk
     = happyFail 0# tk st sts stk
happySpecReduce_0 nt fn j tk st@((action)) sts stk
     = happyGoto nt j tk st (HappyCons (st) (sts)) (fn `HappyStk` stk)

happySpecReduce_1 i fn 0# tk st sts stk
     = happyFail 0# tk st sts stk
happySpecReduce_1 nt fn j tk _ sts@((HappyCons (st@(action)) (_))) (v1`HappyStk`stk')
     = let r = fn v1 in
       happySeq r (happyGoto nt j tk st sts (r `HappyStk` stk'))

happySpecReduce_2 i fn 0# tk st sts stk
     = happyFail 0# tk st sts stk
happySpecReduce_2 nt fn j tk _ (HappyCons (_) (sts@((HappyCons (st@(action)) (_))))) (v1`HappyStk`v2`HappyStk`stk')
     = let r = fn v1 v2 in
       happySeq r (happyGoto nt j tk st sts (r `HappyStk` stk'))

happySpecReduce_3 i fn 0# tk st sts stk
     = happyFail 0# tk st sts stk
happySpecReduce_3 nt fn j tk _ (HappyCons (_) ((HappyCons (_) (sts@((HappyCons (st@(action)) (_))))))) (v1`HappyStk`v2`HappyStk`v3`HappyStk`stk')
     = let r = fn v1 v2 v3 in
       happySeq r (happyGoto nt j tk st sts (r `HappyStk` stk'))

happyReduce k i fn 0# tk st sts stk
     = happyFail 0# tk st sts stk
happyReduce k nt fn j tk st sts stk
     = case happyDrop (k Happy_GHC_Exts.-# (1# :: Happy_GHC_Exts.Int#)) sts of
         sts1@((HappyCons (st1@(action)) (_))) ->
                let r = fn stk in  -- it doesn't hurt to always seq here...
                happyDoSeq r (happyGoto nt j tk st1 sts1 r)

happyMonadReduce k nt fn 0# tk st sts stk
     = happyFail 0# tk st sts stk
happyMonadReduce k nt fn j tk st sts stk =
      case happyDrop k (HappyCons (st) (sts)) of
        sts1@((HappyCons (st1@(action)) (_))) ->
          let drop_stk = happyDropStk k stk in
          happyThen1 (fn stk tk) (\r -> happyGoto nt j tk st1 sts1 (r `HappyStk` drop_stk))

happyMonad2Reduce k nt fn 0# tk st sts stk
     = happyFail 0# tk st sts stk
happyMonad2Reduce k nt fn j tk st sts stk =
      case happyDrop k (HappyCons (st) (sts)) of
        sts1@((HappyCons (st1@(action)) (_))) ->
         let drop_stk = happyDropStk k stk

             off = indexShortOffAddr happyGotoOffsets st1
             off_i = (off Happy_GHC_Exts.+# nt)
             new_state = indexShortOffAddr happyTable off_i



          in
          happyThen1 (fn stk tk) (\r -> happyNewToken new_state sts1 (r `HappyStk` drop_stk))

happyDrop 0# l = l
happyDrop n (HappyCons (_) (t)) = happyDrop (n Happy_GHC_Exts.-# (1# :: Happy_GHC_Exts.Int#)) t

happyDropStk 0# l = l
happyDropStk n (x `HappyStk` xs) = happyDropStk (n Happy_GHC_Exts.-# (1#::Happy_GHC_Exts.Int#)) xs

-----------------------------------------------------------------------------
-- Moving to a new state after a reduction


happyGoto nt j tk st = 
   {- nothing -}
   happyDoAction j tk new_state
   where off = indexShortOffAddr happyGotoOffsets st
         off_i = (off Happy_GHC_Exts.+# nt)
         new_state = indexShortOffAddr happyTable off_i




-----------------------------------------------------------------------------
-- Error recovery (0# is the error token)

-- parse error if we are in recovery and we fail again
happyFail 0# tk old_st _ stk@(x `HappyStk` _) =
     let i = (case Happy_GHC_Exts.unsafeCoerce# x of { (Happy_GHC_Exts.I# (i)) -> i }) in
--      trace "failing" $ 
        happyError_ i tk

{-  We don't need state discarding for our restricted implementation of
    "error".  In fact, it can cause some bogus parses, so I've disabled it
    for now --SDM

-- discard a state
happyFail  0# tk old_st (HappyCons ((action)) (sts)) 
                                                (saved_tok `HappyStk` _ `HappyStk` stk) =
--      trace ("discarding state, depth " ++ show (length stk))  $
        happyDoAction 0# tk action sts ((saved_tok`HappyStk`stk))
-}

-- Enter error recovery: generate an error token,
--                       save the old token and carry on.
happyFail  i tk (action) sts stk =
--      trace "entering error recovery" $
        happyDoAction 0# tk action sts ( (Happy_GHC_Exts.unsafeCoerce# (Happy_GHC_Exts.I# (i))) `HappyStk` stk)

-- Internal happy errors:

notHappyAtAll :: a
notHappyAtAll = error "Internal Happy error\n"

-----------------------------------------------------------------------------
-- Hack to get the typechecker to accept our action functions


happyTcHack :: Happy_GHC_Exts.Int# -> a -> a
happyTcHack x y = y
{-# INLINE happyTcHack #-}


-----------------------------------------------------------------------------
-- Seq-ing.  If the --strict flag is given, then Happy emits 
--      happySeq = happyDoSeq
-- otherwise it emits
--      happySeq = happyDontSeq

happyDoSeq, happyDontSeq :: a -> b -> b
happyDoSeq   a b = a `seq` b
happyDontSeq a b = b

-----------------------------------------------------------------------------
-- Don't inline any functions from the template.  GHC has a nasty habit
-- of deciding to inline happyGoto everywhere, which increases the size of
-- the generated parser quite a bit.


{-# NOINLINE happyDoAction #-}
{-# NOINLINE happyTable #-}
{-# NOINLINE happyCheck #-}
{-# NOINLINE happyActOffsets #-}
{-# NOINLINE happyGotoOffsets #-}
{-# NOINLINE happyDefActions #-}

{-# NOINLINE happyShift #-}
{-# NOINLINE happySpecReduce_0 #-}
{-# NOINLINE happySpecReduce_1 #-}
{-# NOINLINE happySpecReduce_2 #-}
{-# NOINLINE happySpecReduce_3 #-}
{-# NOINLINE happyReduce #-}
{-# NOINLINE happyMonadReduce #-}
{-# NOINLINE happyGoto #-}
{-# NOINLINE happyFail #-}

-- end of Happy Template.
